/**************************************************************************
   fotoxx      digital photo edit program

   Free software licensed under GNU General Public License v.2
   source URL:  kornelix.squarespace.com
***************************************************************************/

#include <fcntl.h>
#include <signal.h>
#include <FreeImage.h>
#include "zfuncs.h"

#define fversion "fotoxx v.6.9.2  2009.05.22"                              //  version and release date
#define flicense "Free software - GNU General Public License v.2"
#define fsource "http://kornelix.squarespace.com"
#define ftranslators "\n Translators:"                                     \
   "\n Stanislas Zeller, Antonio Sánchez, Miguel Bouzada,"                 \
   "\n The Hamsters, Helge Soencksen, Jie Luo"
#define fcredits "\n Programs used: FreeImage, ufraw, exiftool"

//  external references

namespace      image_navi 
{ 
   extern int     xwinW, xwinH;                                            //  thumbnail index window size
   extern int     thumbsize;                                               //  thumbnail image size
}

//  GTK definitions

#define WLAYOUT GTK_LAYOUT(drWin)->bin_window                              //  main window drawing area
#define textwin GTK_TEXT_WINDOW_TEXT                                       //  GDK window of GTK text view
#define nodither GDK_RGB_DITHER_NONE
#define SCROLLWIN GTK_SCROLLED_WINDOW(mScroll)
#define ALWAYS GTK_POLICY_ALWAYS
#define NEVER GTK_POLICY_NEVER
#define dottedline GDK_LINE_DOUBLE_DASH,GDK_CAP_BUTT,GDK_JOIN_MITER
#define PXB GdkPixbuf
#define colorspace GDK_COLORSPACE_RGB

GtkWidget      *mWin, *drWin, *mVbox, *mScroll;                            //  main and drawing window
GtkWidget      *mMbar, *mTbar, *STbar;                                     //  menu bar, tool bar, status bar
GtkWidget      *brightgraph = 0;                                           //  brightness distribution graph
GtkWidget      *drbrightgraph = 0;                                         //  drawing area
GError         **gerror = 0;
GdkGC          *gdkgc = 0;                                                 //  GDK graphics context
GdkColor       black, white;
GdkColormap    *colormap = 0;
uint           maxcolor = 0xffff;
GdkCursor      *arrowcursor = 0;
GdkCursor      *dragcursor = 0;

//  FREEIMAGE definitions

#define FIF FREE_IMAGE_FORMAT                                              //  file types, jpeg or tiff
#define FIB FIBITMAP                                                       //  FreeImage bitmap
#define RR FI_RGBA_RED                                                     //  FIB-24 pixel color order
#define GG FI_RGBA_GREEN
#define BB FI_RGBA_BLUE

//  fotoxx definitions

#define mega (1024 * 1024)
#define track(name) printf("track: %s \n",#name);                          //  debug path tracker
#define ftrash "Desktop/fotoxx-trash"                                      //  trash folder location
#define Pundo_max 50                                                       //  max undo/redo files
#define pixsamp 5000                                                       //  pixel sample size (HDR, pano)

#define maxtag1 30                                                         //  max tag cc for one tag  v.6.8
#define maxtag2 300                                                        //  max tag cc for one image file
#define maxtag3 20000                                                      //  max tag cc for all image files
#define maxtag4 200                                                        //  max tag cc for search tags
#define maxtag5 200                                                        //  max tag cc for recent tags
#define maxntags 1000                                                      //  max tag count for all images
#define maxtagF 300                                                        //  max image search /path*/file*

#define bmpixel(rgb,px,py) ((uint16 *) rgb->bmp+(py*(rgb->ww)+px)*3)       //  return RGB pixel[3] at (px,py)
#define brightness(pix) (0.25*pix[0]+0.65*pix[1]+0.10*pix[2])              //  pixel brightness, 0-64K
#define redness(pix) (25 * pix[0] / (brightness(pix)+1))                   //  pixel redness, 0-100%     v.6.9
#define setredness(pix,redness) (pix[0] = \
   int(redness*(0.65*pix[1]+0.10*pix[2]+1)/(25-0.25*redness)))             //  set pixel redness, 0-100%

#define exif_tags_key "UserComment"                                        //  EXIF key for image tags
#define exif_date_key "DateTimeOriginal"                                   //  EXIF key for image date
#define exif_width_key "ExifImageWidth"                                    //  EXIF key for image width
#define exif_height_key "ExifImageHeight"                                  //  EXIF key for image height
#define exif_orientation_key "Orientation"                                 //  EXIF key for image orientation

pthread_t   tid_fmain = 0;                                                 //  thread ID for main()
char        *topmenu;                                                      //  latest top-menu selection
int         Fexiftool = 0;                                                 //  exiftool program available
int         Fufraw = 0;                                                    //  ufraw program is available
int         Fprintoxx = 0;                                                 //  printoxx program available
int         Fxdgopen = 0;                                                  //  xdg-open program available
int         Fshutdown = 0;                                                 //  app shutdown underway
int         Fdebug = 0;                                                    //  debug flag
int         Wrepaint = 0;                                                  //  request window paint
int         Wpainted = 0;                                                  //  window was repainted
int         Fmodified = 0;                                                 //  image was edited/modified
int         Fpreview = 0;                                                  //  use window image for edits
int         Fblowup = 0;                                                   //  zoom small images to window size
int         SBupdate = 0;                                                  //  request update status bar
int         PercentDone = 0;                                               //  % completion in status bar
int         Fautolens = 0;                                                 //  lens parameter search underway
int         Fsearchlist = 0;                                               //  file list via search tags  v.6.4
int         Fimageturned = 0;                                              //  image was turned when loaded

char        *image_file = 0;                                               //  current image file
double      file_MB;                                                       //  disk file size, MB
int         file_bpp;                                                      //  disk file bits/pixel
char        file_type[8];                                                  //  "jpeg" or "tiff" or "other"
char        *asstagsfile = 0;                                              //  assigned tags file
char        *topdirk = 0;                                                  //  top-level image directory

//  fotoxx RGB bitmaps                                                     //  v.6.5

typedef struct  {                                                          //  RGB bitmap
   char     wmi[8];
   int      ww, hh, bpp;                                                   //  width, height, bits per pixel
   void     *bmp;                                                          //  uint8*/uint16* (bpp=24/48)
}  RGB;

RGB         *Frgb24 = 0;                                                   //  input file 1 bitmap, RGB-24
RGB         *Frgb48 = 0;                                                   //  input file 1 bitmap, RGB-48
RGB         *Grgb48 = 0;                                                   //  input file 2 bitmap, RGB-48
RGB         *Mrgb24 = 0;                                                   //  main window bitmap, RGB-24
RGB         *E1rgb48 = 0;                                                  //  edit bitmap, base image
RGB         *E3rgb48 = 0;                                                  //  edit bitmap, edited image
RGB         *A1rgb48 = 0, *A2rgb48 = 0;                                    //  align image bitmaps (HDR, pano)

int         Fww, Fhh, Gww, Ghh;                                            //  input image dimensions
int         E1ww, E1hh, E3ww, E3hh;                                        //  edit image dimensions
int         A1ww, A1hh, A2ww, A2hh;                                        //  alignment image dimensions
int         Mww = 700, Mhh = 500;                                          //  main window size (defaults)

int         Morgx, Morgy;                                                  //  Mrgb24 origin in drawing window
double      Fzoom = 0;                                                     //  image zoom scale (0 = fit window)
double      Mscale = 1;                                                    //  scale factor, file to window

mutex       bitmaps_lock;                                                  //  lock for accessing RGB bitmaps
int         menu_lock = 0;                                                 //  lock for some menu functions

char        *undo_files = 0;                                               //  undo/redo stack, image files
int         Pundo = 0;                                                     //  undo/redo stack position
int         Pumax = 0;                                                     //  undo/redo stack depth

int         Mbutton = 0;                                                   //  mouse button, 1/3 = left/right
int         LMclick = 0, RMclick = 0;                                      //  mouse left, right click
int         Mxclick, Myclick;                                              //  mouse click position
int         Mxdown, Mydown, Mxdrag, Mydrag;                                //  mouse drag vector
int         Mcapture = 0;                                                  //  mouse captured by edit function
int         KBcapture = 0;                                                 //  KB key captured by edit function
int         KBkey = 0;                                                     //  active keyboard key

int         Ndotlines = 0;                                                 //  dotted lines on top of
int         dotlinex1[1000], dotliney1[1000];                              //    image in window
int         dotlinex2[1000], dotliney2[1000];

zdialog     *zdtags = null;                                                //  edit tags zdialog
zdialog     *zdrename = null;                                              //  rename file zdialog
zdialog     *zdsela = null;                                                //  select area dialog
zdialog     *zdedit = null;                                                //  image edit dialog

//  pano and HDR control data

int      curr_lens = 0;                                                    //  current lens, 0-3
int      lens_cc = 19;                                                     //  lens name cc limit
char     *lens4_name[4];                                                   //  names for 4 lenses
double   lens4_mm[4], lens4_bow[4];                                        //  characteristics for 4 lenses
double   lens_mm, lens_bow;                                                //  current lens characteristics
double   pano_prealign_size = 500;                                         //  pre-align image height
double   pano_image_increase = 1.6;                                        //  image size increase per stage
double   pano_blend_decrease = 0.8;                                        //  blend width decrease per stage
double   pano_blend_min = 0.10;                                            //  minimum blend, 10% image width
double   pano_blend_max = 0.20;                                            //  maximum blend, 20%
double   pano_ystretch = 0.002;                                            //  y-stretch (0.2%/image-half/stage)
double   pano_ystretch_decrease = 0.9;                                     //  decrease per align stage

int      fullSize, alignSize;                                              //  full and align image sizes
int      showRedpix = 0;                                                   //  flag, highlight alignment pixels
int      pxL, pxH, pyL, pyH;                                               //  image overlap rectangle
int      Nalign = 0;                                                       //  counts alignment cycles
int      blendstripe = 0;                                                  //  pano image blend stripe
int      pxM, pxmL, pxmH;                                                  //  defines overlap stripe
int      overlapixs = 0;                                                   //  overlapped pixels count
char     *redpixels = 0;                                                   //  flags edge pixels for image align

double   Bratios1[3][256], Bratios2[3][256];                               //  brightness ratios/color/brightness
double   R12match[65536], G12match[65536], B12match[65536];                //  image1/2 color matching factors 
double   R21match[65536], G21match[65536], B21match[65536];                //  image2/1 color matching factors
double   Radjust, Gadjust, Badjust;                                        //  RGB manual adjustmants
double   xoff, yoff, toff, yst1, yst2;                                     //  align data: x, y, theta, y-stretch
double   xoffB, yoffB, toffB, yst1B, yst2B;                                //  align data: current best values
double   xystep, tstep, xylim, tlim, yststep, ystlim;                      //  align step size and search range
double   matchlev, matchB;                                                 //  image alignment match level

//  GTK functions

int   main(int argc, char * argv[]);                                       //  main program
int   gtkinitfunc(void *data);                                             //  GTK initz. function
int   gtimefunc(void *arg);                                                //  periodic function
int   delete_event();                                                      //  window delete event function
void  destroy_event();                                                     //  window destroy event function
int   mwpaint();                                                           //  window repaint - expose event
void  mwpaint2();                                                          //  window repaint - image modified
void  paint_dotlines();                                                    //  window repaint - dotted lines
void  update_statusbar();                                                  //  update main window status bar
void  mouse_event(GtkWidget *, GdkEventButton *, void *);                  //  mouse event function
int   KBpress(GtkWidget *, GdkEventKey *, void *);                         //  KB key press event function
int   KBrelease(GtkWidget *, GdkEventKey *, void *);                       //  KB key release event
void  menufunc(GtkWidget *, const char *menu);                             //  menu function
void  draw_dotline(int x1, int y1, int x2, int y2);                        //  draw dotted-line, image space
void  erase_dotline(int x1, int y1, int x2, int y2);                       //  erase dotted-line

typedef void CBfunc();                                                     //  callback function type
CBfunc   *mouseCBfunc = 0;                                                 //  current mouse handler function
CBfunc   *KBkeyCBfunc = 0;                                                 //  current KB handler function

//  file functions

void  m_index(GtkWidget *, const char *);                                  //  show thumbnail index window
void  m_index2(char *);                                                    //  converts function types
void  m_open(GtkWidget *, const char *);                                   //  open image file (menu)
void  f_open(const char *file);                                            //  open image file (file)
void  m_raw(GtkWidget *, const char *);                                    //  open RAW file
void  m_prev(GtkWidget *, const char *);                                   //  open previous file
void  m_next(GtkWidget *, const char *);                                   //  open next file
void  m_save(GtkWidget *, const char *);                                   //  save modified image
void  m_print(GtkWidget *, const char *);                                  //  print image file(s)
void  m_trash(GtkWidget *, const char *);                                  //  move image to trash
void  m_rename(GtkWidget *, const char *);                                 //  rename file menu function
void  rename_dialog();                                                     //  start rename file dialog
void  m_quit(GtkWidget *, const char *);                                   //  exit application

//  edit functions

void  m_select(GtkWidget *, const char *);                                 //  select image area to edit
void  m_unselect(GtkWidget *, const char *);                               //  clear selected image area
void  m_flatten(GtkWidget *, const char *);                                //  flatten brightness distribution
void  m_tune(GtkWidget *, const char *);                                   //  brightness / color adjustments
void  m_redeye(GtkWidget *, const char *);                                 //  red-eye removal
void  m_blur(GtkWidget *, const char *);                                   //  blur image
void  m_sharpen(GtkWidget *, const char *);                                //  sharpen image
void  m_denoise(GtkWidget *, const char *);                                //  image noise reduction
void  m_resize(GtkWidget *, const char *);                                 //  resize image
void  m_trim(GtkWidget *, const char *);                                   //  trim image
void  m_rotate(GtkWidget *, const char *);                                 //  rotate image
void  m_unbend(GtkWidget *, const char *);                                 //  fix perspective problems
void  m_WarpA(GtkWidget *, const char *);                                  //  warp image area
void  m_WarpI(GtkWidget *, const char *);                                  //  warp image globally
void  m_colordep(GtkWidget *, const char *);                               //  set color depth 1-16 bits/color
void  m_draw(GtkWidget *, const char *);                                   //  make simulated drawing
void  m_emboss(GtkWidget *, const char *);                                 //  make simulated embossing
void  m_tiles(GtkWidget *, const char *);                                  //  make simulated tiles (pixelate)
void  m_paint(GtkWidget *, const char *);                                  //  edit individual pixels
void  m_HDR(GtkWidget *, const char *);                                    //  make HDR image
void  m_pano(GtkWidget *, const char *);                                   //  make panorama image

//  pano and HDR common functions

int      sigdiff(double d1, double d2, double signf);                      //  test for significant difference
void     get_overlap_region();                                             //  get image overlap rectangle
void     get_Bratios();                                                    //  get brightness ratios for overlap 
void     set_colorfix_factors(int state);                                  //  set color matching factors on/off
void     flag_edge_pixels();                                               //  flag high-contrast pixels
double   match_images();                                                   //  match images in overlap region
double   match_pixels(uint16 *pix1, uint16 *pix2);                         //  match two pixels
uint16 * vpixel(RGB *rgb, double px, double py);                           //  get virtual pixel at (px,py)

//  edit support functions

typedef void * threadfunc(void *);                                         //  edit thread function

int   edit_setup(int preview);                                             //  start new edit transaction
void  edit_cancel();                                                       //  cancel edit
void  edit_done();                                                         //  commit edit, add undo stack
void  edit_undo();                                                         //  undo edit, revert
void  edit_redo();                                                         //  redo the last undo
void  edit_fullsize(int flag);                                             //  convert to full-size bitmaps
void  edit_progress(int done, int goal);                                   //  update status bar progress

void  start_thread(threadfunc func, void *arg);                            //  start a working thread
void  signal_thread();                                                     //  signal work is pending
void  wait_thread_idle();                                                  //  wait for work complete
void  wrapup_thread(int command);                                          //  wait for exit or command exit
void  thread_idle_loop();                                                  //  wait for work or exit command
void  exit_thread();                                                       //  exit thread unconditionally

void  m_undo(GtkWidget *, const char *);                                   //  undo one edit
void  m_redo(GtkWidget *, const char *);                                   //  redo one edit
void  save_undo();                                                         //  undo/redo save function
void  load_undo();                                                         //  undo/redo read function

//  tools functions

void  m_zoom(GtkWidget *, const char *);                                   //  zoom image +/-
void  m_montest(GtkWidget *, const char *);                                //  check monitor
void  m_index_tt(GtkWidget *, const char *);                               //  index tags and thumbnails
void  m_brightgraph(GtkWidget *, const char *);                            //  start brightness dist. graph
void  brightgraph_paint();                                                 //  update brightness dist. graph
void  brightgraph_destroy();                                               //  remove graph window
void  m_clone(GtkWidget *, const char *);                                  //  start another fotoxx instance
void  m_slideshow(GtkWidget *, const char *);                              //  slideshow mode
void  m_RGB(GtkWidget *, const char *);                                    //  show RGB values at mouse click
void  m_parms(GtkWidget *, const char *);                                  //  edit parameters
void  m_lang(GtkWidget *, const char *);                                   //  change language     v.6.2

//  tags and EXIF functions

void  m_edit_tags(GtkWidget *, const char *);                              //  edit tags menu function
void  edit_tags_dialog();                                                  //  start edit tags dialog
void  load_filetags(char *file);                                           //  load tags from an image file
void  update_filetags(char *file);                                         //  write updated tags to image file
void  load_asstags();                                                      //  load all assigned tags
void  update_asstags(char *file, int del = 0);                             //  update assigned tags file
void  m_search_tags(GtkWidget *, const char *);                            //  search images for matching tags
void  m_exif_list(GtkWidget *, const char *);                              //  list EXIF data to popup window
char  ** exif_get(cchar *file, cchar **keys, int nkeys);                   //  get EXIF data for given key(s)
int   exif_set(cchar *file, cchar **keys, cchar **text, int nkeys);        //  set EXIF data for given key(s)
int   exif_copy(cchar *f1, cchar *f2, cchar **k, cchar **t, int nk);       //  copy EXIF data + opt. updates

//  other support functions

void  m_help(GtkWidget *, const char *);                                   //  various help functions
int   load_fotoxx_state();                                                 //  load state from prior session
int   save_fotoxx_state();                                                 //  save state for next session
void  free_resources();                                                    //  free all allocated resources
int   mod_keep();                                                          //  query keep/discard edited image
int   menulock(int lock);                                                  //  lock/unlock menu for some funcs
void  upright_image(int angle);                                            //  upright a turned image
void  FI_error(FIF fif, const char *message);                              //  catch FreeImage error messages

//  RGB bitmap and FI bitmap conversion functions                          //  v.6.5

RGB * RGB_make(int ww, int hh, int bpp);                                   //  initialize RGB bitmap
void  RGB_free(RGB *rgb);                                                  //  free RGB bitmap
RGB * RGB_copy(RGB *rgb);                                                  //  copy RGB bitmap
RGB * RGB_convbpp(RGB *rgb);                                               //  convert from 24/48 to 48/24 bpp
FIB * RGB_FIB(RGB *rgb);                                                   //  convert RGB to FI bitmap
RGB * FIB_RGB(FIB *fib);                                                   //  convert FI to RGB bitmap
PXB * RGB_PXB(RGB *rgb);                                                   //  convert RGB bitmap to pixbuf  v.6.7.2
RGB * PXB_RGB(PXB *pxb);                                                   //  convert pixbuf to RGB bitmap
RGB * RGB_rescale(RGB *rgb, int ww, int hh);                               //  rescale RGB bitmap (ww/hh)
RGB * RGB_rotate(RGB *rgb, double angle);                                  //  rotate RGB bitmap
RGB * image_load(const char *filespec, int bpp);                           //  image file >> RGB-24/48 bitmap

//  translatable strings used in multiple dialogs

const char  *Bsavetoedit;
const char  *Bexiftoolmissing;
const char  *BOK;
const char  *Bcancel;
const char  *Bdone;
const char  *Bclear;
const char  *Bapply;
const char  *Bundo;
const char  *Bredo;
const char  *Bsearch;
const char  *Bstart;
const char  *Bfinish;
const char  *Bshow;
const char  *Bdelete;
const char  *Binvert;
const char  *Bblendwidth;
const char  *Bwidth;
const char  *Bheight;
const char  *Bpercent;
const char  *Bpreset;
const char  *Bproceed;
const char  *Bred;
const char  *Bgreen;
const char  *Bblue;
const char  *Bbrightness;
const char  *Bdarker;
const char  *Blighter;


/**************************************************************************
      main program and GTK/GDK functions
***************************************************************************/

int main(int argc, char *argv[])
{
   GtkWidget      *mFile, *mTags, *mTools, *mEdit, *mHelp;                 //  top level menus
   char           lang[8] = "";
   int8           dashes[2] = { 2, 2 };
   
   printf(fversion "\n");                                                  //  print version
   if (argc > 1 && strEqu(argv[1],"-v")) return 0;

   gtk_init(&argc,&argv);                                                  //  initz. GTK
   zlockInit();

   initz_appfiles("fotoxx",null);                                          //  get app directories
   load_fotoxx_state();                                                    //  restore data from last session

   for (int ii = 1; ii < argc; ii++)                                       //  command line options
   {
      if (strEqu(argv[ii],"-d"))                                           //  -d (debug flag)
            Fdebug = 1;
      else if (strEqu(argv[ii],"-l") && argc > ii+1)                       //  -l language code
            strncpy0(lang,argv[++ii],7);
      else  image_file = strdupz(argv[ii]);                                //  initial file or directory
   }

   ZTXinit(lang);                                                          //  setup translations

   mWin = gtk_window_new(GTK_WINDOW_TOPLEVEL);                             //  create main window
   gtk_window_set_title(GTK_WINDOW(mWin),fversion);
   gtk_window_set_position(GTK_WINDOW(mWin),GTK_WIN_POS_CENTER);
   gtk_window_set_default_size(GTK_WINDOW(mWin),Mww,Mhh);

   mVbox = gtk_vbox_new(0,0);                                              //  add vert. packing box
   gtk_container_add(GTK_CONTAINER(mWin),mVbox);

   mMbar = create_menubar(mVbox,16);                                       //  menus / sub-menus

   mFile = add_menubar_item(mMbar,ZTX("File"),menufunc);
      add_submenu_item(mFile,    ZTX("Index (thumbnails)"),    "m-index.png",    m_index);
      add_submenu_item(mFile,    ZTX("Open RAW File"),         "m-open.png",     m_raw);
      add_submenu_item(mFile,    ZTX("Open Image File"),       "m-open.png",     m_open);
      add_submenu_item(mFile,    ZTX("Save Image File"),       "m-save.png",     m_save);
      add_submenu_item(mFile,    ZTX("Print Image File"),      "m-print.png",    m_print);
      add_submenu_item(mFile,    ZTX("Trash Image File"),      "m-trash.png",    m_trash);
      add_submenu_item(mFile,    ZTX("Rename Image File"),     "m-rename.png",   m_rename);
      add_submenu_item(mFile,    ZTX("Quit fotoxx"),           "m-quit.png",     m_quit);

   mEdit = add_menubar_item(mMbar,ZTX("Edit"),menufunc);
      add_submenu_item(mEdit,    "separator", null, null);                 //  v.6.4
      add_submenu_item(mEdit,    ZTX("Select Area"),           "m-select.png",      m_select);
      add_submenu_item(mEdit,    ZTX("Clear Select Area"),     "m-select.png",      m_unselect);
      add_submenu_item(mEdit,    "separator", null, null);
      add_submenu_item(mEdit,    ZTX("Brightness/Color"), null, null);
      add_submenu_item(mEdit,    ZTX("Flatten Brightness"),    "m-flatten.png",     m_flatten);
      add_submenu_item(mEdit,    ZTX("Tune Image"),            "m-tune.png",        m_tune);
      add_submenu_item(mEdit,    ZTX("Remove Red Eyes"),       "m-redeye.png",      m_redeye);
      add_submenu_item(mEdit,    "separator", null, null);
      add_submenu_item(mEdit,    ZTX("Sharpness"), null, null);
      add_submenu_item(mEdit,    ZTX("Blur Image"),            "m-blur.png",        m_blur);
      add_submenu_item(mEdit,    ZTX("Sharpen Image"),         "m-sharpen.png",     m_sharpen);
      add_submenu_item(mEdit,    ZTX("Reduce Noise"),          "m-denoise.png",     m_denoise);
      add_submenu_item(mEdit,    "separator", null, null);
      add_submenu_item(mEdit,    ZTX("Dimensions"), null, null);
      add_submenu_item(mEdit,    ZTX("Trim Image"),            "m-trim.png",        m_trim);
      add_submenu_item(mEdit,    ZTX("Resize Image"),          "m-resize.png",      m_resize);
      add_submenu_item(mEdit,    ZTX("Rotate Image"),          "m-rotate.png",      m_rotate);
      add_submenu_item(mEdit,    "separator", null, null);
      add_submenu_item(mEdit,    ZTX("Distortion"), null, null);
      add_submenu_item(mEdit,    ZTX("Unbend Image"),          "m-unbend.png",      m_unbend);
      add_submenu_item(mEdit,    ZTX("Warp Area"),             "m-warp.png",        m_WarpA);
      add_submenu_item(mEdit,    ZTX("Warp Image"),            "m-warp.png",        m_WarpI);
      add_submenu_item(mEdit,    "separator", null, null);
      add_submenu_item(mEdit,    ZTX("Special Art Effects"), null, null);
      add_submenu_item(mEdit,    ZTX("Color Depth"),           "m-colordep.png",    m_colordep);
      add_submenu_item(mEdit,    ZTX("Simulate Drawing"),      "m-draw.png",        m_draw);
      add_submenu_item(mEdit,    ZTX("Simulate Embossing"),    "m-emboss.png",      m_emboss);
      add_submenu_item(mEdit,    ZTX("Simulate Tiles"),        "m-tiles.png",       m_tiles);
      add_submenu_item(mEdit,    ZTX("Paint Pixels"),          "m-paint.png",       m_paint);
      add_submenu_item(mEdit,    "separator", null, null);
      add_submenu_item(mEdit,    ZTX("Composite Images"), null, null);
      add_submenu_item(mEdit,    ZTX("Make HDR Image"),        "m-hdr.png",         m_HDR);
      add_submenu_item(mEdit,    ZTX("Make Panorama"),         "m-pano.png",        m_pano);

   mTools = add_menubar_item(mMbar,ZTX("Tools"),menufunc);
      add_submenu_item(mTools,   ZTX("Check Monitor"),         "m-montest.png",     m_montest);
      add_submenu_item(mTools,   ZTX("Index Tags and Thumbs"), "m-index.png",       m_index_tt);
      add_submenu_item(mTools,   ZTX("Brightness Graph"),      "m-distr.png",       m_brightgraph);
      add_submenu_item(mTools,   ZTX("Clone fotoxx"),          "m-clone.png",       m_clone);
      add_submenu_item(mTools,   ZTX("Slide Show"),            "m-slideshow.png",   m_slideshow);
      add_submenu_item(mTools,   ZTX("Show RGB"),              "m-RGB.png",         m_RGB);
      add_submenu_item(mTools,   ZTX("Lens Parameters"),       "m-parms.png",       m_parms);
      add_submenu_item(mTools,   ZTX("Language"),              "m-lang.png",        m_lang);

   mTags = add_menubar_item(mMbar,ZTX("Tags"),menufunc);
      add_submenu_item(mTags,    ZTX("Edit Tags"),          "m-tags.png",        m_edit_tags);
      add_submenu_item(mTags,    ZTX("Search Tags"),        "m-tags.png",        m_search_tags);
      add_submenu_item(mTags,    ZTX("View EXIF Data"),     "m-exif.png",        m_exif_list);

   mHelp = add_menubar_item(mMbar,ZTX("Help"),menufunc);
      add_submenu_item(mHelp,    ZTX("About"),        "m-about.png",       m_help);
      add_submenu_item(mHelp,    ZTX("User Guide"),   "m-userguide.png",   m_help);
      add_submenu_item(mHelp,    "README",            "m-readme.png",      m_help);
      add_submenu_item(mHelp,    ZTX("Change Log"),   "m-changelog.png",   m_help);
      add_submenu_item(mHelp,    ZTX("Translate"),    "m-translate.png",   m_help);
      add_submenu_item(mHelp,    "FreeImage",         "m-FI.png",          m_help);

   mTbar = create_toolbar(mVbox,24);                                       //  toolbar buttons
      add_toolbar_button(mTbar,  ZTX("Index"),  ZTX("Thumbnail Index"),       "index.png",   m_index);
      add_toolbar_button(mTbar,  ZTX("Open"),   ZTX("Open Image File"),       "open.png",    m_open);
      add_toolbar_button(mTbar,  ZTX("Prev"),   ZTX("Open Previous File"),    "prev.png",    m_prev);
      add_toolbar_button(mTbar,  ZTX("Next"),   ZTX("Open Next File"),        "next.png",    m_next);
      add_toolbar_button(mTbar,  ZTX("Save"),   ZTX("Save Image as File"),    "save.png",    m_save);
      add_toolbar_button(mTbar,  ZTX("Undo"),   ZTX("Undo One Edit"),         "undo.png",    m_undo);
      add_toolbar_button(mTbar,  ZTX("Redo"),   ZTX("Redo One Edit"),         "redo.png",    m_redo);
      add_toolbar_button(mTbar,  "Zoom+",       ZTX("Zoom-in (bigger)"),      "zoom+.png",   m_zoom);
      add_toolbar_button(mTbar,  "Zoom-",       ZTX("Zoom-out (smaller)"),    "zoom-.png",   m_zoom);
      add_toolbar_button(mTbar,  ZTX("Trash"),  ZTX("Move Image to Trash"),   "trash.png",   m_trash);
      add_toolbar_button(mTbar,  ZTX("Quit"),   ZTX("Quit fotoxx"),           "quit.png",    m_quit);


   mScroll = gtk_scrolled_window_new(0,0);                                 //  add scrolled window
   gtk_box_pack_start(GTK_BOX(mVbox),mScroll,1,1,0);
   gtk_scrolled_window_set_policy(SCROLLWIN,NEVER,NEVER);

   drWin = gtk_layout_new(0,0);                                            //  drawing window
   gtk_container_add(GTK_CONTAINER(mScroll),drWin);                        //  add to scrolled window

   STbar = create_stbar(mVbox);                                            //  add status bar

   G_SIGNAL(mWin,"delete_event",delete_event,0)                            //  connect signals to windows
   G_SIGNAL(mWin,"destroy",destroy_event,0)
   G_SIGNAL(drWin,"expose-event",mwpaint,0)

   gtk_widget_add_events(drWin,GDK_BUTTON_PRESS_MASK);                     //  connect mouse events
   gtk_widget_add_events(drWin,GDK_BUTTON_RELEASE_MASK);
   gtk_widget_add_events(drWin,GDK_BUTTON_MOTION_MASK);                    //  all buttons   v.6.8
   G_SIGNAL(drWin,"button-press-event",mouse_event,0)
   G_SIGNAL(drWin,"button-release-event",mouse_event,0)
   G_SIGNAL(drWin,"motion-notify-event",mouse_event,0)

   G_SIGNAL(mWin,"key-press-event",KBpress,0)                           	//  connect KB events
   G_SIGNAL(mWin,"key-release-event",KBrelease,0)

   gtk_widget_show_all(mWin);                                              //  show all widgets

   gdkgc = gdk_gc_new(drWin->window);                                      //  initz. graphics context

   black.red = black.green = black.blue = 0;                               //  set up colors black, white
   white.red = white.green = white.blue = maxcolor;
   colormap = gtk_widget_get_colormap(drWin);
   gdk_rgb_find_color(colormap,&black);
   gdk_rgb_find_color(colormap,&white);

   gdk_gc_set_foreground(gdkgc,&black);                                    //  set up to draw dotted lines
   gdk_gc_set_background(gdkgc,&white);
   gdk_gc_set_dashes(gdkgc,0,(gint8 *) dashes,2);
   gdk_gc_set_line_attributes(gdkgc,1,dottedline);

   arrowcursor = gdk_cursor_new(GDK_TOP_LEFT_ARROW);                       //  cursor for selection
   dragcursor = gdk_cursor_new(GDK_CROSSHAIR);                             //  cursor for dragging

   gtk_init_add((GtkFunction) gtkinitfunc,0);                              //  set initz. call from gtk_main()

   gtk_main();                                                             //  process window events
   return 0;
}


/**************************************************************************/

void sighandler(int signal)                                                //  v.6.4.1
{
   zappcrash("segment fault, file %s",image_file);
   return;
}


/**************************************************************************/

//  initial function called from gtk_main() at startup

int gtkinitfunc(void * data)
{
   int               err, flag;
   double            fage;
   char              pid[12], *pp, printoxx[200], command[200];
   const char        *ppc;
   struct stat       statb;
   struct sigaction  sigact;


   Bsavetoedit = ZTX("unknown file type, save as tiff or jpeg to edit");
   Bexiftoolmissing = ZTX("package exiftool is missing");                  //  v.6.9.1
   BOK = ZTX("OK");
   Bcancel = ZTX("Cancel");
   Bdone = ZTX("Done");
   Bclear = ZTX("Clear");
   Bapply = ZTX("Apply");
   Bundo = ZTX("Undo");
   Bredo = ZTX("Redo");
   Bsearch = ZTX("Search");
   Bstart = ZTX("Start");
   Bfinish = ZTX("Finish");
   Bshow = ZTX("Show");
   Bdelete = ZTX("Delete");
   Binvert = ZTX("Invert");
   Bblendwidth = ZTX("Blend Width");
   Bwidth = ZTX("width");
   Bheight = ZTX("height");
   Bpercent = ZTX("percent");
   Bpreset = ZTX("presets");
   Bproceed = ZTX("Proceed");
   Bred = ZTX("red");
   Bgreen = ZTX("green");
   Bblue = ZTX("blue");
   Bbrightness = ZTX("brightness");
   Bdarker = ZTX("darker areas");
   Blighter = ZTX("lighter areas");

   tid_fmain = pthread_self();                                             //  record main() thread ID

   FreeImage_Initialise(FALSE);                                            //  FREEIMAGE setup
   FreeImage_SetOutputMessage(FI_error);
   printf("FreeImage %s \n",FreeImage_GetVersion());

   sigact.sa_handler = sighandler;                                         //  catch segment faults  v.6.4.1
   sigemptyset(&sigact.sa_mask);
   sigact.sa_flags = 0;
   sigaction(SIGSEGV,&sigact,0);

   err = system("exiftool -ver");                                          //  check for exiftool  v.6.9.1
   if (! err) Fexiftool = 1;
   err = system("xdg-open --version");                                     //  check for xdg-open
   if (! err) Fxdgopen = 1;
   err = system("ufraw --version");                                        //  check for xdg-open
   if (! err) Fufraw = 1;

   err = system("printoxx -v");                                            //  check for printoxx
   if (err) {
      pp = getenv("_");                                                    //  look in same place as me
      if (pp) {                                                            //  v.6.2
         strncpy0(printoxx,pp,188);
         pp = strrchr(printoxx,'/');
         if (pp) {
            strcpy(pp,"/printoxx -v");
            err = system(printoxx);
         }
      }
   }
   if (! err) Fprintoxx = 1;
   
   if (! Fexiftool) zmessageACK(ZTX("exiftool is not installed \n"         //  warn user
                               "edited images will lose EXIF data"));

   asstagsfile = zmalloc(200);                                             //  setup assigned tags file
   strncatv(asstagsfile,199,get_zuserdir(),"/assigned_tags",null);         //    home/user/.fotoxx/assigned_tags

   undo_files = zmalloc(200);
   *undo_files = 0;
   strncatv(undo_files,199,get_zuserdir(),"/*_undo_*",null);               //  home/user/.fotoxx/*_undo_*
   flag = 1;
   while ((ppc = SearchWild(undo_files,flag)))                             //  search for undo stack files
   {
      err = stat(ppc,&statb);
      fage = (time(null) - statb.st_mtime) / 3600.0;                       //  file age in hours
      if (fage > 24) {
         printf("old undo file deleted: %s \n",ppc);
         snprintf(command,199,"rm -f %s",ppc);                             //  delete orphaned files   v.6.6
         err = system(command);
      }
   }

   *undo_files = 0;                                                        //  setup undo stack files
   snprintf(pid,11,"/%06d",getpid());
   strncatv(undo_files,199,get_zuserdir(),pid,"_undo_nn",null);            //  home/user/.fotoxx/pppppp_undo_nn

   mutex_init(&bitmaps_lock,0);                                            //  setup lock for edit bitmaps
   
   g_timeout_add(20,gtimefunc,0);                                          //  start periodic function (20 ms)

   if (image_file) {
      char * pp = canonicalize_file_name(image_file);                      //  add cwd if needed
      zfree(image_file);
      image_file = null;
      if (pp) image_file = strdupz(pp);                                    //  change pp to zmalloc
      if (pp) free(pp);
   }
      
   if (image_file) {
      err = stat(image_file,&statb);
      if (! err && S_ISREG(statb.st_mode)) f_open(image_file);             //  open initial file
   }

   return 0;
}


/**************************************************************************/

//  Periodic function - runs every few milliseconds.
//  Avoid any thread usage of gtk/gdk functions.

int gtimefunc(void *arg)
{
   if (Wrepaint) {                                                         //  update drawing window
      Wrepaint = 0;
      mwpaint();
   }

   if (SBupdate)                                                           //  status bar update 
   {
      SBupdate = 0;
      update_statusbar();
   }

   return 1;
}


/**************************************************************************/

//  main window delete_event and destroy signals

int delete_event()                                                         //  main window closed
{
   if (mod_keep()) return 1;                                               //  allow user bailout
   Fshutdown++;                                                            //  shutdown in progress
   save_fotoxx_state();                                                    //  save state for next session
   free_resources();                                                       //  delete undo files
   return 0;
}

void destroy_event()                                                       //  main window destroyed
{
   Fshutdown++;
   printf("main window destroyed \n");
   exit(1);                                                                //  instead of gtk_main_quit();
   return;
}


/**************************************************************************/

//  cause (modified) output image to get repainted immediately
//  this function may be called from threads

void mwpaint2()
{
   Wrepaint++;
   return;
}


/**************************************************************************/

//  paint window when created, exposed, resized, or modified (edited)      //  overhauled  v.6.2

int mwpaint()
{
   GtkAdjustment  *adjustment;
   GdkRectangle   wrect;
   static int     Iww, Ihh, Iwwp, Ihhp;                                    //  image size, previous size
   static int     Lww, Lhh, Dww, Dhh;                                      //  layout size, window size
   static double  centerx, centery, scrollx, scrolly;                      //  image center, scrollbars
   static double  wscale, hscale;
   int            resize = 0, deltax, deltay;
   
   if (Fshutdown) return 0;                                                //  shutdown underway
   if (! Frgb24) {
      gdk_window_clear(WLAYOUT);
      return 0;                                                            //  no image
   }

   gdk_window_freeze_updates(WLAYOUT);                                     //  stop window updates

   if (mutex_trylock(&bitmaps_lock) != 0) {                                //  lock bitmaps
      gdk_window_thaw_updates(WLAYOUT);
      Wrepaint++;                                                          //  cannot, return later  v.6.3
      return 0;
   }
   
   if (LMclick && ! Mcapture) {                                            //  new center from mouse click
      centerx = Mxclick;
      centery = Myclick;
      LMclick = 0;
   }
   else {
      adjustment = gtk_scrolled_window_get_hadjustment(SCROLLWIN);         //  get image center in window
      scrollx = gtk_adjustment_get_value(adjustment);                      //    before any updates
      centerx = (scrollx + 0.5 * Dww - Morgx) / Mscale;
      adjustment = gtk_scrolled_window_get_vadjustment(SCROLLWIN);
      scrolly = gtk_adjustment_get_value(adjustment);
      centery = (scrolly + 0.5 * Dhh - Morgy) / Mscale;
   }
   
   Dww = drWin->allocation.width;                                          //  (new) drawing window size
   Dhh = drWin->allocation.height;
   
   if (E3rgb48) {                                                          //  edit in progress
      Iww = E3ww;
      Ihh = E3hh;
   }
   else {                                                                  //  no edit
      Iww = Fww;
      Ihh = Fhh;
   }

   if (Fzoom == 0) {                                                       //  scale to fit window
      wscale = 1.0 * Dww / Iww;
      hscale = 1.0 * Dhh / Ihh;
      if (wscale < hscale) Mscale = wscale;                                //  use greatest ww/hh ratio
      else  Mscale = hscale;
      if (Iww < Dww && Ihh < Dhh && ! Fblowup) Mscale = 1.0;               //  small image 1x unless Fblowup
   }
   else Mscale = Fzoom;                                                    //  scale to Fzoom level

   Iww = int(Mscale * Iww);                                                //  (new) image display size
   Ihh = int(Mscale * Ihh);

   if (Iww != Iwwp || Ihh != Ihhp)                                         //  image size changed
   {
      Iwwp = Iww;
      Ihhp = Ihh;
      resize = 1;
   }

   if (E3rgb48) {                                                          //  edit in progress
      RGB_free(Mrgb24);
      RGB * temp = RGB_convbpp(E3rgb48);                                   //  convert E3 to 24 bpp
      Mrgb24 = RGB_rescale(temp,Iww,Ihh);                                  //  rescale to image size
      RGB_free(temp);
   }
   else {                                                                  //  no edit
      if (! Mrgb24 || resize) {
         RGB_free(Mrgb24);
         Mrgb24 = RGB_rescale(Frgb24,Iww,Ihh);                             //  rescale Frgb24 to image size
      }
   }

   Mww = Iww;                                                              //  new image size
   Mhh = Ihh;

   if (resize) {
      if (Iww > Dww) Lww = Iww;                                            //  layout = larger of window
      else  Lww = Dww;                                                     //     and image sizes
      if (Ihh > Dhh) Lhh = Ihh;
      else  Lhh = Dhh;
      gtk_layout_set_size(GTK_LAYOUT(drWin),Lww,Lhh);                      //  set new layout size

      wrect.x = wrect.y = 0;                                               //  prevent flicker        v.6.3
      wrect.width = Lww;
      wrect.height = Lhh;
      gdk_window_begin_paint_rect(WLAYOUT,&wrect);
      gdk_window_clear(WLAYOUT);
   }

   if ((Mxdrag || Mydrag) && ! Mcapture) {                                 //  scroll via mouse drag   v.6.2
      deltax = Mxdrag - Mxdown;
      deltay = Mydrag - Mydown;
      wscale = 1.05 * (1.0 - 1.0 * Dww / Iww);                             //  v.6.4
      hscale = 1.05 * (1.0 - 1.0 * Dhh / Ihh);
      centerx += wscale * deltax;
      centery += hscale * deltay;
      Mxdown = Mxdrag;
      Mydown = Mydrag;
      Mxdrag = Mydrag = 0;
   }

   if (Iww > Dww) {
      Dww = drWin->allocation.width;
      scrollx = (centerx * Mscale - 0.5 * Dww);                            //  set scroll to preserve center
      if (scrollx > Iww - Dww) scrollx = Iww - Dww;
      adjustment = gtk_scrolled_window_get_hadjustment(SCROLLWIN);
      gtk_adjustment_set_value(adjustment,scrollx);
      Morgx = 0;
   }
   else { 
      Morgx = int(0.5 * (Dww - Iww));                                      //  or center image in window
      scrollx = 0;
      centerx = 0.5 * Fww;
   }

   if (Ihh > Dhh) {
      Dhh = drWin->allocation.height;
      scrolly = (centery * Mscale - 0.5 * Dhh);
      if (scrolly > Ihh - Dhh) scrolly = Ihh - Dhh;
      adjustment = gtk_scrolled_window_get_vadjustment(SCROLLWIN);
      gtk_adjustment_set_value(adjustment,scrolly);
      Morgy = 0;
   }
   else {
      Morgy = int(0.5 * (Dhh - Ihh));
      scrolly = 0;
      centery = 0.5 * Fhh;
   }

   gdk_draw_rgb_image(WLAYOUT, gdkgc, Morgx, Morgy, Iww, Ihh,              //  draw image to window
                      nodither, (uint8 *) Mrgb24->bmp, Iww*3);
   paint_dotlines();                                                       //  draw dotted line overlays
   
   if (resize) gdk_window_end_paint(WLAYOUT);                              //  v.6.3
   mutex_unlock(&bitmaps_lock);                                            //  unlock bitmaps
   gdk_window_thaw_updates(WLAYOUT);                                       //  thaw window updates

   Wpainted++;                                                             //  notify edit function of repaint
   if (brightgraph) brightgraph_paint();                                   //  update brightness graph
   SBupdate++;                                                             //  update status bar
   return 0;                                                               //  v.6.2
}


/**************************************************************************/

//  refresh dotted lines on top of image after repainting image

void paint_dotlines()
{
   int      ii, x1, y1, x2, y2;

   for (ii = 0; ii < Ndotlines; ii++)
   {
      x1 = dotlinex1[ii];
      y1 = dotliney1[ii];
      x2 = dotlinex2[ii];
      y2 = dotliney2[ii];

      draw_dotline(x1,y1,x2,y2);
   }

   return;
}


/**************************************************************************/

//  update status bar with image data and status

void update_statusbar()
{
   static char    text1[200], text2[100];
   int            bpp, scale;

   if (! image_file) return;
   
   *text1 = *text2 = 0;
   
   if (E3rgb48) {                                                          //  use edit bitmap
      snprintf(text1,199,"%dx%dx%d",E3ww,E3hh,48);
      if (Fpreview) strcat(text1," (preview)");
   }
   else {                                                                  //  file bitmap
      bpp = file_bpp;                                                      //  v.6.6
      if (Fmodified || Pundo) bpp = 48;                                    //  v.6.6
      scale = int(Mscale * 100 + 0.5);
      snprintf(text1,199,"%dx%dx%d  %.2fMB  %d%c",
                  Fww,Fhh,bpp,file_MB,scale,'%');
   }
   
   if (Fimageturned) strcat(text1,"  (turned)");                           //  v.6.8
   
   if (Pundo)                                                              //  edit undo stack depth 
   {
      snprintf(text2,99,"  edits: %d",Pundo);
      strcat(text1,text2);
   }

   if (Nalign)                                                             //  image alignment data 
   {
      snprintf(text2,99,"  align: %d  offsets: %+.1f %+.1f %+.4f %+.1f %+.1f  match: %.5f",
                              Nalign,xoffB,yoffB,toffB,yst1B,yst2B,matchB);
      strcat(text1,text2);
   }

   if (Fautolens)                                                          //  lens parameters (search status)
   {
      snprintf(text2,99,"  lens: %.1f %.2f",lens_mm,lens_bow);
      strcat(text1,text2);
   }

   if (PercentDone > 0 && PercentDone < 100)                               //  % completion       v.6.3
   {
      snprintf(text2,99,"  done: %d%c",PercentDone,'%');
      strcat(text1,text2);
   }

   stbar_message(STbar,text1);
   return;
}


/**************************************************************************/

//  mouse event function - capture buttons and drag movements

void mouse_event(GtkWidget *, GdkEventButton *event, void *)
{
   int mouse_convert(int &xpos, int &ypos);

   static int     bdtime = 0, butime = 0, fdrag = 0, mbusy = 0;
   int            mousex, mousey, button, time, type;

   type = event->type;
   button = event->button;                                                 //  button, 1/3 = left/right
   time = event->time;
   mousex = int(event->x);                                                 //  mouse position
   mousey = int(event->y);
   
   if (type == GDK_MOTION_NOTIFY) {
      if (mbusy) return;                                                   //  discard excess motion events
      mbusy++;
      zmainloop();
      mbusy = 0;
   }

   if (! mouse_convert(mousex,mousey)) return;                             //  convert to image space

   if (type == GDK_BUTTON_PRESS) {                                         //  button down
      bdtime = time;                                                       //  time of button down
      Mxdown = mousex;                                                     //  position at button down time
      Mydown = mousey;
      if (button) {
         fdrag++;                                                          //  possible drag start
         Mbutton = button;
      }
      Mxdrag = Mydrag = 0;
   }

   if (type == GDK_BUTTON_RELEASE) {                                       //  button up
      Mxclick = Myclick  = 0;                                              //  reset click status
      butime = time;                                                       //  time of button up
      if (butime - bdtime < 400) {                                         //  less than 0.4 secs down
         if (Mbutton == 1) LMclick++;                                      //  left mouse click
         if (Mbutton == 3) RMclick++;                                      //  right mouse click
         Mxclick = Mxdown;                                                 //  click = button down position
         Myclick = Mydown;
      }
      Mxdown = Mydown = Mxdrag = Mydrag = fdrag = Mbutton = 0;             //  forget buttons, drag
   }
   
   if (type == GDK_MOTION_NOTIFY && fdrag) {                               //  drag underway
      Mxdrag = mousex;
      Mydrag = mousey;
   }

   if (mouseCBfunc) {                                                      //  pass to handler function
      (* mouseCBfunc)();
      return;
   }

   if (LMclick && ! Mcapture) m_zoom(null, (char *) "+");                  //  if not captured by edit function,
   if (RMclick && ! Mcapture) m_zoom(null, (char *) "-");                  //    handle as zoom request
   if (RMclick && ! Mcapture) RMclick = 0;                                 //  forget right click     v.6.1.2

   if ((Mxdrag || Mydrag) && ! Mcapture) mwpaint();                        //  drag = scroll          v.6.2
   return;
}


//  convert mouse position from window space to image space

int mouse_convert(int &xpos, int &ypos)
{
   xpos = int((xpos - Morgx) / Mscale + 0.5);
   ypos = int((ypos - Morgy) / Mscale + 0.5);
   if (xpos < 0) return 0;
   if (ypos < 0) return 0;

   if (E3rgb48) {                                                          //  bugfix  v.6.4.1
      if (xpos >= E3ww) return 0;
      if (ypos >= E3hh) return 0;
   }
   else {
      if (xpos >= Fww) return 0;
      if (ypos >= Fhh) return 0;
   }
   
   return 1;
}


/**************************************************************************/

//  keyboard event function - some toolbar buttons have KB equivalents
//  GDK key symbols: /usr/include/gtk-2.0/gdk/gdkkeysyms.h

int KBpress(GtkWidget *win, GdkEventKey *event, void *)                    //  prevent propagation of key-press
{                                                                          //    events to toolbar buttons
   return 1;
}

int KBrelease(GtkWidget *win, GdkEventKey *event, void *)
{
   KBkey = event->keyval;

   if (KBkeyCBfunc) {                                                      //  pass to handler function
      (* KBkeyCBfunc)();                                                   //  v.6.5
      return 1;
   }

   if (KBcapture) return 1;                                                //  let function handle it

   if (KBkey == GDK_T) m_index(null,null);                                 //  key T  >>  thumbnail index
   if (KBkey == GDK_t) m_index(null,null);                                          
   if (KBkey == GDK_X) m_index(null,null);                                 //  key X  >>  thumbnail index
   if (KBkey == GDK_x) m_index(null,null);                                          

   if (KBkey == GDK_Left) m_prev(null,null);                               //  arrow keys  >>  prev/next image
   if (KBkey == GDK_Right) m_next(null,null);

   if (KBkey == GDK_plus) m_zoom(null, (char *) "+");                      //  +/- keys  >>  zoom in/out
   if (KBkey == GDK_equal) m_zoom(null, (char *) "+");                     //  = key: same as +
   if (KBkey == GDK_minus) m_zoom(null, (char *) "-");
   if (KBkey == GDK_Z) m_zoom(null, (char *) "Z");                         //  Z key: zoom to 100%
   if (KBkey == GDK_z) m_zoom(null, (char *) "Z");

   if (KBkey == GDK_Escape) m_slideshow(null,null);                        //  escape  >>  exit slideshow
   if (KBkey == GDK_Delete) m_trash(null,null);                            //  delete  >>  trash

   if (KBkey == GDK_R) upright_image(+90);                                 //  keys L, R  >>  rotate
   if (KBkey == GDK_r) upright_image(+90);
   if (KBkey == GDK_L) upright_image(-90);
   if (KBkey == GDK_l) upright_image(-90);

   return 1;
}


/**************************************************************************/

//  process top-level menu entry

void menufunc(GtkWidget *, const char *menu)
{
   topmenu = (char *) menu;                                                //  remember top-level menu in
   return;                                                                 //    case this is needed somewhere
}


/**************************************************************************/

//  draw a dotted line. coordinates are in image space.

void draw_dotline(int x1, int y1, int x2, int y2)
{
   x1 = int(Mscale * x1 + Morgx + 0.5);                                    //  image to window space
   y1 = int(Mscale * y1 + Morgy + 0.5);
   x2 = int(Mscale * x2 + Morgx + 0.5);
   y2 = int(Mscale * y2 + Morgy + 0.5);

   gdk_draw_line(WLAYOUT,gdkgc,x1,y1,x2,y2);
   return;
}


//  erase a dotted line. refresh line path from win_image pixels.

void erase_dotline(int x1, int y1, int x2, int y2)
{
   double   slope;
   int      pxm, pym, pxn, pyn;

   GOFUNC_DATA(1)
   
   x1 = int(Mscale * x1 + 0.5);                                            //  image to window space
   y1 = int(Mscale * y1 + 0.5);
   x2 = int(Mscale * x2 + 0.5);
   y2 = int(Mscale * y2 + 0.5);
   
   if (abs(y2 - y1) > abs(x2 - x1)) {
      slope = 1.0 * (x2 - x1) / (y2 - y1);
      if (y2 > y1) {
         for (pym = y1; pym <= y2; pym++) {
            pxm = int(x1 + slope * (pym - y1) + 0.5);
            GOFUNC(draw_pixel);
         }
      }
      else {
         for (pym = y1; pym >= y2; pym--) {
            pxm = int(x1 + slope * (pym - y1) + 0.5);
            GOFUNC(draw_pixel);
         }
      }
   }
   else {
      slope = 1.0 * (y2 - y1) / (x2 - x1);
      if (x2 > x1) {
         for (pxm = x1; pxm <= x2; pxm++) {
            pym = int(y1 + slope * (pxm - x1) + 0.5);
            GOFUNC(draw_pixel);
         }
      }
      else {
         for (pxm = x1; pxm >= x2; pxm--) {
            pym = int(y1 + slope * (pxm - x1) + 0.5);
            GOFUNC(draw_pixel);
         }
      }
   }

   return;

draw_pixel:
   {
      pxn = pxm - 1;                                                       //  3x3 block around target pixel
      pyn = pym - 1;
      
      if (pxn < 0) pxn = 0;
      if (pxn > Mww-3) pxn = Mww-3;
      if (pyn < 0) pyn = 0;
      if (pyn > Mhh-3) pyn = Mhh-3;
      
      uint8 *pixels = (uint8 *) Mrgb24->bmp + (pyn * Mww + pxn) * 3;
      gdk_draw_rgb_image(WLAYOUT, gdkgc, pxn + Morgx, pyn + Morgy,
                                3, 3, nodither, pixels, Mww * 3);         
      RETURN
   }
}


/**************************************************************************
      file functions
***************************************************************************/

//  display a thumbnail index of images in a separate window

void m_index(GtkWidget *, const char *)
{
   if (image_file) image_xthumbs(image_file,"paint1",0,m_index2);          //  show index window
   else {
      char *pp = get_current_dir_name();                                   //  initz. index and show
      if (pp) {
         image_xthumbs(pp,"init",0,m_index2);
         image_xthumbs(0,"paint1");
         free(pp);
         Fsearchlist = 0;
      }
   }

   return;
}


//  clicked thumbnails will call this function

void  m_index2(char *file)
{
   f_open((const char *) file);
   return;
}


/**************************************************************************/

//  open menu function

void m_open(GtkWidget *, const char *)
{
   Fsearchlist = 0;
   f_open(null);
   return;
}


/**************************************************************************/

//  open a file and initialize FI bitmap

void f_open(const char *filespec)
{
   int            cc, fposn, fcount;
   char           *pp, wtitle[200], fname[100], fdirk[100];
   RGB            *temp24;
   char           **orientation;
   const char     *orientationkey[1] = { exif_orientation_key };
   
   zmondirk("close",null,null);                                            //  reset monitored directory

   if (! menulock(1)) return;                                              //  lock menu
   if (mod_keep()) goto openfail;
   
   if (filespec) filespec = strdupz(filespec);
   else filespec = zgetfile(ZTX("open image file"),image_file,"open");
   if (! filespec) goto openfail;
   
   temp24 = image_load(filespec,24);                                       //  load image as RGB-24 bitmap
   if (! temp24) goto openfail;

   free_resources();                                                       //  free resources for old image file

   mutex_lock(&bitmaps_lock);                                              //  lock bitmaps

   image_file = (char *) filespec;                                         //  setup new image file
   Frgb24 = temp24;
   Fww = Frgb24->ww;
   Fhh = Frgb24->hh;
   
   /// if (file_bpp == 48) Frgb48 = image_load(image_file,48);             //  file is 48 bits/pixel     v.6.6

   pp = strrchr(image_file,'/');                                           //  get image file name
   strncpy0(fname,pp+1,99);
   cc = pp - image_file;
   if (cc < 99) strncpy0(fdirk,image_file,cc+2);                           //  get dirk/path/ if short enough
   else {
      strncpy(fdirk,image_file,96);                                        //  or use /dirk/path...
      strcpy(fdirk+95,"...");
   }
   image_position(image_file,fposn,fcount);                                //  position and count in image index

   Fimageturned = 0;
   orientation = exif_get(image_file,orientationkey,1);                    //  upright turned image
   if (*orientation) {
      if (strstr(*orientation,"90")) {
         upright_image(90);
         Fimageturned = 90;                                                //  v.6.8
      }
      if (strstr(*orientation,"270")) {
         upright_image(270);
         Fimageturned = 270;
      }
   }

   Fzoom = 0;                                                              //  zoom level = fit window

   mutex_unlock(&bitmaps_lock);                                            //  unlock bitmaps

   snprintf(wtitle,199,"%s  %d/%d  %s",fname,fposn,fcount,fdirk);          //  set window title
   gtk_window_set_title(GTK_WINDOW(mWin),wtitle);
   gtk_window_present(GTK_WINDOW(mWin));                                   //  bring main window to front
   if (zdtags) edit_tags_dialog();                                         //  update active tags dialog
   if (zdrename) rename_dialog();                                          //  update active rename dialog

   mwpaint2();                                                             //  refresh main window

   pp = image_xthumbs(image_file,"find",0);                                //  file in current image index?
   if (! pp) {
      image_xthumbs(image_file,"init");                                    //  no, reset index file list
      image_xthumbs(0,"paint2");                                           //  refresh index window if active
      Fsearchlist = 0;
   }
   else zfree(pp);

   if (! Fsearchlist) {                                                    //  start monitoring this directory
      pp = strrchr(image_file,'/');                                        //    if not from search tags
      *pp = 0;                                                             //                v.6.4
      zmondirk("open",image_file,null);
      *pp = '/';
   }

openfail:
   menulock(0);
   mwpaint2();
   return;
}


/**************************************************************************/

//  open a raw image file and convert to tiff-48 

void m_raw(GtkWidget *, const char *)
{
   char           *pp, *rawfile, *outfile;
   char           command[1000];
   int            yn, err;
   struct stat    fstat;

   if (! menulock(1)) return;                                              //  lock menu
   if (mod_keep()) goto openfail;
   
   if (! Fufraw) {
      zmessageACK(ZTX("package ufraw required for this function"));
      goto openfail;
   }
   
   rawfile = zgetfile(ZTX("open RAW file"),image_file,"open");
   if (! rawfile) goto openfail;
   
   err = stat(rawfile,&fstat);
   if (err) {
      zmessageACK(strerror(errno));
      zfree(rawfile);
      goto openfail;
   }

   if (S_ISDIR(fstat.st_mode)) {
      pp = zgetfile(ZTX("open RAW file"),rawfile,"open");
      zfree(rawfile);
      if (! pp) goto openfail;
      rawfile = pp;
   }

   outfile = strdupz(rawfile,5);
   pp = strrchr(rawfile,'.');
   pp = outfile + (pp - rawfile);
   strcpy(pp,".tiff");

   yn = zmessageYN(ZTX("Convert raw file to 48-bit tiff format? \n"
                       " (this may take a while) "));
   if (! yn) {
      zfree(outfile);
      goto openfail;
   }
   
   snprintf(command,999,"ufraw-batch --out-type=tiff --out-depth=16"       //  new ufraw   v.6.5
               " --overwrite --output=\"%s\" \"%s\" ",outfile,rawfile);
   err = system(command);

   if (err) {
      snprintf(command,999,"ufraw-batch --out-type=tiff16"                 //  old ufraw
               " --overwrite --output=\"%s\" \"%s\" ",outfile,rawfile);
      err = system(command);
   }

   if (err) {
      zmessageACK(wstrerror(err));
      zfree(outfile);
      goto openfail;
   }

   menulock(0);                                                            //  bugfix  v.6.3
   image_xthumbs(outfile,"init");                                          //  update image index list
   image_xthumbs(0,"paint2");                                              //  refresh index window if active
   Fsearchlist = 0;
   f_open(outfile);                                                        //  open converted file
   return;

openfail:
   menulock(0);
   mwpaint2();
   return;
}


/**************************************************************************/

//  open previous or next file in same index as last file opened

void m_prev(GtkWidget *, const char *)
{
   if (! image_file) return;

   int mods = 0;
   while (zmondirk("event",null,null) > 0) mods++;                         //  detect directory mods  v.6.4
   if (mods) image_xthumbs(image_file,"init");                             //  refresh file list

   char *pp = image_xthumbs(image_file,"prev");
   if (! pp) return;
   if (image_file_type(pp) == 2) f_open(pp);
   zfree(pp);
   return;
}

void m_next(GtkWidget *, const char *)
{
   if (! image_file) return;

   int mods = 0;
   while (zmondirk("event",null,null) > 0) mods++;                         //  v.6.4
   if (mods) image_xthumbs(image_file,"init");

   char *pp = image_xthumbs(image_file,"next");
   if (! pp) return;
   if (image_file_type(pp) == 2) f_open(pp);
   zfree(pp);
   return;
}


/**************************************************************************/

//  save (modified) image to same or new file

void m_save(GtkWidget *, const char *)
{   
   GtkWidget      *fdialog, *fchooser, *hbox;
   GtkWidget      *label1, *tiff2, *tiff4, *jpeg, *jqlab, *jqval;
   FIB            *fib;
   PXB            *pxb;
   RGB            *rgb24, *rgb48;
   const char     *exifkey[3] = { exif_width_key, exif_height_key, exif_orientation_key };
   const char     *exifdata[3];
   char           wwchar[8], hhchar[8];
   struct stat    fstat;
   char           *outfile, *pp;
   const char     *format;
   int            ii, err, yn, status;
   static char    jquality[8] = "90";                                      //  default jpeg save quality

   if (! image_file) return;

   fdialog = gtk_dialog_new_with_buttons(ZTX("save file"),                 //  build file save dialog
                           GTK_WINDOW(mWin), GTK_DIALOG_MODAL,
                           GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, 
                           GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT, null);

   gtk_window_set_default_size(GTK_WINDOW(fdialog),500,400);

   fchooser = gtk_file_chooser_widget_new(GTK_FILE_CHOOSER_ACTION_SAVE);
   gtk_file_chooser_set_filename(GTK_FILE_CHOOSER(fchooser),image_file);
   gtk_container_add(GTK_CONTAINER(GTK_DIALOG(fdialog)->vbox),fchooser);
   
   hbox = gtk_hbox_new(0,0);
   gtk_container_add(GTK_CONTAINER(GTK_DIALOG(fdialog)->vbox),hbox);
   gtk_box_set_child_packing(GTK_BOX(GTK_DIALOG(fdialog)->vbox),hbox,0,0,10,GTK_PACK_END);

   label1 = gtk_label_new("file type");                                    //  add file type options
   tiff2 = gtk_radio_button_new_with_label(null,"tiff-24");
   tiff4 = gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(tiff2),"tiff-48");
   jpeg = gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(tiff2),"jpeg");
   jqlab = gtk_label_new("jpeg quality");
   jqval = gtk_entry_new();
   gtk_entry_set_width_chars(GTK_ENTRY(jqval),3);
   gtk_box_pack_start(GTK_BOX(hbox),label1,0,0,5);
   gtk_box_pack_start(GTK_BOX(hbox),tiff2,0,0,5);                          //  simplified    v.6.7.1
   gtk_box_pack_start(GTK_BOX(hbox),tiff4,0,0,5);
   gtk_box_pack_start(GTK_BOX(hbox),jpeg,0,0,10);                          //  jpeg quality   v.6.7.2
   gtk_box_pack_start(GTK_BOX(hbox),jqlab,0,0,5);
   gtk_box_pack_start(GTK_BOX(hbox),jqval,0,0,5);
   
   gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(jpeg),1);                //  set default output file type
   gtk_entry_set_text(GTK_ENTRY(jqval),jquality);

   if (strEqu(file_type,"tiff"))
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tiff2),1);
   if (file_bpp == 48)
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tiff4),1);

dialog_run:

   gtk_widget_show_all(fdialog);                                           //  run dialog
   status = gtk_dialog_run(GTK_DIALOG(fdialog));
   if (status != GTK_RESPONSE_ACCEPT) {                                    //  user cancelled
      gtk_widget_destroy(fdialog);
      return;
   }

   outfile = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(fchooser));    //  get user inputs
   if (! outfile) goto dialog_run;

   format = "jpeg";
   if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tiff2)))
      format = "tiff-24";
   if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tiff4)))
      format = "tiff-48";

   if (strEqu(format,"jpeg")) {                                            //  get jpeg quality   v.6.7.2
      ii = atoi(gtk_entry_get_text(GTK_ENTRY(jqval)));
      if (ii < 1 || ii > 100) {
         zmessageACK("jpeg quality must be 1-100");
         goto dialog_run;
      }
      sprintf(jquality,"%d",ii);
   }

   gtk_widget_destroy(fdialog);                                            //  kill dialog

   pp = strdupz(outfile,8);                                                //  existing file extension
   g_free(outfile);
   outfile = pp;
   pp = strrchr(outfile,'/');
   if (pp) pp = strrchr(pp,'.');
   if (! pp) {                                                             //  keep if matches save format
      strcat(outfile,".");                                                 //  else use .jpeg or .tiff
      pp = strrchr(outfile,'.');                                           //  v.6.7.1
   }
   if (strEqu(format,"jpeg") && ! strcmpv(pp,".jpg",".JPG",".jpeg",".JPEG",0))
      strcpy(pp,".jpeg");
   if (strnEqu(format,"tiff",4) && ! strcmpv(pp,".tif",".TIF",".tiff",".TIFF",0))
      strcpy(pp,".tiff");

   err = stat(outfile,&fstat);                                             //  check if file exists
   if (! err) {
      yn = zmessageYN(ZTX("Overwrite file? \n %s"),outfile);               //  confirm overwrite
      if (! yn) {
         zfree(outfile);
         return;
      }
   }

   status = 0;
   fib = 0;
   pxb = 0;

   if (strEqu(format,"jpeg")) {                                            //  save as JPEG file
      if (Frgb48) {
         rgb24 = RGB_convbpp(Frgb48);
         pxb = RGB_PXB(rgb24);
         RGB_free(rgb24);
      }                                                                    //  use GDK pixbuf for jpeg   v.6.7.2
      else  pxb = RGB_PXB(Frgb24);
      status = gdk_pixbuf_save(pxb,outfile,"jpeg",gerror,"quality",jquality,null);
   }
   
   if (strEqu(format,"tiff-24")) {                                         //  save as TIFF-24 file
      if (Frgb48) {
         rgb24 = RGB_convbpp(Frgb48);
         fib = RGB_FIB(rgb24);
         RGB_free(rgb24);
      }
      else  fib = RGB_FIB(Frgb24);
      status = FreeImage_Save(FIF_TIFF,fib,outfile,0);
   }
   
   if (strEqu(format,"tiff-48")) {                                         //  save as TIFF-48 file
      if (Frgb48) fib = RGB_FIB(Frgb48);                                   //  bugfix   v.6.6
      else {
         if (file_bpp == 48) rgb48 = image_load(image_file,48);            //  use original 48 bpp file
         else  rgb48 = RGB_convbpp(Frgb24);
         if (rgb48) {
            fib = RGB_FIB(rgb48);
            RGB_free(rgb48);
         }
      }
      status = FreeImage_Save(FIF_TIFF,fib,outfile,0);
   }

   if (fib) FreeImage_Unload(fib);
   if (pxb) g_object_unref(pxb);

   if (! status) {
      zmessageACK(ZTX("unable to save image"));
      zfree(outfile);
      return;
   }
   
   snprintf(wwchar,6,"%d",Fww);
   snprintf(hhchar,6,"%d",Fhh);                                            //  copy EXIF data from source file
   exifdata[0] = wwchar;                                                   //    with changes for width, height,
   exifdata[1] = hhchar;                                                   //      and orientation
   exifdata[2] = "";                                                       //  assume saved upright     v.6.2
   err = exif_copy(image_file,outfile,exifkey,exifdata,3);
   if (err) zmessageACK(ZTX("unable to copy EXIF data"));
   else {
      load_filetags(outfile);                                              //  update assigned tags file
      update_asstags(outfile);
   }
   
   Fmodified = Pundo = 0;                                                  //  reset file modified status
   if (! Fsearchlist) image_xthumbs(outfile,"init");                       //  update image index list
   image_xthumbs(0,"paint2");                                              //  refresh index window if active
   gtk_window_present(GTK_WINDOW(mWin));                                   //  bring main to foreground

   if (strEqu(outfile,image_file)) {                                       //  if same file was replaced,
      stat(outfile,&fstat);                                                //  update file size
      file_MB = 1.0 * fstat.st_size / mega;
      SBupdate++;                                                          //  update status bar
   }

   zfree(outfile);
   return;
}


/**************************************************************************/

//  print image files

void m_print(GtkWidget *, const char *)
{
   int      err;
   char     command[1000];
   
   if (! image_file) return;

   if (! Fprintoxx) {
      zmessageACK(ZTX("printoxx program not found (see user guide)"));
      return;
   }

   snprintf(command,999,"printoxx \"%s\" &",image_file);                   //  send curr. file to printoxx
   err = system(command);
   if (err) zmessLogACK(wstrerror(err));
   return;
}


/**************************************************************************/

//  Delete image file - move image_file to trash.
//  Trash has no standard location, so use a trash folder on the desktop.
//  User must delete or move to the official trash bin.

void m_trash(GtkWidget *, const char *)
{
   int            err, yn;
   char           command[1000], trashdir[100];
   struct stat    trstat;

   if (! image_file) return;                                               //  nothing to trash
   
   err = stat(image_file,&trstat);                                         //  get file status
   if (err) {
      zmessLogACK(strerror(errno));
      return;
   }

   if (! (trstat.st_mode & S_IWUSR)) {                                     //  check permission
      yn = zmessageYN(ZTX("move read-only file to trash?"));
      if (! yn) return;
      trstat.st_mode |= S_IWUSR;
      chmod(image_file,trstat.st_mode);
   }
   
   snprintf(trashdir,99,"%s/%s",getenv("HOME"),ftrash);                    //  get full fotoxx trash file
   
   trstat.st_mode = 0;
   err = stat(trashdir,&trstat);
   if (! S_ISDIR(trstat.st_mode)) {
      snprintf(command,999,"mkdir -m 0750 \"%s\"",trashdir);
      err = system(command);
      if (err) {
         zmessLogACK(ZTX("cannot create trash folder: %s"),wstrerror(err));
         return;
      }
   }

   snprintf(command,999,"cp \"%s\" \"%s\" ",image_file,trashdir);          //  copy image file to trash
   err = system(command);
   if (err) {
      zmessLogACK(ZTX("error: %s"),wstrerror(err));
      return;
   }

   snprintf(command,999,"rm \"%s\"",image_file);                           //  delete image file
   err = system(command);
   if (err) {
      zmessLogACK(ZTX("error: %s"),wstrerror(err));
      return;
   }
   
   update_asstags(image_file,1);                                           //  delete in assigned tags file
   if (! Fsearchlist) image_xthumbs(image_file,"init");                    //  reset image file list
   image_xthumbs(0,"paint2");                                              //  refresh index window if active
   m_next(null,null);                                                      //  step to next file if there

   return;
}


/**************************************************************************/

//  rename menu function

char     rename_old[100] = "";
char     rename_new[100] = "";

void  m_rename(GtkWidget *, const char *)
{
   rename_dialog();                                                        //  activate rename dialog
   return;
}
   

//  activate rename dialog, stuff data from current file

void rename_dialog()
{
   int rename_dialog_event(zdialog *zd, const char *event);
   int rename_dialog_compl(zdialog *zd, int zstat);
   
   char     *pdir, *pfile, *pext;

   if (! image_file) return;

   if (! zdrename)                                                         //  restart dialog
   {
      zdrename = zdialog_new(ZTX("Rename Image File"),mWin,Bcancel,0);
      zdialog_add_widget(zdrename,"hbox","hb1","dialog",0,"space=10");
      zdialog_add_widget(zdrename,"vbox","vb1","hb1",0,"homog|space=5");
      zdialog_add_widget(zdrename,"vbox","vb2","hb1",0,"homog|expand");

      zdialog_add_widget(zdrename,"button","Bold","vb1",ZTX("old name"));
      zdialog_add_widget(zdrename,"button","Bnew","vb1",ZTX("rename to"));
      zdialog_add_widget(zdrename,"button","Bprev","vb1",ZTX("previous"));

      zdialog_add_widget(zdrename,"hbox","hb21","vb2",0);                  //  [ old name ] [ oldname  ]
      zdialog_add_widget(zdrename,"hbox","hb22","vb2",0);                  //  [ new name ] [ newname  ] [+1]
      zdialog_add_widget(zdrename,"hbox","hb23","vb2",0);                  //  [ previous ] [ prevname ]

      zdialog_add_widget(zdrename,"label","Lold","hb21");
      zdialog_add_widget(zdrename,"entry","Enew","hb22",0,"expand|scc=30");
      zdialog_add_widget(zdrename,"button","B+1","hb22"," +1 ","space=5");
      zdialog_add_widget(zdrename,"label","Lprev","hb23");

      zdialog_run(zdrename,rename_dialog_event,rename_dialog_compl);       //  start dialog
   }

   parsefile(image_file,&pdir,&pfile,&pext);
   strncpy0(rename_old,pfile,99);
   strncpy0(rename_new,pfile,99);
   zdialog_stuff(zdrename,"Lold",rename_old);                              //  current file name
   zdialog_stuff(zdrename,"Enew",rename_new);                              //  entered file name

   return;
}


//  rename dialog event and completion functions

int rename_dialog_event(zdialog *zd, const char *event)
{
   char           *pp, *pdir, *pfile, *pext, command[2000];
   int            nseq, digits, ccp, ccn, ccx, err;
   struct stat    statb;

   if (strEqu(event,"Bold"))                                               //  reset to current file name
      zdialog_stuff(zd,"Enew",rename_old);

   if (strEqu(event,"Bprev")) {                                            //  previous name >> new name
      zdialog_fetch(zd,"Lprev",rename_new,99);
      zdialog_stuff(zd,"Enew",rename_new);
   }

   if (strEqu(event,"B+1"))                                                //  increment sequence number
   {
      zdialog_fetch(zd,"Enew",rename_new,94);                              //  get entered filename
      pp = rename_new + strlen(rename_new);
      digits = 0;
      while (pp[-1] >= '0' && pp[-1] <= '9') {
         pp--;                                                             //  look for NNN in filenameNNN
         digits++;
      }
      nseq = 1 + atoi(pp);                                                 //  NNN + 1
      if (nseq > 9999) nseq = 0;
      if (digits < 2) digits = 2;                                          //  keep digit count if enough
      if (nseq > 99 && digits < 3) digits = 3;                             //  use leading zeros           v.6.2
      if (nseq > 999 && digits < 4) digits = 4;
      snprintf(pp,digits+1,"%0*d",digits,nseq);
      zdialog_stuff(zd,"Enew",rename_new);
   }

   if (strEqu(event,"Bnew")) 
   {
      parsefile(image_file,&pdir,&pfile,&pext);                            //  existing /directories/file.ext

      zdialog_fetch(zd,"Enew",rename_new,94);                              //  new file name from user

      ccp = strlen(pdir);                                                  //  length of /directories/
      ccn = strlen(rename_new);                                            //  length of file
      if (pext) ccx = strlen(pext);                                        //  length of .ext
      else ccx = 0;

      pp = zmalloc(ccp + ccn + ccx + 1);                                   //  put it all together
      strncpy(pp,image_file,ccp);                                          //   /directories/file.ext
      strcpy(pp+ccp,rename_new);
      if (ccx) strcpy(pp+ccp+ccn,pext);
      
      err = stat(pp,&statb);                                               //  check for new name exists
      if (! err) {
         zmessageACK(ZTX("the target file already exists"));
         zfree(pp);
         return 0;
      }
      
      snprintf(command,1999,"cp \"%s\" \"%s\"",image_file,pp);             //  copy to new file
      err = system(command);
      if (err) {
         zmessageACK(ZTX("rename failed \n %s"),wstrerror(err));
         zfree(pp);
         return 0;
      }

      zdialog_stuff(zd,"Lprev",rename_new);                                //  set previous name in dialog

      load_filetags(pp);                                                   //  update assigned tags file
      update_asstags(pp);
      zfree(pp);

      pp = strdupz(image_file);                                            //  save file name to be deleted
      m_next(null,null);                                                   //  move to image_file + 1
      snprintf(command,999,"rm \"%s\"",pp);                                //  delete old file
      err = system(command);
      zfree(pp);

      if (! Fsearchlist) image_xthumbs(image_file,"init");                 //  update image index list
      image_xthumbs(0,"paint2");                                           //  refresh index window if active
      gtk_window_present(GTK_WINDOW(mWin));                                //  bring main to foreground
   }

   return 0;
}


int rename_dialog_compl(zdialog *zd, int zstat)
{
   zdialog_free(zdrename);                                                 //  kill dialog
   zdrename = null;
   return 0;
}


/**************************************************************************/

//  forced quit - can cause running function to crash

void m_quit(GtkWidget *, const char *)
{
   if (image_file) update_filetags(image_file);                            //  commit tag changes, if any
   if (mod_keep()) return;                                                 //  keep or discard pending changes
   printf("quit \n");
   Fshutdown++;
   save_fotoxx_state();                                                    //  save state for next session
   free_resources();                                                       //  delete temp files
   gtk_main_quit();                                                        //  gone forever
   return;
}


/**************************************************************************
      image edit functions
***************************************************************************/

/**************************************************************************

   Select an area within the current image within which subsequent
   edit functions will be carried out. If no selected area has been
   defined, edit functions will apply to the entire image.
   
***************************************************************************/

int      sa_npg, sa_maxpg = 1000, sa_pgx[1000], sa_pgy[1000];              //  select area polygon points
int      sa_blend;                                                         //  select area blend width
int      sa_stat = 0;                                                      //  select area status
int      sa_exist = 0;                                                     //  select area exists or not
int      sa_invert = 0;                                                    //  select area inverted or not
char     *sa_inside = 0;                                                   //  map pixels inside select area

struct   sa_pix1 { int16  px, py, dist; };                                 //  map pixels in select area
sa_pix1  *sa_pix = 0;
int      sa_npix;

void   select_mousefunc();                                                 //  select area mouse event handler
void   select_area_show();                                                 //  draw select area dotted-lines
int    select_dialog_event(zdialog*, const char *event);
int    select_dialog_compl(zdialog*, int zstat);


//  dialog to outline an image area - may run parallel with edit functions

void m_select(GtkWidget *, const char *)
{
   const char  *title = ZTX("image area for following edits");
   const char  *helptext = ZTX("Use mouse drag and click to enclose an area.\n"
                               "Use right click to undo the last line segment");

   if (! image_file) return;                                               //  no image
   if (zdsela) return;                                                     //  already active
   
   zdsela = zdialog_new(title,mWin,BOK,Bcancel,null);
   zdialog_add_widget(zdsela,"label","labhelp","dialog",helptext,"space=10");
   zdialog_add_widget(zdsela,"hbox","hb1","dialog",0,"space=10");
   zdialog_add_widget(zdsela,"button","start","hb1",Bstart);
   zdialog_add_widget(zdsela,"button","finish","hb1",Bfinish);
   zdialog_add_widget(zdsela,"button","show","hb1",Bshow);
   zdialog_add_widget(zdsela,"button","delete","hb1",Bdelete);
   zdialog_add_widget(zdsela,"button","invert","hb1",Binvert);
   zdialog_add_widget(zdsela,"hsep","hsep1","dialog");
   zdialog_add_widget(zdsela,"hbox","hb2","dialog",0,"space=10");
   zdialog_add_widget(zdsela,"label","labblend","hb2",Bblendwidth);
   zdialog_add_widget(zdsela,"hscale","blendwidth","hb2","0|300|1|0","expand");

   zdialog_run(zdsela,select_dialog_event,select_dialog_compl);            //  run dialog - parallel
   return;
}

int select_dialog_compl(zdialog *zd, int zstat)
{
   mouseCBfunc = 0;                                                        //  disconnect mouse function
   Mcapture = 0;
   gdk_window_set_cursor(drWin->window,0);                                 //  restore normal cursor
   if (zstat != 1) m_unselect(null,null);                                  //  cancel, remove select area
   if (sa_npg < 3) m_unselect(null,null);                                  //  not enough points
   zdialog_free(zdsela);                                                   //  kill dialog
   zdsela = null;
   return 0;
}

int select_dialog_event(zdialog *zd, const char *event)
{
   int sel_area_inside(int px, int py);                                    //  is point inside select area
   int sel_area_distance(int px, int py);                                  //  distance to nearest polygon edge

   const char  *delwarn = ZTX("delete existing area?");
   int         ii, yn, inside, px, py, cc, dist;
   int         minpgx, minpgy, maxpgx, maxpgy;
   int         Nrect, Ndone;
   const char  *eventx = event;

   if (strEqu(eventx,"start")) 
   {
      if (sa_stat != 0) {
         yn = zmessageYN(delwarn);
         if (! yn) return 0;
      }

      if (Fpreview) edit_fullsize(1);                                      //  use full-size bitmaps  v.6.2
      m_unselect(null,null);                                               //  erase prior area
      Mcapture++;                                                          //  mouse captured for me
      mouseCBfunc = select_mousefunc;                                      //  connect mouse function
      gdk_window_set_cursor(drWin->window,dragcursor);                     //  set drag cursor
   }

   if (strEqu(eventx,"show")) select_area_show();

   if (strEqu(eventx,"delete")) 
   {
      if (sa_npg > 30) {
         yn = zmessageYN(delwarn);                                         //  warn if significant work
         if (! yn) return 0;                                               //    will be deleted
      }
      m_unselect(null,null);                                               //  delete area
      mouseCBfunc = 0;                                                     //  disconnect mouse function
      Mcapture = 0;
      gdk_window_set_cursor(drWin->window,0);                              //  restore normal cursor
      mwpaint2();
   }
   
   if (strEqu(eventx,"invert") && sa_invert)                               //  invert selected area
   {
      zfree(sa_pix);                                                       //  if already inverted, free memory
      sa_pix = 0;                                                          //    and handle as "finish"
      sa_invert = 0;
      eventx = "finish";
      zfree(sa_inside);                                                    //  memory leak   v.6.3
      sa_inside = 0;
   }
         
   if (strEqu(eventx,"finish")) 
   {
      mouseCBfunc = 0;                                                     //  disconnect mouse function
      Mcapture = 0;
      gdk_window_set_cursor(drWin->window,0);                              //  restore normal cursor

      if (sa_npg < 3) {
         m_unselect(null,null);                                            //  not enough points
         return 0;
      }
      
      mutex_lock(&bitmaps_lock);                                           //  temp. prevent edits   v.6.3

      ii = sa_npg - 1;
      if (sa_pgx[ii] != sa_pgx[0] || sa_pgy[ii] != sa_pgy[0]) {            //  if not connected already
         ii++;                                                             //  make the connection
         sa_pgx[ii] = sa_pgx[0];                                           //  last line end = 1st point
         sa_pgy[ii] = sa_pgy[0];
      }

      sa_npg = ii;                                                         //  line count
      sa_pgx[ii+1] = sa_pgy[ii+1] = 0;                                     //  last point + 1 = zeros

      minpgx = maxpgx = sa_pgx[0];
      minpgy = maxpgy = sa_pgy[0];

      for (ii = 0; ii < sa_npg; ii++)                                      //  get rectangle enclosing polygon
      {
         if (sa_pgx[ii] < minpgx) minpgx = sa_pgx[ii];
         if (sa_pgx[ii] > maxpgx) maxpgx = sa_pgx[ii];
         if (sa_pgy[ii] < minpgy) minpgy = sa_pgy[ii];
         if (sa_pgy[ii] > maxpgy) maxpgy = sa_pgy[ii];
      }
      
      Nrect = (maxpgx - minpgx) * (maxpgy - minpgy);                       //  allocate enough memory for 
      cc = Nrect * sizeof(sa_pix1);                                        //    all pixels in rectangle
      sa_pix = (sa_pix1 *) zmalloc(cc);
      
      cc = Fww * Fhh;                                                      //  set up to map all pixels
      sa_inside = zmalloc(cc);                                             //    inside select area
      memset(sa_inside,0,cc);                                              //       v.6.3

      Ndone = sa_npix = 0;

      for (py = minpgy; py < maxpgy; py++)                                 //  loop all pixels in rectangle
      for (px = minpgx; px < maxpgx; px++)
      {
         edit_progress(Ndone++,Nrect);                                     //  v.6.3
         inside = sel_area_inside(px,py);                                  //  check if pixel inside polygon
         if (! inside) continue;
         dist = sel_area_distance(px,py);                                  //  distance to nearest edge
         sa_pix[sa_npix].px = px;
         sa_pix[sa_npix].py = py;                                          //  save pixel and distance
         sa_pix[sa_npix].dist = dist;
         sa_npix++;
         ii = py * Fww + px;                                               //  mark pixel inside area  v.6.3
         sa_inside[ii] = 1;
      }

      edit_progress(0,0);
      sa_stat = 3;                                                         //  area select finished
      sa_exist = 1;
      mutex_unlock(&bitmaps_lock);
      select_area_show();
   }

   if (strEqu(eventx,"blendwidth")) {                                      //  blend width changed
      zdialog_fetch(zd,"blendwidth",sa_blend);                             //  update sa_blend
      zdialog_send_event(zdedit,eventx);                                   //  notify active edit dialog  v.6.5
   }

   if (strEqu(eventx,"invert"))                                            //  invert selected area
   {
      if (! sa_exist) return 0;

      mutex_lock(&bitmaps_lock);                                           //  temp. prevent edits   v.6.3

      zfree(sa_pix);                                                       //  memory leak         v.6.3
      ii = Fhh * Fww * sizeof(sa_pix1);                                    //  allocate enough memory for 
      sa_pix = (sa_pix1 *) zmalloc(ii);                                    //    all pixels in image

      Nrect = Fww * Fhh;
      Ndone = sa_npix = 0;

      for (py = 0; py < Fhh; py++)                                         //  invert sa_inside flags   v.6.3
      for (px = 0; px < Fww; px++)                                         //  sa_pix array = outside pixels
      {
         edit_progress(Ndone++,Nrect);                                     //  v.6.3
         ii = py * Fww + px;
         if (sa_inside[ii]) sa_inside[ii] = 0;
         else {
            sa_inside[ii] = 1;
            dist = sel_area_distance(px,py);                               //  distance to nearest edge
            sa_pix[sa_npix].px = px;
            sa_pix[sa_npix].py = py;                                       //  save pixel and distance
            sa_pix[sa_npix].dist = dist;
            sa_npix++;
         }
      }

      edit_progress(0,0);
      sa_invert = 1;
      mutex_unlock(&bitmaps_lock);
      select_area_show();
   }

   return 0;
}


//  select area mouse function - draw lines, compute enclosed pixels

void select_mousefunc()
{
   int            ii, jj;
   static int     mdrag;
   
   if (sa_npg == 0) mdrag = 0;
   
   if (Mxdrag || Mydrag)                                                   //  drag underway
   {
      mdrag = 1;
      
      if (Mxdown == Mxdrag && Mydown == Mydrag) return;                    //  no drag progress

      if (sa_npg == 0) {                                                   //  set 1st point if none
         sa_pgx[0] = Mxdown;
         sa_pgy[0] = Mydown;
         sa_npg = 1;
         sa_pgx[1] = sa_pgy[1] = -1;                                       //  no drag line
      }
      else {
         ii = sa_npg - 1;                                                  //  drag underway
         jj = sa_npg;
         if (sa_pgx[jj] != -1)                                             //  erase old drag line
            erase_dotline(sa_pgx[ii],sa_pgy[ii],sa_pgx[jj],sa_pgy[jj]);
         sa_pgx[jj] = Mxdrag;                                              //  next point = curr. mouse posn.
         sa_pgy[jj] = Mydrag;
         draw_dotline(sa_pgx[ii],sa_pgy[ii],sa_pgx[jj],sa_pgy[jj]);        //  draw new drag line
      }
   }

   else if (mdrag || LMclick)                                              //  drag end or left click
   {
      if (sa_npg > sa_maxpg-2) {
         zmessageACK(ZTX("too many points"));
         return;
      }

      if (LMclick) {
         sa_pgx[sa_npg] = Mxclick;                                         //  position at click
         sa_pgy[sa_npg] = Myclick;
      }

      sa_npg++;                                                            //  new point added

      LMclick = RMclick = mdrag = 0;                                       //  reset clicks, mouse drag
      
      if (sa_npg > 1) {
         ii = sa_npg - 1;                                                  //  eliminate duplication
         jj = sa_npg - 2;
         if (sa_pgx[ii] == sa_pgx[jj] && sa_pgy[ii] == sa_pgy[jj]) 
            sa_npg--;
      }

      if (sa_npg > 1) {
         ii = sa_npg - 2;                                                  //  draw new last line
         jj = sa_npg - 1;
         draw_dotline(sa_pgx[ii],sa_pgy[ii],sa_pgx[jj],sa_pgy[jj]);
      }
      
      ii = sa_npg;
      sa_pgx[ii] = sa_pgy[ii] = -1;                                        //  no drag line
   }

   else if (RMclick)                                                       //  right mouse click 
   {
      RMclick = LMclick = mdrag = 0;                                       //  reset clicks, mouse drag

      ii = sa_npg - 2;
      jj = sa_npg - 1;
      if (sa_npg > 1)                                                      //  erase last line
         erase_dotline(sa_pgx[ii],sa_pgy[ii],sa_pgx[jj],sa_pgy[jj]);

      sa_npg--;                                                            //  reduce line count
      if (sa_npg <= 1) sa_npg = 0;

      ii = sa_npg;
      sa_pgx[ii] = sa_pgy[ii] = -1;                                        //  no drag line
   }
      
   if (Wpainted) select_area_show();                                       //  draw all lines
   Wpainted = 0;

   return;
}


//  draw all dotted lines around selected area

void select_area_show()
{
   int         ii, jj;
   
   mwpaint();                                                              //  v.6.3

   for (ii = 0; ii < sa_npg; ii++)
   {
      jj = ii + 1;
      if (sa_pgx[jj] == -1) break;
      draw_dotline(sa_pgx[ii],sa_pgy[ii],sa_pgx[jj],sa_pgy[jj]);           //  last or curr. drag line
   }
   
   return;
}


//  determine if pixel at (px,py) is inside the area selection polygon

int sel_area_inside(int px, int py)
{
   int  sel_area_inside2(int px, int py, int nth);
   int      ii, inside = 0;

   for (ii = 0; ii < sa_npg; ii++)                                         //  count intersections of line up
      if (sel_area_inside2(px,py,ii) == 2) inside = 1 - inside;            //    from (px,py) to a polygon face

   return inside;
}

int sel_area_inside2(int px, int py, int nth)                              //  determine if point (px,py) is
{                                                                          //    above or below nth polygon face
   int      x1, y1, x2, y2;
   double   yp;
   
   x1 = sa_pgx[nth];                                                       //  line segment of nth face
   y1 = sa_pgy[nth];
   x2 = sa_pgx[nth+1];
   y2 = sa_pgy[nth+1];
   
   if (x1 == x2) return 0;                                                 //  line is vertical

   if (x1 < x2) {
      if (px <= x1) return 0;                                              //  point is left or right of line
      if (px > x2) return 0;
   }
   else { 
      if (px <= x2) return 0;
      if (px > x1) return 0;
   }

   if (py <= y1 && py <= y2) return 1;                                     //  point is above or below both ends
   if (py >= y1 && py >= y2) return 2;
   
   yp = (px*(y2-y1) + (x2*y1-x1*y2))/(1.0*(x2-x1));                        //  intersect of vertical line x = px
   if (py < yp) return 1;                                                  //  above
   if (py > yp) return 2;                                                  //  below
   return 0;                                                               //  dead on
}


//  get distance from pixel (px,py) to nearest polygon face

int sel_area_distance(int px, int py)
{
   int  sel_area_distance2(int px, int py, int nth);
   int      ii, dd, mindd;

   mindd = sel_area_distance2(px,py,0);                                    //  distance^2 to 0th polygon face

   for (ii = 1; ii < sa_npg; ii++)
   {
      dd = sel_area_distance2(px,py,ii);                                   //  distance^2 to Nth polygon face
      if (dd < mindd) mindd = dd;                                          //  remember smallest distance^2
   }
   
   mindd = int(sqrt(mindd));                                               //  return smallest distance
   return  mindd;
}

int sel_area_distance2(int px, int py, int nth)                            //  get distance^2 from point (px,py)
{                                                                          //    to nth polygon face
   int      x1, y1, x2, y2;
   int      a, b, c, d;
   int      h1, h2, ls, temp;
   int      dist2;
   int      huge = 1000000;                                                //  distance = 1000

   x1 = sa_pgx[nth];                                                       //  line segment of nth face
   y1 = sa_pgy[nth];
   x2 = sa_pgx[nth+1];
   y2 = sa_pgy[nth+1];
   
   if (x1 < 5 && x2 < 5) return huge;
   if (y1 < 5 && y2 < 5) return huge;                                      //  if line segment hugs image edge,
   if (Fww - x1 < 5 && Fww - x2 < 5) return huge;                          //    then it is logically distant
   if (Fhh - y1 < 5 && Fhh - y2 < 5) return huge;
   
   a = y1-y2;                                                              //  get min. distance ^2 from point
   b = x2-x1;                                                              //    to entire line of line segment
   c = x1*y2 - y1*x2;                                                      //  (intersect may be outside segment)
   d = (a*px + b*py + c);
   dist2 = int(1.0 * d*d / (a*a + b*b));                                   //  (can overflow 32 bits)
   
   h1 = (x1-px)*(x1-px) + (y1-py)*(y1-py);                                 //  point to (x1,y1) ^2
   h2 = (x2-px)*(x2-px) + (y2-py)*(y2-py);                                 //  point to (x2,y2) ^2
   ls = (x1-x2)*(x1-x2) + (y1-y2)*(y1-y2);                                 //  segment length ^2
   if (h1 > h2) { temp = h2; h2 = h1; h1 = temp; }                         //  re-order, h1 <= h2
   
   if (h1 < 0 || dist2 < 0) {
      printf("px py nth: %d %d %d \n",px,py,nth);                          //  algebra failure     ***
      printf("x1 y1 x2 y2: %d %d %d %d \n",x1,y1,x2,y2);
      printf("h1 dist2: %d %d \n",h1,dist2);
      printf("sel_area_distance2() failure \n");
      return 0;
   }

   if (h2 - dist2 < ls) return dist2;                                      //  intersect in segment, use dist2
   else return h1;                                                         //  distance ^2 to nearest end point
}


//  clear selected image area, free memory

void m_unselect(GtkWidget *, const char *)
{
   if (sa_exist) mwpaint2();
   sa_npg = sa_npix = sa_stat = sa_exist = sa_invert = sa_blend = 0;
   if (sa_pix) zfree(sa_pix);
   if (sa_inside) zfree(sa_inside);
   sa_pix = 0;
   sa_inside = 0;
   mwpaint2();
   return;
}


/**************************************************************************/

//  flatten brightness distribution

#define  flatNT 4                                                          //  no. of working threads
int      flatten_done[flatNT];
double   flatten_brdist[65536];
double   flatten_value = 0;                                                //  flatten value, 0 - 100%

void m_flatten(GtkWidget *, const char *)
{
   int    flatten_dialog_event(zdialog* zd, const char *event);
   int    flatten_dialog_compl(zdialog* zd, int zstat);
   void * flatten_thread(void *);

   const char  *title = ZTX("Flatten Brightness Distribution");

   if (! edit_setup(1)) return;                                            //  setup edit: preview

   zdedit = zdialog_new(title,mWin,Bundo,Bredo,Bdone,Bcancel,null);        //  flatten distribution dialog
   zdialog_add_widget(zdedit,"hbox","hb1","dialog",0,"space=15");
   zdialog_add_widget(zdedit,"label","labfd","hb1",ZTX("Flatten"),"space=5");
   zdialog_add_widget(zdedit,"hscale","flatten","hb1","0|100|1|0","expand");

   zdialog_resize(zdedit,300,0);
   zdialog_run(zdedit,flatten_dialog_event,flatten_dialog_compl);          //  run dialog - parallel
   
   flatten_value = 0;
   start_thread(flatten_thread,0);                                         //  start working thread
   return;
}


//  flatten dialog event and completion functions

int flatten_dialog_event(zdialog *zd, const char *event)                   //  flatten dialog event function
{
   zdialog_fetch(zd,"flatten",flatten_value);                              //  get slider value
   signal_thread();                                                        //  trigger update thread
   return 1;
}


int flatten_dialog_compl(zdialog *zd, int zstat)                           //  flatten dialog completion function
{
   if (zstat == 1) edit_undo();                                            //  undo
   else if (zstat == 2) edit_redo();                                       //  redo
   else if (zstat == 3) edit_done();                                       //  done
   else edit_cancel();                                                     //  cancel or destroy
   return 0;
}


//  thread function - use multiple working threads

void * flatten_thread(void *)
{
   void  * flatten_thread4(void *arg);

   int         px, py, ii, index[flatNT];
   double      bright1;
   uint16      *pix1;

   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request

      for (ii = 0; ii < 65536; ii++)                                       //  clear brightness distribution data
         flatten_brdist[ii] = 0;

      if (sa_exist)                                                        //  process selected area
      {
         for (ii = 0; ii < sa_npix; ii++)                                  //  process enclosed pixels
         {
            px = sa_pix[ii].px;                                            //  compute brightness distribution
            py = sa_pix[ii].py;
            pix1 = bmpixel(E1rgb48,px,py);
            bright1 = brightness(pix1);
            flatten_brdist[int(bright1)]++;
         }
         
         for (ii = 1; ii < 65536; ii++)                                    //  cumulative brightness distribution
            flatten_brdist[ii] += flatten_brdist[ii-1];                    //   0 ... sa_npix

         for (ii = 0; ii < 65536; ii++)
            flatten_brdist[ii] = flatten_brdist[ii]                        //  multiplier per brightness level
                               / sa_npix * 65536.0 / (ii + 1);
      }

      else                                                                 //  process whole image
      {
         for (py = 0; py < E1hh; py++)                                     //  compute brightness distribution
         for (px = 0; px < E1ww; px++)
         {
            pix1 = bmpixel(E1rgb48,px,py);
            bright1 = brightness(pix1);
            flatten_brdist[int(bright1)]++;
         }
         
         for (ii = 1; ii < 65536; ii++)                                    //  cumulative brightness distribution
            flatten_brdist[ii] += flatten_brdist[ii-1];                    //   0 ... (ww1 * hh1)

         for (ii = 0; ii < 65536; ii++)
            flatten_brdist[ii] = flatten_brdist[ii]                        //  multiplier per brightness level 
                               / (E1ww * E1hh) * 65536.0 / (ii + 1);
      }
      
      for (ii = 0; ii < flatNT; ii++)                                      //  start flatNT threads
      {
         flatten_done[ii] = 0;
         index[ii] = ii;
         start_detached_thread(flatten_thread4,&index[ii]);
      }

      for (ii = 0; ii < flatNT; ii++)                                      //  wait for all done
         while (flatten_done[ii] == 0) zsleep(0.01);

      Fmodified = 1;
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  never executed, stop g++ warning
}


void * flatten_thread4(void *arg)                                          //  thread function
{
   int         index = *((int *) (arg));
   int         px, py, ii, dist;
   double      fold = 1, fnew = 0;

   GOFUNC_DATA(1)

   fnew = 0.01 * flatten_value;                                            //  0.0 - 1.0  how much to flatten
   fold = 1.0 - fnew;                                                      //  1.0 - 0.0  how much to retain

   if (sa_exist)                                                           //  process selected area
   {
      for (ii = index; ii < sa_npix; ii += flatNT)                         //  process all enclosed pixels
      {
         px = sa_pix[ii].px;                                               //  flatten brightness distribution
         py = sa_pix[ii].py;
         dist = sa_pix[ii].dist;
         GOFUNC(flatten_1pix)
      }
   }

   else
   {
      dist = sa_blend = 0;
      for (py = index; py < E1hh; py += flatNT)                            //  flatten brightness distribution
      for (px = 0; px < E1ww; px++)
         GOFUNC(flatten_1pix)
   }

   flatten_done[index] = 1;
   pthread_exit(0);

flatten_1pix:                                                              //  process one pixel
   {
      uint16      *pix1, *pix3;
      double      dold, dnew, cmax;
      double      red1, green1, blue1, red3, green3, blue3;
      double      bright1, bright2;

      pix1 = bmpixel(E1rgb48,px,py);                                       //  input pixel
      pix3 = bmpixel(E3rgb48,px,py);                                       //  output pixel
         
      red1 = pix1[0];
      green1 = pix1[1];
      blue1 = pix1[2];

      bright1 = brightness(pix1);                                          //  input brightness
      bright2 = flatten_brdist[int(bright1)];                              //  output brightness adjustment

      red3 = bright2 * red1;                                               //  flattened brightness
      green3 = bright2 * green1;
      blue3 = bright2 * blue1;

      red3 = fnew * red3 + fold * red1;                                    //  blend new and old brightness
      green3 = fnew * green3 + fold * green1;
      blue3 = fnew * blue3 + fold * blue1;

      if (dist < sa_blend) {                                               //  blend over distance sa_blend
         dnew = 1.0 * dist / sa_blend;
         dold = 1.0 - dnew;
         red3 = dnew * red3 + dold * red1;
         green3 = dnew * green3 + dold * green1;
         blue3 = dnew * blue3 + dold * blue1;
      }

      cmax = red3;                                                         //  stop overflow, keep color balance
      if (green3 > cmax) cmax = green3;
      if (blue3 > cmax) cmax = blue3;
      if (cmax > 65535) {
         cmax = 65535 / cmax;
         red3 = red3 * cmax;
         green3 = green3 * cmax;
         blue3 = blue3 * cmax;
      }

      pix3[0] = int(red3 + 0.5);
      pix3[1] = int(green3 + 0.5);
      pix3[2] = int(blue3 + 0.5);
      
      RETURN
   }
}


/**************************************************************************/

//  brightness / color / contrast adjustment

int   tune_curve_adjust(void *,GdkEventButton *);                          //  mouse events in drawing area
int   tune_curve_draw();                                                   //  draw curve in drawing area

GtkWidget   *tune_drawarea;                                                //  drawing area for curve
int         tune_thread_busy[4];
int         tune_ii;                                                       //  ii = current spline curve
int         tune_nap[6];                                                   //  no. anchor points for 6 curves
double      tune_apx[6][50], tune_apy[6][50];                              //  anchor points for 6 curves
double      tune_dat[6][100];                                              //  data points for 6 curves


void m_tune(GtkWidget *, const char *)                                     //  overhauled  v.6.8
{
   int   tune_dialog_event(zdialog *zd, cchar *event);
   int   tune_dialog_compl(zdialog *zd, int zstat);
   void  *tune_thread(void *);

   const char  *title = ZTX("Adjust Brightness and Color");

   if (! edit_setup(1)) return;                                            //  setup edit: preview
   
   for (int ii = 0; ii < 6; ii++)
   {                                                                       //  initz. all curves to flat
      tune_nap[ii] = 3;
      tune_apx[ii][0] = 0;                                                 //  3 anchor points: 
      tune_apy[ii][0] = 50;                                                //    (0,50) (50,50) (100,50)
      tune_apx[ii][1] = 50;
      tune_apy[ii][1] = 50;
      tune_apx[ii][2] = 100;
      tune_apy[ii][2] = 50;
      
      for (int jj = 0; jj < 100; jj++)                                     //  initz. curve data points
         tune_dat[ii][jj] = 50;
   }
   
   tune_ii = 0;                                                            //  default curve = brightness

/***
       --------------------------------------------
      |                                            |
      |           curve drawing area               |
      |                                            |
       --------------------------------------------
       darker areas                   lighter areas

      [+++] [---] [+ -] [- +] [+-+] [-+-]

      (o) brightness
      (o) color intensity    [reset 1]  [reset all]
      (o) color saturation
      (o) color balance    (o) red  (o) green  (o) blue

***/

   zdedit = zdialog_new(title,mWin,Bundo,Bredo,Bdone,Bcancel,null);        //  create dialog

   zdialog_add_widget(zdedit,"frame","fr1","dialog",0,"expand");
   zdialog_add_widget(zdedit,"hbox","hba","dialog");
   zdialog_add_widget(zdedit,"label","labda","hba",Bdarker,"space=5");
   zdialog_add_widget(zdedit,"label","labspa","hba",0,"expand");
   zdialog_add_widget(zdedit,"label","labba","hba",Blighter,"space=5");
   zdialog_add_widget(zdedit,"hbox","hbb","dialog",0,"space=5");
   zdialog_add_widget(zdedit,"button","b +++","hbb","+++");
   zdialog_add_widget(zdedit,"button","b ---","hbb"," - - - ");
   zdialog_add_widget(zdedit,"button","b +-", "hbb"," + -  ");
   zdialog_add_widget(zdedit,"button","b -+", "hbb","  - + ");
   zdialog_add_widget(zdedit,"button","b +-+","hbb","+ - +");
   zdialog_add_widget(zdedit,"button","b -+-","hbb"," - + - ");
   zdialog_add_widget(zdedit,"hbox","hb2","dialog");
   zdialog_add_widget(zdedit,"vbox","vb21","hb2");
   zdialog_add_widget(zdedit,"vbox","vb22","hb2");
   zdialog_add_widget(zdedit,"radio","radbr","vb21",Bbrightness);
   zdialog_add_widget(zdedit,"radio","radcol","vb21",ZTX("color intensity"));
   zdialog_add_widget(zdedit,"radio","radsat","vb21",ZTX("color saturation"));
   zdialog_add_widget(zdedit,"radio","radbal","vb21",ZTX("color balance"));
   zdialog_add_widget(zdedit,"hbox","hbrs","vb22",0,"space=5");
   zdialog_add_widget(zdedit,"label","labsprs","hbrs",0,"space=20");
   zdialog_add_widget(zdedit,"button","reset1","hbrs",ZTX(" reset 1 "));
   zdialog_add_widget(zdedit,"button","resetA","hbrs",ZTX("reset all"));
   zdialog_add_widget(zdedit,"label","labsprgb","vb22",0,"expand");
   zdialog_add_widget(zdedit,"hbox","hbrgb","vb22");
   zdialog_add_widget(zdedit,"radio","radR","hbrgb",Bred,"space=10");
   zdialog_add_widget(zdedit,"radio","radG","hbrgb",Bgreen,"space=5");
   zdialog_add_widget(zdedit,"radio","radB","hbrgb",Bblue,"space=5");
   
   GtkWidget *frame = zdialog_widget(zdedit,"fr1");                        //  add drawing area to frame
   tune_drawarea = gtk_drawing_area_new();
   gtk_container_add(GTK_CONTAINER(frame),tune_drawarea);

   gtk_widget_add_events(tune_drawarea,GDK_BUTTON_PRESS_MASK);             //  connect drawing area events
   gtk_widget_add_events(tune_drawarea,GDK_BUTTON_RELEASE_MASK);
   gtk_widget_add_events(tune_drawarea,GDK_BUTTON1_MOTION_MASK); 
   G_SIGNAL(tune_drawarea,"motion-notify-event",tune_curve_adjust,0)
   G_SIGNAL(tune_drawarea,"button-press-event",tune_curve_adjust,0)
   G_SIGNAL(tune_drawarea,"expose-event",tune_curve_draw,0)
   
   zdialog_stuff(zdedit,"radbr",1);                                        //  stuff defaults
   zdialog_stuff(zdedit,"radR",1);
   
   zdialog_resize(zdedit,0,350);
   zdialog_run(zdedit,tune_dialog_event,tune_dialog_compl);                //  run dialog - parallel
   start_thread(tune_thread,0);                                            //  start working thread
   return;
}


//  tune dialog event and completion functions

int tune_dialog_event(zdialog *zd, const char *event)
{
   int      ii, jj, curve = -1;
   double   px, py;
   
   if (strnEqu(event,"rad",3))                                             //  new choice of curve
   {
      ii = strcmpv(event,"radbr","radcol","radsat","radR","radG","radB",0);
      if (ii > 0) curve = ii - 1;
      
      if (strstr("radR radG radB",event))                                  //  if RGB, set color balance
         zdialog_stuff(zd,"radbal",1);
      
      if (strEqu(event,"radbal")) {                                        //  if color balance,
         zdialog_fetch(zd,"radR",ii);                                      //    get current RGB selection
         if (ii) curve = 3;
         zdialog_fetch(zd,"radG",ii);
         if (ii) curve = 4;
         zdialog_fetch(zd,"radB",ii);
         if (ii) curve = 5;
      }

      if (curve >= 0 && curve != tune_ii) {
         tune_ii = curve;                                                  //  set new curve
         tune_curve_draw();                                                //  redraw curve
      }
   }
   
   if (strnEqu(event,"b ",2))                                              //  button to move entire curve
   {
      ii = tune_ii;
      
      for (jj = 0; jj < tune_nap[ii]; jj++)
      {
         px = tune_apx[ii][jj];
         py = tune_apy[ii][jj];
      
         if (strEqu(event,"b +++")) py += 10;
         if (strEqu(event,"b ---")) py -= 10;
         if (strEqu(event,"b +-"))  py += 10.0 - 0.2 * px;
         if (strEqu(event,"b -+"))  py -= 10.0 - 0.2 * px;
         if (strEqu(event,"b +-+")) py -= 5 - 0.2 * abs(px-50);
         if (strEqu(event,"b -+-")) py += 5 - 0.2 * abs(px-50);

         if (py > 100) py = 100;
         if (py < 0) py = 0;
         tune_apy[ii][jj] = py;
      }

      tune_curve_draw();                                                   //  redraw curve
      signal_thread();                                                     //  trigger image update
   }
   
   if (strEqu(event,"reset1")) 
   {
      ii = tune_ii;                                                        //  current curve
      tune_nap[ii] = 3;
      tune_apx[ii][0] = 0;                                                 //  3 anchor points: 
      tune_apy[ii][0] = 50;                                                //    (0,50) (50,50) (100,50)
      tune_apx[ii][1] = 50;
      tune_apy[ii][1] = 50;
      tune_apx[ii][2] = 100;
      tune_apy[ii][2] = 50;
      
      for (int jj = 0; jj < 100; jj++)                                     //  initz. curve data points
         tune_dat[ii][jj] = 50;
      
      tune_curve_draw();                                                   //  update dialog curve
      signal_thread();                                                     //  update image
   }
   
   if (strEqu(event,"resetA")) 
   {
      for (int ii = 0; ii < 6; ii++)                                       //  do all curves
      {
         tune_nap[ii] = 3;
         tune_apx[ii][0] = 0;
         tune_apy[ii][0] = 50;
         tune_apx[ii][1] = 50;
         tune_apy[ii][1] = 50;
         tune_apx[ii][2] = 100;
         tune_apy[ii][2] = 50;
         
         for (int jj = 0; jj < 100; jj++)
            tune_dat[ii][jj] = 50;
      }
      
      tune_curve_draw();
      signal_thread();
   }
   
   if (strEqu(event,"blendwidth")) signal_thread();                        //  select area blend width change

   return 1;
}


int tune_dialog_compl(zdialog *zd, int zstat)                              //  tune dialog completion function
{
   if (zstat == 1) edit_undo();                                            //  undo
   else if (zstat == 2) edit_redo();                                       //  redo
   else if (zstat == 3) edit_done();                                       //  done
   else edit_cancel();                                                     //  cancel or destroy
   return 0;
}


//  Add, delete, or move anchor points to brightness curve using mouse.

int tune_curve_adjust(void *,GdkEventButton *event)
{
   int         ww, hh, px, py;
   int         kk, ii, jj, newjj, minjj = -1;
   int         mx, my, button, evtype;
   double      dist2, mindist2 = 1000000;
   double      xval, yval;
   
   ii = tune_ii;                                                           //  get current curve

   if (tune_nap[ii] > 49) {
      zmessageACK(ZTX("exceed 50 anchor points"));
      return 0;
   }
   
   mx = int(event->x);                                                     //  mouse position in drawing area
   my = int(event->y);
   evtype = event->type;
   button = event->button;
   ww = tune_drawarea->allocation.width;                                   //  drawing area size
   hh = tune_drawarea->allocation.height;
   
   for (jj = 0; jj < tune_nap[ii]; jj++)                                   //  find closest anchor point
   {
      xval = tune_apx[ii][jj];
      yval = spline2(xval);
      px = int(0.01 * ww * xval);                                          //  0 - ww
      py = int(hh - 0.01 * hh * yval + 0.5);                               //  0 - hh
      dist2 = (px-mx)*(px-mx) + (py-my)*(py-my);
      if (dist2 < mindist2) {
         mindist2 = dist2;
         minjj = jj;
      }
   }

   if (minjj < 0) return 0;                                                //  huh?
   
   if (evtype == GDK_BUTTON_PRESS && button == 3) {                        //  richt click, remove anchor point
      if (mindist2 > 25) return 0;
      if (tune_nap[ii] < 3) return 0;
      for (kk = minjj; kk < tune_nap[ii] -1; kk++) {
         tune_apx[ii][kk] = tune_apx[ii][kk+1];
         tune_apy[ii][kk] = tune_apy[ii][kk+1];
      }
      tune_nap[ii]--;

      tune_curve_draw();                                                   //  regen and redraw curve
      signal_thread();                                                     //  trigger image update
      return 0;
   }

//  drag or left click, move nearby anchor point to mouse position,
//  or add a new anchor point if nothing near enough

   xval = 100.0 * mx / ww;                                                 //  0 - 100
   yval = 100.0 * (hh - my) / hh;                                          //  0 - 100

   if (xval < 0 || xval > 100) return 0;                                   //  v.6.8
   if (yval < 0 || yval > 100) return 0;

   if (mindist2 < 100) {                                                   //  existing point < 10 pixels away
      jj = minjj;
      if (jj < tune_nap[ii] - 1 && tune_apx[ii][jj+1] - xval < 5)          //  disallow < 5 x-pixels
         return 0;                                                         //    to next or prior point
      if (jj > 0 && xval - tune_apx[ii][jj-1] < 5) return 0;
      newjj = minjj;                                                       //  point to be moved
   }
   else                                                                    //  > 10 pixels, add a point
   {
      for (jj = 0; jj < tune_nap[ii]; jj++)
         if (xval <= tune_apx[ii][jj]) break;                              //  find point with next higher x

      if (jj < tune_nap[ii] && tune_apx[ii][jj] - xval < 5) return 0;      //  disallow < 5 pixels
      if (jj > 0 && xval - tune_apx[ii][jj-1] < 5) return 0;               //    to next or prior point

      for (kk = tune_nap[ii]; kk > jj; kk--) {                             //  make hole for new point
         tune_apx[ii][kk] = tune_apx[ii][kk-1];
         tune_apy[ii][kk] = tune_apy[ii][kk-1];
      }

      tune_nap[ii]++;                                                      //  up point count
      newjj = jj;                                                          //  point to be added
   }

   tune_apx[ii][newjj] = xval;                                             //  coordinates of new or moved point
   tune_apy[ii][newjj] = yval;
   
   tune_curve_draw();                                                      //  regen and redraw the curve
   signal_thread();                                                        //  trigger image update
   return 0;
}


//  Draw brightness curve based on defined spline anchor points.

int tune_curve_draw()
{
   int         ww, hh, px, py;
   int         ii, jj, jjx, jjy;
   double      xval, yval;
   
   ii = tune_ii;                                                           //  current curve

   ww = tune_drawarea->allocation.width;                                   //  drawing area size
   hh = tune_drawarea->allocation.height;
   if (ww < 50 || hh < 20) return 0;

   spline1(tune_nap[ii],tune_apx[ii],tune_apy[ii]);                        //  make curve fitting anchor points

   gdk_window_clear(tune_drawarea->window);                                //  clear window

   for (px = 0; px < ww; px++)                                             //  generate all points for curve
   {
      xval = 100.0 * px / ww;
      yval = spline2(xval);
      py = int(hh - 0.01 * hh * yval + 0.5);
      gdk_draw_point(tune_drawarea->window,gdkgc,px,py);
   }
   
   for (jj = 0; jj < tune_nap[ii]; jj++)                                   //  draw boxes at anchor points
   {
      xval = tune_apx[ii][jj];
      yval = spline2(xval);
      px = int(0.01 * ww * xval);
      py = int(hh - 0.01 * hh * yval + 0.5);
      for (jjx = -2; jjx < 3; jjx++)
      for (jjy = -2; jjy < 3; jjy++) {
         if (px+jjx < 0 || px+jjx >= ww) continue;
         if (py+jjy < 0 || py+jjy >= hh) continue;
         gdk_draw_point(tune_drawarea->window,gdkgc,px+jjx,py+jjy);
      }
   }

   for (jj = 0; jj < 100; jj++)                                            //  save 100 curve data points
      tune_dat[ii][jj] = spline2(jj);

   return 0;
}


//  Update image based on latest settings of all dialog controls.

void * tune_thread(void *)
{
   void * tune_thread4(void *arg);

   static int     index[4];

   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request
      
      for (int ii = 0; ii < 4; ii++) {                                     //  start 4 worker threads
         tune_thread_busy[ii] = 1;
         index[ii] = ii;
         start_detached_thread(tune_thread4,&index[ii]);
      }
      
      for (int ii = 0; ii < 4; ii++)                                       //  wait for all to complete
         while (tune_thread_busy[ii]) zsleep(0.01);

      Fmodified = 1;
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  never executed, stop g++ warning
}


void * tune_thread4(void *arg)
{
   int      px, py, ii, dist;
   int      index = *((int *) arg);

   GOFUNC_DATA(1)
   
   if (sa_exist)                                                           //  process selected area
   {
      for (ii = index; ii < sa_npix; ii += 4)                              //  process all enclosed pixels
      {
         px = sa_pix[ii].px;
         py = sa_pix[ii].py;
         dist = sa_pix[ii].dist;
         GOFUNC(tune1pix)
      }
   }

   else                                                                    //  process whole image
   {
      dist = sa_blend = 0;
      for (py = index; py < E1hh; py += 4)
      for (px = 0; px < E1ww; px++)
         GOFUNC(tune1pix)
   }

   tune_thread_busy[index] = 0;
   pthread_exit(0);


tune1pix:                                                                  //  process one pixel
   {
      uint16      *pix1, *pix3;
      double      red1, green1, blue1, red3, green3, blue3;
      double      brin, brout;
      int         curveindex;

      pix1 = bmpixel(E1rgb48,px,py);                                       //  input pixel
      pix3 = bmpixel(E3rgb48,px,py);                                       //  output pixel
      
      red1 = red3 = pix1[0];
      green1 = green3 = pix1[1];
      blue1 = blue3 = pix1[2];

      brin = red3;                                                         //  brin = brightest color
      if (green3 > brin) brin = green3;
      if (blue3 > brin) brin = blue3;
      
      curveindex = int(brin/656);                                          //  index into curve data, 0-99

/* ------------------------------------------------------------------------

      brightness curve values:
           0 = dark
          50 = normal, unchanged
         100 = 200% brightness, clipped
*/

      brout = tune_dat[0][curveindex];                                     //  brightness factor, 0 - 99
      
      if (brout < 49 || brout > 51)
      {
         brout = brout / 50.0;                                             //  0 - 2.0
         if (brout * brin > 65535.0) brout = 65535.0 / brin;               //  reduce if necessary
         
         red3 = red3 * brout;                                              //  apply to all colors
         green3 = green3 * brout;
         blue3 = blue3 * brout;
      }
      
/* ------------------------------------------------------------------------

      color intensity curve values:
           0 = no color (grey scale)  
          50 = normal, unchanged
         100 = highest color

      50 >> 0: move all RGB values to their mean: (R+G+B)/3
      50 >> 100: increase all RGB values by same factor
      
      In the 2nd case, the movement is greater for darker pixels
*/

      double   red50, green50, blue50, red100, green100, blue100;
      double   rgb0, max50, min50, color, bright, ramper;
      
      brout = tune_dat[1][curveindex];                                     //  brightness factor, 0 - 99
      
      if (brout < 49 || brout > 51)
      {
         red50 = red3;                                                     //  50%  color values (normal)
         green50 = green3;
         blue50 = blue3;
         
         rgb0 = (red50 + green50 + blue50) / 3;                            //  0%  color values (grey scale)
            
         max50 = min50 = red50;
         if (green50 > max50) max50 = green50;                             //  get max/min normal color values
         else if (green50 < min50) min50 = green50;
         if (blue50 > max50) max50 = blue50;
         else if (blue50 < min50) min50 = blue50;
         
         color = (max50 - min50) * 1.0 / (max50 + 1);                      //  gray .. color       0 .. 1
         color = sqrt(color);                                              //  accelerated curve   0 .. 1
         bright = max50 / 65535.0;                                         //  dark .. bright      0 .. 1
         bright = sqrt(bright);                                            //  accelerated curve   0 .. 1
         ramper = 1 - color + bright * color;                              //  1 - color * (1 - bright)
         ramper = 1.0 / ramper;                                            //  large if color high and bright low

         red100 = int(red50 * ramper);                                     //  100%  color values (max)
         green100 = int(green50 * ramper);
         blue100 = int(blue50 * ramper);
         
         if (brout < 50) 
         {
            red3 = rgb0 + (brout) * 0.02 * (red50 - rgb0);                 //  compute new color value
            green3 = rgb0 + (brout) * 0.02 * (green50 - rgb0);
            blue3 = rgb0 + (brout) * 0.02 * (blue50 - rgb0);
         }

         if (brout > 50)
         {
            red3 = red50 + (brout - 50) * 0.02 * (red100 - red50);
            green3 = green50 + (brout - 50) * 0.02 * (green100 - green50);
            blue3 = blue50 + (brout - 50) * 0.02 * (blue100 - blue50);
         }
      }

/* ------------------------------------------------------------------------

      color saturation curve values:
           0 = no color saturation (gray scale)
          50 = normal (initial unmodified RGB)
         100 = max. color saturation

      50 >> 0: move all RGB values to their mean: (R+G+B)/3
      50 >> 100: increase RGB spread until one color is 0 or 65535

      In both cases, the average of RGB is not changed.
*/

      double   rinc, ginc, binc, scale;
      double   spread, spread1, spread2;
      
      brout = tune_dat[2][curveindex];                                     //  saturation factor, 0 - 99
      
      if (brout < 49 || brout > 51)
      {
         spread = brout - 50;                                              //  -50  to  0  to  50
         spread1 = 0.02 * spread + 1;                                      //    0  to  1  to   2
         spread2 = spread1 - 1.0;                                          //   -1  to  0  to   1

         red50 = red3;                                                     //  50%  color values (normal)
         green50 = green3;
         blue50 = blue3;
         
         rgb0 = (red50 + green50 + blue50 + 1) / 3;

         rinc = red50 - rgb0;
         ginc = green50 - rgb0;
         binc = blue50 - rgb0;
         
         scale = 1.0;

      again:
         rinc = scale * rinc;
         ginc = scale * ginc;
         binc = scale * binc;

         red100 = red50 + rinc;
         green100 = green50 + ginc;
         blue100 = blue50 + binc;

         if (red100 > 65535) { scale = (65535.0 - red50) / rinc;  goto again; }
         if (red100 < 0) { scale = -1.0 * red50 / rinc; goto again; }
         if (green100 > 65535) { scale = (65535.0 - green50) / ginc; goto again; }
         if (green100 < 0) { scale = -1.0 * green50 / ginc; goto again; }
         if (blue100 > 65535) { scale = (65535.0 - blue50) / binc; goto again; }
         if (blue100 < 0) { scale = -1.0 * blue50 / binc; goto again; }

         if (spread < 0) {                                                 //  make mid-scale == original RGB
            red3 = int(rgb0 + spread1 * (red50 - rgb0));
            green3 = int(rgb0 + spread1 * (green50 - rgb0));
            blue3 = int(rgb0 + spread1 * (blue50 - rgb0));
         }
         else {
            red3 = int(red50 + spread2 * (red100 - red50));
            green3 = int(green50 + spread2 * (green100 - green50));
            blue3 = int(blue50 + spread2 * (blue100 - blue50));
         }
      }

/* ------------------------------------------------------------------------

      color balance curve values:
           0 = 0.5 * original color
          50 = unmodified
         100 = 1.5 * original color, clipped
*/

      brout = tune_dat[3][curveindex];
      if (brout < 49 || brout > 51) red3 = red3 * 0.01 * (brout + 50);
      brout = tune_dat[4][curveindex];
      if (brout < 49 || brout > 51) green3 = green3 * 0.01 * (brout + 50);
      brout = tune_dat[5][curveindex];
      if (brout < 49 || brout > 51) blue3 = blue3 * 0.01 * (brout + 50);


//    if working within a select area, blend changes over distance from edge
      
      double      dold, dnew;

      if (dist < sa_blend) {
         dnew = 1.0 * dist / sa_blend;
         dold = 1.0 - dnew;
         red3 = dnew * red3 + dold * red1;
         green3 = dnew * green3 + dold * green1;
         blue3 = dnew * blue3 + dold * blue1;
      }
      
      if (red3 > 65535) red3 = 65535;                                      //  clip overflows
      if (green3 > 65535) green3 = 65535;
      if (blue3 > 65535) blue3 = 65535;

      pix3[0] = int(red3);
      pix3[1] = int(green3);
      pix3[2] = int(blue3);
      
      RETURN
   }
}


/**************************************************************************/

//  red eye removal function

struct sredmem {                                                           //  red-eye struct in memory
   char        type, space[3];
   int         cx, cy, ww, hh, rad, clicks;
   double      thresh, tstep;
};
sredmem  redmem[100];                                                      //  store up to 100 red-eyes
int      Nredmem = 0, maxredmem = 100;


void m_redeye(GtkWidget *, const char *)
{
   void     redeye_mousefunc();
   int      redeye_dialog_compl(zdialog *zd, int zstat);

   const char  *redeye_message 
         = ZTX("Method 1:\n"
               "  Left-click on red-eye to darken.\n"
               "Method 2:\n"
               "  Drag down and right to enclose red-eye.\n"
               "  Left-click on red-eye to darken.\n"
               "Undo red-eye:\n"
               "  Right-click on red-eye.");

   if (! edit_setup(0)) return;                                            //  setup edit: no preview

   zdedit = zdialog_new(ZTX("red eye reduction"),mWin,Bdone,Bcancel,null);
   zdialog_add_widget(zdedit,"label","lab1","dialog",redeye_message);
   zdialog_run(zdedit,0,redeye_dialog_compl);                              //  run dialog, parallel mode

   Nredmem = 0;
   mouseCBfunc = redeye_mousefunc;                                         //  connect mouse function
   Mcapture = 1;                                                           //  capture mouse clicks
   return;
}


//  dialog completion callback function

int redeye_dialog_compl(zdialog *zd, int zstat)
{
   mouseCBfunc = 0;                                                        //  disconnect mouse
   Mcapture = 0;
   if (Nredmem > 0) Fmodified = 1;
   if (zstat == 1) edit_done();
   else edit_cancel();
   return 0;
}


//  mouse functions to define, darken, and undo red-eyes

int      redeye_createF(int px, int py);                                   //  create 1-click red-eye (type F)
int      redeye_createR(int px, int py, int ww, int hh);                   //  create robust red-eye (type R)
void     redeye_darken(int ii);                                            //  darken red-eye
void     redeye_distr(int ii);                                             //  build pixel redness distribution
int      redeye_find(int px, int py);                                      //  find red-eye at mouse position
void     redeye_remove(int ii);                                            //  remove red-eye at mouse position
int      redeye_radlim(int cx, int cy);                                    //  compute red-eye radius limit


void redeye_mousefunc()
{
   int         ii, px, py, ww, hh;

   if (Nredmem == maxredmem) {
      zmessageACK("%d red-eye limit reached",maxredmem);                   //  too many red-eyes
      return;
   }

   if (LMclick)                                                            //  left mouse click
   {
      LMclick = 0;

      px = Mxclick;                                                        //  click position
      py = Myclick;
      if (px < 0 || px > E3ww-1 || py < 0 || py > E3hh-1) return;          //  outside image area

      ii = redeye_find(px,py);                                             //  find existing red-eye
      if (ii < 0) ii = redeye_createF(px,py);                              //  or create new type F
      redeye_darken(ii);                                                   //  darken red-eye
      mwpaint2();
   }
   
   if (RMclick)                                                            //  right mouse click
   {
      RMclick = 0;
      px = Mxclick;                                                        //  click position
      py = Myclick;
      ii = redeye_find(px,py);                                             //  find red-eye
      if (ii >= 0) redeye_remove(ii);                                      //  if found, remove
      mwpaint2();
   }

   if (Mxdrag || Mydrag)                                                   //  mouse drag underway
   {
      px = Mxdown;                                                         //  initial position
      py = Mydown;
      ww = Mxdrag - Mxdown;                                                //  increment
      hh = Mydrag - Mydown;
      if (ww < 2 && hh < 2) return;
      if (ww < 2) ww = 2;
      if (hh < 2) hh = 2;
      if (px < 1) px = 1;                                                  //  keep within image area
      if (py < 1) py = 1;      
      if (px + ww > E3ww-1) ww = E3ww-1 - px;
      if (py + hh > E3hh-1) hh = E3hh-1 - py;
      ii = redeye_find(px,py);                                             //  find existing red-eye
      if (ii >= 0) redeye_remove(ii);                                      //  remove it
      mwpaint();                                                           //  avoid delayed update
      ii = redeye_createR(px,py,ww,hh);                                    //  create new red-eye type R
   }
}


//  create type F redeye (1-click automatic)

int redeye_createF(int cx, int cy)
{
   int         cx0, cy0, cx1, cy1, px, py, rad, radlim;
   int         loops, ii, sumx, sumy;
   int         Tnpix, Rnpix, R2npix, npix;
   double      rd, rcx, rcy, redpart;
   double      Tsum, Rsum, R2sum, Tavg, Ravg, R2avg;
   uint16      *ppix;
   
   cx0 = cx;
   cy0 = cy;
   
   for (loops = 0; loops < 5; loops++)
   {
      cx1 = cx;
      cy1 = cy;

      radlim = redeye_radlim(cx,cy);                                       //  radius limit (image edge)
      Tsum = Tavg = Ravg = Tnpix = 0;

      for (rad = 0; rad < radlim-2; rad++)                                 //  find red-eye radius from (cx,cy)
      {
         Rsum = Rnpix = 0;
         R2sum = R2npix = 0;

         for (py = cy-rad-2; py <= cy+rad+2; py++)
         for (px = cx-rad-2; px <= cx+rad+2; px++)
         {
            rd = sqrt((px-cx)*(px-cx) + (py-cy)*(py-cy));
            ppix = bmpixel(E3rgb48,px,py);
            redpart = redness(ppix);

            if (rd <= rad + 0.5 && rd > rad - 0.5) {                       //  accum. redness at rad
               Rsum += redpart;
               Rnpix++;
            }
            else if (rd <= rad + 2.5 && rd > rad + 1.5) {                  //  accum. redness at rad + 2
               R2sum += redpart;
               R2npix++;
            }
         }
         
         Tsum += Rsum;                                                     //  accum. redness over 0-rad
         Tnpix += Rnpix;
         Tavg = Tsum / Tnpix;
         Ravg = Rsum / Rnpix;
         R2avg = R2sum / R2npix;
         if (R2avg > Ravg || Ravg > Tavg) continue;
         if ((Ravg - R2avg) < 0.1 * (Tavg - Ravg)) break;
      }
      
      sumx = sumy = npix = 0;
      rad = int(1.2 * rad + 1);
      if (rad > radlim) rad = radlim;
      
      for (py = cy-rad; py <= cy+rad; py++)                                //  compute center of gravity for
      for (px = cx-rad; px <= cx+rad; px++)                                //   pixels within rad of (cx,cy)
      {                                                                    //    with high redness
         rd = sqrt((px-cx)*(px-cx) + (py-cy)*(py-cy));
         if (rd > rad + 0.5) continue;
         ppix = bmpixel(E3rgb48,px,py);
         if (redness(ppix) < Tavg) continue;
         sumx += (px - cx);
         sumy += (py - cy);
         npix++;
      }

      rcx = cx + 1.0 * sumx / npix;                                        //  new center of red-eye
      rcy = cy + 1.0 * sumy / npix;
      if (fabs(cx0 - rcx) > rad) break;                                    //  give up if big movement   v.6.9
      if (fabs(cy0 - rcy) > rad) break;
      cx = int(rcx + 0.5);
      cy = int(rcy + 0.5);
      if (cx == cx1 && cy == cy1) break;                                   //  done if no change
   }

   radlim = redeye_radlim(cx,cy);
   if (rad > radlim) rad = radlim;

   ii = Nredmem++;                                                         //  add red-eye to memory
   redmem[ii].type = 'F';
   redmem[ii].cx = cx;
   redmem[ii].cy = cy;
   redmem[ii].rad = rad;
   redmem[ii].clicks = 0;
   redmem[ii].thresh = 0;
   return ii;
}


//  create type R red-eye (drag an ellipse over red-eye area)

int redeye_createR(int cx, int cy, int ww, int hh)
{
   int      mpx, mpy, mww, mhh;
   int      rad, radlim;
   
   ww = 2 * ww;
   hh = 2 * hh;
   mpx = int((cx-ww/2) * Mscale + Morgx + 0.5);                            //  image to window space
   mpy = int((cy-hh/2) * Mscale + Morgy + 0.5);
   mww = int(ww * Mscale);
   mhh = int(hh * Mscale);
   gdk_draw_arc(WLAYOUT,gdkgc,0,mpx,mpy,mww,mhh,0,64*360);                 //  draw ellipse

   if (ww > hh) rad = ww / 2;
   else rad = hh / 2;
   radlim = redeye_radlim(cx,cy);
   if (rad > radlim) rad = radlim;

   int ii = Nredmem++;                                                     //  add red-eye to memory
   redmem[ii].type = 'R';
   redmem[ii].cx = cx;
   redmem[ii].cy = cy;
   redmem[ii].ww = ww;
   redmem[ii].hh = hh;
   redmem[ii].rad = rad;
   redmem[ii].clicks = 0;
   redmem[ii].thresh = 0;
   return ii;
}


//  darken a red-eye and increase click count

void redeye_darken(int ii)
{
   int         cx, cy, ww, hh, px, py, rad, clicks;
   double      rd, thresh, tstep;
   char        type;
   uint16      *ppix;

   type = redmem[ii].type;
   cx = redmem[ii].cx;
   cy = redmem[ii].cy;
   ww = redmem[ii].ww;
   hh = redmem[ii].hh;
   rad = redmem[ii].rad;
   thresh = redmem[ii].thresh;
   tstep = redmem[ii].tstep;
   clicks = redmem[ii].clicks++;
   
   if (thresh == 0)                                                        //  1st click 
   {
      redeye_distr(ii);                                                    //  get pixel redness distribution
      thresh = redmem[ii].thresh;                                          //  initial redness threshhold
      tstep = redmem[ii].tstep;                                            //  redness step size
   }

   tstep = (thresh - tstep) / thresh;                                      //  convert to reduction factor
   thresh = thresh * pow(tstep,clicks);                                    //  reduce threshhold by total clicks

   for (py = cy-rad; py <= cy+rad; py++)                                   //  darken pixels over threshhold
   for (px = cx-rad; px <= cx+rad; px++)
   {
      if (type == 'R') {
         if (px < cx - ww/2) continue;
         if (px > cx + ww/2) continue;
         if (py < cy - hh/2) continue;
         if (py > cy + hh/2) continue;
      }
      rd = sqrt((px-cx)*(px-cx) + (py-cy)*(py-cy));
      if (rd > rad + 0.5) continue;
      ppix = bmpixel(E3rgb48,px,py);
      if (redness(ppix) > thresh) setredness(ppix,thresh);                 //  set pixel redness = threshhold
   }

   return;
}


//  Build a distribution of redness for a red-eye. Use this information 
//  to set initial threshhold and step size for stepwise darkening.

void redeye_distr(int ii)
{
   int         cx, cy, ww, hh, rad, px, py;
   int         bin, npix, dbins[20], bsum, blim;
   double      rd, maxred, minred, redpart, dbase, dstep;
   char        type;
   uint16      *ppix;
   
   type = redmem[ii].type;
   cx = redmem[ii].cx;
   cy = redmem[ii].cy;
   ww = redmem[ii].ww;
   hh = redmem[ii].hh;
   rad = redmem[ii].rad;
   
   maxred = 0;
   minred = 100;

   for (py = cy-rad; py <= cy+rad; py++)
   for (px = cx-rad; px <= cx+rad; px++)
   {
      if (type == 'R') {
         if (px < cx - ww/2) continue;
         if (px > cx + ww/2) continue;
         if (py < cy - hh/2) continue;
         if (py > cy + hh/2) continue;
      }
      rd = sqrt((px-cx)*(px-cx) + (py-cy)*(py-cy));
      if (rd > rad + 0.5) continue;
      ppix = bmpixel(E3rgb48,px,py);
      redpart = redness(ppix);
      if (redpart > maxred) maxred = redpart;
      if (redpart < minred) minred = redpart;
   }
   
   dbase = minred;
   dstep = (maxred - minred) / 19.99;

   for (bin = 0; bin < 20; bin++) dbins[bin] = 0;
   npix = 0;

   for (py = cy-rad; py <= cy+rad; py++)
   for (px = cx-rad; px <= cx+rad; px++)
   {
      if (type == 'R') {
         if (px < cx - ww/2) continue;
         if (px > cx + ww/2) continue;
         if (py < cy - hh/2) continue;
         if (py > cy + hh/2) continue;
      }
      rd = sqrt((px-cx)*(px-cx) + (py-cy)*(py-cy));
      if (rd > rad + 0.5) continue;
      ppix = bmpixel(E3rgb48,px,py);
      redpart = redness(ppix);
      bin = int((redpart - dbase) / dstep);
      ++dbins[bin];
      ++npix;
   }
   
   bsum = 0;
   blim = int(0.5 * npix);

   for (bin = 0; bin < 20; bin++)                                          //  find redness level for 50% of
   {                                                                       //    pixels within red-eye radius
      bsum += dbins[bin];
      if (bsum > blim) break;
   }

   redmem[ii].thresh = dbase + dstep * bin;                                //  initial redness threshhold
   redmem[ii].tstep = dstep;                                               //  redness step (5% of range)   v.6.9

   return;
}


//  find a red-eye (nearly) overlapping the mouse click position

int redeye_find(int cx, int cy)
{
   for (int ii = 0; ii < Nredmem; ii++)
   {
      if (cx > redmem[ii].cx - 2 * redmem[ii].rad && 
          cx < redmem[ii].cx + 2 * redmem[ii].rad &&
          cy > redmem[ii].cy - 2 * redmem[ii].rad && 
          cy < redmem[ii].cy + 2 * redmem[ii].rad) 
            return ii;                                                     //  found
   }
   return -1;                                                              //  not found
}


//  remove a red-eye from memory

void redeye_remove(int ii)
{
   int      cx, cy, rad, px, py;
   uint16   *pix1, *pix3;

   cx = redmem[ii].cx;
   cy = redmem[ii].cy;
   rad = redmem[ii].rad;

   for (px = cx-rad; px <= cx+rad; px++)
   for (py = cy-rad; py <= cy+rad; py++)
   {
      pix1 = bmpixel(E1rgb48,px,py);
      pix3 = bmpixel(E3rgb48,px,py);
      pix3[0] = pix1[0];
      pix3[1] = pix1[1];
      pix3[2] = pix1[2];
   }
   
   for (ii++; ii < Nredmem; ii++) 
      redmem[ii-1] = redmem[ii];
   Nredmem--;
}


//  compute red-eye radius limit: smaller of 100 and nearest image edge

int redeye_radlim(int cx, int cy)
{
   int radlim = 100;
   if (cx < 100) radlim = cx;
   if (E3ww-1 - cx < 100) radlim = E3ww-1 - cx;
   if (cy < 100) radlim = cy;
   if (E3hh-1 - cy < 100) radlim = E3hh-1 - cy;
   return radlim;
}


/**************************************************************************/

//  image blur function 

int         blur_radius;
double      blur_weight[100][100];                                         //  up to blur radius = 99   v.6.3
int         blur_thread_busy[4];
int         blur_Npixels, blur_pixdone;


void m_blur(GtkWidget *, const char *)
{
   int    blur_dialog_event(zdialog *zd, const char *event);
   int    blur_dialog_compl(zdialog *zd, int zstat);
   void * blur_thread(void *);

   if (! edit_setup(0)) return;                                            //  setup edit: no preview

   zdedit = zdialog_new(ZTX("set blur radius"),mWin,Bdone,Bcancel,null);
   zdialog_add_widget(zdedit,"hbox","hb2","dialog",0,"space=10");
   zdialog_add_widget(zdedit,"label","labrad","hb2",ZTX("blur radius"),"space=5");
   zdialog_add_widget(zdedit,"spin","radius","hb2","0|99|1|1","space=5");
   zdialog_add_widget(zdedit,"button","apply","hb2",Bapply,"space=5");

   zdialog_run(zdedit,blur_dialog_event,blur_dialog_compl);                //  start dialog
   
   blur_radius = 1;
   start_thread(blur_thread,0);                                            //  start working thread
   return;
}


//  blur dialog event and completion callback functions

int blur_dialog_compl(zdialog * zd, int zstat)
{
   if (zstat == 1) edit_done();                                            //  done
   else edit_cancel();                                                     //  cancel or destroy
   return 0;
}


int blur_dialog_event(zdialog * zd, const char *event)
{
   if (strNeq(event,"apply")) return 0;

   zdialog_fetch(zd,"radius",blur_radius);                                 //  get blur radius

   if (blur_radius == 0) {
      if (Fmodified) edit_undo();                                          //  restore original image
      Fmodified = 0;
      return 0;
   }

   signal_thread();                                                        //  trigger working thread
   wait_thread_idle();                                                     //  wait for completion
   mwpaint2(); 
   return 1;
}


//  image blur thread function

void * blur_thread(void *)
{
   void * blur_thread4(void *arg);

   int            dx, dy, rad, rad2;
   double         m, d, w, sum;
   static int     index[4];
  
   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request

      rad = blur_radius;
      rad2 = rad * rad;

      for (dx = 0; dx <= rad; dx++)                                        //  clear weights array
      for (dy = 0; dy <= rad; dy++)
         blur_weight[dx][dy] = 0;

      for (dx = -rad; dx <= rad; dx++)                                     //  blur_weight[dx][dy] = no. of pixels
      for (dy = -rad; dy <= rad; dy++)                                     //    at distance (dx,dy) from center
         ++blur_weight[abs(dx)][abs(dy)];

      m = sqrt(rad2 + rad2);                                               //  corner pixel distance from center
      sum = 0;

      for (dx = 0; dx <= rad; dx++)                                        //  compute weight of pixel
      for (dy = 0; dy <= rad; dy++)                                        //    at distance dx, dy
      {
         d = sqrt(dx*dx + dy*dy);
         w = (m + 1 - d) / m;
         w = w * w;
         sum += blur_weight[dx][dy] * w;
         blur_weight[dx][dy] = w;
      }

      for (dx = 0; dx <= rad; dx++)                                        //  make weights add up to 1.0
      for (dy = 0; dy <= rad; dy++)
         blur_weight[dx][dy] = blur_weight[dx][dy] / sum;
         
      if (sa_exist) blur_Npixels = sa_npix;
      else  blur_Npixels = E3ww * E3hh;
      blur_pixdone = 0;

      for (int ii = 0; ii < 4; ii++) {                                     //  start 4 worker threads
         blur_thread_busy[ii] = 1;
         index[ii] = ii;
         start_detached_thread(blur_thread4,&index[ii]);
      }
      
      for (int ii = 0; ii < 4; ii++)                                       //  wait for all to complete
      while (blur_thread_busy[ii]) 
      {
         zsleep(0.01);
         edit_progress(blur_pixdone,blur_Npixels);
      }

      edit_progress(0,0);
      Fmodified = 1;
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  never executed, stop g++ warning
}


void * blur_thread4(void *arg)
{
   int            index = *((int *) arg);
   int            ii, jj, px, py, dist = 0, dx, dy, adx, ady, rad;
   double         red, green, blue;
   double         weight1, weight2, f1, f2;
   uint16         *pix1, *pix3, *pixN;
   
   GOFUNC_DATA(1)

   if (! sa_exist)                                                         //  process entire image
   {
      for (py = index; py < E3hh-1; py += 4)                               //  loop all image pixels
      for (px = 0; px < E3ww-1; px++)
         GOFUNC(blur_pixel)
   }

   if (sa_exist)                                                           //  process selected area
   {
      for (ii = index; ii < sa_npix; ii += 4)                              //  process all enclosed pixels
      {
         px = sa_pix[ii].px;
         py = sa_pix[ii].py;
         dist = sa_pix[ii].dist;
         GOFUNC(blur_pixel)
      }
   }

   blur_thread_busy[index] = 0;
   return 0;


blur_pixel:
   {
      pix1 = bmpixel(E1rgb48,px,py);                                       //  source pixel
      pix3 = bmpixel(E3rgb48,px,py);                                       //  target pixel
      
      rad = blur_radius;
      red = green = blue = 0;
      weight2 = 0.0;
      
      if (! sa_exist) 
      {
         for (dy = -rad; dy <= rad; dy++)                                  //  loop neighbor pixels within radius
         for (dx = -rad; dx <= rad; dx++)
         {
            if (px+dx < 0 || px+dx > E3ww-1) continue;                     //  omit pixels off edge   v.6.3
            if (py+dy < 0 || py+dy > E3hh-1) continue;
            adx = abs(dx);
            ady = abs(dy);
            pixN = pix1 + (dy * E3ww + dx) * 3;
            weight1 = blur_weight[adx][ady];                               //  weight at distance (dx,dy)
            weight2 += weight1;
            red += pixN[0] * weight1;                                      //  accumulate contributions
            green += pixN[1] * weight1;
            blue += pixN[2] * weight1;
         }

         red = red / weight2;                                              //  weighted average   v.6.3
         green = green / weight2;
         blue = blue / weight2;

         pix3[0] = int(red);
         pix3[1] = int(green);
         pix3[2] = int(blue);
      }
      
      if (sa_exist)
      {
         for (dy = -rad; dy <= rad; dy++)                                  //  loop neighbor pixels within radius
         for (dx = -rad; dx <= rad; dx++)
         {
            if (px+dx < 0 || px+dx > E3ww-1) continue;                     //  omit pixels off edge
            if (py+dy < 0 || py+dy > E3hh-1) continue;
            jj = (py+dy) * E3ww + (px+dx);
            if (! sa_inside[jj]) continue;                                 //  omit pixels outside area   v.6.3
            adx = abs(dx);
            ady = abs(dy);
            pixN = pix1 + (dy * E3ww + dx) * 3;
            weight1 = blur_weight[adx][ady];                               //  weight at distance (dx,dy)
            weight2 += weight1;
            red += pixN[0] * weight1;                                      //  accumulate contributions
            green += pixN[1] * weight1;
            blue += pixN[2] * weight1;
         }
         
         red = red / weight2;                                              //  weighted average
         green = green / weight2;
         blue = blue / weight2;

         if (dist < sa_blend) {                                            //  blend changes over sa_blend
            f1 = 1.0 * dist / sa_blend;
            f2 = 1.0 - f1;
            red = f1 * red + f2 * pix1[0];
            green = f1 * green + f2 * pix1[1];
            blue = f1 * blue + f2 * pix1[2];
         }

         pix3[0] = int(red);
         pix3[1] = int(green);
         pix3[2] = int(blue);
      }

      blur_pixdone++;
      RETURN
   }

   return 0;                                                               //  not executed, stop gcc warning
}


/**************************************************************************/

//  image sharpening function

int      sharpED_cycles;
int      sharpED_reduce;
int      sharpED_thresh;
int      sharpUM_radius;
int      sharpUM_amount;
int      sharpUM_thresh;
int      sharpLP_radius;
int      sharpLP_amount;
char     sharp_function[4];


void m_sharpen(GtkWidget *, const char *)
{
   int      sharp_dialog_event(zdialog *zd, const char *event);
   int      sharp_dialog_compl(zdialog *zd, int zstat);
   void *   sharp_thread(void *);

   if (! edit_setup(0)) return;                                            //  setup edit: no preview

   zdedit = zdialog_new(ZTX("sharpen image"),mWin,Bundo,Bdone,Bcancel,null); 

   zdialog_add_widget(zdedit,"hbox","hb2","dialog",0,"space=5");
   zdialog_add_widget(zdedit,"vbox","vb21","hb2",0,"space=5");
   zdialog_add_widget(zdedit,"vbox","vb22","hb2",0,"homog|space=5");
   zdialog_add_widget(zdedit,"vbox","vb23","hb2",0,"homog|space=5");
   zdialog_add_widget(zdedit,"button","ED","vb21",ZTX("edge detection"),"space=5");
   zdialog_add_widget(zdedit,"label","lab21","vb22",ZTX("cycles"));
   zdialog_add_widget(zdedit,"label","lab22","vb22",ZTX("reduce"));
   zdialog_add_widget(zdedit,"label","lab23","vb22",ZTX("threshold"));
   zdialog_add_widget(zdedit,"spin","cyclesED","vb23","1|30|1|10");
   zdialog_add_widget(zdedit,"spin","reduceED","vb23","50|95|1|80");
   zdialog_add_widget(zdedit,"spin","threshED","vb23","1|99|1|1");

   zdialog_add_widget(zdedit,"hsep","sep4","dialog");
   zdialog_add_widget(zdedit,"hbox","hb4","dialog",0,"space=5");
   zdialog_add_widget(zdedit,"vbox","vb41","hb4",0,"space=5");
   zdialog_add_widget(zdedit,"vbox","vb42","hb4",0,"homog|space=5");
   zdialog_add_widget(zdedit,"vbox","vb43","hb4",0,"homog|space=5");
   zdialog_add_widget(zdedit,"button","UM","vb41",ZTX("unsharp mask"),"space=5");
   zdialog_add_widget(zdedit,"label","lab41","vb42",ZTX("radius"));
   zdialog_add_widget(zdedit,"label","lab42","vb42",ZTX("amount"));
   zdialog_add_widget(zdedit,"label","lab43","vb42",ZTX("threshold"));
   zdialog_add_widget(zdedit,"spin","radiusUM","vb43","1|9|1|2");
   zdialog_add_widget(zdedit,"spin","amountUM","vb43","1|200|1|100");
   zdialog_add_widget(zdedit,"spin","threshUM","vb43","1|99|1|1");

   zdialog_add_widget(zdedit,"hsep","sep5","dialog");
   zdialog_add_widget(zdedit,"hbox","hb5","dialog",0,"space=5");
   zdialog_add_widget(zdedit,"vbox","vb51","hb5",0,"space=5");
   zdialog_add_widget(zdedit,"vbox","vb52","hb5",0,"homog|space=5");
   zdialog_add_widget(zdedit,"vbox","vb53","hb5",0,"homog|space=5");
   zdialog_add_widget(zdedit,"button","LP","vb51","Laplacian","space=5");
   zdialog_add_widget(zdedit,"label","lab51","vb52",ZTX("radius"));
   zdialog_add_widget(zdedit,"label","lab52","vb52",ZTX("amount"));
   zdialog_add_widget(zdedit,"spin","radiusLP","vb53","1|9|1|1");
   zdialog_add_widget(zdedit,"spin","amountLP","vb53","1|100|1|50");

   zdialog_run(zdedit,sharp_dialog_event,sharp_dialog_compl);              //  run dialog, parallel
   *sharp_function = 0;
   start_thread(sharp_thread,0);                                           //  start working thread
   return;
}


//  sharpen dialog event and completion callback functions

int sharp_dialog_compl(zdialog *zd, int zstat)
{
   if (zstat == 1) edit_undo();
   else if (zstat == 2) edit_done();
   else edit_cancel();
   return 0;
}


int sharp_dialog_event(zdialog *zd, const char *event)
{
   edit_undo();                                                            //  restore original image

   if (strcmpv(event,"ED","UM","LP",null))
   {
      zdialog_fetch(zd,"cyclesED",sharpED_cycles);                         //  get all input values
      zdialog_fetch(zd,"reduceED",sharpED_reduce);
      zdialog_fetch(zd,"threshED",sharpED_thresh);
      zdialog_fetch(zd,"radiusUM",sharpUM_radius);
      zdialog_fetch(zd,"amountUM",sharpUM_amount);
      zdialog_fetch(zd,"threshUM",sharpUM_thresh);
      zdialog_fetch(zd,"radiusLP",sharpLP_radius);
      zdialog_fetch(zd,"amountLP",sharpLP_amount);

      strcpy(sharp_function,event);                                        //  pass to working thread
      signal_thread();
      wait_thread_idle();
   }

   return 0;
}


//  sharpen image thread function

void * sharp_thread(void *)
{
   int    sharp_imageED();
   int    sharp_imageUM();
   int    sharp_imageLP();
   
   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request
      
      if (strEqu(sharp_function,"ED")) sharp_imageED();
      if (strEqu(sharp_function,"UM")) sharp_imageUM();                    //  do requested function
      if (strEqu(sharp_function,"LP")) sharp_imageLP();

      Fmodified = 1;
      mwpaint2(); 
   }

   return 0;                                                               //  never executed, stop g++ warning
}


//  image sharpen function by edge detection and compression

int sharp_imageED()
{
   int      sharp_thresh1 = 100;                                           //  initial threshold
   double   sharp_thresh2 = 0.01 * sharpED_reduce;                         //  decline rate
   int      ii, px, py, dist = 0, cycles, thresh;
   
   GOFUNC_DATA(1)
   
   thresh = sharp_thresh1;
    
   for (cycles = 0; cycles < sharpED_cycles; cycles++)
   {
      if (cycles > 0) thresh = int(thresh * sharp_thresh2);
      

      if (! sa_exist)                                                      //  process entire image
      {
         for (py = 2; py < E3hh-2; py++)
         for (px = 2; px < E3ww-2; px++)                                   //  loop all pixels
            GOFUNC(sharp_pixel)
      }

      if (sa_exist)                                                        //  selected area to sharpen
      {
         for (ii = 0; ii < sa_npix; ii++)                                  //  process all enclosed pixels
         {
            px = sa_pix[ii].px;
            py = sa_pix[ii].py;
            dist = sa_pix[ii].dist;
            if (px < 2 || px >= E3ww-2) continue;                          //  omit pixels on edge
            if (py < 2 || py >= E3hh-2) continue;
            GOFUNC(sharp_pixel)
         }
      }

      edit_progress(cycles,sharpED_cycles);                                //  v.6.3
   }

   edit_progress(0,0);
   return 1;

sharp_pixel:                                                               //  revised
   {
      int      dd, rgb, pthresh;
      int      dx[4] = { -1, 0, 1, 1 };                                    //  4 directions: NW N NE E
      int      dy[4] = { -1, -1, -1, 0 };
      int      pv2, pv2u, pv2d, pv2uu, pv2dd, pvdiff;
      uint16   *pix1, *pix3, *pix3u, *pix3d, *pix3uu, *pix3dd;
      double   f1, f2;
      
      pthresh = sharpED_thresh;                                            //  pthresh = larger
      if (thresh > pthresh) pthresh = thresh;

      pix1 = bmpixel(E1rgb48,px,py);                                       //  input pixel
      pix3 = bmpixel(E3rgb48,px,py);                                       //  output pixel

      for (dd = 0; dd < 4; dd++)                                           //  4 directions
      {
         pix3u = pix3 + (dy[dd] * E3ww + dx[dd]) * 3;                      //  upstream pixel
         pix3d = pix3 - (dy[dd] * E3ww - dx[dd]) * 3;                      //  downstream pixel

         for (rgb = 0; rgb < 3; rgb++)                                     //  loop 3 RGB colors
         {
            pv2 = pix3[rgb];
            pv2u = pix3u[rgb];                                             //  brightness difference
            pv2d = pix3d[rgb];                                             //    across target pixel

            pvdiff = pv2d - pv2u;
            if (pvdiff < 0) pvdiff = -pvdiff;
            if (pvdiff < 256 * pthresh) continue;                          //  brightness slope < threshold

            if (pv2u < pv2 && pv2 < pv2d)                                  //  slope up, monotone
            {
               pix3uu = pix3u + (dy[dd] * E3ww + dx[dd]) * 3;              //  upstream of upstream pixel
               pix3dd = pix3d - (dy[dd] * E3ww - dx[dd]) * 3;              //  downstream of downstream
               pv2uu = pix3uu[rgb];
               pv2dd = pix3dd[rgb];

               if (pv2uu >= pv2u) {                                        //  shift focus of changes to
                  pix3u = pix3;                                            //    avoid up/down/up jaggies
                  pv2u = pv2;
               }
               
               if (pv2dd <= pv2d) {
                  pix3d = pix3;
                  pv2d = pv2;
               }
                  
               if (pv2u > 256) pv2u -= 256;
               if (pv2d < 65279) pv2d += 256;
            }
            
            else if (pv2u > pv2 && pv2 > pv2d)                             //  slope down, monotone
            {
               pix3uu = pix3u + (dy[dd] * E3ww + dx[dd]) * 3;
               pix3dd = pix3d - (dy[dd] * E3ww - dx[dd]) * 3;
               pv2uu = pix3uu[rgb];
               pv2dd = pix3dd[rgb];

               if (pv2uu <= pv2u) {
                  pix3u = pix3;
                  pv2u = pv2;
               }
               
               if (pv2dd >= pv2d) {
                  pix3d = pix3;
                  pv2d = pv2;
               }

               if (pv2d > 256) pv2d -= 256;
               if (pv2u < 65279) pv2u += 256;
            }

            else continue;                                                 //  slope too small

            if (sa_exist && dist < sa_blend) {                             //  if area selection, blend pixel
               f1 = 1.0 * dist / sa_blend;                                 //    changes over sa_blend
               f2 = 1.0 - f1;
               pv2u = int(f1 * pv2u + f2 * pix1[rgb]);
               pv2d = int(f1 * pv2d + f2 * pix1[rgb]);
            }

            pix3u[rgb] = pv2u;                                             //  modified brightness values
            pix3d[rgb] = pv2d;                                             //    >> image3 pixel
         }
      }
      RETURN
   }
}


//  image sharpen function using unsharp mask

int sharp_imageUM()
{
   int         ii, px, py, dist = 0, dx, dy, rgb;
   int         incr, cval, mean;
   int         Npixels, Ndone;
   int         rad = sharpUM_radius;
   uint16      *pix1, *pix3, *pixN;
   double      rad2, f1, f2;
   
   GOFUNC_DATA(1)
   
   rad2 = 2 * rad + 1;
   rad2 = 1.0 / (rad2 * rad2);                                             //  1 / area of unsharp mask

   if (sa_exist) Npixels = sa_npix;
   else  Npixels = E3ww * E3hh;
   Ndone = 0;

   if (! sa_exist)                                                         //  process entire image
   {
      for (py = rad; py < E3hh-rad; py++)
      for (px = rad; px < E3ww-rad; px++)                                  //  loop all image3 pixels
      {
         GOFUNC(sharp_pixel)
         edit_progress(Ndone++,Npixels);                                   //  v.6.3
      }
   }

   if (sa_exist)                                                           //  selected area to sharpen
   {
      for (ii = 0; ii < sa_npix; ii++)                                     //  process all enclosed pixels
      {
         px = sa_pix[ii].px;
         py = sa_pix[ii].py;
         dist = sa_pix[ii].dist;
         if (px < rad || px >= E3ww-rad) continue;                         //  omit pixels on edge
         if (py < rad || py >= E3hh-rad) continue;
         GOFUNC(sharp_pixel)
         edit_progress(Ndone++,Npixels);                                   //  v.6.3
      }
   }

   edit_progress(0,0);
   return 1;

sharp_pixel:
   {
      pix1 = bmpixel(E1rgb48,px,py);                                       //  input pixel
      pix3 = bmpixel(E3rgb48,px,py);                                       //  output pixel

      for (rgb = 0; rgb < 3; rgb++)                                        //  loop 3 RGB colors
      {
         mean = 0;

         for (dy = -rad; dy <= rad; dy++)                                  //  loop neighbor pixels within radius
         for (dx = -rad; dx <= rad; dx++)                                  //  outer loop y
         {
            pixN = pix1 + (dy * E3ww + dx) * 3;                            //  compute mean brightness
            mean += pixN[rgb];
         }

         mean = int(mean * rad2 + 0.5);
         
         cval = pix3[rgb];                                                 //  pixel brightness
         incr = cval - mean;                                               //  - mean of neighbors
         if (abs(incr) < sharpUM_thresh) continue;                         //  if < threshold, skip
         incr = incr * sharpUM_amount / 100;                               //  reduce to sharp_amount
         cval = cval + incr;                                               //  add back to pixel
         if (cval < 0) cval = 0;
         if (cval > 65535) cval = 65535;
         
         if (sa_exist && dist < sa_blend) {                                //  if area selection, blend pixel
            f1 = 1.0 * dist / sa_blend;                                    //    changes over sa_blend
            f2 = 1.0 - f1;
            cval = int(f1 * cval + f2 * pix1[rgb]);
         }
         
         pix3[rgb] = cval;
      }
      
      RETURN
   }
}


//  image sharpen function using Lapacian method

int sharp_imageLP()
{
   int         ii, jj, px, py, dist = 0, dx, dy, rgb;
   int         rad = sharpLP_radius;
   int         sumpix[3], maxrgb;
   int         Npixels, Ndone;
   double      rad2, kern, kernsum;
   double      scale, scale1, scale2;
   double      kernel[19][19];
   double      f1, f2;
   uint16      *pix1, *pix3, *pixN;
   
   GOFUNC_DATA(1)
   
   rad2 = rad * rad;
   kernsum = 0;

   for (dy = -rad; dy <= rad; dy++)                                        //  kernel, Gaussian distribution
   for (dx = -rad; dx <= rad; dx++)
   {
      ii = dx + rad;
      jj = dy + rad;
      kern = (0.5 / rad2) * exp( -(dx*dx + dy*dy) / (2 * rad2));
      if (dx || dy) {
         kernel[ii][jj] = -kern;                                           //  surrounding cells < 0
         kernsum += kern;
      }
   }
   
   kernel[rad][rad] = kernsum + 1.0;                                       //  middle cell, total = +1
   
   scale1 = (100 - sharpLP_amount) / 100.0;

   if (sa_exist) Npixels = sa_npix;
   else  Npixels = E3ww * E3hh;
   Ndone = 0;

   if (! sa_exist)                                                         //  process entire image
   {
      for (py = rad; py < E3hh-rad; py++)
      for (px = rad; px < E3ww-rad; px++)                                  //  loop all image3 pixels
      {
         GOFUNC(sharp_pixel)
         edit_progress(Ndone++,Npixels);                                   //  v.6.3
      }
   }

   if (sa_exist)                                                           //  selected area to sharpen
   {
      for (ii = 0; ii < sa_npix; ii++)                                     //  process all enclosed pixels
      {
         px = sa_pix[ii].px;
         py = sa_pix[ii].py;
         dist = sa_pix[ii].dist;
         if (px < rad || px >= E3ww-rad) continue;                         //  omit pixels on edge
         if (py < rad || py >= E3hh-rad) continue;
         GOFUNC(sharp_pixel)
         edit_progress(Ndone++,Npixels);                                   //  v.6.3
      }
   }

   edit_progress(0,0);   
   return 1;

sharp_pixel:
   {
      pix1 = bmpixel(E1rgb48,px,py);                                       //  input pixel
      pix3 = bmpixel(E3rgb48,px,py);                                       //  output pixel

      sumpix[0] = sumpix[1] = sumpix[2] = 0;      

      for (dy = -rad; dy <= rad; dy++)                                     //  loop surrounding block of pixels
      for (dx = -rad; dx <= rad; dx++)                                     //  outer loop y
      {
         pixN = pix1 + (dy * E3ww + dx) * 3;
         kern = kernel[dx+rad][dy+rad];
         for (rgb = 0; rgb < 3; rgb++)
            sumpix[rgb] += int(kern * pixN[rgb] - 0.5);                    //  round
      }
      
      maxrgb = sumpix[0];
      if (sumpix[1] > maxrgb) maxrgb = sumpix[1];
      if (sumpix[2] > maxrgb) maxrgb = sumpix[2];
      
      if (maxrgb > 65535) scale = 65535.0 / maxrgb;
      else scale = 1.0;
      
      if (sumpix[0] < 0) sumpix[0] = 0;
      if (sumpix[1] < 0) sumpix[1] = 0;
      if (sumpix[2] < 0) sumpix[2] = 0;
      
      scale2 = (1.0 - scale1) * scale;

      pix3[0] = int(scale1 * pix1[0] + scale2 * sumpix[0]);
      pix3[1] = int(scale1 * pix1[1] + scale2 * sumpix[1]);
      pix3[2] = int(scale1 * pix1[2] + scale2 * sumpix[2]);
      
      if (sa_exist && dist < sa_blend) {                                   //  if area selection, blend pixel
         f1 = 1.0 * dist / sa_blend;                                       //    changes over sa_blend
         f2 = 1.0 - f1;
         pix3[0] = int(f1 * pix3[0] + f2 * pix1[0]);
         pix3[1] = int(f1 * pix3[1] + f2 * pix1[1]);
         pix3[2] = int(f1 * pix3[2] + f2 * pix1[2]);
      }

      RETURN
   }
}


/**************************************************************************/

//  image noise reduction

int      denoise_method = 5;
int      denoise_radius = 4;


void m_denoise(GtkWidget *, const char *)
{
   int denoise_dialog_event(zdialog *zd, const char *event);               //  dialog event function
   int denoise_dialog_compl(zdialog *zd, int zstat);                       //  dialog completion function
   void * denoise_thread(void *);

   const char  *denoise_message = ZTX(" Press the reduce button to \n"
                                      " reduce noise in small steps. \n"
                                      " Use reset to start over.");

   if (! edit_setup(0)) return;                                            //  setup edit: no preview

   zdedit = zdialog_new(ZTX("noise reduction"),mWin,Bdone,Bcancel,null);
   zdialog_add_widget(zdedit,"label","lab1","dialog",denoise_message,"space=5");
   zdialog_add_widget(zdedit,"hbox","hb1","dialog",0,"space=5");
   zdialog_add_widget(zdedit,"label","labalg","hb1",ZTX("algorithm"),"space=5");
   zdialog_add_widget(zdedit,"combo","method","hb1",0,"space=5|expand");
   zdialog_add_widget(zdedit,"hbox","hb2","dialog",0,"space=5");
   zdialog_add_widget(zdedit,"label","labrad","hb2",ZTX("radius"),"space=5");
   zdialog_add_widget(zdedit,"spin","radius","hb2","1|9|1|4","space=5");
   zdialog_add_widget(zdedit,"hbox","hb3","dialog",0,"space=5");
   zdialog_add_widget(zdedit,"button","reduce","hb3",ZTX("reduce"),"space=5");
   zdialog_add_widget(zdedit,"button","Undo","hb3",Bundo,"space=5");
   
   zdialog_cb_app(zdedit,"method","flatten outlyers by color");
   zdialog_cb_app(zdedit,"method","set median brightness by color");
   zdialog_cb_app(zdedit,"method","top hat filter by color");
   zdialog_stuff(zdedit,"method","top hat filter by color");

   zdialog_run(zdedit,denoise_dialog_event,denoise_dialog_compl);          //  run dialog
   start_thread(denoise_thread,0);                                         //  start working thread
   return;
}


//  denoise dialog event and completion callback functions

int denoise_dialog_compl(zdialog *zd, int zstat)
{
   if (zstat == 1) edit_done();                                            //  done
   else edit_cancel();                                                     //  cancel or destroy
   return 0;
}


int denoise_dialog_event(zdialog * zd, const char *event)
{
   char     method[40];

   if (strEqu(event,"radius")) 
      zdialog_fetch(zd,"radius",denoise_radius);

   if (strEqu(event,"method")) 
   {
      zdialog_fetch(zd,"method",method,39);

      if (strEqu(method,"flatten outlyers by color")) {
         denoise_method = 1;
         denoise_radius = 1;
      }

      if (strEqu(method,"set median brightness by color")) {
         denoise_method = 4;
         denoise_radius = 2;
      }

      if (strEqu(method,"top hat filter by color")) {
         denoise_method = 5;
         denoise_radius = 4;
      }
      
      zdialog_stuff(zd,"radius",denoise_radius);
   }
   
   if (strEqu(event,"reduce")) {
      signal_thread();                                                     //  trigger update thread
      wait_thread_idle();                                                  //  wait for thread done
   }

   if (strEqu(event,"Undo")) edit_undo();
   return 1;
}


//  image noise reduction thread

void * denoise_thread(void *)
{
   RGB         *E9rgb48;
   uint16      *pixN, *pix1, *pix3, *pix9;
   int         ii, px, py, dist, rad;
   int         Npixels, Ndone;
   double      f1, f2;

   GOFUNC_DATA(1)
   
   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request

      E9rgb48 = RGB_copy(E3rgb48);                                         //  copy image3 >> image9
                                                                           //  image3 is reference source
      rad = denoise_radius;                                                //  image9 will be modified

      if (sa_exist) Npixels = sa_npix;
      else  Npixels = E3ww * E3hh;
      Ndone = 0;

      if (sa_exist)
      {
         for (ii = 0; ii < sa_npix; ii++)                                  //  process pixels in select area
         {
            px = sa_pix[ii].px;
            py = sa_pix[ii].py;
            dist = sa_pix[ii].dist;
            if (px < rad || px >= E3ww-rad) continue;
            if (py < rad || py >= E3hh-rad) continue;
            pix1 = bmpixel(E1rgb48,px,py);                                 //  source pixel
            pix3 = bmpixel(E3rgb48,px,py);                                 //  source pixel
            pix9 = bmpixel(E9rgb48,px,py);                                 //  target pixel
            if (denoise_method == 1) GOFUNC(denoise_func_1)
            if (denoise_method == 4) GOFUNC(denoise_func_4)
            if (denoise_method == 5) GOFUNC(denoise_func_5)

            if (dist < sa_blend) {                                         //  blend changes over sa_blend
               f1 = 1.0 * dist / sa_blend;
               f2 = 1.0 - f1;
               pix9[0] = int(f1 * pix9[0] + f2 * pix1[0]);
               pix9[1] = int(f1 * pix9[1] + f2 * pix1[1]);
               pix9[2] = int(f1 * pix9[2] + f2 * pix1[2]);
            }

            edit_progress(Ndone++,Npixels);
         }
      }

      if (! sa_exist)
      {
         for (py = rad; py < E3hh-rad; py++)                               //  loop all image3 pixels
         for (px = rad; px < E3ww-rad; px++)
         {
            pix3 = bmpixel(E3rgb48,px,py);                                 //  source pixel
            pix9 = bmpixel(E9rgb48,px,py);                                 //  target pixel
            if (denoise_method == 1) GOFUNC(denoise_func_1)
            if (denoise_method == 4) GOFUNC(denoise_func_4)
            if (denoise_method == 5) GOFUNC(denoise_func_5)
            edit_progress(Ndone++,Npixels);
         }
      }
      
      edit_progress(0,0);

      mutex_lock(&bitmaps_lock);                                           //  v.6.3
      RGB_free(E3rgb48);                                                   //  image9 >> image3
      E3rgb48 = E9rgb48;
      mutex_unlock(&bitmaps_lock);

      Fmodified = 1;
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  never executed, stop g++ warning


//  noise reduction algorithms

   int         dy, dx;
   int         min0, min1, min2, max0, max1, max2;
   int         ns, rgb, bsortN[400];


//  flatten outlyers within radius, by color 

denoise_func_1:
   {
      min0 = min1 = min2 = 65535;
      max0 = max1 = max2 = 0;

      for (dy = -rad; dy <= rad; dy++)                                     //  loop surrounding pixels
      for (dx = -rad; dx <= rad; dx++)
      {
         if (dy == 0 && dx == 0) continue;                                 //  skip self

         pixN = pix3 + (dy * E3ww + dx) * 3;
         if (pixN[0] < min0) min0 = pixN[0];                               //  find min and max per color
         if (pixN[0] > max0) max0 = pixN[0];
         if (pixN[1] < min1) min1 = pixN[1];
         if (pixN[1] > max1) max1 = pixN[1];
         if (pixN[2] < min2) min2 = pixN[2];
         if (pixN[2] > max2) max2 = pixN[2];
      }
      
      if (pix3[0] <= min0 && min0 < 65279) pix9[0] = min0 + 256;           //  if outlier, flatten
      if (pix3[0] >= max0 && max0 > 256) pix9[0] = max0 - 256;
      if (pix3[1] <= min1 && min1 < 65279) pix9[1] = min1 + 256;
      if (pix3[1] >= max1 && max1 > 256) pix9[1] = max1 - 256;
      if (pix3[2] <= min2 && min2 < 65279) pix9[2] = min2 + 256;
      if (pix3[2] >= max2 && max2 > 256) pix9[2] = max2 - 256;
      
      RETURN
   }

//  use median brightness for pixels within radius

denoise_func_4:
   {
      for (rgb = 0; rgb < 3; rgb++)                                        //  loop all RGB colors
      {
         ns = 0;

         for (dy = -rad; dy <= rad; dy++)                                  //  loop surrounding pixels
         for (dx = -rad; dx <= rad; dx++)                                  //  get brightness values
         {
            pixN = pix3 + (dy * E3ww + dx) * 3;
            bsortN[ns] = pixN[rgb];
            ns++;
         }

         HeapSort(bsortN,ns);
         pix9[rgb] = bsortN[ns/2];                                         //  median brightness of 2*ns pixels
      }

      RETURN
   }

//  modified top hat filter: execute with increasing radius from 1 to limit
//  detect outlyer by comparing with pixels in outer radius

denoise_func_5:
   for (rad = 1; rad <= denoise_radius; rad++)
   for (int loops = 0; loops < 2; loops++)
   {
      min0 = min1 = min2 = 65535;
      max0 = max1 = max2 = 0;

      for (dy = -rad; dy <= rad; dy++)                                     //  loop all pixels within rad
      for (dx = -rad; dx <= rad; dx++)
      {
         if (dx > -rad && dx < rad) continue;                              //  skip inner pixels
         if (dy > -rad && dy < rad) continue;

         pixN = pix3 + (dy * E3ww + dx) * 3;
         if (pixN[0] < min0) min0 = pixN[0];                               //  find min and max per color
         if (pixN[0] > max0) max0 = pixN[0];                               //    among outermost pixels
         if (pixN[1] < min1) min1 = pixN[1];
         if (pixN[1] > max1) max1 = pixN[1];
         if (pixN[2] < min2) min2 = pixN[2];
         if (pixN[2] > max2) max2 = pixN[2];
      }
      
      if (pix3[0] < min0 && pix9[0] < 65279) pix9[0] += 256;               //  if central pixel is outlyer,
      if (pix3[0] > max0 && pix9[0] > 256) pix9[0] -= 256;                 //    moderate its values
      if (pix3[1] < min1 && pix9[1] < 65279) pix9[1] += 256;
      if (pix3[1] > max1 && pix9[1] > 256) pix9[1] -= 256;
      if (pix3[2] < min2 && pix9[2] < 65279) pix9[2] += 256;
      if (pix3[2] > max2 && pix9[2] > 256) pix9[2] -= 256;
   }

   RETURN
}


/**************************************************************************/

//  Resize or rescale image.
//
//  Output pixels are composites of input pixels, e.g. 2/3 size means 
//  that 3x3 input pixels are mapped into 2x2 output pixels, and an 
//  image size of 1000 x 600 becomes 667 x 400.


int      resize_ww0, resize_hh0, resize_ww1, resize_hh1;


void m_resize(GtkWidget *, const char *)
{
   int      resize_dialog_event(zdialog *zd, cchar * event);
   int      resize_dialog_compl(zdialog *zd, int zstat);
   void   * resize_thread(void *);

   const char  *title = ZTX("resize image");
   const char  *lockmess = ZTX("lock aspect ratio");

   if (! edit_setup(0)) return;                                            //  setup edit: no preview

   m_unselect(null,null);                                                  //  remove select area

   zdedit = zdialog_new(title,mWin,Bapply,Bcancel,null);
   zdialog_add_widget(zdedit,"hbox","hb1","dialog",0,"space=10");
   zdialog_add_widget(zdedit,"vbox","vb11","hb1",0,"homog|space=5");
   zdialog_add_widget(zdedit,"vbox","vb12","hb1",0,"homog|space=5");
   zdialog_add_widget(zdedit,"vbox","vb13","hb1",0,"homog|space=5");
   zdialog_add_widget(zdedit,"label","placeholder","vb11",0);              //             pixels       percent
   zdialog_add_widget(zdedit,"label","labw","vb11",Bwidth);                //    width    [______]     [______]
   zdialog_add_widget(zdedit,"label","labh","vb11",Bheight);               //    height   [______]     [______]
   zdialog_add_widget(zdedit,"label","labpix","vb12","pixels");            //
   zdialog_add_widget(zdedit,"spin","wpix","vb12","20|9999|1|0");          //    presets  [2/3] [1/2] [1/3] [1/4] 
   zdialog_add_widget(zdedit,"spin","hpix","vb12","20|9999|1|0");          //
   zdialog_add_widget(zdedit,"label","labpct","vb13",Bpercent);            //    [_] lock width/height ratio
   zdialog_add_widget(zdedit,"spin","wpct","vb13","1|500|0.1|100");        //
   zdialog_add_widget(zdedit,"spin","hpct","vb13","1|500|0.1|100");        //       [  done  ]  [ cancel ]  
   zdialog_add_widget(zdedit,"hbox","hb2","dialog",0,"space=5");
   zdialog_add_widget(zdedit,"label","preset","hb2",Bpreset,"space=5");
   zdialog_add_widget(zdedit,"button","b 3/4","hb2"," 3/4 ");
   zdialog_add_widget(zdedit,"button","b 2/3","hb2"," 2/3 ");
   zdialog_add_widget(zdedit,"button","b 1/2","hb2"," 1/2 ");
   zdialog_add_widget(zdedit,"button","b 1/3","hb2"," 1/3 ");
   zdialog_add_widget(zdedit,"button","b 1/4","hb2"," 1/4 ");
   zdialog_add_widget(zdedit,"check","lock","dialog",lockmess);

   resize_ww0 = Frgb48->ww;                                                //  original width, height
   resize_hh0 = Frgb48->hh;
   zdialog_stuff(zdedit,"wpix",resize_ww0);
   zdialog_stuff(zdedit,"hpix",resize_hh0);
   zdialog_stuff(zdedit,"lock",1);
   
   zdialog_run(zdedit,resize_dialog_event,resize_dialog_compl);            //  run dialog - parallel
   start_thread(resize_thread,0);                                          //  start working thread
   return;
}


//  resize dialog event function

int resize_dialog_event(zdialog *zd, const char * event)
{
   int         lock;
   double      wpct1, hpct1;

   zdialog_fetch(zd,"wpix",resize_ww1);                                    //  get all widget values
   zdialog_fetch(zd,"hpix",resize_hh1);
   zdialog_fetch(zd,"wpct",wpct1);
   zdialog_fetch(zd,"hpct",hpct1);
   zdialog_fetch(zd,"lock",lock);
   
   if (strEqu(event,"b 3/4")) {
      resize_ww1 = (3 * resize_ww0 + 3) / 4;
      resize_hh1 = (3 * resize_hh0 + 3) / 4;
   }
   
   if (strEqu(event,"b 2/3")) {
      resize_ww1 = (2 * resize_ww0 + 2) / 3;
      resize_hh1 = (2 * resize_hh0 + 2) / 3;
   }
   
   if (strEqu(event,"b 1/2")) {
      resize_ww1 = (resize_ww0 + 1) / 2;
      resize_hh1 = (resize_hh0 + 1) / 2;
   }
   
   if (strEqu(event,"b 1/3")) {
      resize_ww1 = (resize_ww0 + 2) / 3;
      resize_hh1 = (resize_hh0 + 2) / 3;
   }
   
   if (strEqu(event,"b 1/4")) {
      resize_ww1 = (resize_ww0 + 3) / 4;
      resize_hh1 = (resize_hh0 + 3) / 4;
   }

   if (strEqu(event,"wpct"))                                               //  width % - set pixel width
      resize_ww1 = int(wpct1 / 100.0 * resize_ww0 + 0.5);

   if (strEqu(event,"hpct"))                                               //  height % - set pixel height
      resize_hh1 = int(hpct1 / 100.0 * resize_hh0 + 0.5);
   
   if (lock && event[0] == 'w')                                            //  preserve width/height ratio
      resize_hh1 = int(resize_ww1 * (1.0 * resize_hh0 / resize_ww0) + 0.5);
   if (lock && event[0] == 'h') 
      resize_ww1 = int(resize_hh1 * (1.0 * resize_ww0 / resize_hh0) + 0.5);
   
   hpct1 = 100.0 * resize_hh1 / resize_hh0;                                //  set percents to match pixels
   wpct1 = 100.0 * resize_ww1 / resize_ww0;
   
   zdialog_stuff(zd,"wpix",resize_ww1);                                    //  index all widget values
   zdialog_stuff(zd,"hpix",resize_hh1);
   zdialog_stuff(zd,"wpct",wpct1);
   zdialog_stuff(zd,"hpct",hpct1);
   
   return 1;
}


//  resize dialog completion function

int resize_dialog_compl(zdialog *zd, int zstat)
{
   if (zstat == 1) {
      signal_thread();
      edit_done();
   }
   else edit_cancel();
   return 0;
}


//  resize image based on dialog controls.

void * resize_thread(void *)
{
   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request

      if (resize_ww1 != resize_ww0 || resize_hh1 != resize_hh0) 
      {                                                                    //  rescale to target size
         mutex_lock(&bitmaps_lock);
         RGB_free(E3rgb48);
         E3rgb48 = RGB_rescale(Frgb48,resize_ww1,resize_hh1);
         E3ww = resize_ww1;
         E3hh = resize_hh1;
         Fmodified = 1;
         mutex_unlock(&bitmaps_lock);
      }

      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  never executed, stop g++ warning
}


/**************************************************************************/

//  trim image - use mouse to select image region to retain

int      trimx1, trimy1, trimx2, trimy2;                                   //  trim rectangle
int      trimww, trimhh, trimwwp, trimhhp;
double   trimR;																				//  trim aspect ratio


void m_trim(GtkWidget *, const char *)
{
   void     trim_mousefunc();
   int      trim_dialog_compl(zdialog *zd, int zstat);
   void   * trim_thread(void *);

   const char     *trim_message = ZTX("click or drag trim margins");

   if (! edit_setup(0)) return;                                            //  setup edit: no preview

   m_unselect(null,null);                                                  //  remove select area

   zdedit = zdialog_new(ZTX("trim image"),mWin,ZTX("Trim"),Bcancel,null);  //  dialog to get user inuts
   zdialog_add_widget(zdedit,"label","lab1","dialog",trim_message);
   zdialog_add_widget(zdedit,"hsep","sep1","dialog");
   zdialog_add_widget(zdedit,"label","labw","dialog","width: xxx");
   zdialog_add_widget(zdedit,"label","labh","dialog","height: xxx");
   zdialog_add_widget(zdedit,"label","labr","dialog","ratio: x.xx");

   zdialog_run(zdedit,0,trim_dialog_compl);                                //  run dialog, parallel

   mouseCBfunc = trim_mousefunc;                                           //  connect mouse function
   Mcapture++;
   
   trimx1 = int(0.2 * E3ww);                                               //  start with 20% trim margins
   trimy1 = int(0.2 * E3hh);
   trimx2 = int(0.8 * E3ww);
   trimy2 = int(0.8 * E3hh);
   trimww = trimx2 - trimx1;
   trimhh = trimy2 - trimy1;
   trimwwp = trimhhp = 0;

   trim_mousefunc();                                                       //  draw dotted lines
   start_thread(trim_thread,0);                                            //  start working thread
   return;
}


//  trim dialog completion callback function

int trim_dialog_compl(zdialog *zd, int zstat)
{
   mouseCBfunc = 0;                                                        //  disconnect mouse
   Mcapture = 0;

   if (zstat == 1) {
      signal_thread();                                                     //  triger update thread
      edit_done();                                                         //  commit changes
   }

   else edit_cancel();
   return 0;
}


//  trim thread function

void * trim_thread(void *)
{
   int      px1, py1, px2, py2;
   uint16   *pix1, *pix3;

   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request

      if (trimx1 + trimww > E3rgb48->ww) trimww = E3ww - trimx1;
      if (trimy1 + trimhh > E3rgb48->hh) trimhh = E3hh - trimy1;
      
      mutex_lock(&bitmaps_lock);
      RGB_free(E3rgb48);
      E3rgb48 = RGB_make(trimww,trimhh,48);                                //  new bitmap with requested size
      E3ww = trimww;
      E3hh = trimhh;
      
      for (py1 = trimy1; py1 < trimy2; py1++)                              //  copy pixels
      for (px1 = trimx1; px1 < trimx2; px1++)
      {
         px2 = px1 - trimx1;
         py2 = py1 - trimy1;
         pix1 = bmpixel(E1rgb48,px1,py1);
         pix3 = bmpixel(E3rgb48,px2,py2);
         pix3[0] = pix1[0];
         pix3[1] = pix1[1];
         pix3[2] = pix1[2];
      }

      Fmodified = 1;
      mutex_unlock(&bitmaps_lock);
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  never executed, stop g++ warning
}


//  trim mouse function

void trim_mousefunc()
{
   int         mpx, mpy;
   char        text[20];
   
   if (LMclick || Mxdrag || Mydrag)                                        //  mouse click or drag
   {
      if (LMclick) {
         mpx = Mxclick;
         mpy = Myclick;
         LMclick = 0;
      }
      else {
         mpx = Mxdrag;
         mpy = Mydrag;
      }

      if (abs(mpx - trimx1) < abs(mpx - trimx2)) trimx1 = mpx;             //  new trim rectangle from mouse posn.
      else trimx2 = mpx;
      if (abs(mpy - trimy1) < abs(mpy - trimy2)) trimy1 = mpy;
      else trimy2 = mpy;

      if (trimx1 > trimx2-10) trimx1 = trimx2-10;                          //  sanity limits
      if (trimy1 > trimy2-10) trimy1 = trimy2-10;
      if (trimx1 < 0) trimx1 = 0;
      if (trimy1 < 0) trimy1 = 0;
      if (trimx2 > E3ww) trimx2 = E3ww;
      if (trimy2 > E3hh) trimy2 = E3hh;
      trimww = trimx2 - trimx1;
      trimhh = trimy2 - trimy1;
   }

   if (trimww == trimwwp && trimhh == trimhhp) return;                     //  detect changes
   trimwwp = trimww;
   trimhhp = trimhh;
   
   dotlinex1[0] = trimx1;                                                  //  update dotted rectangle   v.6.2
   dotliney1[0] = trimy1;
   dotlinex2[0] = trimx2;
   dotliney2[0] = trimy1;

   dotlinex1[1] = trimx2;
   dotliney1[1] = trimy1;
   dotlinex2[1] = trimx2;
   dotliney2[1] = trimy2;

   dotlinex1[2] = trimx2;
   dotliney1[2] = trimy2;
   dotlinex2[2] = trimx1;
   dotliney2[2] = trimy2;

   dotlinex1[3] = trimx1;
   dotliney1[3] = trimy2;
   dotlinex2[3] = trimx1;
   dotliney2[3] = trimy1;

   Ndotlines = 4;
   mwpaint2();
   
   if (trimww > trimhh) trimR = 1.0 * trimww / trimhh;                     //  update dialog data
   else  trimR = 1.0 * trimhh / trimww;
   snprintf(text,19,"width: %d",trimww);
   zdialog_stuff(zdedit,"labw",text);
   snprintf(text,19,"height: %d",trimhh);
   zdialog_stuff(zdedit,"labh",text);
   snprintf(text,19,"ratio: %.2f",trimR);
   zdialog_stuff(zdedit,"labr",text);
   
   return;
}


/**************************************************************************/

//  rotate image through any arbitrary angle

double      rotate_angle = 0;                                              //  E3 rotatation vs. F
double      rotate_delta = 0;
int         rotate_trim = 0;


void m_rotate(GtkWidget *, const char *menu)                               //  menu function
{
   int    rotate_dialog_event(zdialog *zd, const char * event);
   int    rotate_dialog_compl(zdialog *zd, int zstat);
   void * rotate_thread(void *);
   void   rotate_mousefunc();
   
   const char  *rotmess = "Use buttons or drag right edge with mouse.";

   m_unselect(null,null);                                                  //  remove select area first  v.6.2

   if (! edit_setup(1)) return;                                            //  setup edit: use preview   v.6.2

   zdedit = zdialog_new(ZTX("rotate image"),mWin,Bundo,Bdone,Bcancel,null);
   zdialog_add_widget(zdedit,"label","labrot","dialog",ZTX(rotmess),"space=5");
   zdialog_add_widget(zdedit,"label","labdeg","dialog",ZTX("degrees"),"space=5");
   zdialog_add_widget(zdedit,"hbox","hb1","dialog",0,"homog|space=5");
   zdialog_add_widget(zdedit,"vbox","vb1","hb1",0,"space=5");
   zdialog_add_widget(zdedit,"vbox","vb2","hb1",0,"space=5");
   zdialog_add_widget(zdedit,"vbox","vb3","hb1",0,"space=5");
   zdialog_add_widget(zdedit,"vbox","vb4","hb1",0,"space=5");
   zdialog_add_widget(zdedit,"button"," +0.1  ","vb1"," + 0.1 ");          //  button name is increment to use
   zdialog_add_widget(zdedit,"button"," -0.1  ","vb1"," - 0.1 ");
   zdialog_add_widget(zdedit,"button"," +1.0  ","vb2"," + 1   ");
   zdialog_add_widget(zdedit,"button"," -1.0  ","vb2"," - 1   ");
   zdialog_add_widget(zdedit,"button"," +10.0 ","vb3"," + 10  ");
   zdialog_add_widget(zdedit,"button"," -10.0 ","vb3"," - 10  ");
   zdialog_add_widget(zdedit,"button"," +90.0 ","vb4"," + 90  ");
   zdialog_add_widget(zdedit,"button"," -90.0 ","vb4"," - 90  ");
   zdialog_add_widget(zdedit,"hbox","hb2","dialog",0,"space=10");
   zdialog_add_widget(zdedit,"button","trim","hb2",ZTX("Trim"),"space=10");

   zdialog_run(zdedit,rotate_dialog_event,rotate_dialog_compl);            //  run dialog - parallel

   mouseCBfunc = rotate_mousefunc;                                         //  connect mouse function
   Mcapture++;

   gdk_window_set_cursor(drWin->window,dragcursor);                        //  set drag cursor
   rotate_angle = rotate_delta = rotate_trim = 0;
   start_thread(rotate_thread,0);                                          //  start working thread
   return;
}


//  rotate dialog event and completion callback functions

int rotate_dialog_event(zdialog *zd, const char * event)
{
   int         err;
   double      incr;
   char        text[20];
   
   if (strEqu(event,"trim")) {
      rotate_trim = 1;
   }

   if (strpbrk(event,"+-")) {
      err = convSD(event,incr);                                            //  button name is increment to use
      if (err) return 0;
      rotate_delta += incr;
   }

   zdialog_stuff(zd,"labdeg","computing");
   signal_thread();
   wait_thread_idle();

   sprintf(text,ZTX("degrees: %.1f"),rotate_angle);                        //  update dialog angle display
   zdialog_stuff(zd,"labdeg",text);
   return 1;
}


int rotate_dialog_compl(zdialog *zd, int zstat)
{
   if (zstat == 1) {                                                       //  undo, dialog stays active
      edit_undo();                                                         //     v.6.1.2
      rotate_angle = rotate_delta = rotate_trim = 0;
      return 0;
   }

   if (zstat == 2) {
      rotate_delta = rotate_angle;                                         //  rotate main image   v.6.2
      rotate_angle = 0;
      edit_done();
   }
 
   else edit_cancel();

   rotate_angle = rotate_delta = rotate_trim = 0;

   gdk_window_set_cursor(drWin->window,0);                                 //  restore normal cursor
   mouseCBfunc = 0;                                                        //  disconnect mouse
   Mcapture = 0;
   return 0;
}


//  rotate mouse function - drag right edge of image up/down for rotation

void rotate_mousefunc()
{
   static int     mpx0 = 0, mpy0 = 0;
   static int     mpy1, mpy2, dist;

   if (! Mxdrag && ! Mydrag) return;                                       //  no drag underway
   if (Mxdrag < 0.8 * E3ww) return;                                        //  not right edge of image

   if (Mxdown != mpx0 || Mydown != mpy0) {
      mpx0 = Mxdown;                                                       //  new drag started
      mpy0 = mpy1 = Mydown;
   }
   
   mpy2 = Mydrag;
   dist = mpy2 - mpy1;                                                     //  drag distance
   mpy1 = mpy2;                                                            //  reset origin for next time
   if (! dist) return;

   rotate_delta = 30.0 * dist / E3ww;                                      //  convert to angle
   rotate_dialog_event(zdedit,"mouse");
   return;
}


//  rotate thread function

void * rotate_thread(void *)
{
   int         px3, py3, px9, py9;
   int         wwcut, hhcut, ww, hh;
   double      radians;
   uint16      *pix3, *pix9;
   RGB         *E9rgb48;
   
   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request

      mutex_lock(&bitmaps_lock);

      if (rotate_delta) 
      {
         rotate_angle += rotate_delta;                                     //  accum. net rotation   
         rotate_delta = 0;                                                 //    from dialog widget
         
         if (rotate_angle >= 360) rotate_angle -=360;
         if (rotate_angle <= -360) rotate_angle +=360;
         if (fabs(rotate_angle) < 0.01) rotate_angle = 0;

         if (! rotate_angle) {
            RGB_free(E3rgb48);                                             //  E1 >> E3
            E3rgb48 = RGB_copy(E1rgb48);
            E3ww = E1ww;
            E3hh = E1hh;
            Fmodified = 0;
         }
         
         if (rotate_angle) {
            RGB_free(E3rgb48);
            E3rgb48 = RGB_rotate(E1rgb48,rotate_angle);                    //  E3 is rotated E1
            E3ww = E3rgb48->ww;
            E3hh = E3rgb48->hh;
            Fmodified = 1;
         }
      }

      if (rotate_trim)
      {                                                                    //  auto trim
         radians = fabs(rotate_angle / 57.296);
         wwcut = int(E3rgb48->hh * sin(radians) + 1);                      //  amount to trim
         hhcut = int(E3rgb48->ww * sin(radians) + 1);
         ww = E3rgb48->ww - 2 * wwcut;
         hh = E3rgb48->hh - 2 * hhcut;
         E9rgb48 = RGB_make(ww,hh,48);
         
         for (py3 = hhcut; py3 < E3hh-hhcut; py3++)                        //  E9 = trimmed E3
         for (px3 = wwcut; px3 < E3ww-wwcut; px3++)
         {
            px9 = px3 - wwcut;
            py9 = py3 - hhcut;
            pix3 = bmpixel(E3rgb48,px3,py3);
            pix9 = bmpixel(E9rgb48,px9,py9);
            pix9[0] = pix3[0];
            pix9[1] = pix3[1];
            pix9[2] = pix3[2];
         }

         RGB_free(E3rgb48);                                                //  E3 = E9
         E3rgb48 = E9rgb48;
         E3ww = ww;
         E3hh = hh;
      }
      
      mutex_unlock(&bitmaps_lock);
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  never executed, stop g++ warning
}


/**************************************************************************/

//  unbend an image
//  straighten curvature added by pano or improve perspective

int      unb_horz, unb_vert;                                               //  unbend values from dialog
double   unb_x1, unb_x2, unb_y1, unb_y2;                                   //  unbend axes scaled 0 to 1


//  menu function

void m_unbend(GtkWidget *, const char *)                                   //  overhaul for preview   v.6.2
{
   int      unbend_dialog_event(zdialog* zd, const char *event);
   int      unbend_dialog_compl(zdialog* zd, int zstat);
   void   * unbend_thread(void *);
   void     unbend_mousefunc();

   m_unselect(null,null);                                                  //  remove select area first

   if (! edit_setup(1)) return;                                            //  setup edit: preview

   zdedit = zdialog_new(ZTX("unbend panorama image"),mWin,Bdone,Bcancel,null);
   zdialog_add_widget(zdedit,"hbox","hb1","dialog",0,"space=10");
   zdialog_add_widget(zdedit,"vbox","vb1","hb1",0,"homog|space=10");
   zdialog_add_widget(zdedit,"vbox","vb2","hb1",0,"homog|space=10");
   zdialog_add_widget(zdedit,"spin","spvert","vb1","-30|30|1|0");
   zdialog_add_widget(zdedit,"spin","sphorz","vb1","-20|20|1|0");
   zdialog_add_widget(zdedit,"label","labvert","vb2",ZTX("vertical unbend"));
   zdialog_add_widget(zdedit,"label","labhorz","vb2",ZTX("horizontal unbend"));
   
   zdialog_resize(zdedit,260,0);
   zdialog_run(zdedit,unbend_dialog_event,unbend_dialog_compl);            //  run dialog, parallel

   unb_x1 = unb_x2 = unb_y1 = unb_y2 = 0.5;                                //  initial axes thru image middle
   unb_horz = unb_vert = 0;                                                //  v.6.3

   Mcapture = 1;
   mouseCBfunc = unbend_mousefunc;                                         //  connect mouse function
   
   start_thread(unbend_thread,0);                                          //  start working thread
   signal_thread();

   return;
}


//  dialog event and completion functions

int unbend_dialog_event(zdialog *zd, const char *event)
{
   zdialog_fetch(zd,"spvert",unb_vert);                                    //  get new unbend values
   zdialog_fetch(zd,"sphorz",unb_horz);
   signal_thread();                                                        //  trigger thread
   return 1;
}


int unbend_dialog_compl(zdialog *zd, int zstat)
{
   mouseCBfunc = 0;                                                        //  disconnect mouse
   Mcapture = 0;
   LMclick = RMclick = 0;

   if (zstat != 1) {
      edit_cancel();                                                       //  canceled
      return 0;
   }

   if (unb_vert || unb_horz) Fmodified = 1;                                //  image3 modified
   else Fmodified = 0;
   
   edit_done();                                                            //  commit changes to image3
   return 1;
}


//  unbend mouse function                                                  //  adjustable axes

void unbend_mousefunc()   
{
   const char  *close;
   double      dist1, dist2;
   double      mpx = 0, mpy = 0;
   
   if (Mxclick || Myclick) {                                               //  mouse clicked
      mpx = Mxclick;
      mpy = Myclick;
      Mxclick = Myclick = 0;
   }
   
   if (Mxdrag || Mydrag) {                                                 //  mouse dragged
      mpx = Mxdrag;
      mpy = Mydrag;
   }
   
   if (! mpx && ! mpy) return;

   mpx = 1.0 * mpx / E3ww;                                                 //  scale mouse position 0 to 1
   mpy = 1.0 * mpy / E3hh;

   if (mpx < 0.2 || mpx > 0.8 ) {                                          //  check reasonable position
      if (mpy < 0.1 || mpy > 0.9) return;
   }
   else if (mpy < 0.2 || mpy > 0.8) {
      if (mpx < 0.1 || mpx > 0.9) return;
   }
   else return;

   close = "?";                                                            //  find closest axis end-point
   dist1 = 2;

   dist2 = mpx * mpx + (mpy-unb_y1) * (mpy-unb_y1);
   if (dist2 < dist1) {
      dist1 = dist2;
      close = "left";
   }

   dist2 = (1-mpx) * (1-mpx) + (mpy-unb_y2) * (mpy-unb_y2);
   if (dist2 < dist1) {
      dist1 = dist2;
      close = "right";
   }

   dist2 = (mpx-unb_x1) * (mpx-unb_x1) + mpy * mpy;
   if (dist2 < dist1) {
      dist1 = dist2;
      close = "top";
   }

   dist2 = (mpx-unb_x2) * (mpx-unb_x2) + (1-mpy) * (1-mpy);
   if (dist2 < dist1) {
      dist1 = dist2;
      close = "bottom";
   }
   
   if (strEqu(close,"left")) unb_y1 = mpy;                                 //  set new axis end-point
   if (strEqu(close,"right")) unb_y2 = mpy;
   if (strEqu(close,"top")) unb_x1 = mpx;
   if (strEqu(close,"bottom")) unb_x2 = mpx;

   signal_thread();                                                        //  trigger thread 

   return ;
}


//  unbend thread function

void * unbend_thread(void *)
{
   int            vert, horz;
   int            px3, py3, cx3, cy3;
   int            hx1, hy1, hx2, hy2, vx1, vy1, vx2, vy2;
   double         px1, py1, dispx, dispx2, dispy;
   uint16        *pix1, *pix3;

   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request

      vert = int(unb_vert * 0.01 * E3hh);                                  //  convert % to pixels
      horz = int(unb_horz * 0.005 * E3ww);

      hx1 = 0;                                                             //  scale axes to E3ww/hh
      hy1 = unb_y1 * E3hh;
      hx2 = E3ww;
      hy2 = unb_y2 * E3hh;

      vx1 = unb_x1 * E3ww;
      vy1 = 0;
      vx2 = unb_x2 * E3ww;
      vy2 = E3hh;

      dotlinex1[0] = hx1;                                                  //  dotted lines on window
      dotliney1[0] = hy1;
      dotlinex2[0] = hx2;
      dotliney2[0] = hy2;

      dotlinex1[1] = vx1;
      dotliney1[1] = vy1;
      dotlinex2[1] = vx2;
      dotliney2[1] = vy2;
      
      Ndotlines = 2;

      for (py3 = 0; py3 < E3hh; py3++)                                     //  step through F3 pixels
      for (px3 = 0; px3 < E3ww; px3++)
      {
         pix3 = bmpixel(E3rgb48,px3,py3);                                  //  output pixel

         cx3 = vx1 + (vx2 - vx1) * py3 / E3hh;                             //  center of unbend
         cy3 = hy1 + (hy2 - hy1) * px3 / E3ww;
         dispx = 2.0 * (px3 - cx3) / E3ww;                                 //  -1.0 ..  0.0 .. +1.0 (roughly)
         dispy = 2.0 * (py3 - cy3) / E3hh;                                 //  -1.0 ..  0.0 .. +1.0
         dispx2 = dispx * dispx - 0.5;                                     //  +0.5 .. -0.5 .. +0.5  curved

         px1 = px3 + dispx * dispy * horz;                                 //  input virtual pixel, x
         py1 = py3 - dispy * dispx2 * vert;                                //  input virtual pixel, y
         pix1 = vpixel(E1rgb48,px1,py1);                                   //  input virtual pixel

         if (pix1) {
            pix3[0] = pix1[0];                                             //  input pixel >> output pixel
            pix3[1] = pix1[1];
            pix3[2] = pix1[2];
         }
         else pix3[0] = pix3[1] = pix3[2] = 0;
      }
      
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  never executed, stop g++ warning
}


/**************************************************************************/

//  warp/distort area - select image area and pull with mouse

float       *WarpAx, *WarpAy;                                              //  memory of all displaced pixels
float       WarpAmem[4][100];                                              //  undo memory, last 100 warps
int         NWarpA;                                                        //  WarpA mem count

void  WarpA_warpfunc(float wdx, float wdy, float wdw, float wdh, int acc);


//  menu function

void m_WarpA(GtkWidget *, const char *)
{
   int      WarpA_dialog_event(zdialog *zd, const char *event);
   int      WarpA_dialog_compl(zdialog *zd, int zstat);

   const char  *WarpA_message = 
         ZTX(" Select an area to warp using menu select function. \n"      //  v.6.3
             " Press [start warp] and pull area with mouse. \n"
             " Make multiple mouse pulls until satisfied. \n"
             " When finished, select another area or press [done]."); 
   
   int         px, py, ii;

   if (! edit_setup(0)) return;                                            //  setup edit: no preview

   zdedit = zdialog_new(ZTX("warp image in selected area"),mWin,Bdone,Bcancel,null);
   zdialog_add_widget(zdedit,"label","lab1","dialog",WarpA_message,"space=5");
   zdialog_add_widget(zdedit,"hbox","hb1","dialog",0,"space=10");
   zdialog_add_widget(zdedit,"button","swarp","hb1",ZTX("start warp"),"space=5");
   zdialog_add_widget(zdedit,"button","undlast","hb1",ZTX("undo last"),"space=5");
   zdialog_add_widget(zdedit,"button","undall","hb1",ZTX("undo all"),"space=5");

   zdialog_run(zdedit,WarpA_dialog_event,WarpA_dialog_compl);              //  run dialog

   WarpAx = (float *) zmalloc(E3ww * E3hh * sizeof(float));                //  get memory for pixel displacements
   WarpAy = (float *) zmalloc(E3ww * E3hh * sizeof(float));
   
   NWarpA = 0;                                                             //  no warp data

   for (py = 0; py < E3hh; py++)                                           //  no pixel displacements
   for (px = 0; px < E3ww; px++)
   {
      ii = py * E3ww + px;
      WarpAx[ii] = WarpAy[ii] = 0.0;
   }
   
   return;
}


//  warp dialog event and completion callback functions

int WarpA_dialog_compl(zdialog * zd, int zstat)
{
   if (NWarpA) Fmodified = 1;
   if (zstat == 1) edit_done();
   else edit_cancel();

   mouseCBfunc = 0;                                                        //  disconnect mouse
   Mcapture = 0;
   gdk_window_set_cursor(drWin->window,0);                                 //  restore normal cursor
   zfree(WarpAx);                                                          //  release undo memory
   zfree(WarpAy);
   return 0;
}


int WarpA_dialog_event(zdialog * zd, const char *event)
{
   void     WarpA_mousefunc(void);

   int         px, py, ii;
   float       wdx, wdy, wdw, wdh;
   uint16      *pix1, *pix3;

   if (strEqu(event,"swarp")) {                                            //  start warp
      if (! sa_exist) {
         zmessageACK(ZTX("must select area first"));
         return 0;
      }

      gdk_window_set_cursor(drWin->window,dragcursor);                     //  set drag cursor
      mouseCBfunc = WarpA_mousefunc;                                       //  connect mouse function
      Mcapture++;
   }

   if (strEqu(event,"undlast")) {
      if (NWarpA) {                                                        //  undo most recent warp
         ii = --NWarpA;
         wdx = WarpAmem[0][ii];
         wdy = WarpAmem[1][ii];
         wdw = WarpAmem[2][ii];
         wdh = WarpAmem[3][ii];
         WarpA_warpfunc(wdx,wdy,-wdw,-wdh,0);                              //  unwarp image
         WarpA_warpfunc(wdx,wdy,-wdw,-wdh,1);                              //  unwarp memory
      }
   }

   if (strEqu(event,"undall")) {                                           //  undo all warps
      for (ii = 0; ii < sa_npix; ii++)                                     //  process all enclosed pixels
      {
         px = sa_pix[ii].px;
         py = sa_pix[ii].py;
         pix1 = bmpixel(E1rgb48,px,py);                                    //  image1 pixel >> image3
         pix3 = bmpixel(E3rgb48,px,py);
         pix3[0] = pix1[0];
         pix3[1] = pix1[1];
         pix3[2] = pix1[2];
      }

      for (py = 0; py < E3hh; py++)                                        //  reset pixel displacements
      for (px = 0; px < E3ww; px++)
      {
         ii = py * E3ww + px;
         WarpAx[ii] = WarpAy[ii] = 0.0;
      }

      NWarpA = 0;                                                          //  erase undo memory
      mwpaint2(); 
   }

   return 1;
}


//  warp mouse function

void  WarpA_mousefunc(void)
{
   static float   wdx, wdy, wdw, wdh;
   static int     ii, warped = 0;

   if (Mxdrag || Mydrag)                                                   //  mouse drag underway
   {
      wdx = Mxdown;                                                        //  drag origin, image coordinates
      wdy = Mydown;
      wdw = Mxdrag - Mxdown;                                               //  drag increment
      wdh = Mydrag - Mydown;
      WarpA_warpfunc(wdx,wdy,wdw,wdh,0);                                   //  warp image
      warped = 1;
      return;
   }
   
   else if (warped) 
   {
      warped = 0;
      WarpA_warpfunc(wdx,wdy,wdw,wdh,1);                                   //  drag done, add to warp memory

      if (NWarpA == 100)                                                   //  if full, throw away oldest
      {
         NWarpA = 99;
         for (ii = 0; ii < NWarpA; ii++)
         {
            WarpAmem[0][ii] = WarpAmem[0][ii+1];
            WarpAmem[1][ii] = WarpAmem[1][ii+1];
            WarpAmem[2][ii] = WarpAmem[2][ii+1];
            WarpAmem[3][ii] = WarpAmem[3][ii+1];
         }
      }

      ii = NWarpA;
      WarpAmem[0][ii] = wdx;                                               //  save warp for undo
      WarpAmem[1][ii] = wdy;
      WarpAmem[2][ii] = wdw;
      WarpAmem[3][ii] = wdh;
      NWarpA++;
   }
   
   return;
}


//  warp image and accumulate warp memory

void  WarpA_warpfunc(float wdx, float wdy, float wdw, float wdh, int acc)
{
   int            ii, jj, px, py;
   double         ddx, ddy, dpe, dpm, mag, dispx, dispy;
   uint16         *pix1, *pix3;

   for (ii = 0; ii < sa_npix; ii++)                                        //  process all enclosed pixels
   {
      px = sa_pix[ii].px;
      py = sa_pix[ii].py;
      dpe = sa_pix[ii].dist;                                               //  distance from area edge

      ddx = (px - wdx);                                                    //  distance from drag origin
      ddy = (py - wdy);
      dpm = sqrt(ddx*ddx + ddy*ddy);

      if (dpm < 1) mag = 1;
      else mag = dpe / (dpe + dpm);                                        //  magnification, 0...1
      mag = mag * mag;

      dispx = -wdw * mag;                                                  //  warp = drag * magnification
      dispy = -wdh * mag;
      
      jj = py * E3ww + px;

      if (acc) {                                                           //  mouse drag done,
         WarpAx[jj] += dispx;                                              //    accumulate warp memory
         WarpAy[jj] += dispy;
         continue;
      }

      dispx += WarpAx[jj];                                                 //  add this warp to prior
      dispy += WarpAy[jj];

      pix1 = vpixel(E1rgb48,px+dispx,py+dispy);                            //  input virtual pixel    v.6.6
      if (! pix1) continue;
      pix3 = bmpixel(E3rgb48,px,py);                                       //  output pixel
      pix3[0] = pix1[0];
      pix3[1] = pix1[1];
      pix3[2] = pix1[2];
   }

   mwpaint2();                                                             //  update window
   return;
}


/**************************************************************************/

//  warp/distort whole image
//  fix perspective problems (e.g. curved walls, leaning buildings)

float       *WarpIx, *WarpIy;                                              //  memory of all dragged pixels
float       WarpImem[4][100];                                              //  undo memory, last 100 drags
int         NWarpI;                                                        //  WarpImem count
int         WarpIdrag;
int         WarpIww, WarpIhh;

void  WarpI_warpfunc(float wdx, float wdy, float wdw, float wdh, int acc);


//  menu function

void m_WarpI(GtkWidget *, const char *)                                    //  v.6.1
{
   int      WarpI_dialog_event(zdialog *zd, const char *event);
   int      WarpI_dialog_compl(zdialog *zd, int zstat);
   void     WarpI_mousefunc(void);

   const char  *WarpI_message = 
         ZTX(" Pull on an image edge using the mouse. \n"
             " Make multiple mouse pulls until satisfied. \n"
             " When finished, press [done]."); 
   
   int         px, py, ii;

   m_unselect(null,null);                                                  //  remove select area first  v.6.2

   if (! edit_setup(1)) return;                                            //  setup edit: use preview

   zdedit = zdialog_new(ZTX("Fix Image Perspective"),mWin,Bdone,Bcancel,null);
   zdialog_add_widget(zdedit,"label","lab1","dialog",WarpI_message,"space=5");
   zdialog_add_widget(zdedit,"hbox","hb1","dialog",0,"space=10");
   zdialog_add_widget(zdedit,"button","undlast","hb1",ZTX("undo last"),"space=5");
   zdialog_add_widget(zdedit,"button","undall","hb1",ZTX("undo all"),"space=5");

   zdialog_run(zdedit,WarpI_dialog_event,WarpI_dialog_compl);              //  run dialog

   NWarpI = WarpIdrag = 0;                                                 //  no drag data

   WarpIx = (float *) zmalloc(E3ww * E3hh * sizeof(float));                //  get memory for pixel displacements
   WarpIy = (float *) zmalloc(E3ww * E3hh * sizeof(float));
   
   for (py = 0; py < E3hh; py++)                                           //  no pixel displacements
   for (px = 0; px < E3ww; px++)
   {
      ii = py * E3ww + px;
      WarpIx[ii] = WarpIy[ii] = 0.0;
   }
   
   WarpIww = E3ww;                                                         //  preview dimensions
   WarpIhh = E3hh;
   
   gdk_window_set_cursor(drWin->window,dragcursor);                        //  set drag cursor
   mouseCBfunc = WarpI_mousefunc;                                          //  connect mouse function
   Mcapture++;

   return;
}


//  WarpI dialog event and completion callback functions

int WarpI_dialog_compl(zdialog * zd, int zstat)
{
   int         fpx, fpy, epx, epy, ii;
   double      scale, dispx, dispy;
   uint16      *pix1, *pix3;

   if (zstat != 1) edit_cancel();
   else if (NWarpI == 0) edit_cancel();
   
   else 
   {
      edit_fullsize(0);                                                    //  get full-size E1/E3

      scale = 1.0 * (E3ww + E3hh) / (WarpIww + WarpIhh);

      for (fpy = 0; fpy < E3hh; fpy++)                                     //  scale net pixel displacements
      for (fpx = 0; fpx < E3ww; fpx++)                                     //    to full image size
      {
         epx = WarpIww * fpx / E3ww;
         epy = WarpIhh * fpy / E3hh;
         ii = epy * WarpIww + epx;
         dispx = WarpIx[ii] * scale;
         dispy = WarpIy[ii] * scale;

         pix1 = vpixel(E1rgb48,fpx+dispx,fpy+dispy);                       //  input virtual pixel
         pix3 = bmpixel(E3rgb48,fpx,fpy);                                  //  output pixel
         if (! pix1) pix3[0] = pix3[1] = pix3[2] = 0;
         else {
            pix3[0] = pix1[0];
            pix3[1] = pix1[1];
            pix3[2] = pix1[2];
         }
      }

      edit_done();
   }

   mouseCBfunc = 0;                                                        //  disconnect mouse
   Mcapture = 0;
   gdk_window_set_cursor(drWin->window,0);                                 //  restore normal cursor
   zfree(WarpIx);                                                          //  release memory
   zfree(WarpIy);
   return 0;
}


int WarpI_dialog_event(zdialog * zd, const char *event)
{
   int         px, py, ii;
   float       wdx, wdy, wdw, wdh;
   
   if (strEqu(event,"undlast")) 
   {
      if (NWarpI == 1) event = "undall";
      else if (NWarpI) {                                                   //  undo most recent drag
         ii = --NWarpI;
         wdx = WarpImem[0][ii];
         wdy = WarpImem[1][ii];
         wdw = WarpImem[2][ii];
         wdh = WarpImem[3][ii];
         WarpI_warpfunc(wdx,wdy,-wdw,-wdh,0);                              //  undrag image
         WarpI_warpfunc(wdx,wdy,-wdw,-wdh,1);                              //  undrag memory
      }
   }

   if (strEqu(event,"undall"))                                             //  undo all drags
   {
      NWarpI = 0;                                                          //  erase undo memory

      for (py = 0; py < E3hh; py++)                                        //  reset pixel displacements
      for (px = 0; px < E3ww; px++)
      {
         ii = py * E3ww + px;
         WarpIx[ii] = WarpIy[ii] = 0.0;
      }

      edit_undo();                                                         //  restore image1
      Fmodified = 0;                                                       //  v.6.3
   }

   return 1;
}


//  WarpI mouse function

void  WarpI_mousefunc(void)
{
   static float   wdx, wdy, wdw, wdh;
   int            ii;

   if (Mxdrag || Mydrag)                                                   //  mouse drag underway
   {
      wdx = Mxdown * Mscale;                                               //  drag origin, window coordinates
      wdy = Mydown * Mscale;
      wdw = (Mxdrag - Mxdown) * Mscale;                                    //  drag increment
      wdh = (Mydrag - Mydown) * Mscale;
      WarpI_warpfunc(wdx,wdy,wdw,wdh,0);                                   //  drag image
      WarpIdrag = 1;
      return;
   }
   
   else if (WarpIdrag) 
   {
      WarpIdrag = 0;
      WarpI_warpfunc(wdx,wdy,wdw,wdh,1);                                   //  drag done, add to memory

      if (NWarpI == 100)                                                   //  if full, throw away oldest
      {
         NWarpI = 99;
         for (ii = 0; ii < NWarpI; ii++)
         {
            WarpImem[0][ii] = WarpImem[0][ii+1];
            WarpImem[1][ii] = WarpImem[1][ii+1];
            WarpImem[2][ii] = WarpImem[2][ii+1];
            WarpImem[3][ii] = WarpImem[3][ii+1];
         }
      }

      ii = NWarpI;
      WarpImem[0][ii] = wdx;                                               //  save drag for undo
      WarpImem[1][ii] = wdy;
      WarpImem[2][ii] = wdw;
      WarpImem[3][ii] = wdh;
      NWarpI++;
   }
   
   return;
}


//  warp image and accumulate warp memory

void  WarpI_warpfunc(float wdx, float wdy, float wdw, float wdh, int acc)
{
   int         ii, px, py;
   double      farw, farh, pwd, mag, dispx, dispy;
   double      farpw, farph, farp;
   uint16      *pix1, *pix3;

   farw = E3ww;                                                            //  farthest edges from (wdx,wdy)
   if (wdx > E3ww - wdx) farw = 0;
   farh = E3hh;
   if (wdy > E3hh - wdy) farh = 0;

   for (py = 0; py < E3hh; py++)                                           //  process all pixels
   for (px = 0; px < E3ww; px++)
   {
      farpw = abs(px - farw);                                              //  (px,py) distance to farw/farh
      farph = abs(py - farh);
      if (farpw < farph) farp = farpw;
      else farp = farph;

      pwd = sqrt((px-wdx)*(px-wdx) + (py-wdy)*(py-wdy));                   //  (px,py) distance to (wdx,wdy)

      mag = 1.0 - pwd / (pwd + farp);                                      //  relative movement for (px,py)
      dispx = -wdw * mag;                                                  //  displacement = drag * mag
      dispy = -wdh * mag;
      
      ii = py * E3ww + px;

      if (acc) {                                                           //  drag done, accumulate drag sum
         WarpIx[ii] += dispx;
         WarpIy[ii] += dispy;
         continue;
      }

      dispx += WarpIx[ii];                                                 //  add this drag to prior sum
      dispy += WarpIy[ii];

      pix1 = vpixel(E1rgb48,px+dispx,py+dispy);                            //  input virtual pixel
      pix3 = bmpixel(E3rgb48,px,py);                                       //  output pixel
      if (! pix1) pix3[0] = pix3[1] = pix3[2] = 0;
      else {
         pix3[0] = pix1[0];
         pix3[1] = pix1[1];
         pix3[2] = pix1[2];
      }
   }

   Fmodified = 1;                                                          //  v.6.3
   mwpaint2();                                                             //  update window
   return;
}


/**************************************************************************/

//  image color-depth reduction

int      color_depth = 16;                                                 //  bits per RGB color


void m_colordep(GtkWidget *, const char *)
{
   int      colordep_dialog_event(zdialog *zd, const char *event);
   int      colordep_dialog_compl(zdialog *zd, int zstat);
   void   * colordep_thread(void *);

   const char  *colmess = ZTX(" Set color depth to 1-16 bits");
   
   if (! edit_setup(1)) return;                                            //  setup edit: preview

   zdedit = zdialog_new(ZTX("Set Color Depth"),mWin,Bdone,Bcancel,null);
   zdialog_add_widget(zdedit,"label","lab1","dialog",colmess,"space=5");
   zdialog_add_widget(zdedit,"hbox","hb1","dialog",0,"space=10");
   zdialog_add_widget(zdedit,"spin","colors","hb1","1|16|1|16","space=5");
   zdialog_add_widget(zdedit,"button","Undo","hb1",Bundo,"space=5");
   zdialog_add_widget(zdedit,"button","Redo","hb1",Bredo,"space=5");

   zdialog_run(zdedit,colordep_dialog_event,colordep_dialog_compl);        //  run dialog - parallel
   
   color_depth = 16;
   start_thread(colordep_thread,0);                                        //  start working thread
   return;
}


//  colors dialog event and completion callback functions

int colordep_dialog_event(zdialog *zd, const char *event)
{
   if (strcmpv(event,"colors","blendwidth",0)) {
      zdialog_fetch(zd,"colors",color_depth);
      signal_thread();
   }
   
   if (strEqu(event,"Undo")) edit_undo();
   if (strEqu(event,"Redo")) edit_redo();

   return 0;
}


int colordep_dialog_compl(zdialog * zd, int zstat)
{
   if (zstat == 1) edit_done();
   else edit_cancel();
   return 0;
}


//  image color depth thread function

void * colordep_thread(void *)
{
   int         ii, px, py, rgb, dist;
   uint16      m1, m2, m3, val1, val3;
   uint16      *pix1, *pix3;
   double      f1, f2;
   
   while (true)
   {
      thread_idle_loop();
   
      m1 = 0xFFFF << (16 - color_depth);                                   //  5 > 1111100000000000
      m2 = 0x8000 >> color_depth;                                          //  5 > 0000010000000000
      m3 = 0xFFFF >> color_depth;                                          //  5 > 0000011111111111

      if (! sa_exist)                                                      //  process entire image
      {
         for (py = 0; py < E3hh; py++)
         for (px = 0; px < E3ww; px++)
         {
            pix1 = bmpixel(E1rgb48,px,py);                                 //  input pixel
            pix3 = bmpixel(E3rgb48,px,py);                                 //  output pixel
            
            for (rgb = 0; rgb < 3; rgb++)
            {
               val1 = pix1[rgb];
               val3 = val1 & m1;
               if (val1 & m2) val3 = val3 | m3;
               pix3[rgb] = val3;
            }
         }
      }

      if (sa_exist)                                                        //  process select area
      {
         for (ii = 0; ii < sa_npix; ii++)
         {
            px = sa_pix[ii].px;
            py = sa_pix[ii].py;
            dist = sa_pix[ii].dist;
            pix1 = bmpixel(E1rgb48,px,py);                                 //  input pixel
            pix3 = bmpixel(E3rgb48,px,py);                                 //  output pixel

            for (rgb = 0; rgb < 3; rgb++)
            {
               val1 = pix1[rgb];
               val3 = val1 & m1;
               if (val1 & m2) val3 = val3 | m3;
               if (dist < sa_blend) {                                      //  if area selection, blend pixel
                  f2 = 1.0 * dist / sa_blend;                              //    changes over distance sa_blend
                  f1 = 1.0 - f2;
                  val3 = int(f1 * val1 + f2 * val3);
               }
               pix3[rgb] = val3;
            }
         }
      }

      Fmodified = 1;
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  never executed, stop g++ warning
}


/**************************************************************************/

//  convert image to simulate a drawing

#define  drawNT 4                                                          //  no. of working threads

int      draw_done[drawNT];
double   draw_trfunc[256];
int      draw_contrast;
int      draw_threshold;
int      draw_pixcon;
int      draw_reverse;
uint8    *draw_pixcon_map = 0;

void m_draw(GtkWidget *, const char *)                                     //  v.6.7
{
   int    draw_dialog_event(zdialog* zd, const char *event);
   int    draw_dialog_compl(zdialog* zd, int zstat);
   void * draw_thread(void *);

   const char  *title = ZTX("Simulate Drawing");
   uint16      *pix1, *pix2;
   int         ii, px, py, qx, qy;
   int         red, green, blue, con, maxcon;

   if (! edit_setup(0)) return;                                            //  setup edit: no preview

   draw_pixcon_map = (uint8 *) zmalloc(E1ww*E1hh);                         //  set up pixel contrast map
   memset(draw_pixcon_map,0,E1ww*E1hh);

   for (py = 1; py < E1hh-1; py++)                                         //  scan image pixels
   for (px = 1; px < E1ww-1; px++)
   {
      pix1 = bmpixel(E1rgb48,px,py);                                       //  pixel at (px,py)
      red = pix1[0];                                                       //  pixel RGB levels
      green = pix1[1];
      blue = pix1[2];
      maxcon = 0;

      for (qy = py-1; qy < py+2; qy++)                                     //  loop 3x3 block of neighbor
      for (qx = px-1; qx < px+2; qx++)                                     //    pixels around pix1
      {
         pix2 = bmpixel(E1rgb48,qx,qy);                                    //  find max. contrast with
         con = abs(red-pix2[0]) + abs(green-pix2[1]) + abs(blue-pix2[2]);  //    neighbor pixel
         if (con > maxcon) maxcon = con;
      }

      ii = py * E1ww + px;
      draw_pixcon_map[ii] = (maxcon/3) >> 8;                               //  contrast for (px,py) 0-255
   }

   zdedit = zdialog_new(title,mWin,Bundo,Bredo,Bdone,Bcancel,null);        //  setup drawing dialog
   zdialog_add_widget(zdedit,"hbox","hb1","dialog",0,"space=10");
   zdialog_add_widget(zdedit,"vbox","vb1","hb1",0,"homog");
   zdialog_add_widget(zdedit,"vbox","vb2","hb1",0,"homog|expand");
   zdialog_add_widget(zdedit,"label","lab1","vb1",ZTX("contrast"));
   zdialog_add_widget(zdedit,"label","lab2","vb1",ZTX("threshold"));
   zdialog_add_widget(zdedit,"label","lab3","vb1",ZTX("outlines"));
   zdialog_add_widget(zdedit,"hscale","contrast","vb2","0|100|1|0","expand");
   zdialog_add_widget(zdedit,"hscale","threshold","vb2","0|100|1|0","expand");
   zdialog_add_widget(zdedit,"hscale","pixcon","vb2","0|255|1|0","expand");
   zdialog_add_widget(zdedit,"hbox","hb4","dialog");
   zdialog_add_widget(zdedit,"radio","pencil","hb4",ZTX("pencil"),"space=10");
   zdialog_add_widget(zdedit,"radio","chalk","hb4",ZTX("chalk"),"space=10");

   zdialog_run(zdedit,draw_dialog_event,draw_dialog_compl);                //  run dialog - parallel
   
   start_thread(draw_thread,0);                                            //  start working thread
   return;
}


//  draw dialog event and completion functions

int draw_dialog_event(zdialog *zd, const char *event)                      //  draw dialog event function
{
   zdialog_fetch(zd,"contrast",draw_contrast);                             //  get slider values
   zdialog_fetch(zd,"threshold",draw_threshold);
   zdialog_fetch(zd,"pixcon",draw_pixcon);
   zdialog_fetch(zd,"chalk",draw_reverse);
   signal_thread();                                                        //  trigger update thread
   return 1;
}


int draw_dialog_compl(zdialog *zd, int zstat)                              //  draw dialog completion function
{
   if (zstat == 1) { edit_undo(); return 0; }                              //  undo
   else if (zstat == 2) { edit_redo(); return 0; }                         //  redo
   else if (zstat == 3) edit_done();                                       //  done
   else edit_cancel();                                                     //  cancel or destroy
   zfree(draw_pixcon_map);
   return 0;
}


//  thread function - use multiple working threads

void * draw_thread(void *)
{
   void  * draw_thread4(void *arg);

   int         ii, index[drawNT];
   double      threshold, contrast, trf;
   
   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request

      threshold = 0.01 * draw_threshold;                                   //  range 0 to 1
      contrast = 0.01 * draw_contrast;                                     //  range 0 to 1
      
      for (ii = 0; ii < 256; ii++)                                         //  brightness transfer function
      {
         trf = 1.0 - 0.003906 * (256 - ii) * contrast;                     //  ramp-up from 0-1 to 1
         if (ii < 256 * threshold) trf = 0;                                //  0 if below threshold
         draw_trfunc[ii] = trf;
      }
      
      for (ii = 0; ii < drawNT; ii++)                                      //  start drawNT threads
      {
         draw_done[ii] = 0;
         index[ii] = ii;
         start_detached_thread(draw_thread4,&index[ii]);
      }

      for (ii = 0; ii < drawNT; ii++)                                      //  wait for all done
         while (draw_done[ii] == 0) zsleep(0.01);

      Fmodified = 1;
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  never executed, stop g++ warning
}


void * draw_thread4(void *arg)                                             //  thread function
{
   int         index = *((int *) (arg));
   int         px, py, ii, dist;
   double      pixcon;

   GOFUNC_DATA(1)

   pixcon = draw_pixcon / 255.0;                                           //  0-1 linear ramp
   pixcon = 255 * pixcon * pixcon * pixcon;                                //  0-255 cubic ramp

   if (sa_exist)                                                           //  process selected area
   {
      for (ii = index; ii < sa_npix; ii += drawNT)                         //  process all enclosed pixels
      {
         px = sa_pix[ii].px;
         py = sa_pix[ii].py;
         dist = sa_pix[ii].dist;
         GOFUNC(draw_1pix)
      }
   }

   else
   {
      dist = sa_blend = 0;
      for (py = index; py < E1hh; py += drawNT)                            //  process all pixels
      for (px = 0; px < E1ww; px++)
         GOFUNC(draw_1pix)
   }

   draw_done[index] = 1;
   pthread_exit(0);

draw_1pix:                                                                 //  process one pixel
   {
      uint16      *pix1, *pix3;
      int         bright1, bright2;
      int         red1, green1, blue1;
      int         red3, green3, blue3;
      double      dold, dnew;
      
      if (! px || ! py) RETURN
      if (px > E1ww-2 || py > E1hh-2) RETURN
      
      pix1 = bmpixel(E1rgb48,px,py);                                       //  input pixel
      pix3 = bmpixel(E3rgb48,px,py);                                       //  output pixel
         
      red1 = pix1[0];
      green1 = pix1[1];
      blue1 = pix1[2];
      
      bright1 = ((red1 + green1 + blue1) / 3) >> 8;                        //  old brightness  0-255
      bright2 = bright1 * draw_trfunc[bright1];                            //  new brightness  0-255
      
      ii = py * E1ww + px;
      if (draw_pixcon_map[ii] < pixcon) bright2 = 255;

      if (pixcon > 1 && bright2 > draw_threshold) bright2 = 255;           //  empirical !!!

      if (draw_reverse) bright2 = 255 - bright2;                           //  negate if "chalk"

      red3 = green3 = blue3 = bright2 << 8;                                //  gray scale, new brightness

      if (dist < sa_blend) {                                               //  blend over distance sa_blend
         dnew = 1.0 * dist / sa_blend;
         dold = 1.0 - dnew;
         red3 = dnew * red3 + dold * red1;
         green3 = dnew * green3 + dold * green1;
         blue3 = dnew * blue3 + dold * blue1;
      }

      pix3[0] = red3;
      pix3[1] = green3;
      pix3[2] = blue3;
      
      RETURN
   }
}


/**************************************************************************/

//  convert image to simulate an embossing

#define  embossNT 4                                                        //  no. of working threads

int      emboss_done[embossNT];
int      emboss_radius, emboss_color;
double   emboss_depth;
double   emboss_kernel[20][20];                                            //  support radius <= 9

void m_emboss(GtkWidget *, const char *)                                   //  v.6.7
{
   int    emboss_dialog_event(zdialog* zd, const char *event);
   int    emboss_dialog_compl(zdialog* zd, int zstat);
   void * emboss_thread(void *);

   const char  *title = ZTX("Simulate embossing");

   if (! edit_setup(0)) return;                                            //  setup edit: no preview

   zdedit = zdialog_new(title,mWin,Bundo,Bredo,Bdone,Bcancel,null);        //  setup embossing dialog
   zdialog_add_widget(zdedit,"hbox","hb1","dialog",0,"space=10");
   zdialog_add_widget(zdedit,"label","lab1","hb1",ZTX("radius"),"space=5");
   zdialog_add_widget(zdedit,"spin","radius","hb1","0|9|1|0");
   zdialog_add_widget(zdedit,"label","lab2","hb1",ZTX("depth"),"space=5");
   zdialog_add_widget(zdedit,"spin","depth","hb1","0|99|1|0");
   zdialog_add_widget(zdedit,"check","color","hb1",ZTX("color"),"space=8");

   zdialog_run(zdedit,emboss_dialog_event,emboss_dialog_compl);            //  run dialog - parallel
   
   start_thread(emboss_thread,0);                                          //  start working thread
   return;
}


//  emboss dialog event and completion functions

int emboss_dialog_event(zdialog *zd, const char *event)                    //  emboss dialog event function
{
   zdialog_fetch(zd,"radius",emboss_radius);                               //  get user inputs
   zdialog_fetch(zd,"depth",emboss_depth);
   zdialog_fetch(zd,"color",emboss_color);
   signal_thread();                                                        //  trigger update thread
   return 1;
}


int emboss_dialog_compl(zdialog *zd, int zstat)                            //  emboss dialog completion function
{
   if (zstat == 1) edit_undo();                                            //  undo
   else if (zstat == 2) edit_redo();                                       //  redo
   else if (zstat == 3) edit_done();                                       //  done
   else edit_cancel();                                                     //  cancel or destroy
   return 0;
}


//  thread function - use multiple working threads

void * emboss_thread(void *)
{
   void  * emboss_thread4(void *arg);

   int         ii, index[embossNT];
   int         dx, dy, rad;
   double      depth, kern, coeff;
   
   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request
      
      rad = emboss_radius;
      depth = emboss_depth;

      coeff = 0.1 * depth / (rad * rad + 1);

      for (dy = -rad; dy <= rad; dy++)                                     //  build kernel with radius and depth
      for (dx = -rad; dx <= rad; dx++)
      {
         kern = coeff * (dx + dy);
         emboss_kernel[dx+rad][dy+rad] = kern;
      }
      
      emboss_kernel[rad][rad] = 1;                                         //  kernel center cell = 1

      for (ii = 0; ii < embossNT; ii++)                                    //  start embossNT threads
      {
         emboss_done[ii] = 0;
         index[ii] = ii;
         start_detached_thread(emboss_thread4,&index[ii]);
      }

      for (ii = 0; ii < embossNT; ii++)                                    //  wait for all done
         while (emboss_done[ii] == 0) zsleep(0.01);

      Fmodified = 1;
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  never executed, stop g++ warning
}


void * emboss_thread4(void *arg)                                           //  thread function
{
   int         index = *((int *) (arg));
   int         px, py, ii, dist;

   GOFUNC_DATA(1)

   if (sa_exist)                                                           //  process selected area
   {
      for (ii = index; ii < sa_npix; ii += embossNT)                       //  process all enclosed pixels
      {
         px = sa_pix[ii].px;
         py = sa_pix[ii].py;
         dist = sa_pix[ii].dist;
         GOFUNC(emboss_1pix)
      }
   }

   else
   {
      dist = sa_blend = 0;
      for (py = index; py < E1hh; py += embossNT)                          //  process all pixels
      for (px = 0; px < E1ww; px++)
         GOFUNC(emboss_1pix)
   }

   emboss_done[index] = 1;
   pthread_exit(0);

emboss_1pix:                                                               //  process one pixel
   {
      uint16      *pix1, *pix3, *pixN;
      int         bright1, bright3;
      int         rgb, dx, dy, rad;
      double      sumpix, kern, dold, dnew;
      
      rad = emboss_radius;

      if (px < rad || py < rad) RETURN
      if (px > E3ww-rad-1 || py > E3hh-rad-1) RETURN
      
      pix1 = bmpixel(E1rgb48,px,py);                                       //  input pixel
      pix3 = bmpixel(E3rgb48,px,py);                                       //  output pixel
      
      if (emboss_color)                                                    //  keep color   v.6.9
      {
         for (rgb = 0; rgb < 3; rgb++)
         {      
            sumpix = 0;
            
            for (dy = -rad; dy <= rad; dy++)                               //  loop surrounding block of pixels
            for (dx = -rad; dx <= rad; dx++)
            {
               pixN = pix1 + (dy * E1ww + dx) * 3;
               kern = emboss_kernel[dx+rad][dy+rad];
               sumpix += kern * pixN[rgb];
         
               bright1 = pix1[rgb];
               bright3 = sumpix;
               if (bright3 < 0) bright3 = 0;
               if (bright3 > 65535) bright3 = 65535;

               if (dist < sa_blend) {                                      //  blend over distance sa_blend
                  dnew = 1.0 * dist / sa_blend;
                  dold = 1.0 - dnew;
                  bright3 = dnew * bright3 + dold * bright1;
               }

               pix3[rgb] = bright3;
            }
         }
      }
      
      else                                                                 //  use gray scale
      {
         sumpix = 0;
            
         for (dy = -rad; dy <= rad; dy++)                                  //  loop surrounding block of pixels
         for (dx = -rad; dx <= rad; dx++)
         {
            pixN = pix1 + (dy * E1ww + dx) * 3;
            kern = emboss_kernel[dx+rad][dy+rad];
            sumpix += kern * (pixN[0] + pixN[1] + pixN[2]);
         }
         
         bright1 = 0.3333 * (pix1[0] + pix1[1] + pix1[2]);
         bright3 = 0.3333 * sumpix;
         if (bright3 < 0) bright3 = 0;
         if (bright3 > 65535) bright3 = 65535;
         
         if (dist < sa_blend) {                                            //  blend over distance sa_blend
            dnew = 1.0 * dist / sa_blend;
            dold = 1.0 - dnew;
            bright3 = dnew * bright3 + dold * bright1;
         }

         pix3[0] = pix3[1] = pix3[2] = bright3;
      }

      RETURN
   }
}


/**************************************************************************/

//  make simulated tiles function 

int         tile_size, tile_gap;
int         tile_Npixels, tile_pixdone;
uint16      *tile_pixmap = 0;


void m_tiles(GtkWidget *, const char *)                                    //  new  v.6.8
{
   int    tile_dialog_event(zdialog *zd, const char *event);
   int    tile_dialog_compl(zdialog *zd, int zstat);
   void * tile_thread(void *);

   if (! edit_setup(0)) return;                                            //  setup edit: no preview

   zdedit = zdialog_new(ZTX("set tile and gap size"),mWin,Bdone,Bcancel,null);
   zdialog_add_widget(zdedit,"hbox","hb1","dialog",0,"space=5");
   zdialog_add_widget(zdedit,"label","labt","hb1",ZTX("tile size"),"space=5");
   zdialog_add_widget(zdedit,"spin","size","hb1","1|99|1|5","space=5");
   zdialog_add_widget(zdedit,"button","apply","hb1",Bapply,"space=10");
   zdialog_add_widget(zdedit,"hbox","hb2","dialog",0,"space=5");
   zdialog_add_widget(zdedit,"label","labg","hb2",ZTX("tile gap"),"space=5");
   zdialog_add_widget(zdedit,"spin","gap","hb2","0|9|1|1","space=5");

   zdialog_run(zdedit,tile_dialog_event,tile_dialog_compl);                //  start dialog

   tile_size = 5;
   tile_gap = 1;

   tile_pixmap = (uint16 *) zmalloc(E1ww*E1hh*6);                          //  set up pixel color map
   memset(tile_pixmap,0,E1ww*E1hh*6);

   start_thread(tile_thread,0);                                            //  start working thread
   return;
}


//  tiles dialog event and completion callback functions

int tile_dialog_compl(zdialog * zd, int zstat)
{
   if (zstat == 1) edit_done();                                            //  done
   else edit_cancel();                                                     //  cancel or destroy
   zfree(tile_pixmap);
   return 0;
}


int tile_dialog_event(zdialog * zd, const char *event)
{
   if (strNeq(event,"apply")) return 0;

   zdialog_fetch(zd,"size",tile_size);                                     //  get tile size
   zdialog_fetch(zd,"gap",tile_gap);                                       //  get tile gap 

   if (tile_size < 2) {
      if (Fmodified) edit_undo();                                          //  restore original image
      Fmodified = 0;
      return 0;
   }
   
   signal_thread();                                                        //  trigger working thread
   wait_thread_idle();                                                     //  wait for completion
   mwpaint2(); 
   return 1;
}


//  image tiles thread function

void * tile_thread(void *)
{
   int         sg, gg;
   int         sumpix, red, green, blue;
   int         ii, jj, px, py, qx, qy, dist;
   uint16      *pix1, *pix3;
  
   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request

      sg = tile_size + tile_gap;
      gg = tile_gap;

      for (py = 0; py < E1hh; py += sg)                                    //  initz. pixel color map for
      for (px = 0; px < E1ww; px += sg)                                    //    given pixel size
      {
         sumpix = red = green = blue = 0;

         for (qy = py + gg; qy < py + sg; qy++)                            //  get mean color for pixel block
         for (qx = px + gg; qx < px + sg; qx++)
         {
            if (qy > E1hh-1 || qx > E1ww-1) continue;

            pix1 = bmpixel(E1rgb48,qx,qy);
            red += pix1[0];
            green += pix1[1];
            blue += pix1[2];
            sumpix++;
         }

         if (sumpix) {
            red = (red / sumpix);
            green = (green / sumpix);
            blue = (blue / sumpix);
         }
         
         for (qy = py; qy < py + sg; qy++)                                 //  set color for pixels in block
         for (qx = px; qx < px + sg; qx++)
         {
            if (qy > E1hh-1 || qx > E1ww-1) continue;
            
            jj = (qy * E1ww + qx) * 3;

            if (qx-px < gg || qy-py < gg) {
               tile_pixmap[jj] = tile_pixmap[jj+1] = tile_pixmap[jj+2] = 0;
               continue;
            }

            tile_pixmap[jj] = red;
            tile_pixmap[jj+1] = green;
            tile_pixmap[jj+2] = blue;
         }
      }

      if (! sa_exist)                                                      //  process entire image
      {
         for (py = 0; py < E3hh-1; py++)                                   //  loop all image pixels
         for (px = 0; px < E3ww-1; px++)
         {
            pix3 = bmpixel(E3rgb48,px,py);                                 //  target pixel
            jj = (py * E3ww + px) * 3;                                     //  color map for (px,py)
            pix3[0] = tile_pixmap[jj];
            pix3[1] = tile_pixmap[jj+1];
            pix3[2] = tile_pixmap[jj+2];
         }
      }

      if (sa_exist)                                                        //  process selected area
      {
         for (ii = 0; ii < sa_npix; ii++)                                  //  process all enclosed pixels
         {
            px = sa_pix[ii].px;
            py = sa_pix[ii].py;
            dist = sa_pix[ii].dist;
            pix3 = bmpixel(E3rgb48,px,py);
            jj = (py * E3ww + px) * 3;
            pix3[0] = tile_pixmap[jj];
            pix3[1] = tile_pixmap[jj+1];
            pix3[2] = tile_pixmap[jj+2];
         }
      }

      Fmodified = 1;
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  never executed, stop g++ warning
}


/**************************************************************************/

//  paint function - edit individual pixels

int      paint_RGB[3];
int      paint_mode = 0;
int      paint_radius;
int      paint_blend;
double   paint_kernel[200][200];                                           //  support radius <= 99


void m_paint(GtkWidget *, const char *)                                    //  v.6.7
{
   int   paint_dialog_event(zdialog* zd, const char *event);
   int   paint_dialog_compl(zdialog* zd, int zstat);
   void  paint_mousefunc();

   const char  *title = ZTX("Edit Pixels");

   if (! edit_setup(0)) return;                                            //  setup edit: no preview

   zdedit = zdialog_new(title,mWin,Bundo,Bredo,Bdone,Bcancel,null);        //  setup paint dialog
   
   zdialog_add_widget(zdedit,"hbox","hbc","dialog",0,"space=5");
   zdialog_add_widget(zdedit,"label","labc","hbc",ZTX("color"),"space=8");
   zdialog_add_widget(zdedit,"colorbutt","color","hbc","100|100|100");
   zdialog_add_widget(zdedit,"label","labspc","hbc",0,"space=10");
   zdialog_add_widget(zdedit,"radio","rad1","hbc",ZTX("pick"),"space=3");
   zdialog_add_widget(zdedit,"radio","rad2","hbc",ZTX("paint"),"space=3");
   zdialog_add_widget(zdedit,"radio","rad3","hbc",ZTX("erase"),"space=3");
   zdialog_add_widget(zdedit,"hbox","hbr","dialog",0,"space=5");
   zdialog_add_widget(zdedit,"label","labr","hbr",ZTX("radius"),"space=5");
   zdialog_add_widget(zdedit,"spin","radius","hbr","1|99|1|1");
   zdialog_add_widget(zdedit,"label","labspr","hbr",0,"space=10");
   zdialog_add_widget(zdedit,"label","labb","hbr",ZTX("blend"),"space=5");
   zdialog_add_widget(zdedit,"spin","blend","hbr","0|99|1|1");

   zdialog_run(zdedit,paint_dialog_event,paint_dialog_compl);              //  run dialog - parallel
   zdialog_send_event(zdedit,"radius");                                    //  get kernel initialized
   
   paint_mode = 1;                                                         //  mode = color pixels
   mouseCBfunc = paint_mousefunc;                                          //  connect mouse function
   Mcapture = 1;                                                           //  capture mouse clicks
   return;
}


//  paint dialog event and completion functions

int paint_dialog_event(zdialog *zd, const char *event)                     //  paint dialog event function
{
   char        color[20];
   const char  *pp;
   int         radius, blend, dx, dy, brad;
   double      r1, r2, kern;
   
   zdialog_fetch(zd,"rad1",brad);                                          //  v.6.8
   if (brad) paint_mode = 1;
   zdialog_fetch(zd,"rad2",brad);
   if (brad) paint_mode = 2;
   zdialog_fetch(zd,"rad3",brad);
   if (brad) paint_mode = 3;

   if (strEqu(event,"color")) {
      zdialog_fetch(zd,"color",color,19);                                  //  get color from color wheel
      pp = strField(color,"|",1);
      if (pp) paint_RGB[0] = atoi(pp);
      pp = strField(color,"|",2);
      if (pp) paint_RGB[1] = atoi(pp);
      pp = strField(color,"|",3);
      if (pp) paint_RGB[2] = atoi(pp);
   }
   
   if (strEqu(event,"radius") || strEqu(event,"blend"))                    //  get new radius and blend
   {
      zdialog_fetch(zd,"radius",radius);
      zdialog_fetch(zd,"blend",blend);

      paint_radius = radius;
      paint_blend = blend;

      for (dy = -radius; dy <= radius; dy++)                               //  build kernel   v.6.8
      for (dx = -radius; dx <= radius; dx++)
      {
         r1 = sqrt(dx*dx + dy*dy);
         r2 = 1.0 * radius / (radius + blend);
         kern = r2 * (1.0 - r1 * blend / radius / radius);
         if (kern < 0) kern = 0;
         if (kern > 1) kern = 1;
         paint_kernel[dx+radius][dy+radius] = kern;
      }
   }
   return 1;
}


int paint_dialog_compl(zdialog *zd, int zstat)                             //  paint dialog completion function
{
   if (zstat == 1) { edit_undo(); return 0; }                              //  undo
   else if (zstat == 2) { edit_redo(); return 0; }                         //  redo
   else if (zstat == 3) edit_done();                                       //  done
   else edit_cancel();                                                     //  cancel or destroy

   Mcapture = 0;                                                           //  disconnect mouse
   mouseCBfunc = 0;
   return 0;
}


void paint_mousefunc()                                                     //  mouse function
{
   int         px = -1, py = -1;
   int         radius, dx, dy;
   int         red, green, blue;
   double      kern;
   char        color[20];
   uint16      *ppix1, *ppix3;

   GOFUNC_DATA(1)

   if (LMclick) {                                                          //  mouse click
      px = Mxclick;
      py = Myclick;
   }

   if (Mxdrag || Mydrag) {                                                 //  drag in progress
      px = Mxdrag;
      py = Mydrag;
   }

   if (px < 0 || px > E3ww-1 || py < 0 || py > E3hh-1) {                   //  outside image area
      LMclick = RMclick = Mxdrag = Mydrag = 0;
      return;
   }
   
   if (LMclick)                                                            //  left mouse click
   {
      if (paint_mode == 1)                                                 //  pick new color from image
      {
         ppix3 = bmpixel(E3rgb48,px,py);
         paint_RGB[0] = ppix3[0] / 256;
         paint_RGB[1] = ppix3[1] / 256;
         paint_RGB[2] = ppix3[2] / 256;
         snprintf(color,19,"%d|%d|%d",paint_RGB[0],paint_RGB[1],paint_RGB[2]);
         if (zdedit) zdialog_stuff(zdedit,"color",color);
      }
      else  GOFUNC(do_pixels)                                              //  paint or erase
   }
   
   if (Mxdrag || Mydrag) GOFUNC(do_pixels)                                 //  drag in progress
   
   LMclick = RMclick = Mxdrag = Mydrag = 0;
   mwpaint2();
   return;

do_pixels:
   {
      radius = paint_radius;

      if (px < radius || px > E3ww-radius-1) RETURN
      if (py < radius || py > E3hh-radius-1) RETURN
      
      red = 256 * paint_RGB[0];
      green = 256 * paint_RGB[1];
      blue = 256 * paint_RGB[2];

      for (dy = -radius; dy <= radius; dy++)                               //  loop surrounding block of pixels
      for (dx = -radius; dx <= radius; dx++)
      {
         kern = paint_kernel[dx+radius][dy+radius];
         ppix1 = bmpixel(E1rgb48,(px+dx),(py+dy));                         //  original image pixel
         ppix3 = bmpixel(E3rgb48,(px+dx),(py+dy));                         //  edited image pixel

         if (paint_mode == 2)                                              //  color pixels with blend
         {
            ppix3[0] = kern * red + (1.0 - kern) * ppix3[0];
            ppix3[1] = kern * green + (1.0 - kern) * ppix3[1];
            ppix3[2] = kern * blue + (1.0 - kern) * ppix3[2];
            Fmodified = 1;
         }

         if (paint_mode == 3)                                              //  restore org. pixels with blend
         {
            ppix3[0] = kern * ppix1[0] + (1.0 - kern) * ppix3[0];
            ppix3[1] = kern * ppix1[1] + (1.0 - kern) * ppix3[1];
            ppix3[2] = kern * ppix1[2] + (1.0 - kern) * ppix3[2];
         }
      }

      RETURN
   }
}


/**************************************************************************/

//    Make an HDR (high dynamic range) image from two images of the same 
//    subject with different exposure levels. The HDR image has expanded 
//    visibility of detail in both the brightest and darkest areas.


void *   HDR_align_thread(void *);
void *   HDR_dialog_thread(void *);
void     HDR_combine(int ww);

int      HDR_brcurve_adjust(GtkWidget *, GdkEventButton *);
int      HDR_brcurve_draw(GtkWidget *);

int      HDR_align_stat = 0;

int      HDR_br_np = 0;                                                    //  up to 100 anchor points
double   HDR_br_px[100], HDR_br_py[100];                                   //    for image weight graph
double   HDRweights[256];                                                  //  curve, weight per brightness


//  menu function

void m_HDR(GtkWidget *, const char *)
{
   char        *file2 = 0;

   if (! edit_setup(0)) return;                                            //  setup edit: no preview

   m_unselect(null,null);                                                  //  remove select area
   
   file2 = zgetfile(ZTX("select image to combine"),image_file,"open");     //  get 2nd HDR image
   if (! file2) {
      edit_cancel();
      return;
   }

   Grgb48 = image_load(file2,48);                                          //  load and validate image
   if (! Grgb48) {
      edit_cancel();
      return;
   }

   Gww = Grgb48->ww;
   Ghh = Grgb48->hh;
   
   if (Fww != Gww || Fhh != Ghh) {
      zmessageACK(ZTX("2nd image not same size as 1st image"));
      RGB_free(Grgb48);
      Grgb48 = 0;
      edit_cancel();
      return;
   }

   start_thread(HDR_align_thread,0);                                       //  start thread to align images
   wrapup_thread(0);                                                       //  wait for thread exit

   if (HDR_align_stat != 1) {                                              //  check thread exit status
      edit_cancel();
      Nalign = 0;                                                          //  reset align counter
      RGB_free(A1rgb48);                                                   //  free alignment images
      RGB_free(A2rgb48);
      RGB_free(Grgb48);                                                    //  free 2nd input image bitmap
      if (redpixels) zfree(redpixels);                                     //  free edge-pixel flags
      redpixels = 0;
      return;
   }   
   
   int    HDR_dialog_event(zdialog *zd, const char *name);                 //  dialog for user adjustment
   int    HDR_dialog_compl(zdialog *zd, int zstat);

   const char  *weightmess = ZTX("\n image weights per brightness level \n");
   double      bravg;

   zdedit = zdialog_new(ZTX("HDR image weights"),mWin,Bdone,Bcancel,null);
   zdialog_add_widget(zdedit,"label","labt","dialog",weightmess);
   zdialog_add_widget(zdedit,"hbox","hb1","dialog",0,"expand");
   zdialog_add_widget(zdedit,"vbox","vb1","hb1",0,"space=5");
   zdialog_add_widget(zdedit,"vbox","vb2","hb1",0,"expand|space=10");

   zdialog_add_widget(zdedit,"label","lab11","vb1","input images");
   zdialog_add_widget(zdedit,"hbox","hb12","vb1",0,"expand");
   zdialog_add_widget(zdedit,"vbox","vb11","hb12",0,"expand|space=10");
   zdialog_add_widget(zdedit,"vbox","vb12","hb12",0,"expand");
   zdialog_add_widget(zdedit,"label","lab111","vb11","image 1");
   zdialog_add_widget(zdedit,"label","lab112","vb11",0,"expand");
   zdialog_add_widget(zdedit,"label","lab113","vb11","image 2");
   zdialog_add_widget(zdedit,"label","lab121","vb12","100 %");
   zdialog_add_widget(zdedit,"label","lab122","vb12",0,"expand");
   zdialog_add_widget(zdedit,"label","lab123","vb12","50/50","expand");
   zdialog_add_widget(zdedit,"label","lab124","vb12",0,"expand");
   zdialog_add_widget(zdedit,"label","lab125","vb12","100 %");
   zdialog_add_widget(zdedit,"label","lab13","vb1"," ");

   zdialog_add_widget(zdedit,"label","lab21","vb2","output image");
   zdialog_add_widget(zdedit,"frame","fr22","vb2",0,"expand");
   zdialog_add_widget(zdedit,"hbox","hb23","vb2",0);
   zdialog_add_widget(zdedit,"label","lab231","hb23",Bdarker);
   zdialog_add_widget(zdedit,"label","lab232","hb23",0,"expand");
   zdialog_add_widget(zdedit,"label","lab233","hb23",Blighter);

   GtkWidget *brframe = zdialog_widget(zdedit,"fr22");                     //  add drawing area to frame
   GtkWidget *brcurve = gtk_drawing_area_new();
   gtk_container_add(GTK_CONTAINER(brframe),brcurve);

   gtk_widget_add_events(brcurve,GDK_BUTTON_PRESS_MASK);                   //  connect drawing area events
   gtk_widget_add_events(brcurve,GDK_BUTTON_RELEASE_MASK);
   gtk_widget_add_events(brcurve,GDK_BUTTON1_MOTION_MASK); 
   G_SIGNAL(brcurve,"motion-notify-event",HDR_brcurve_adjust,0)
   G_SIGNAL(brcurve,"button-press-event",HDR_brcurve_adjust,0)
   G_SIGNAL(brcurve,"expose-event",HDR_brcurve_draw,0)

   bravg = 0;                                                              //  get mean brightness ratio
   for (int ii = 0; ii < 256; ii++)
      bravg = bravg + Bratios2[0][ii] + Bratios2[1][ii] + Bratios2[2][ii];
   bravg = bravg / 256 / 3;
   
   HDR_br_np = 3;                                                          //  initz. image weights curve,
   HDR_br_px[0] = 5;                                                       //    3 spline anchor points
   HDR_br_px[1] = 127;
   HDR_br_px[2] = 250;

   if (bravg < 1) {                                                        //  image1 < image2, ramp up
      HDR_br_py[0] = 0.02;
      HDR_br_py[1] = 0.50;
      HDR_br_py[2] = 0.98;
   }
   else {                                                                  //  image1 > image2, ramp down
      HDR_br_py[0] = 0.98;
      HDR_br_py[1] = 0.50;
      HDR_br_py[2] = 0.02;
   }
   
   spline1(HDR_br_np,HDR_br_px,HDR_br_py);                                 //  make curve fitting anchor points
   
   zdialog_resize(zdedit,400,270);
   zdialog_run(zdedit,HDR_dialog_event,HDR_dialog_compl);                  //  run dialog, parallel
   start_thread(HDR_dialog_thread,0);                                      //  start compute thread
   signal_thread();
   return;
}


//  dialog event and completion functions for HDR image level adjustment

int HDR_dialog_event(zdialog *zd, const char *event)
{
   for (int ii = 0; ii < 256; ii++)                                        //  compute weight curve for 
      HDRweights[ii] = spline2(ii);                                        //    all 256 brightness levels

   signal_thread();                                                        //  signal thread to update
   return 1;
}


int HDR_dialog_compl(zdialog *zd, int zstat)
{
   if (zstat != 1) edit_cancel();                                          //  user cancel
   else edit_done();
   Nalign = 0;                                                             //  reset align counter
   RGB_free(A1rgb48);                                                      //  free alignment images
   RGB_free(A2rgb48);
   RGB_free(Grgb48);                                                       //  free 2nd input image bitmap
   if (redpixels) zfree(redpixels);                                        //  free edge-pixel flags
   redpixels = 0;
   return 0;
}


//  thread to update image brightness levels
//  runs asynchronously to dialog updates

void * HDR_dialog_thread(void *)
{
   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request
      HDR_combine(1);                                                      //  combine images >> E3rgb48
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  never executed, stop g++ warning
}


//  Add, delete, or move anchor points to weights curve using mouse.

int HDR_brcurve_adjust(GtkWidget *brcurve, GdkEventButton *event)
{
   int         ww, hh, px, py;
   int         kk, ii, newii, minii = -1;
   int         mx, my, button, evtype;
   double      dist2, mindist2 = 1000000;
   double      xval, yval;
   
   if (HDR_br_np > 99) {
      zmessageACK(ZTX("exceed 100 anchor points"));
      return 0;
   }
   
   mx = int(event->x);                                                     //  mouse position in drawing area
   my = int(event->y);
   evtype = event->type;
   button = event->button;
   ww = brcurve->allocation.width;                                         //  drawing area size
   hh = brcurve->allocation.height;
   
   for (ii = 0; ii < HDR_br_np; ii++)                                      //  find closest anchor point
   {
      xval = HDR_br_px[ii];
      yval = spline2(xval);
      px = int(0.00392 * ww * xval);                                       //  0 - ww
      py = int(hh - hh * yval + 0.5);                                      //  0 - hh
      dist2 = (px-mx)*(px-mx) + (py-my)*(py-my);
      if (dist2 < mindist2) {
         mindist2 = dist2;
         minii = ii;
      }
   }

   if (minii < 0) return 0;                                                //  huh?
   
   if (evtype == GDK_BUTTON_PRESS && button == 3) {                        //  right click, remove anchor point
      if (mindist2 > 25) return 0;
      if (HDR_br_np < 3) return 0;
      for (kk = minii; kk < HDR_br_np -1; kk++) {
         HDR_br_px[kk] = HDR_br_px[kk+1];
         HDR_br_py[kk] = HDR_br_py[kk+1];
      }
      HDR_br_np--;

      HDR_brcurve_draw(brcurve);                                           //  regen and redraw the curve
      return 0;
   }

//  drag or left click, move nearby anchor point to mouse position,
//  or add a new anchor point if nothing near enough

   xval = 255.0 * mx / ww;                                                 //  0 - 255
   yval = 1.0 * (hh - my) / hh;                                            //  0 - 1.0

   if (xval < 0 || xval > 255) return 0;                                   //  v.6.8
   if (yval < 0 || yval > 1.0) return 0;

   if (mindist2 < 100) {                                                   //  existing point < 10 pixels away
      ii = minii;
      if (ii < HDR_br_np - 1 && HDR_br_px[ii+1] - xval < 5) return 0;      //  disallow < 5 x-pixels
      if (ii > 0 && xval - HDR_br_px[ii-1] < 5) return 0;                  //    to next or prior point
      newii = minii;                                                       //  this point to be moved
   }
   else                                                                    //  > 10 pixels, add a point
   {
      for (ii = 0; ii < HDR_br_np; ii++)
         if (xval <= HDR_br_px[ii]) break;                                 //  find point with next higher x

      if (ii < HDR_br_np && HDR_br_px[ii] - xval < 5) return 0;            //  disallow < 5 x-pixels
      if (ii > 0 && xval - HDR_br_px[ii-1] < 5) return 0;                  //    to next or prior point

      for (kk = HDR_br_np; kk > ii; kk--) {                                //  make hole for new point
         HDR_br_px[kk] = HDR_br_px[kk-1];
         HDR_br_py[kk] = HDR_br_py[kk-1];
      }

      HDR_br_np++;                                                         //  up point count
      newii = ii;                                                          //  this point to be added
   }

   HDR_br_px[newii] = xval;                                                //  coordinates of new or moved point
   HDR_br_py[newii] = yval;
   
   HDR_brcurve_draw(brcurve);                                              //  regen and redraw the curve
   return 0;
}


//  Draw brightness curve based on defined spline anchor points.

int HDR_brcurve_draw(GtkWidget *brcurve)
{
   int         ww, hh, px, py;
   int         ii, iix, iiy;
   double      xval, yval;

   ww = brcurve->allocation.width;                                         //  drawing area size
   hh = brcurve->allocation.height;
   if (ww < 50 || hh < 20) return 0;

   spline1(HDR_br_np,HDR_br_px,HDR_br_py);                                 //  make curve fitting anchor points

   gdk_window_clear(brcurve->window);                                      //  clear window

   for (px = 0; px < ww; px++)                                             //  generate all points for curve
   {
      xval = 255.0 * px / ww;
      yval = spline2(xval);
      py = int(hh - hh * yval + 0.5);
      gdk_draw_point(brcurve->window,gdkgc,px,py);
   }
   
   for (ii = 0; ii < HDR_br_np; ii++)                                      //  draw boxes at anchor points
   {
      xval = HDR_br_px[ii];
      yval = spline2(xval);
      px = int(0.00392 * ww * xval);
      py = int(hh - hh * yval + 0.5);
      for (iix = -2; iix < 3; iix++)
      for (iiy = -2; iiy < 3; iiy++) {
         if (px+iix < 0 || px+iix >= ww) continue;
         if (py+iiy < 0 || py+iiy >= hh) continue;
         gdk_draw_point(brcurve->window,gdkgc,px+iix,py+iiy);
      }
   }

   HDR_dialog_event(0,0);                                                  //  trigger image update
   return 0;
}


//  HDR image align thread, combine Frgb48 + Grgb48 >> E3rgb48

void * HDR_align_thread(void *)
{
   double      xfL, xfH, yfL, yfH, tfL, tfH;
   int         firstpass = 1, lastpass = 0;

   HDR_align_stat = 0;                                                     //  no status yet
   Radjust = Gadjust = Badjust = 1.0;                                      //  no manual color adjustments
   Nalign = 1;                                                             //  alignment in progress
   showRedpix = 1;                                                         //  highlight alignment pixels
   Fzoom = 0;                                                              //  fit to window if big
   Fblowup = 1;                                                            //  scale up to window if small
   
   fullSize = Fww;                                                         //  full image size
   if (Fhh > Fww) fullSize = Fhh;                                          //  (largest dimension)

   alignSize = 140;                                                        //  initial alignment image size
   if (alignSize > fullSize) alignSize = fullSize;
   A1rgb48 = A2rgb48 = 0;

   xoff = yoff = toff = 0;                                                 //  initial offsets = 0
   blendstripe = 0;                                                        //  overlap region is 100%

   yst1 = yst2 = yst1B = yst2B = 0;                                        //  disable pano stuff

   while (true)                                                            //  next alignment stage / image size
   {   
      A1ww = Fww * alignSize / fullSize;                                   //  align width, height in same ratio
      A1hh = Fhh * alignSize / fullSize;
      A2ww = A1ww;
      A2hh = A1hh;
      
      if (! lastpass) 
      {
         RGB_free(A1rgb48);                                                //  align images = scaled input images
         RGB_free(A2rgb48);
         A1rgb48 = RGB_rescale(Frgb48,A1ww,A1hh);
         A2rgb48 = RGB_rescale(Grgb48,A2ww,A2hh);

         get_overlap_region();                                             //  get image overlap area
         get_Bratios();                                                    //  get image brightness ratios
         set_colorfix_factors(1);                                          //  compute color matching factors
         flag_edge_pixels();                                               //  flag high-contrast pixels

         mutex_lock(&bitmaps_lock);
         RGB_free(E3rgb48);                                                //  resize output image
         E3rgb48 = RGB_make(A1ww,A1hh,48);
         E3ww = A1ww;
         E3hh = A1hh;
         HDR_combine(0);                                                   //  combine images >> E3rgb48
         mutex_unlock(&bitmaps_lock);

         mwpaint2();                                                       //  update window
      }

      xylim = 2;                                                           //  search range from prior stage:
      xystep = 1;                                                          //    -2 -1 0 +1 +2 pixels

      if (firstpass) xylim = 0.05 * alignSize;                             //  1st stage search range, huge

      if (lastpass) {
         xylim = 1;                                                        //  final stage search range:
         xystep = 0.5;                                                     //    -1.0 -0.5 0.0 +0.5 +1.0
      }

      tlim = xylim / alignSize / 2;                                        //  theta max offset, radians
      tstep = xystep / alignSize / 2;                                      //  theta step size

      xfL = xoff - xylim;
      xfH = xoff + xylim + xystep/2;
      yfL = yoff - xylim;
      yfH = yoff + xylim + xystep/2;
      tfL = toff - tlim;
      tfH = toff + tlim + tstep/2;

      xoffB = xoff;
      yoffB = yoff;
      toffB = toff;
      
      matchB = match_images();                                             //  set base match level

      for (xoff = xfL; xoff < xfH; xoff += xystep)                         //  test all offset dimensions
      for (yoff = yfL; yoff < yfH; yoff += xystep)                         //    in all combinations
      for (toff = tfL; toff < tfH; toff += tstep)
      {
         matchlev = match_images();
         if (sigdiff(matchlev,matchB,0.00001) > 0) {
            matchB = matchlev;
            xoffB = xoff;
            yoffB = yoff;
            toffB = toff;
         }

         Nalign++;                                                         //  count alignment tests
         SBupdate++;                                                       //  update status bar
      }

      xoff = xoffB;                                                        //  recover best offsets
      yoff = yoffB;
      toff = toffB;

      HDR_combine(0);                                                      //  combine images >> E3rgb48
      mwpaint2();                                                          //  update window
     
      firstpass = 0;
      if (lastpass) break;                                                 //  done

      if (alignSize == fullSize) {                                         //  full size image was aligned
         lastpass++;                                                       //  one more pass
         continue;
      }

      double R = alignSize;
      alignSize = 2 * alignSize;                                           //  next larger image size
      if (alignSize > 0.8 * fullSize) alignSize = fullSize;                //  if near goal, jump to it now
      R = alignSize / R;                                                   //  ratio of new / old image size
      xoff = R * xoff;                                                     //  adjust offsets for image size
      yoff = R * yoff;
   }

   Fmodified = 1;                                                          //  image is modified
   showRedpix = 0;                                                         //  stop red pixel highlights
   Fblowup = 0;                                                            //  reset forced image scaling
   HDR_align_stat = 1;                                                     //  signal success
   exit_thread();
   return 0;                                                               //  never executed, stop g++ warning
}


//  Combine images A1rgb48 and A2rgb48 using weights in HDRweights[256].
//  Output is to E3rgb48 (not reallocated).

void HDR_combine(int weight)
{
   int         px3, py3, ii;
   double      px1, py1, px2, py2;
   double      sintf = sin(toff), costf = cos(toff);
   double      br1, br2, brm;
   uint16     *pix1, *pix2, *pix3;
   
   for (py3 = 0; py3 < A1hh; py3++)                                        //  step through A1rgb48 pixels
   for (px3 = 0; px3 < A1ww; px3++)
   {
      px1 = costf * px3 - sintf * (py3 - yoff);                            //  A1rgb48 pixel, after offsets
      py1 = costf * py3 + sintf * (px3 - xoff);
      pix1 = vpixel(A1rgb48,px1,py1);

      px2 = costf * (px3 - xoff) + sintf * (py3 - yoff);                   //  corresponding A2rgb48 pixel
      py2 = costf * (py3 - yoff) - sintf * (px3 - xoff);
      pix2 = vpixel(A2rgb48,px2,py2);

      pix3 = bmpixel(E3rgb48,px3,py3);                                     //  output pixel

      if (! pix1 || ! pix2) {                                              //  if non-overlapping pixel,
         pix3[0] = pix3[1] = pix3[2] = 0;                                  //    set output pixel black
         continue;
      }
      
      if (weight) {
         br1 = pix1[0] + pix1[1] + pix1[2];                                //  image1 pixel brightness
         br2 = pix2[0] + pix2[1] + pix2[2];                                //  image2
         brm = (br1 + br2) * 0.000651;                                     //  mean, 0 to 255.98
         br1 = HDRweights[int(brm)];                                       //  image1 weight
         br2 = 1.0 - br1;                                                  //  image2 weight
         
         pix3[0] = int(pix1[0] * br1 + pix2[0] * br2);                     //  build output pixel
         pix3[1] = int(pix1[1] * br1 + pix2[1] * br2);
         pix3[2] = int(pix1[2] * br1 + pix2[2] * br2);
      }

      else {
         pix3[0] = (pix1[0] + pix2[0]) / 2;                                //  output pixel is simple average
         pix3[1] = (pix1[1] + pix2[1]) / 2;
         pix3[2] = (pix1[2] + pix2[2]) / 2;
      }

      if (showRedpix && pix2) {                                            //  highlight alignment pixels
         ii = py3 * A1ww + px3;
         if (redpixels[ii]) {
            pix3[0] = 65535;
            pix3[1] = pix3[2] = 0;
         }
      }
   }

   return;
}


/**************************************************************************/

//  panorama function - stitch two images together

void     pano_prealign();                                                  //  manual pre-align
void     pano_autolens();                                                  //  auto optimize lens parameters
void *   pano_align_thread(void *);                                        //  auto align images
void     pano_final_adjust();                                              //  manual final adjustment
void     pano_get_align_images(int newf, int strf);                        //  scale and curve images for align
void     pano_combine(int fcolor);                                         //  combine images

int      pano_stat;                                                        //  dialog and thread status
int      pano_automatch;                                                   //  auto color matching on/off
RGB      *A1cache, *A2cache;                                               //  cached alignment images


void m_pano(GtkWidget *, const char *)
{
   char        *file2 = 0;

   Grgb48 = A1rgb48 = A2rgb48 = A1cache = A2cache = 0;

   if (! edit_setup(0)) return;                                            //  setup edit: no preview

   m_unselect(null,null);                                                  //  remove select area
   
   file2 = zgetfile(ZTX("select image to combine"),image_file,"open");     //  2nd or next pano file
   if (! file2) goto pano_cancel;

   Grgb48 = image_load(file2,48);                                          //  load and validate image
   if (! Grgb48) goto pano_cancel;
   Gww = Grgb48->ww;
   Ghh = Grgb48->hh;
   
   Radjust = Gadjust = Badjust = 1.0;                                      //  no manual color adjustments
   Nalign = 1;                                                             //  alignment in progress
   showRedpix = 0;                                                         //  no pixel highlight (yet)
   Fblowup = 1;                                                            //  scale-up small image to window

   pano_prealign();                                                        //  do manual pre-align
   if (pano_stat != 1) goto pano_cancel;

   if (overlapixs < alignSize * 20) {                                      //  need > 20 pixel overlap
      zmessageACK(ZTX("too little overlap, cannot align"));
      goto pano_cancel;
   }

   start_thread(pano_align_thread,0);                                      //  start thread to align images
   wrapup_thread(0);                                                       //  wait for thread exit
   if (pano_stat != 1) goto pano_cancel;
   
   pano_final_adjust();                                                    //  do manual final adjustments
   if (pano_stat != 1) goto pano_cancel;
   Fmodified = 1;
   edit_done();
   goto pano_cleanup;
   
pano_cancel:
   edit_cancel();

pano_cleanup:
   Nalign = 0;
   if (file2) zfree(file2);
   RGB_free(Grgb48);
   RGB_free(A1rgb48);
   RGB_free(A2rgb48);
   RGB_free(A1cache);
   RGB_free(A2cache);
   return;
}


//  perform manual pre-align of image2 to image1
//  return offsets: xoff, yoff, toff
//  lens_mm and lens_bow may also be altered

void pano_prealign()   
{  
   int      pano_prealign_event(zdialog *zd, const char *event);           //  dialog event function
   int      pano_prealign_compl(zdialog *zd, int zstat);                   //  dialog completion function
   void *   pano_prealign_thread(void *);                                  //  working thread

   const char  *align_mess = ZTX("drag right image into rough alignment with left \n"
                                      " to rotate, drag right edge up or down");
   const char  *proceed_mess = ZTX("merge the images together");
   const char  *search_mess = ZTX("auto-search lens mm and bow");

   zdedit = zdialog_new(ZTX("pre-align images"),mWin,Bcancel,null);
   zdialog_add_widget(zdedit,"label","lab1","dialog",align_mess,"space=5");
   zdialog_add_widget(zdedit,"hbox","hb1","dialog",0,"space=5");
   zdialog_add_widget(zdedit,"spin","spmm","hb1","22|200|0.1|35","space=5");
   zdialog_add_widget(zdedit,"label","labmm","hb1",ZTX("lens mm"));
   zdialog_add_widget(zdedit,"hbox","hb2","dialog",0,"space=5");
   zdialog_add_widget(zdedit,"spin","spbow","hb2","-9|9|0.01|0","space=5");
   zdialog_add_widget(zdedit,"label","labbow","hb2",ZTX("lens bow"));
   zdialog_add_widget(zdedit,"hbox","hb3","dialog",0,"space=5");
   zdialog_add_widget(zdedit,"button","proceed","hb3",Bproceed,"space=5");
   zdialog_add_widget(zdedit,"label","labproceed","hb3",proceed_mess);
   zdialog_add_widget(zdedit,"hbox","hb4","dialog",0,"space=5");
   zdialog_add_widget(zdedit,"button","search","hb4",Bsearch,"space=5");
   zdialog_add_widget(zdedit,"label","labsearch","hb4",search_mess);
   
   lens_mm = lens4_mm[curr_lens];                                          //  initz. curr. lens parameters
   lens_bow = lens4_bow[curr_lens];
   zdialog_stuff(zdedit,"spmm",lens_mm);
   zdialog_stuff(zdedit,"spbow",lens_bow);

   zdialog_run(zdedit,pano_prealign_event,pano_prealign_compl);            //  run dialog, parallel
   
   pano_stat = -1;
   start_thread(pano_prealign_thread,0);                                   //  start working thread
   wrapup_thread(0);                                                       //  wait for completion
   return;
}


int pano_prealign_event(zdialog *zd, const char *event)                    //  dialog event callback function
{
   zdialog_fetch(zd,"spmm",lens_mm);                                       //  get revised lens data
   zdialog_fetch(zd,"spbow",lens_bow);
   if (strEqu(event,"search")) Fautolens = 1;                              //  trigger auto-lens function

   if (strEqu(event,"proceed")) {                                          //  proceed with pano
      pano_stat = 1;                                                       //  signal success
      wrapup_thread(0);                                                    //  wait for thread exit
      zdialog_free(zdedit);                                                //  kill dialog
      zdedit = null;
   }
   return 0;
}


int pano_prealign_compl(zdialog *zd, int zstat)                            //  dialog completion callback function
{                                                                          //  (cancel only)
   pano_stat = 0;                                                          //  signal cancel
   wrapup_thread(0);                                                       //  wait for thread exit
   zdialog_free(zdedit);                                                   //  kill dialog
   zdedit = null;
   return 0;
}


void * pano_prealign_thread(void *)                                        //  prealign working thread
{
   int         mx0, my0, mx, my, key;                                      //  mouse drag origin, position
   double      lens_mm0, lens_bow0;
   double      mlever = 0.5;
   double      dtoff;

   yst1 = yst1B = yst2 = yst2B = 0;                                        //  initial y-stretch offsets
   yoff = yoffB = toff = toffB = 0;                                        //  initial y and theta offsets

   fullSize = Ghh;                                                         //  full size = image2 height
   alignSize = int(pano_prealign_size);                                    //  prealign image size
   pano_get_align_images(1,0);                                             //  get prealign images and curve them

   xoff = xoffB = 0.8 * A1ww;                                              //  initial x offset (20% overlap)
   blendstripe = int(A1ww - xoff);                                         //  initial blendstripe = all overlap
   get_overlap_region();                                                   //  get image overlap area
   pano_combine(0);                                                        //  combine images

   lens_mm0 = lens_mm;                                                     //  to detect changes
   lens_bow0 = lens_bow;

   mx0 = my0 = 0;                                                          //  no drag in progress
   Mcapture = KBcapture = 1;                                               //  capture mouse drag and KB keys

   while (pano_stat == -1)                                                 //  loop and align until done
   {
      zsleep(0.05);
      
      if (Fautolens) {
         pano_autolens();                                                  //  get lens parameters
         zdialog_stuff(zdedit,"spmm",lens_mm);                             //  update dialog
         zdialog_stuff(zdedit,"spbow",lens_bow);
      }
         
      if (lens_mm != lens_mm0 || lens_bow != lens_bow0) {                  //  change in lens parameters
         lens_mm0 = lens_mm;
         lens_bow0 = lens_bow;
         pano_get_align_images(0,0);
         pano_combine(0);                                                  //  show combined images
      }
      
      if (KBkey) {                                                         //  KB input
         key = KBkey;
         KBkey = 0;
         if (key == GDK_Left)  xoff -= 0.2;                                //  tweak alignment offsets
         if (key == GDK_Right) xoff += 0.2;
         if (key == GDK_Up)    yoff -= 0.2;
         if (key == GDK_Down)  yoff += 0.2;
         if (key == GDK_r)     toff += 0.0002;
         if (key == GDK_l)     toff -= 0.0002;

         blendstripe = int(A1ww - xoff);                                   //  entire overlap
         get_overlap_region();
         pano_combine(0);                                                  //  show combined images
      }

      if (! Mxdrag && ! Mydrag) {
         mx0 = my0 = 0;                                                    //  no drag in progress
         goto idle;
      }

      mx = Mxdrag;                                                         //  mouse position in image
      my = Mydrag;

      if (mx < xoff || mx > xoff + A2ww || my > E3hh) {                    //  if mouse not in image2 area,
         mx0 = my0 = 0;                                                    //    no drag in progress
         goto idle;
      }

      if (mx0 == 0 && my0 == 0) {                                          //  new drag origin
         mx0 = mx;
         my0 = my;
         goto idle;
      }
      
      if (mx == mx0 && my == my0) goto idle;                               //  no movement

      if (mx > xoff + 0.8 * A2ww) {                                        //  near right edge, theta drag 
         dtoff = mlever * (my - my0) / A2ww;                               //  delta theta, radians
         toff += dtoff;
         xoff += dtoff * (A1hh + yoff);                                    //  change center of rotation
         yoff -= dtoff * (A1ww - xoff);                                    //    to middle of overlap area
      }
      else  {                                                              //  x/y drag
         xoff += mlever * (mx - mx0);                                      //  image2 offsets / mouse leverage
         yoff += mlever * (my - my0);
      }

      mx0 = mx;                                                            //  reset drag origin
      my0 = my;

      if (xoff > A1ww) xoff = A1ww;                                        //  limit nonsense
      if (xoff < 0.3 * A1ww) xoff = 0.3 * A1ww;
      if (yoff < -0.5 * A2hh) yoff = -0.5 * A2hh;
      if (yoff > 0.5 * A1hh) yoff = 0.5 * A1hh;
      if (toff < -0.20) toff = -0.20;
      if (toff > 0.20) toff = 0.20;

      blendstripe = int(A1ww - xoff);                                      //  entire overlap
      get_overlap_region();
      pano_combine(0);                                                     //  show combined images
      continue;

idle:                                                                      //  spare time activity
      get_overlap_region();                                                //  get image overlap area
      get_Bratios();                                                       //  get brightness ratios in overlap 
      set_colorfix_factors(1);                                             //  compute color matching factors
      flag_edge_pixels();                                                  //  flag edge pixels in overlap 
      matchB = match_images();                                             //  match images
      xoffB = xoff;
      yoffB = yoff;
      toffB = toff;
      SBupdate++;                                                          //  update status bar
   }

   Fzoom = 0;                                                              //  reset in case user zoomed in
   KBcapture = Mcapture = 0;
   exit_thread();
   return 0;                                                               //  never executed, stop g++ warning
}


//  optimize lens parameters
//  assumes a good starting point since search ranges are limited
//  inputs and outputs: lens_mm, lens_bow, xoff, yoff, toff

void pano_autolens()
{
   double   mm_range, bow_range, xoff_range, yoff_range, toff_range;
   double   squeeze, xoff_rfinal, rnum;
   double   lens_mmB, lens_bowB;
   int      counter = 0;
   
   mm_range = 0.1 * lens_mm;                                               //  set initial search ranges
   bow_range = 0.5 * lens_bow;
   if (bow_range < 1) bow_range = 1;
   xoff_range = 7;
   yoff_range = 7;
   toff_range = 0.01;

   xoff_rfinal = 0.2;                                                      //  final xoff range - when to quit
   
   Nalign = 0;
   showRedpix = 1;
   pano_get_align_images(0,0);
   get_overlap_region();
   get_Bratios();
   set_colorfix_factors(1);
   flag_edge_pixels();
   pano_combine(0);

   lens_mmB = lens_mm;                                                     //  initial best fit = current data
   lens_bowB = lens_bow;
   xoffB = xoff;
   yoffB = yoff;
   toffB = toff;
   matchB = match_images();
   
   while (true)
   {
      srand48(time(0) + counter++);
      squeeze = 0.9;                                                       //  search range reduction factor

      lens_mm = lens_mmB + mm_range * (drand48() - 0.5);                   //  new random lens factors
      lens_bow = lens_bowB + bow_range * (drand48() - 0.5);                //     within search range
      pano_get_align_images(0,0);
         
      for (int ii = 0; ii < 300; ii++)
      {                                                                    
         rnum = drand48();

         if (rnum < 0.33)                                                  //  random change some alignment offset 
            xoff = xoffB + xoff_range * (drand48() - 0.5);                 //    within search range
         else if (rnum < 0.67)
            yoff = yoffB + yoff_range * (drand48() - 0.5);
         else
            toff = toffB + toff_range * (drand48() - 0.5);
      
         matchlev = match_images();                                        //  test quality of image alignment
         if (sigdiff(matchlev,matchB,0.00001) > 0) {
            get_overlap_region();                                          //  better: reset alignment data
            get_Bratios();
            set_colorfix_factors(1);
            flag_edge_pixels();
            matchB = match_images();                                       //  save new best fit
            lens_mmB = lens_mm;
            lens_bowB = lens_bow;
            xoffB = xoff;
            yoffB = yoff;
            toffB = toff;
            pano_combine(0);                                               //  update window
            squeeze = 1;                                                   //  keep same search range as long
            break;                                                         //    as improvements are found
         }

         Nalign++;
         SBupdate++;
         if (pano_stat != -1) goto done;
      }

      if (xoff_range < xoff_rfinal) goto done;

      mm_range = squeeze * mm_range;                                       //  reduce search range if no 
      if (mm_range < 0.02 * lens_mmB) mm_range = 0.02 * lens_mmB;          //    improvements were found
      bow_range = squeeze * bow_range;
      if (bow_range < 0.1 * lens_bowB) bow_range = 0.1 * lens_bowB;
      if (bow_range < 0.2) bow_range = 0.2;
      xoff_range = squeeze * xoff_range;
      yoff_range = squeeze * yoff_range;
      toff_range = squeeze * toff_range;
   }

done:
   lens_mm = lens_mmB;                                                     //  set best alignment found
   lens_bow = lens_bowB;
   xoff = xoffB;
   yoff = yoffB;
   toff = toffB;
   Fautolens = 0;
   showRedpix = 0;
   pano_get_align_images(0,0);
   pano_combine(0);
   SBupdate++;
   return;
}


//  Thread function for combining A1 + A2 >> E3

void * pano_align_thread(void *)
{
   int         firstpass = 1, lastpass = 0;
   double      alignR;
   double      xfL, xfH, yfL, yfH, tfL, tfH;
   double      ystlim, yst1L, yst1H, yst2L, yst2H;
   
   pano_stat = 0;
   Fzoom = 0;                                                              //  fit to window if big
   Fblowup = 1;                                                            //  scale up to window if small
   showRedpix = 1;                                                         //  highlight alignment pixels
   ystlim = pano_ystretch;                                                 //  set initial y-stretch limit

   while (true)
   {
      alignR = alignSize;
      if (firstpass) alignSize = 140;
      else if (lastpass) alignSize = fullSize;
      else  alignSize = int(pano_image_increase * alignSize);              //  next larger image size
      if (alignSize > 0.8 * fullSize) alignSize = fullSize;                //  if near goal, jump to it now
      alignR = alignSize / alignR;                                         //  ratio of new / old image size

      xoff = alignR * xoff;                                                //  adjust offsets for new image size
      yoff = alignR * yoff;
      toff = toff;
      yst1 = alignR * yst1;
      yst2 = alignR * yst2;
      blendstripe = int(alignR * blendstripe);

      if (! firstpass) {
         blendstripe = int(blendstripe * pano_blend_decrease);             //  reduce image comparison width
         if (blendstripe < pano_blend_min * alignSize)                     //  keep within range
            blendstripe = int(pano_blend_min * alignSize);
         if (blendstripe > pano_blend_max * alignSize) 
            blendstripe = int(pano_blend_max * alignSize);
      }
      
      xylim = 3;                                                           //  +/- search range, centered on
      xystep = 1;                                                          //    results from prior stage
      
      if (firstpass) {
         xylim = alignSize * 0.032;                                        //  3.2 % error range in pre-alignment
         xystep = 0.8;                                                     //  -3.2 -2.4 -1.6 -0.8  0  +0.8 ...
      }

      if (lastpass) {
         xylim = 1;                                                        //  final stage pixel search steps
         xystep = 0.5;                                                     //   -1.0 -0.5 0.0 +0.5 +1.0
      }

      tlim = xylim / alignSize / 2;                                        //  theta max offset, radians
      tstep = xystep / alignSize / 2;                                      //  theta step size

      if (! lastpass) pano_get_align_images(1,0);                          //  get new alignment images

      get_overlap_region();                                                //  get image overlap area
      get_Bratios();                                                       //  get image bright. ratios in overlap 
      set_colorfix_factors(1);                                             //  compute color matching factors
      flag_edge_pixels();                                                  //  flag high-contrast pixels in blend
      pano_combine(0);                                                     //  combine images
      
      xfL = xoff - xylim;                                                  //  set x/y/t search ranges, step sizes
      xfH = xoff + xylim + xystep/2;
      yfL = yoff - xylim;
      yfH = yoff + xylim + xystep/2;
      tfL = toff - tlim;
      tfH = toff + tlim + tstep/2;

      xoffB = xoff;                                                        //  initial offsets = best so far
      yoffB = yoff;
      toffB = toff;
      yst1B = yst1;
      yst2B = yst2;

      matchB = match_images();                                             //  set base match level

      for (xoff = xfL; xoff < xfH; xoff += xystep)                         //  test x, y, theta offsets
      for (yoff = yfL; yoff < yfH; yoff += xystep)                         //    in all possible combinations
      for (toff = tfL; toff < tfH; toff += tstep)
      {
         matchlev = match_images();
         if (sigdiff(matchlev,matchB,0.00001) > 0) {                       //  remember best alignment and offsets
            matchB = matchlev;
            xoffB = xoff;
            yoffB = yoff;
            toffB = toff;
         }

         Nalign++;
         SBupdate++;                                                       //  update status bar
      }
      
      xoff = xoffB;                                                        //  recover best offsets
      yoff = yoffB;
      toff = toffB;
      
      pano_combine(0);                                                     //  combine images >> E3rgb48
      
      if (! firstpass)                                                     //  do image2 y-stretch alignment
      {
         yststep = 0.5;

         yst1L = yst1 - ystlim * A2hh;                                     //  set y-stretch limits
         yst1H = yst1 + ystlim * A2hh + yststep/2;
         yst2L = yst2 - ystlim * A2hh;
         yst2H = yst2 + ystlim * A2hh + yststep/2;

         ystlim = ystlim * pano_ystretch_decrease;                         //  reduce for next pass
         
         for (yst1 = yst1L; yst1 < yst1H; yst1 += yststep)                 //  search upper image half
         {
            pano_get_align_images(0,1);                                    //  curve and y-stretch
            matchlev = match_images();
            if (sigdiff(matchlev,matchB,0.00001) > 0) {                    //  remember best align and y-stretch 
               matchB = matchlev;
               yst1B = yst1;
            }

            Nalign++;
            SBupdate++;
         }
         
         yst1 = yst1B;                                                     //  restore best y-stretch offset
         pano_get_align_images(0,1);

         for (yst2 = yst2L; yst2 < yst2H; yst2 += yststep)                 //  search lower half
         {
            pano_get_align_images(0,2);
            matchlev = match_images();
            if (sigdiff(matchlev,matchB,0.00001) > 0) {
               matchB = matchlev;
               yst2B = yst2;
            }

            Nalign++;
            SBupdate++;
         }
         
         yst2 = yst2B;                                                     //  restore best offset
         pano_get_align_images(0,0);
         pano_combine(0);
      }

      firstpass = 0;
      if (lastpass) break;
      if (alignSize == fullSize) lastpass = 1;                             //  one more pass, reduced step size
   }

   pano_stat = 1;                                                          //  signal success
   showRedpix = 0;                                                         //  pixel highlights off
   Fzoom = Fblowup = 0;                                                    //  reset image scaling
   exit_thread();
   return 0;                                                               //  never executed, stop g++ warning
}


//  do manual adjustment of brightness, color, blend width

void pano_final_adjust()
{
   int      pano_adjust_event(zdialog *zd, const char *event);             //  dialog event function
   int      pano_adjust_compl(zdialog *zd, int zstat);                     //  dialog completion function
   void *   pano_adjust_thread(void *);
   
   int      zstat;

   const char  *adjmessage = ZTX("\n match brightness and color");
   
   pano_automatch = 1;                                                     //  init. auto color match on
   blendstripe = 1;
   
   get_overlap_region();
   get_Bratios();                                                          //  get overlap area color ratios
   set_colorfix_factors(pano_automatch);                                   //  set initial color matching
   pano_combine(1);                                                        //  show final results

   zdedit = zdialog_new(ZTX("match images"),mWin,Bdone,Bcancel,null);      //  color adjustment dialog

   zdialog_add_widget(zdedit,"label","lab0","dialog",adjmessage,"space=10");
   zdialog_add_widget(zdedit,"hbox","hb1","dialog",0,"space=10");          //  match brightness and color
   zdialog_add_widget(zdedit,"vbox","vb1","hb1",0,"homog");
   zdialog_add_widget(zdedit,"vbox","vb2","hb1",0,"homog");                //  red           [ 100 ]
   zdialog_add_widget(zdedit,"label","lab1","vb1",Bred,"space=7");         //  green         [ 100 ]
   zdialog_add_widget(zdedit,"label","lab2","vb1",Bgreen,"space=7");       //  blue          [ 100 ]
   zdialog_add_widget(zdedit,"label","lab3","vb1",Bblue,"space=7");        //  brightness    [ 100 ]
   zdialog_add_widget(zdedit,"label","lab4","vb1",Bbrightness,"space=7");  //  blend width   [  0  ]
   zdialog_add_widget(zdedit,"label","lab5","vb1",Bblendwidth,"space=7");  //
   zdialog_add_widget(zdedit,"spin","spred","vb2","50|200|0.1|100");       //  [ apply ]  [ auto ]  off
   zdialog_add_widget(zdedit,"spin","spgreen","vb2","50|200|0.1|100");
   zdialog_add_widget(zdedit,"spin","spblue","vb2","50|200|0.1|100");
   zdialog_add_widget(zdedit,"spin","spbright","vb2","50|200|0.1|100");
   zdialog_add_widget(zdedit,"spin","spblend","vb2","1|200|1|1");
   zdialog_add_widget(zdedit,"hbox","hb2","dialog",0,"space=5");
   zdialog_add_widget(zdedit,"button","apply","hb2",Bapply,"space=5");
   zdialog_add_widget(zdedit,"button","auto","hb2",ZTX("auto"));
   zdialog_add_widget(zdedit,"label","labauto","hb2","on");
   
   zstat = zdialog_run(zdedit,pano_adjust_event,pano_adjust_compl);        //  run dialog, parallel
   
   start_thread(pano_adjust_thread,0);                                     //  start thread
   wrapup_thread(0);                                                       //  wait for completion
   return;
}


//  thread function - stall return from final_adjust until dialog done

void * pano_adjust_thread(void *)
{
   while (true) thread_idle_loop();                                        //  wait for work or exit request
   return 0;                                                               //  never executed, stop g++ warning
}


//  dialog completion function

int pano_adjust_compl(zdialog *zd, int zstat)
{
   zdialog_free(zdedit);                                                   //  kill dialog
   zdedit = null;
   if (zstat == 1) pano_stat = 1;
   else pano_stat = 0;
   wrapup_thread(8);                                                       //  kill thread
   return 0;
}


//  dialog event function
//  A1 + A2 >> E3 under control of spin buttons

int pano_adjust_event(zdialog *zd, const char *event)
{
   double      red, green, blue, bright, bright2;
   
   if (strEqu(event,"auto")) 
   {
      pano_automatch = 1 - pano_automatch;                                 //  toggle color automatch state
      set_colorfix_factors(pano_automatch);                                //  corresp. color matching factors
      pano_combine(1);                                                     //  combine images and update window
      if (pano_automatch) zdialog_stuff(zd,"labauto","on");
      else zdialog_stuff(zd,"labauto","off");
      return 1;
   }

   if (strNeq(event,"apply")) return 0;                                    //  wait for apply button

   zdialog_fetch(zd,"spred",red);                                          //  get color adjustments
   zdialog_fetch(zd,"spgreen",green);
   zdialog_fetch(zd,"spblue",blue);
   zdialog_fetch(zd,"spbright",bright);                                    //  brightness adjustment
   zdialog_fetch(zd,"spblend",blendstripe);                                //  blend width

   bright2 = (red + green + blue) / 3;                                     //  RGB brightness
   bright = bright / bright2;                                              //  bright setpoint / RGB brightness
   red = red * bright;                                                     //  adjust RGB brightness
   green = green * bright;
   blue = blue * bright;
   
   bright = (red + green + blue) / 3;
   zdialog_stuff(zd,"spred",red);                                          //  force back into consistency
   zdialog_stuff(zd,"spgreen",green);
   zdialog_stuff(zd,"spblue",blue);
   zdialog_stuff(zd,"spbright",bright);

   Radjust = red / 100;                                                    //  normalize 0.5 ... 2.0
   Gadjust = green / 100;
   Badjust = blue / 100;

   get_overlap_region();
   get_Bratios();                                                          //  color ratios for overlap region
   set_colorfix_factors(pano_automatch);                                   //  corresp. color matching factors
   pano_combine(1);                                                        //  combine and update window

   return 1;
}


//  create scaled and curved alignment images in A1rgb48, A2rgb48
//  newf:  create new alignment images from scratch
//  strf:  0: curve both images, y-stretch both halves of image2
//         1: curve image2 only, y-stretch upper half only
//         2: curve image2 only, y-stretch lower half only

void pano_get_align_images(int newf, int strf)
{
   uint16      *pix, *pixc;
   int         pxc, pyc, pxcL, pxcH, pycL, pycH;
   double      px, py, yst;
   double      xdisp, ydisp, xshrink, yshrink;
   double      costf = cos(toff), sintf = sin(toff);
   double      lens_curve, pcurve, pyadj, pbow, R;

   double   mm[12] =    { 20, 22, 25,   26,  28,  32,   40,   48,   60,  80,   100, 200 };
   double   curve[12] = { 35, 19, 10.1, 8.7, 6.8, 5.26, 3.42, 2.32, 1.7, 1.35, 1.2, 1.1 };

   if (newf) 
   {
      A1ww = Fww * alignSize / fullSize;                                   //  size of alignment images
      A1hh = Fhh * alignSize / fullSize;
      A2ww = Gww * alignSize / fullSize;
      A2hh = Ghh * alignSize / fullSize;
      
      RGB_free(A1cache);                                                   //  align images = scaled input images
      RGB_free(A2cache);
      A1cache = RGB_rescale(Frgb48,A1ww,A1hh);
      A2cache = RGB_rescale(Grgb48,A2ww,A2hh);

      RGB_free(A1rgb48);                                                   //  A1/A2 will be curved
      A1rgb48 = RGB_make(A1ww,A1hh,48);

      RGB_free(A2rgb48);
      A2rgb48 = RGB_make(A2ww,A2hh,48);

      spline1(12,mm,curve);                                                //  always initialize   bugfix v.6.2
   }
   
   lens_curve = spline2(lens_mm);
   pcurve = lens_curve * 0.01 * A2ww;                                      //  curve % to pixels
   pyadj = 1.41;                                                           //  y adjustment factor
   pbow = 0.01 * A2ww * lens_bow;                                          //  lens_bow % to pixels

   R = 1.0 * A2hh / A2ww;
   if (R > 1) pcurve = pcurve / R / R;                                     //  adjust for vertical format
   
   if (strf == 0 && A2ww <= 0.8 * A1ww)                                    //  image already curved, copy as-is
      A1rgb48 = RGB_copy(A1cache);                                         //  (apparently 2+ images already)

   if (strf == 0 && A2ww > 0.8 * A1ww)                                     //  curve image1
   {
      for (pyc = 0; pyc < A1hh; pyc++)
      for (pxc = 0; pxc < A1ww; pxc++)
      {
         xdisp = (pxc - A1ww/2.0) / (A1ww/2.0);                            //  -1 .. 0 .. +1
         ydisp = (pyc - A1hh/2.0) / (A1ww/2.0);                            //  -h/w .. 0 .. +h/w
         xshrink = xdisp * xdisp * xdisp;
         yshrink = pyadj * ydisp * xdisp * xdisp;
         
         px = pxc + pcurve * xshrink;                                      //  x shrink pixels
         py = pyc + pcurve * yshrink;                                      //  y shrink pixels
         px -= pbow * xdisp * ydisp * ydisp;                               //  x bow pixels

         pix = vpixel(A1cache,px,py);                                      //  virtual pixel in uncurved image
         pixc = bmpixel(A1rgb48,pxc,pyc);                                  //  destination pixel in curved image
         if (pix) {  
            pixc[0] = pix[0];
            pixc[1] = pix[1];
            pixc[2] = pix[2];
         }
         else pixc[0] = pixc[1] = pixc[2] = 0;
      }
   }
   
   pycL = 0;
   pycH = A2hh;
   if (strf == 1) pycH = A2hh / 2;                                         //  upper half only
   if (strf == 2) pycL = A2hh / 2;                                         //  lower half only

   pxcL = 0;
   pxcH = A2ww;

   if (strf) {                                                             //  during y-stretch alignment,
      pxcL = int((pxmL - xoff) * costf + (pyL - yoff) * sintf);            //    stay within blend stripe
      if (pxcL < 0) pxcL = 0;
      pxcH = int((pxmH - xoff) * costf + (pyH - yoff) * sintf);
      if (pxcH > A2ww/2) pxcH = A2ww/2;
   }

   for (pyc = pycL; pyc < pycH; pyc++)                                     //  curve and y-stretch image2
   for (pxc = pxcL; pxc < pxcH; pxc++)
   {
      xdisp = (pxc - A2ww/2.0) / (A2ww/2.0);
      ydisp = (pyc - A2hh/2.0) / (A2ww/2.0);
      xshrink = xdisp * xdisp * xdisp;
      yshrink = pyadj * ydisp * xdisp * xdisp;

      px = pxc + pcurve * xshrink;
      py = pyc + pcurve * yshrink;
      px -= pbow * xdisp * ydisp * ydisp;

      if (pyc < A2hh / 2) yst = yst1;                                      //  do y-stretch
      else yst = yst2;
      if (yst) py += yst * ydisp;                                          //  linear better than quadratic

      pix = vpixel(A2cache,px,py);
      pixc = bmpixel(A2rgb48,pxc,pyc);
      if (pix) {  
         pixc[0] = pix[0];
         pixc[1] = pix[1];
         pixc[2] = pix[2];
      }
      else pixc[0] = pixc[1] = pixc[2] = 0;
   }

   return;
}


//  combine images: A1rgb48 + A2rgb48  >>  E3rgb48

void pano_combine(int fcolor)
{
   int            px3, py3, ii, max;
   int            red1, green1, blue1;
   int            red2, green2, blue2;
   int            red3, green3, blue3;
   uint16         *pix1, *pix2, *pix3;
   double         ww, px1, py1, px2, py2, f1, f2;
   double         costf = cos(toff), sintf = sin(toff);

   mutex_lock(&bitmaps_lock);
   
   ww = xoff + A2ww;                                                       //  combined width
   if (toff < 0) ww -= A2hh * toff;                                        //  adjust for theta
   E3ww = int(ww+1);
   E3hh = A1rgb48->hh;
   
   RGB_free(E3rgb48);                                                      //  allocate output bitmap
   E3rgb48 = RGB_make(E3ww,E3hh,48);
   
   overlapixs = 0;                                                         //  counts overlapping pixels
   red1 = green1 = blue1 = 0;                                              //  suppress compiler warnings
   red2 = green2 = blue2 = 0;
   
   for (py3 = 0; py3 < E3hh; py3++)                                        //  step through E3 rows
   for (px3 = 0; px3 < E3ww; px3++)                                        //  step through E3 pixels in row
   {
      pix1 = pix2 = 0;
      red3 = green3 = blue3 = 0;

      if (px3 < pxmH) {
         px1 = costf * px3 - sintf * (py3 - yoff);                         //  A1 pixel, after offsets  
         py1 = costf * py3 + sintf * (px3 - xoff);
         pix1 = vpixel(A1rgb48,px1,py1);
      }

      if (px3 >= pxmL) {
         px2 = costf * (px3 - xoff) + sintf * (py3 - yoff);                //  A2 pixel, after offsets
         py2 = costf * (py3 - yoff) - sintf * (px3 - xoff);
         pix2 = vpixel(A2rgb48,px2,py2);
      }

      if (pix1) {
         red1 = pix1[0];
         green1 = pix1[1];
         blue1 = pix1[2];
         if (!red1 && !green1 && !blue1) pix1 = 0;                         //  ignore black pixels
      }

      if (pix2) {
         red2 = pix2[0];
         green2 = pix2[1];
         blue2 = pix2[2];
         if (!red2 && !green2 && !blue2) pix2 = 0;
      }

      if (fcolor) {                                                        //  brightness compensation  
         if (pix1) {                                                       //  (auto + manual adjustments)
            red1 = int(R12match[red1]);
            green1 = int(G12match[green1]);
            blue1 = int(B12match[blue1]);
            if (red1 > 65535 || green1 > 65535 || blue1 > 65535) {
               max = red1;
               if (green1 > max) max = green1;
               if (blue1 > max) max = blue1;
               f1 = 65535.0 / max;
               red1 = int(red1 * f1);
               green1 = int(green1 * f1);
               blue1 = int(blue1 * f1);
            }
         }

         if (pix2) {                                                       //  adjust both images
            red2 = int(R21match[red2]);                                    //    in opposite directions
            green2 = int(G21match[green2]);
            blue2 = int(B21match[blue2]);
            if (red2 > 65535 || green2 > 65535 || blue2 > 65535) {
               max = red2;
               if (green2 > max) max = green2;
               if (blue2 > max) max = blue2;
               f1 = 65535.0 / max;
               red2 = int(red2 * f1);
               green2 = int(green2 * f1);
               blue2 = int(blue2 * f1);
            }
         }
      }

      if (pix1) {
         if (! pix2) {
            red3 = red1;                                                   //  use image1 pixel
            green3 = green1;
            blue3 = blue1; 
         }
         else {                                                            //  use blended pixel
            if (fcolor) {
               if (blendstripe == 0) f1 = 1.0;
               else f1 = 1.0 * (pxmH - px3) / blendstripe;                 //  use progressive blend
               f2 = 1.0 - f1;
               red3 = int(f1 * red1 + f2 * red2);
               green3 = int(f1 * green1 + f2 * green2);
               blue3 = int(f1 * blue1 + f2 * blue2);
            }
            else {                                                         //  use 50/50 mix
               red3 = (red1 + red2) / 2;
               green3 = (green1 + green2) / 2;
               blue3 = (blue1 + blue2) / 2;
            }
            overlapixs++;                                                  //  count overlapped pixels
         }
      }

      else if (pix2) {
            red3 = red2;                                                   //  use image2 pixel
            green3 = green2;
            blue3 = blue2; 
      }

      pix3 = bmpixel(E3rgb48,px3,py3);                                     //  output pixel
      pix3[0] = red3;
      pix3[1] = green3;
      pix3[2] = blue3;
      
      if (showRedpix && pix1 && pix2) {                                    //  highlight alignment pixels
         ii = py3 * A1ww + px3;
         if (redpixels[ii]) {
            pix3[0] = 65535;
            pix3[1] = pix3[2] = 0;
         }
      }
   }

   mutex_unlock(&bitmaps_lock);
   mwpaint2();                                                             //  update window
   return;
}


/**************************************************************************
    HDR and pano shared functions (image matching, alignment, overlay)
***************************************************************************/

/**************************************************************************/

//  compare two doubles for significant difference
//  return:  0  difference not significant
//          +1  d1 > d2
//          -1  d1 < d2

int sigdiff(double d1, double d2, double signf)
{
   double diff = fabs(d1-d2);
   if (diff == 0.0) return 0;
   diff = diff / (fabs(d1) + fabs(d2));
   if (diff < signf) return 0;
   if (d1 > d2) return 1;
   else return -1;
}


/**************************************************************************/

//  Get the rectangle containing the overlap region of two images
//  outputs:   pxL pxH pyL pyH   total image overlap rectangle
//             pxmL pxmH         reduced to overlap match stripe

void get_overlap_region()
{
   int         pxL2, pyL2;

   pxL = 0;
   if (xoff > 0) pxL = int(xoff);

   pxH = A1ww;
   if (pxH > xoff + A2ww) pxH = int(xoff + A2ww);
   
   pyL = 0;
   if (yoff > 0) pyL = int(yoff);
   
   pyH = A1hh;
   if (pyH > yoff + A2hh) pyH = int(yoff + A2hh);

   if (toff > 0) {
      pyL2 = int(yoff + toff * (pxH - pxL));
      if (pyL2 > pyL) pyL = pyL2;
   }

   if (toff < 0) {   
      pxL2 = int(xoff - toff * (pyH - pyL));
      if (pxL2 > pxL) pxL = pxL2;
   }
   
   if (blendstripe > 0) {
      pxM = (pxL + pxH) / 2;                                               //  midpoint of overlap
      pxmL = pxM - int(blendstripe/2);                                     //  overlap stripe x-range for
      pxmH = pxM + int(blendstripe/2);                                     //    image matching & alignment
      if (pxmL < pxL) pxmL = pxL;
      if (pxmH > pxH) pxmH = pxH;
   }
   else {                                                                  //  use whole range
      pxmL = pxL;
      pxmH = pxH;
   }
   
   return;
}


/**************************************************************************/

//  Compute brightness ratio by color for overlapping image areas.         //  compute both image directions
//    (image2 is overlayed on image1, offset by xoff, yoff, toff)
//
//  Outputs: 
//    Bratios1[rgb][ii] = image2/image1 brightness ratio for color rgb
//                        and image1 brightness ii
//    Bratios2[rgb][ii] = image1/image2 brightness ratio for color rgb
//                        and image2 brightness ii

void get_Bratios()
{
   uint16      *pix1, *pix2;
   int         px, py, ii, jj, rgb, npix, npix1, npix2, npix3;
   int         brdist1[3][256], brdist2[3][256];
   double      px1, py1, px2, py2;
   double      brlev1[3][256], brlev2[3][256];
   double      costf = cos(toff), sintf = sin(toff);
   double      a1, a2, b1, b2, bratio = 1;
   double      s8 = 1.0 / 256.0;

   for (rgb = 0; rgb < 3; rgb++)                                           //  clear distributions
   for (ii = 0; ii < 256; ii++)
      brdist1[rgb][ii] = brdist2[rgb][ii] = 0;
      
   npix = 0;

   for (py = pyL; py < pyH; py++)                                          //  scan image1/image2 pixels parallel
   for (px = pxL; px < pxH; px++)                                          //  use entire overlap area
   {
      px1 = costf * px - sintf * (py - yoff);                              //  image1 pixel, after offsets
      py1 = costf * py + sintf * (px - xoff);
      pix1 = vpixel(A1rgb48,px1,py1);
      if (! pix1) continue;                                                //  does not exist
      if (!pix1[0] && !pix1[1] && !pix1[2]) continue;                      //  ignore black pixels

      px2 = costf * (px - xoff) + sintf * (py - yoff);                     //  corresponding image2 pixel
      py2 = costf * (py - yoff) - sintf * (px - xoff);
      pix2 = vpixel(A2rgb48,px2,py2);
      if (! pix2) continue;                                                //  does not exist
      if (!pix2[0] && !pix2[1] && !pix2[2]) continue;                      //  ignore black pixels

      ++npix;                                                              //  count overlapping pixels
      
      for (rgb = 0; rgb < 3; rgb++)                                        //  accumulate distributions
      {                                                                    //    by color in 256 bins
         ++brdist1[rgb][int(s8*pix1[rgb])];
         ++brdist2[rgb][int(s8*pix2[rgb])];
      }
   }
   
   npix1 = npix / 256;                                                     //  1/256th of total pixels
   
   for (rgb = 0; rgb < 3; rgb++)                                           //  get brlev1[rgb][N] = mean bright
   for (ii = jj = 0; jj < 256; jj++)                                       //    for Nth group of image1 pixels
   {                                                                       //      for color rgb
      brlev1[rgb][jj] = 0;
      npix2 = npix1;                                                       //  1/256th of total pixels

      while (npix2 > 0 && ii < 256)                                        //  next 1/256th group from distr,
      {
         npix3 = brdist1[rgb][ii];
         if (npix3 == 0) { ++ii; continue; }
         if (npix3 > npix2) npix3 = npix2;
         brlev1[rgb][jj] += ii * npix3;                                    //  brightness * (pixels with)
         brdist1[rgb][ii] -= npix3;
         npix2 -= npix3;
      }

      brlev1[rgb][jj] = brlev1[rgb][jj] / npix1;                           //  mean brightness for group, 0-255
   }

   for (rgb = 0; rgb < 3; rgb++)                                           //  do same for image2
   for (ii = jj = 0; jj < 256; jj++)
   {
      brlev2[rgb][jj] = 0;
      npix2 = npix1;

      while (npix2 > 0 && ii < 256)
      {
         npix3 = brdist2[rgb][ii];
         if (npix3 == 0) { ++ii; continue; }
         if (npix3 > npix2) npix3 = npix2;
         brlev2[rgb][jj] += ii * npix3;
         brdist2[rgb][ii] -= npix3;
         npix2 -= npix3;
      }

      brlev2[rgb][jj] = brlev2[rgb][jj] / npix1;
   }

   for (rgb = 0; rgb < 3; rgb++)                                           //  color
   for (ii = jj = 0; ii < 256; ii++)                                       //  brlev1 brightness, 0 to 255
   {                                                                       //  bugfix  v.6.4
      if (ii == 0) bratio = 1;
      while (ii > brlev2[rgb][jj] && jj < 256) ++jj;                       //  find matching brlev2 brightness
      a2 = brlev2[rgb][jj];                                                //  next higher value
      b2 = brlev1[rgb][jj];
      if (a2 > 0 && b2 > 0) {
         if (jj > 0) {
            a1 = brlev2[rgb][jj-1];                                        //  next lower value
            b1 = brlev1[rgb][jj-1];
         }
         else   a1 = b1 = 0;
         if (ii == 0)  bratio = b2 / a2;
         else   bratio = (b1 + (ii-a1)/(a2-a1) * (b2-b1)) / ii;            //  interpolate
      }

      if (bratio < 0.2) bratio = 0.2;                                      //  contain outlyers
      if (bratio > 5) bratio = 5;
      Bratios2[rgb][ii] = bratio;
   }

   for (rgb = 0; rgb < 3; rgb++)                                           //  color
   for (ii = jj = 0; ii < 256; ii++)                                       //  brlev2 brightness, 0 to 255
   {                                                                       //  bugfix  v.6.4
      if (ii == 0) bratio = 1;
      while (ii > brlev1[rgb][jj] && jj < 256) ++jj;                       //  find matching brlev1 brightness
      a2 = brlev1[rgb][jj];                                                //  next higher value
      b2 = brlev2[rgb][jj];
      if (a2 > 0 && b2 > 0) {
         if (jj > 0) {
            a1 = brlev1[rgb][jj-1];                                        //  next lower value
            b1 = brlev2[rgb][jj-1];
         }
         else   a1 = b1 = 0;
         if (ii == 0)  bratio = b2 / a2;
         else   bratio = (b1 + (ii-a1)/(a2-a1) * (b2-b1)) / ii;            //  interpolate
      }

      if (bratio < 0.2) bratio = 0.2;                                      //  contain outlyers
      if (bratio > 5) bratio = 5;
      Bratios1[rgb][ii] = bratio;
   }

   return;
}


/**************************************************************************/

//  Set color matching factors
//     on:   Bratios are used for color matching the two images
//     off:  Bratios are not used - use 1.0 instead
//  In both cases, manual settings Radjust/Gadjust/Badjust are used

void set_colorfix_factors(int state)
{
   unsigned      ii, jj;

   if (state) 
   {
      for (ii = 0; ii < 65536; ii++)
      {
         jj = ii >> 8;

         R12match[ii] = sqrt(Bratios1[0][jj]) / Radjust * ii;              //  use sqrt(ratio) so that adjustment
         G12match[ii] = sqrt(Bratios1[1][jj]) / Gadjust * ii;              //    can be applied to both images
         B12match[ii] = sqrt(Bratios1[2][jj]) / Badjust * ii;              //      (image1 down, image2 up)

         R21match[ii] = sqrt(Bratios2[0][jj]) * Radjust * ii;
         G21match[ii] = sqrt(Bratios2[1][jj]) * Gadjust * ii;
         B21match[ii] = sqrt(Bratios2[2][jj]) * Badjust * ii;
      }
   }

   else 
   {
      for (ii = 0; ii < 65536; ii++)
      {
         R12match[ii] = 1.0 / Radjust * ii;
         G12match[ii] = 1.0 / Gadjust * ii;
         B12match[ii] = 1.0 / Badjust * ii;

         R21match[ii] = Radjust * ii;
         G21match[ii] = Gadjust * ii;
         B21match[ii] = Badjust * ii;
      }
   }
   
   return;
}


/**************************************************************************/

//  find pixels of greatest contrast within overlap area
//  flag high-contrast pixels to use in each image compare region

void flag_edge_pixels()
{
   void  flag_edge_pixels2(int pxL, int pxH, int pyL, int pyH, int samp);

   int      samp = pixsamp / 4;
   
   if (redpixels) zfree(redpixels);                                        //  clear flags for alignment pixels
   redpixels = zmalloc(A1ww*A1hh);
   memset(redpixels,0,A1ww*A1hh);

   if (blendstripe == 0) {                                                 //  HDR: use 4 regions
      flag_edge_pixels2(pxL, pxH/2, pyL, pyH/2, samp);
      flag_edge_pixels2(pxH/2, pxH, pyL, pyH/2, samp);
      flag_edge_pixels2(pxL, pxH/2, pyH/2, pyH, samp);
      flag_edge_pixels2(pxH/2, pxH, pyH/2, pyH, samp);
   }
   else {                                                                  //  pano: 4 regions in overlap stripe
      flag_edge_pixels2(pxmL, pxmH, pyL, pyH/4, samp);
      flag_edge_pixels2(pxmL, pxmH, pyH/4, pyH/2, samp);
      flag_edge_pixels2(pxmL, pxmH, pyH/2, pyH-pyH/4, samp);
      flag_edge_pixels2(pxmL, pxmH, pyH-pyH/4, pyH, samp);
   }

   return;
}


//  Find the highest contrast pixels meeting sample size
//  within the specified sub-region of image overlap.

void flag_edge_pixels2(int pxL, int pxH, int pyL, int pyH, int samp)
{
   int         px, py, ii, jj, npix;
   int         red1, green1, blue1, red2, green2, blue2, tcon;
   int         Hdist[256], Vdist[256], Hmin, Vmin;
   double      costf = cos(toff), sintf = sin(toff);
   double      px1, py1, s8 = 1.0 / 769.0;
   uchar       *Hcon, *Vcon;
   uint16      *pix1, *pix2;
   
   npix = (pxH - pxL) * (pyH - pyL);                                       //  overlapping pixels
   if (npix < 100) return;                                                 //  insignificant
   if (samp > npix / 4) samp = npix / 4;                                   //  use max. 1/4 of pixels

   Hcon = (uchar *) zmalloc(npix);                                         //  horizontal pixel contrast 0-255
   Vcon = (uchar *) zmalloc(npix);                                         //  vertical pixel contrast 0-255

   for (py = pyL; py < pyH-2; py++)                                        //  scan image pixels in sub-region
   for (px = pxL; px < pxH-2; px++)
   {
      ii = (py-pyL) * (pxH-pxL) + (px-pxL);
      Hcon[ii] = Vcon[ii] = 0;                                             //  horiz. = vert. contrast = 0

      px1 = costf * px - sintf * (py - yoff);                              //  image1 pixel
      py1 = costf * py + sintf * (px - xoff);
      pix1 = vpixel(A1rgb48,px1,py1);
      if (! pix1) continue;                                                //  does not exist
      red1 = pix1[0];
      green1 = pix1[1];
      blue1 = pix1[2];
      if (!red1 && !green1 && !blue1) continue;                            //  ignore if black

      pix2 = vpixel(A1rgb48,px1+4,py1);                                    //  4 pixels to right
      if (! pix2) continue;                                                //  reject if off edge
      if (! pix2[0] && ! pix2[1] && ! pix2[2]) continue;
      pix2 = vpixel(A1rgb48,px1+2,py1);                                    //  2 pixels to right
      red2 = pix2[0];
      green2 = pix2[1];
      blue2 = pix2[2];
      tcon = abs(red1-red2) + abs(green1-green2) + abs(blue1-blue2);       //  horizontal contrast
      Hcon[ii] = int(tcon * s8);                                           //    0 - 255

      pix2 = vpixel(A1rgb48,px1,py1+4);                                    //  4 pixels below
      if (! pix2) continue;
      if (! pix2[0] && ! pix2[1] && ! pix2[2]) continue;
      pix2 = vpixel(A1rgb48,px1,py1+2);                                    //  2 pixels below
      red2 = pix2[0];
      green2 = pix2[1];
      blue2 = pix2[2];
      tcon = abs(red1-red2) + abs(green1-green2) + abs(blue1-blue2);       //  vertical contrast
      Vcon[ii] = int(tcon * s8);
   }

   for (ii = 0; ii < 256; ii++) Hdist[ii] = Vdist[ii] = 0;                 //  clear contrast distributions

   for (py = pyL; py < pyH-2; py++)                                        //  scan image pixels
   for (px = pxL; px < pxH-2; px++)
   {                                                                       //  build contrast distributions
      ii = (py-pyL) * (pxH-pxL) + (px-pxL);
      ++Hdist[Hcon[ii]];
      ++Vdist[Vcon[ii]];
   }
   
   for (npix = 0, ii = 255; ii > 0; ii--)                                  //  find minimum contrast needed to get
   {                                                                       //    enough pixels for sample size
      npix += Hdist[ii];                                                   //      (horizontal contrast pixels)
      if (npix > samp) break; 
   }
   Hmin = ii; 

   for (npix = 0, ii = 255; ii > 0; ii--)                                  //  (verticle contrast pixels)
   {
      npix += Vdist[ii];
      if (npix > samp) break;
   }
   Vmin = ii;
   
   for (py = pyL; py < pyH-2; py++)                                        //  scan image pixels
   for (px = pxL; px < pxH-2; px++)
   {
      ii = (py-pyL) * (pxH-pxL) + (px-pxL);
      jj = py * A1ww + px;

      if (Hcon[ii] > Hmin) {
         redpixels[jj] = 1;                                                //  flag horizontal group of 3
         redpixels[jj+1] = 1;
         redpixels[jj+2] = 1;
      }

      if (Vcon[ii] > Vmin) {
         redpixels[jj] = 1;                                                //  flag verticle group of 3
         redpixels[jj+A1ww] = 1;
         redpixels[jj+2*A1ww] = 1;
      }
   }
   
   zfree(Hcon);
   zfree(Vcon);
   return;
}


/**************************************************************************/

//  Compare two images in overlapping areas.
//  (image2 is overlayed on image1, offset by xoff, yoff, toff).
//  If blendstripe > 0, compare verticle stripe in the middle.
//  Use pixels with contrast > minimum needed to reach sample size.
//  return: 1 = perfect match, 0 = total mismatch (black/white)

double match_images()
{
   uint16      *pix1, *pix2;
   int         px, py, ii, pixcount, ymid;
   double      px1, py1, px2, py2;
   double      costf = cos(toff), sintf = sin(toff);
   double      match, cmatch, maxcmatch;
   double      yspan, weight;
   
   cmatch = maxcmatch = pixcount = 0;

   if (blendstripe > 0) goto match_overlap;

   for (py = pyL; py < pyH; py++)                                          //  scan image1/image2 pixels parallel
   for (px = pxL; px < pxH; px++)
   {
      ii = py * A1ww + px;                                                 //  skip low-contrast pixels
      if (! redpixels[ii]) continue;

      px1 = costf * px - sintf * (py - yoff);                              //  image1 pixel
      py1 = costf * py + sintf * (px - xoff);
      pix1 = vpixel(A1rgb48,px1,py1);
      if (! pix1) continue;                                                //  does not exist
      if (!pix1[0] && !pix1[1] && !pix1[2]) continue;                      //  ignore black pixels

      px2 = costf * (px - xoff) + sintf * (py - yoff);                     //  corresponding image2 pixel
      py2 = costf * (py - yoff) - sintf * (px - xoff);
      pix2 = vpixel(A2rgb48,px2,py2);
      if (! pix2) continue;
      if (!pix2[0] && !pix2[1] && !pix2[2]) continue;

      match = match_pixels(pix1,pix2);                                     //  compare brightness adjusted
      cmatch += match;                                                     //  accumulate total match
      pixcount++;
   }

   return cmatch / pixcount;

match_overlap:

   if (pxM > A1ww) return 0;                                               //  overlap runs off of image, no match

   yspan = 1.0 / (pyH - pyL);                                              //  prepare params for weight calc.
   ymid = (pyL + pyH) / 2;                                                 //  (weight central pixels more)

   for (py = pyL; py < pyH; py++)                                          //  step through image1 pixels, rows
   {
      weight = 1 - yspan * abs(py - ymid);                                 //  weight: 0.5 .. 1.0 .. 0.5

      for (px = pxmL; px < pxmH; px++)                                     //  step through image1 pixels, cols
      {
         ii = py * A1ww + px;                                              //  skip low-contrast pixels
         if (! redpixels[ii]) continue;

         px1 = costf * px - sintf * (py - yoff);                           //  image1 pixel
         py1 = costf * py + sintf * (px - xoff);
         pix1 = vpixel(A1rgb48,px1,py1);
         if (! pix1) continue;                                             //  does not exist
         if (!pix1[0] && !pix1[1] && !pix1[2]) continue;                   //  ignore black pixels

         px2 = costf * (px - xoff) + sintf * (py - yoff);                  //  corresponding image2 pixel
         py2 = costf * (py - yoff) - sintf * (px - xoff);
         pix2 = vpixel(A2rgb48,px2,py2);
         if (! pix2) continue;
         if (!pix2[0] && !pix2[1] && !pix2[2]) continue;

         match = match_pixels(pix1,pix2);                                  //  compare brightness adjusted
         cmatch += match * weight;                                         //  accumulate total match
         maxcmatch += weight;
         pixcount++;
      }
   }

   return cmatch / maxcmatch;
}


/**************************************************************************/

//  Compare 2 pixels using precalculated brightness ratios
//  1.0 = perfect match   0 = total mismatch (black/white)

double match_pixels(uint16 *pix1, uint16 *pix2)
{
   double      red1, green1, blue1, red2, green2, blue2;
   double      reddiff, greendiff, bluediff, match;
   double      ff = 1.0 / 65536.0;

   red1 = R12match[pix1[0]];
   green1 = G12match[pix1[1]];
   blue1 = B12match[pix1[2]];

   red2 = R21match[pix2[0]];
   green2 = G21match[pix2[1]];
   blue2 = B21match[pix2[2]];

   reddiff = ff * fabs(red1-red2);                                         //  0 = perfect match
   greendiff = ff * fabs(green1-green2);                                   //  1 = total mismatch
   bluediff = ff * fabs(blue1-blue2);
   
   match = (1.0 - reddiff) * (1.0 - greendiff) * (1.0 - bluediff);
   return match;
}


/**************************************************************************/

//  Get a virtual pixel at location (px,py) (real) in an RGB-48 bitmap.
//  Get the overlapping real pixels and build a composite.

uint16 * vpixel(RGB *rgb, double px, double py)
{
   int            ww, hh, px0, py0;
   uint16         *ppix, *pix0, *pix1, *pix2, *pix3;
   double         f0, f1, f2, f3;
   double         red, green, blue;
   static int     which = 0;
   static uint16  rpix[6];
   
   ww = rgb->ww;
   hh = rgb->hh;
   ppix = (uint16 *) rgb->bmp;

   which = 3 - which;                                                      //  2 alternating return pixels

   px0 = int(px);                                                          //  pixel containing (px,py)
   py0 = int(py);

   if (px0 < 1 || py0 < 1) return 0;
   if (px0 > ww-3 || py0 > hh-3) return 0;
   
   pix0 = ppix + (py0 * ww + px0) * 3;                                     //  4 pixels based at (px0,py0)
   pix1 = pix0 + ww * 3;
   pix2 = pix0 + 3;
   pix3 = pix0 + ww * 3 + 3;

   f0 = (px0+1 - px) * (py0+1 - py);                                       //  overlap of (px,py)
   f1 = (px0+1 - px) * (py - py0);                                         //   in each of the 4 pixels
   f2 = (px - px0) * (py0+1 - py);
   f3 = (px - px0) * (py - py0);
   
   red =   f0 * pix0[0] + f1 * pix1[0] + f2 * pix2[0] + f3 * pix3[0];      //  sum the weighted inputs
   green = f0 * pix0[1] + f1 * pix1[1] + f2 * pix2[1] + f3 * pix3[1];
   blue =  f0 * pix0[2] + f1 * pix1[2] + f2 * pix2[2] + f3 * pix3[2];
   
   rpix[which] = int(red);
   rpix[which+1] = int(green);
   rpix[which+2] = int(blue);

   return rpix + which;
}


/**************************************************************************

   edit transaction and thread support functions         overhauled  v.6.2

edit transaction management
   edit_setup()                     start new edit - copy E3 > E1
   edit_cancel()                    cancel edit - E1 > E3, delete E1
   edit_done()                      commit edit - add to undo stack
   edit_undo()                      undo edit - E1 > E3
   edit_redo()                      redo edit - run thread again
   edit_fullsize(flag)              convert preview to full-size bitmaps

main level thread management
   start_thread(func,arg)           start thread running
   signal_thread()                  signal thread that work is pending
   wait_thread_idle()               wait for pending work complete
   wrapup_thread(command)           wait for exit or command thread exit

thread program
   thread_idle_loop()               wait for pending work, exit if commanded
   exit_thread()                    exit thread unconditionally
   
thread_status (thread ownership
   0     no thread is running
   1     thread is running and idle (no work)
   2     thread is working
   0     thread has exited

thread_command (main program ownership)
   0     idle, no work pending
   8     exit when pending work is done
   9     exit now, unconditionally

thread_pend    work requested counter
thread_done    work done counter
thread_hwm     high water mark

***************************************************************************/


int      edit_busy = 0;
int      thread_command = 0, thread_status = 0;
int      thread_pend = 0, thread_done = 0, thread_hwm = 0;


/**************************************************************************/

//  Setup for a new edit transaction
//  Create E1 (edit input) and E3 (edit output) bitmaps from
//  previous edit output (Frgb48) or image file (Frgb24/48).

int edit_setup(int preview)
{
   if (! image_file) return 0;                                             //  no open image file
   if (! menulock(1)) return 0;                                            //  lock menu
   
   if (Pundo > Pundo_max) {
      zmessageACK(ZTX("too many undo buffers, please save image"));
      return 0;
   }

   if (Fimageturned) {                                                     //  un-turn if needed   v.6.8
      upright_image(360 - Fimageturned);
      Fimageturned = 0;
   }

   mutex_lock(&bitmaps_lock);                                              //  lock bitmaps

   if (preview) Fpreview = 1;                                              //  set preview mode: edit
   else Fpreview = 0;                                                      //    window or full image bitmap
   if (sa_exist) Fpreview = 0;                                             //  not if select area is active
   if (Fpreview) Fzoom = 0;                                                //  v.6.3

   if (! Frgb48) Frgb48 = image_load(image_file,48);                       //  create Frgb48 if not already
   if (! Frgb48) {
      mutex_unlock(&bitmaps_lock);
      menulock(0);
      return 0;
   }
   
   RGB_free(E1rgb48);                                                      //  free prior edit bitmaps
   RGB_free(E3rgb48);

   if (Fpreview)                                                           //  edit bitmaps are window-size
      E1rgb48 = RGB_rescale(Frgb48,Mww,Mhh);                               //    E1rgb48 = Frgb48 scaled to window
   else E1rgb48 = RGB_copy(Frgb48);                                        //  edit bitmaps are full-size
   E3rgb48 = RGB_copy(E1rgb48);                                            //  E1 >> E3

   E1ww = E3ww = E1rgb48->ww;
   E1hh = E3hh = E1rgb48->hh;

   save_undo();                                                            //  save Frgb48 in undo stack
   Fmodified = 0;                                                          //  image not modified yet
   thread_command = thread_status = 0;                                     //  no thread running
   thread_pend = thread_done = thread_hwm = 0;                             //  no work pending or done

   mutex_unlock(&bitmaps_lock);
   mwpaint2();                                                             //  v.6.8
   return 1;
}


/**************************************************************************/

//  process edit dialog [cancel]

void edit_cancel()
{
   if (edit_busy) return;                                                  //  v.6.8
   edit_busy++;

   wrapup_thread(9);                                                       //  tell thread to quit now

   if (zdedit) {                                                           //  v.6.6
      zdialog_free(zdedit);                                                //  kill dialog
      zdedit = null;
   }

   mutex_lock(&bitmaps_lock);
   RGB_free(E1rgb48);                                                      //  free edit bitmaps E1, E3
   E1rgb48 = 0;
   RGB_free(E3rgb48);
   E3rgb48 = 0;
   RGB_free(Mrgb24);
   Mrgb24 = 0;
   E1ww = E3ww = Mww = 0;
   mutex_unlock(&bitmaps_lock);

   Fmodified = Fpreview = 0;                                               //  reset flags
   Ndotlines = 0;
   menulock(0);                                                            //  unlock menu
   mwpaint2();                                                             //  refresh window
   edit_busy = 0;
   return;
}   


/**************************************************************************/

//  process edit dialog [done]  
//  E3rgb48 >> Frgb48 >> Frgb24

void edit_done()
{
   if (edit_busy) return;                                                  //  v.6.8
   edit_busy++;

   if (Fpreview && Fmodified) edit_fullsize(0);                            //  update full image
   wrapup_thread(8);                                                       //  wait for thread done

   if (zdedit) {                                                           //  v.6.6
      zdialog_free(zdedit);                                                //  kill dialog
      zdedit = null;
   }
   
   mutex_lock(&bitmaps_lock);

   if (Fmodified) {
      RGB_free(Frgb48);                                                    //  memory leak   v.6.8
      Frgb48 = RGB_copy(E3rgb48);                                          //  E3 >> Frgb48
      RGB_free(Frgb24);
      Frgb24 = RGB_convbpp(Frgb48);                                        //  Frgb48 >> Frgb24
      Fww = Frgb24->ww;
      Fhh = Frgb24->hh;
      RGB_free(Mrgb24);
      Mrgb24 = 0;
      Mww = 0;
      Pundo++;
      Pumax = Pundo;
      save_undo();                                                         //  save next undo state
   }

   RGB_free(E1rgb48);                                                      //  free edit bitmaps
   RGB_free(E3rgb48);
   E1rgb48 = E3rgb48 = 0;
   E1ww = E3ww = 0;
   
   Fmodified = Fpreview = 0;                                               //  reset flags
   Ndotlines = 0;
   mutex_unlock(&bitmaps_lock);
   menulock(0);                                                            //  unlock menu
   mwpaint2();                                                             //  update window
   edit_busy = 0;
   return;
}


/**************************************************************************/

//  process edit dialog [undo] and [redo] (not the toolbar functions)

void edit_undo()
{
   uint16   *pix1, *pix3;
   int     px, py;
   
   if (! Fmodified) return;                                                //  v.6.3

   if (edit_busy) return;                                                  //  v.6.8
   edit_busy++;
   
   if (sa_exist)                                                           //  select area exists
   {
      for (int ii = 0; ii < sa_npix; ii++)                                 //  restore enclosed pixels
      {
         px = sa_pix[ii].px;
         py = sa_pix[ii].py;
         pix1 = bmpixel(E1rgb48,px,py);
         pix3 = bmpixel(E3rgb48,px,py);
         pix3[0] = pix1[0];
         pix3[1] = pix1[1];
         pix3[2] = pix1[2];
      }
   }

   else 
   {                                                                       //  restore entire image
      mutex_lock(&bitmaps_lock);
      RGB_free(E3rgb48);                                                   //  (size may have changed)
      E3rgb48 = RGB_copy(E1rgb48);
      E3ww = E1ww;
      E3hh = E1hh;
      RGB_free(Mrgb24);
      Mrgb24 = 0;
      Mww = 0;
      Fmodified = 0;                                                       //  reset image modified status
      mutex_unlock(&bitmaps_lock);
   }

   mwpaint2();                                                             //  refresh window
   edit_busy = 0;
   return;
}


void edit_redo()
{
   if (edit_busy) return;                                                  //  v.6.8
   edit_busy++;
   signal_thread();                                                        //  start thread working
   Fmodified = 1;
   edit_busy = 0;
   return;
}


/**************************************************************************/

//  Convert from preview mode (window-size bitmaps) to full-size bitmaps.
//  If flag is set, copy E3 >> E1 to make last edit the new baseline
//  (select area started when Fpreview was set).

void edit_fullsize(int flag)                                               //  v.6.2
{
   if (! Fpreview) return;
   Fpreview = 0;

   mutex_lock(&bitmaps_lock);
   RGB_free(E1rgb48);                                                      //  free preview bitmaps
   RGB_free(E3rgb48);
   E1rgb48 = RGB_copy(Frgb48);                                             //  make full-size bitmaps
   E3rgb48 = RGB_copy(Frgb48);
   E1ww = E3ww = Fww;
   E1hh = E3hh = Fhh;
   mutex_unlock(&bitmaps_lock);

   signal_thread();                                                        //  reinstate edits
   wait_thread_idle();
   return;
}


/**************************************************************************/

//  update progress for long-running functions on status bar
//  update in steps of 1%
//  call with args = 0 when done

void edit_progress(int done, int goal)
{
   static int     ppercentdone = 0;
   
   if (goal == 0) {
      PercentDone = ppercentdone = 0;
      if (pthread_equal(tid_fmain,pthread_self())) update_statusbar();
      else SBupdate++;
      return;
   }

   PercentDone = 100.0 * done / goal;
   if (PercentDone - ppercentdone < 1) return;

   ppercentdone = PercentDone;
   if (pthread_equal(tid_fmain,pthread_self())) update_statusbar();
   else SBupdate++;
   return;
}


/**************************************************************************/

//  start thread that does the edit work

void start_thread(threadfunc func, void *arg)
{
   thread_status = 1;                                                      //  thread is running
   thread_command = thread_pend = thread_done = thread_hwm = 0;            //  nothing pending
   start_detached_thread(func,arg);
   return;
}


//  signal thread that work is pending

void signal_thread()
{
   if (thread_status > 0) thread_pend++;                                   //  v.6.2
   return;
}


//  wait for edit thread to complete pending work and become idle

void wait_thread_idle()
{
   while (thread_pend > thread_done)
   {
      if (! thread_status) break;                                          //  no thread running
      zmainloop();
      zsleep(0.05);
   }
   
   return;
}


//  wait for thread exit or command thread exit
//  status = 0    wait for normal completion
//           8    finish pending work and exit
//           9    quit, exit now

void wrapup_thread(int command)
{
   thread_command = command;                                               //  tell thread to quit or finish

   while (thread_status > 0)                                               //  wait for thread to finish
   {                                                                       //    pending work and exit
      zmainloop();
      zsleep(0.05);
   }

   return;
}


//  idle loop for threads - wait for work request or exit command
//  thread work loop includes call to thread_idle_loop()

void thread_idle_loop()
{
   thread_status = 1;                                                      //  status = idle
   thread_done = thread_hwm;                                               //  work done = high-water mark

   while (true)
   {
      if (thread_command == 9) exit_thread();                              //  quit now command
      if (thread_command == 8)                                             //  finish work and exit
         if (thread_pend <= thread_done) exit_thread();
      if (thread_pend > thread_done) break;                                //  wait for work request
      zsleep(0.01);
   }
   
   thread_hwm = thread_pend;                                               //  set high-water mark
   thread_status = 2;                                                      //  thread is working
   return;                                                                 //  loop to thread
}


//  thread exit unconditionally

void exit_thread()
{
   thread_pend = thread_done = thread_hwm = 0;
   thread_status = 0;
   pthread_exit(0);
}


/**************************************************************************/

//  [undo] menu function - reinstate previous edit in undo/redo stack

void m_undo(GtkWidget *, const char *)
{
   if (Pundo == 0) return;
   if (! menulock(1)) return;
   Pundo--;
   load_undo();
   menulock(0);
   return;
}


//  [redo] menu function - reinstate next edit in undo/redo stack

void m_redo(GtkWidget *, const char *)
{
   if (Pundo == Pumax) return;
   if (! menulock(1)) return;
   Pundo++;
   load_undo();
   menulock(0);
   return;
}


//  Save Frgb48 to undo/redo file stack
//  stack position = Pundo

void save_undo()
{
   char     *pp, buff[24];
   int      fid, cc, cc2;

   pp = strstr(undo_files,"_undo_");
   if (! pp) zappcrash("undo/redo stack corrupted 1");
   snprintf(pp+6,3,"%02d",Pundo);
   
   fid = open(undo_files,O_WRONLY|O_CREAT|O_TRUNC,0640);
   if (! fid) zappcrash("undo/redo stack corrupted 2");

   snprintf(buff,24," %05d %05d fotoxx ",Fww,Fhh);
   cc = write(fid,buff,20);
   if (cc != 20) zappcrash("undo/redo stack corrupted 3");
   
   cc = Fww * Fhh * 6;
   cc2 = write(fid,Frgb48->bmp,cc);
   if (cc2 != cc) zappcrash("undo/redo stack corrupted 4");

   close(fid);
   return;
}


//  Load Frgb48 from undo/redo file stack
//  stack position = Pundo

void load_undo()
{
   char     *pp, buff[24], fotoxx[8];
   int      fid, ww, hh, cc, cc2;

   pp = strstr(undo_files,"_undo_");
   if (! pp) zappcrash("undo/redo stack corrupted 1");
   snprintf(pp+6,3,"%02d",Pundo);
   
   fid = open(undo_files,O_RDONLY);
   if (! fid) zappcrash("undo/redo stack corrupted 2");
   
   *fotoxx = 0;
   cc = read(fid,buff,20);
   sscanf(buff," %d %d %8s ",&ww, &hh, fotoxx);
   if (! strEqu(fotoxx,"fotoxx")) zappcrash("undo/redo stack corrupted 4");

   mutex_lock(&bitmaps_lock);                                              //  v.6.3

   RGB_free(Frgb48);
   Frgb48 = RGB_make(ww,hh,48);
   cc = ww * hh * 6;
   cc2 = read(fid,Frgb48->bmp,cc);
   if (cc2 != cc) zappcrash("undo/redo stack corrupted 5");
   close(fid);
   
   RGB_free(Frgb24);
   Frgb24 = RGB_convbpp(Frgb48);
   Fww = ww;
   Fhh = hh;
   RGB_free(Mrgb24);                                                       //  v.6.8
   Mrgb24 = 0;
   Mww = 0;

   mutex_unlock(&bitmaps_lock);
   mwpaint2();
   return;
}


/**************************************************************************
      tools functions
***************************************************************************/

/**************************************************************************/

//  set new image zoom level or magnification

void m_zoom(GtkWidget *, const char *menu)
{
   int      ii, Dww, Dhh;
   char     zoom;
   double   scalew, scaleh, fitscale;
   double   scales[9] = { 0.125, 0.176, 0.25, 0.354, 0.5, 0.71, 1.0, 1.41, 2.0 };
   
   if (strnEqu(menu,"Zoom",4)) zoom = menu[4];                             //  get + or -
   else  zoom = *menu;
   
   Dww = drWin->allocation.width;                                          //  drawing window size
   Dhh = drWin->allocation.height;

   if (Fww > Dww || Fhh > Dhh) {                                           //  get window fit scale
      scalew = 1.0 * Dww / Fww;
      scaleh = 1.0 * Dhh / Fhh;
      if (scalew < scaleh) fitscale = scalew;
      else fitscale = scaleh;
   }
   else fitscale = 1.0;                                                    //  if image < window use 100%
   
   if (zoom == '+') {                                                      //  zoom bigger
      if (! Fzoom) Fzoom = fitscale / 1.2;
      Fzoom = Fzoom * sqrt(2.0);                                           //  new scale: 41% bigger
      for (ii = 0; ii < 9; ii++)
         if (Fzoom < 1.01 * scales[ii]) break;                             //  next higher scale in table
      if (ii == 9) ii = 8;
      Fzoom = scales[ii];
      if (Fzoom < fitscale) Fzoom = 0;                                     //  image < window
   }

   if (zoom == '-') Fzoom = 0;                                             //  zoom to fit window

   if (zoom == 'Z') {
      if (Fzoom != 0) Fzoom = 0;                                           //  toggle 100% and fit window
      else  Fzoom = 1;
   }

   mwpaint2();                                                             //  refresh window
   return;
}


/**************************************************************************/

//  monitor test function

void m_montest(GtkWidget *, const char *)
{
   uint8       *pixel;
   int         red, green, blue;
   int         row, col, row1, row2;
   int         ww = 800, hh = 500;
   
   if (mod_keep()) return;
   if (! menulock(1)) return;

   /// free_resources();                                                   //  v.6.6
   mutex_lock(&bitmaps_lock);

   RGB_free(Frgb24);
   Frgb24 = RGB_make(ww,hh,24);
   Fww = ww;
   Fhh = hh;
   file_bpp = 24;
   file_MB = 0;

   for (red = 0; red <= 1; red++)
   for (green = 0; green <= 1; green++)
   for (blue = 0; blue <= 1; blue++)
   {
      row1 = 4 * red + 2 * green + blue;                                   //  row 0 to 7
      row1 = row1 * hh / 8;                                                //  stripe, 1/8 of image
      row2 = row1 + hh / 8;
      
      for (row = row1; row < row2; row++)
      for (col = 0; col < ww; col++)
      {
         pixel = (uint8 *) Frgb24->bmp + (row * ww + col) * 3;
         pixel[0] = red * 256 * col / ww;
         pixel[1] = green * 256 * col / ww;
         pixel[2] = blue * 256 * col / ww;
      }
   }

   Fzoom = 0;                                                              //  scale to window
   gtk_window_set_title(GTK_WINDOW(mWin),"monitor check");
   mutex_unlock(&bitmaps_lock);
   mwpaint2();                                                             //  repaint window
   menulock(0);
   return;
}


/**************************************************************************/

//  Rebuild assigned tags index and thumbnail files.
//  Process all image files within given top-level directory.

void m_index_tt(GtkWidget *, const char *)
{
   PXB         *pixbuf;
   char        *filespec1, *filespec2, *thumbdirk;
   char        *subdirk, stbartext[200];
   char        *pp, **ppv;
   const char  *exifkeys[2] = { exif_date_key, exif_tags_key };
   char        *imagedate, *imagetags;
   int         contx = 0, err, cc, fcount = 0;
   FILE        *fid;

   if (! Fexiftool) {                                                      //  exiftool is required
      zmessageACK(Bexiftoolmissing);
      return;
   }

   pp = zgetfile(ZTX("select top image directory"),topdirk,"folder");
   if (! pp) return;
   if (topdirk) zfree(topdirk);
   topdirk = pp;

   if (! menulock(1)) return;

   fid = fopen(asstagsfile,"w");                                           //  open assigned tags file
   if (! fid) goto tagsfile_error;
   
   while ((subdirk = command_output(contx,"find \"%s\" -type d",topdirk))) //  get all subdirectories   v.6.9.2
   {
      pp = strrchr(subdirk,'/');
      if (pp && strEqu(pp,"/.thumbnails")) {                               //  ignore .thumbnails
         zfree(subdirk);
         continue;
      }

      image_xthumbs(subdirk,"init");                                       //  get all image files in directory
      filespec1 = image_xthumbs(subdirk,"first");
      Fsearchlist = 0;
      thumbdirk = 0;

      while (filespec1)
      {
         if (image_file_type(filespec1) == 2) 
         {
            if (! thumbdirk) {                                             //  when first image file found,
               thumbdirk = strdupz(subdirk,20);                            //    create .thumbnails if needed
               strcat(thumbdirk,"/.thumbnails");
               err = mkdir(thumbdirk,0751);
               if (err && errno != EEXIST) 
                  zappcrash(ZTX("cannot create %s \n %s"),thumbdirk,strerror(errno));
            }

            pixbuf = image_thumbnail(filespec1);                           //  create thumbnail for image file
            if (pixbuf) g_object_unref(pixbuf);
            

            ppv = exif_get(filespec1,exifkeys,2);

            imagedate = ppv[0];
            imagetags = ppv[1];

            if (imagedate && strlen(imagedate) > 9) {
               imagedate[10] = 0;                                          //  truncate to yyyy:mm:dd
               fprintf(fid,"date: %s\n",imagedate);
               zfree(imagedate);
            }
            else fprintf(fid,"date: null\n");                              //  no date >> "null"
            
            if (imagetags && strlen(imagetags)) {
               fprintf(fid,"tags: %s\n",imagetags);
               zfree(imagetags);
            }
            else fprintf(fid,"tags: null\n");                              //  no tags >> "null"
               
            cc = fprintf(fid,"file: %s\n\n",filespec1);                    //  output image filespec
            if (! cc) goto tagsfile_error;

            snprintf(stbartext,199,"%5d %s",++fcount,filespec1);           //  update status bar
            stbar_message(STbar,stbartext);
         }
         
         filespec2 = image_xthumbs(filespec1,"next");                      //  next image file
         zfree(filespec1);
         filespec1 = filespec2;
      }

      zfree(subdirk);
   }

   err = fclose(fid);                                                      //  close tag file
   if (err) goto tagsfile_error;

   snprintf(stbartext,199,"%5d completed",fcount);
   stbar_message(STbar,stbartext);
   menulock(0);
   return;

tagsfile_error:
   zappcrash(ZTX("assigned tags file error: %s"),strerror(errno));
   menulock(0);
   return;
}


/**************************************************************************/

//  create or update brightness distribution graph

void m_brightgraph(GtkWidget *, const char *)                              //  menu function
{
   if (! Mrgb24) return;

   if (brightgraph) {
      brightgraph_paint();
      return;
   }

   brightgraph = gtk_window_new(GTK_WINDOW_TOPLEVEL);
   gtk_window_set_title(GTK_WINDOW(brightgraph),ZTX("brightness distribution"));
   gtk_window_set_transient_for(GTK_WINDOW(brightgraph),GTK_WINDOW(mWin));
   gtk_window_set_default_size(GTK_WINDOW(brightgraph),300,200);
   gtk_window_set_position(GTK_WINDOW(brightgraph),GTK_WIN_POS_MOUSE);
   
   drbrightgraph = gtk_drawing_area_new();
   gtk_container_add(GTK_CONTAINER(brightgraph),drbrightgraph);

   G_SIGNAL(brightgraph,"destroy",brightgraph_destroy,0)
   G_SIGNAL(drbrightgraph,"expose-event",brightgraph_paint,0)

   gtk_widget_show_all(brightgraph);
   
   return;
}


void brightgraph_paint()                                                   //  paint graph window
{
   int         brdist[20], nbins = 20;
   int         px, py, ii;
   int         winww, winhh;
   int         ww, hh, orgx, orgy;
   int         dist_maxbin = 0;
   uint8       *pixel;
   double      bright;

   if (! brightgraph) return;
   if (! Mrgb24) return;

   for (ii = 0; ii < nbins; ii++)                                          //  clear brightness distribution
      brdist[ii] = 0;

   mutex_lock(&bitmaps_lock);

   for (py = 0; py < Mhh; py++)                                            //  compute brightness distribution
   for (px = 0; px < Mww; px++)                                            //    for image in visible window
   {
      pixel = (uint8 *) Mrgb24->bmp + (py * Mww + px) * 3;
      bright = brightness(pixel);                                          //  0 to 256
      brdist[int(bright / 256 * nbins)]++;                                 //  0 to nbins
   }

   mutex_unlock(&bitmaps_lock);

   gdk_window_clear(drbrightgraph->window);

   winww = drbrightgraph->allocation.width;
   winhh = drbrightgraph->allocation.height;
   
   for (ii = 0; ii < nbins; ii++)
      if (brdist[ii] > dist_maxbin) dist_maxbin = brdist[ii];

   for (ii = 0; ii < nbins; ii++)
   {
      ww = winww / nbins;
      hh = int(0.9 * winhh * brdist[ii] / dist_maxbin);
      orgx = ii * ww;
      orgy = winhh - hh;
      gdk_draw_rectangle(drbrightgraph->window,gdkgc,1,orgx,orgy,ww,hh);
   }
   
   return;
}


void brightgraph_destroy()                                                 //  delete window
{
   if (brightgraph) gtk_widget_destroy(brightgraph);
   brightgraph = 0;
   return;
}


/**************************************************************************/

//  start a new instance of fotoxx in parallel

void m_clone(GtkWidget *, const char *)
{
   char     command[300];
   int      ignore;
   
   strcpy(command,"fotoxx");
   if (image_file) strncatv(command,299," \"",image_file,"\" &",null);
   else strcat(command," &");
   ignore = system(command);
   return;
}


/**************************************************************************/

//  enter or leave slideshow mode

void m_slideshow(GtkWidget *, const char *)
{
   int            key;
   static int     SSmode = 0, ww, hh;
   
   key = KBkey;                                                            //  get KB key and reset
   KBkey = 0;

   if (! SSmode) 
   {
      if (key == GDK_Escape) return;
      gtk_window_get_size(GTK_WINDOW(mWin),&ww,&hh);
      gtk_widget_hide_all(GTK_WIDGET(mMbar));                              //  enter slide show mode
      gtk_widget_hide_all(GTK_WIDGET(mTbar));
      gtk_widget_hide_all(GTK_WIDGET(STbar));
      gtk_window_fullscreen(GTK_WINDOW(mWin));
      SSmode = 1;
   }

   else
   {
      gtk_window_unfullscreen(GTK_WINDOW(mWin));                           //  leave slide show mode
      gtk_window_resize(GTK_WINDOW(mWin),ww,hh);
      gtk_widget_show_all(GTK_WIDGET(mMbar));
      gtk_widget_show_all(GTK_WIDGET(mTbar));
      gtk_widget_show_all(GTK_WIDGET(STbar));
      SSmode = 0;
   }

   Fzoom = 0;                                                              //  fit image to window
   Fblowup = SSmode;                                                       //  blow-up small images if SS mode
   mwpaint2(); 
   return;
}


/**************************************************************************/

//  show RGB values for pixel at mouse click

CBfunc   *RGB_func_save;
int      RGB_mcap_save;


void m_RGB(GtkWidget *, const char *)                                      //  menu function
{
   void RGB_mousefunc();

   if (! Frgb24) return;                                                   //  no image
   RGB_func_save = mouseCBfunc;
   RGB_mcap_save = Mcapture;
   mouseCBfunc = RGB_mousefunc;                                            //  connect mouse function
   Mcapture = 1;                                                           //  capture mouse clicks
   return;
}

void RGB_mousefunc()                                                       //  mouse function
{
   int         px, py;
   double      red, green, blue;
   double      fbright, fred;
   char        work[90];
   uint8       *ppix24;
   uint16      *ppix48;
   
   if (! Frgb24) return;

   if (LMclick)                                                            //  left mouse click
   {
      LMclick = 0;
      px = Mxclick;                                                        //  click position
      py = Myclick;
      
      if (E3rgb48) {                                                       //  use current image being edited
         if (px < 0 || px > E3ww-1 ||                                      //  outside image area     v.6.4
             py < 0 || py > E3hh-1) return;
         ppix48 = bmpixel(E3rgb48,px,py);                                  //  bugfix: * Mscale removed  v.6.7
         red = ppix48[0] / 256.0;
         green = ppix48[1] / 256.0;
         blue = ppix48[2] / 256.0;
         fbright = brightness(ppix48) / 256.0;
         fred = redness(ppix48);
      }

      else if (Frgb48) {                                                   //  use edited image
         if (px < 0 || px > Fww-1 || 
             py < 0 || py > Fhh-1) return;
         ppix48 = bmpixel(Frgb48,px,py);
         red = ppix48[0] / 256.0;
         green = ppix48[1] / 256.0;
         blue = ppix48[2] / 256.0;
         fbright = brightness(ppix48) / 256.0;
         fred = redness(ppix48);
      }

      else  {                                                              //  use 24 bpp image
         if (px < 0 || px > Fww-1 || 
             py < 0 || py > Fhh-1) return;
         ppix24 = (uint8 *) Frgb24->bmp + (py * Fww + px) * 3;
         red = ppix24[0];
         green = ppix24[1];
         blue = ppix24[2];
         fbright = brightness(ppix24);
         fred = redness(ppix24);
      }
      
      snprintf(work,89,"RGB: %6.3f %6.3f %6.3f  brightness: %6.3f  redness: %.1f",
                    red, green, blue, fbright, fred);
      stbar_message(STbar,work);
   }
   
   if (RMclick)
   {
      RMclick = 0;                                                         //  v.6.1.2
      Mcapture = RGB_mcap_save;
      mouseCBfunc = RGB_func_save;                                         //  disconnect mouse
   }
   
   return;
}


/**************************************************************************/

//  choose or set lens parameters for panoramas

void  m_parms(GtkWidget *, const char *)
{
   zdialog     *zd;
   int         ii, zstat, radb;
   char        text[20];
   
   zd = zdialog_new(ZTX("Lens Parameters"),mWin,Bapply,Bcancel,0);
   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=5");
   
   zdialog_add_widget(zd,"vbox","vb1","hb1",0,"space=5|homog");            //        Lens    mm    bow
   zdialog_add_widget(zd,"vbox","vb2","hb1",0,"space=5|homog");            //   (o)  name1   30    0.33
   zdialog_add_widget(zd,"vbox","vb3","hb1",0,"space=5|homog");            //   (x)  name2   40    0.22
   zdialog_add_widget(zd,"vbox","vb4","hb1",0,"space=5|homog");            //   (o)  name3   45    0.28
   zdialog_add_widget(zd,"label","space","vb1");                           //   (o)  name4   50    0.44
   zdialog_add_widget(zd,"radio","radb0","vb1",0);                         //
   zdialog_add_widget(zd,"radio","radb1","vb1",0);                         //         [apply]  [cancel]
   zdialog_add_widget(zd,"radio","radb2","vb1",0);
   zdialog_add_widget(zd,"radio","radb3","vb1",0);
   zdialog_add_widget(zd,"label","lname","vb2","Lens name");
   zdialog_add_widget(zd,"entry","name0","vb2","scc=10");
   zdialog_add_widget(zd,"entry","name1","vb2","scc=10");
   zdialog_add_widget(zd,"entry","name2","vb2","scc=10");
   zdialog_add_widget(zd,"entry","name3","vb2","scc=10");
   zdialog_add_widget(zd,"label","lmm","vb3","Lens mm");
   zdialog_add_widget(zd,"entry","mm0","vb3","0","scc=5");
   zdialog_add_widget(zd,"entry","mm1","vb3","0","scc=5");
   zdialog_add_widget(zd,"entry","mm2","vb3","0","scc=5");
   zdialog_add_widget(zd,"entry","mm3","vb3","0","scc=5");
   zdialog_add_widget(zd,"label","lbow","vb4","Lens bow");
   zdialog_add_widget(zd,"entry","bow0","vb4","0.0","scc=6");
   zdialog_add_widget(zd,"entry","bow1","vb4","0.0","scc=6");
   zdialog_add_widget(zd,"entry","bow2","vb4","0.0","scc=6");
   zdialog_add_widget(zd,"entry","bow3","vb4","0.0","scc=6");
   
   for (ii = 0; ii < 4; ii++)                                              //  stuff lens data into dialog
   {
      snprintf(text,20,"name%d",ii);
      zdialog_stuff(zd,text,lens4_name[ii]);
      snprintf(text,20,"mm%d",ii);
      zdialog_stuff(zd,text,lens4_mm[ii]);
      snprintf(text,20,"bow%d",ii);
      zdialog_stuff(zd,text,lens4_bow[ii]);
   }

   snprintf(text,20,"radb%d",curr_lens);                                   //  current lens = selected
   zdialog_stuff(zd,text,1);

   zstat = zdialog_run(zd,0,0);                                            //  run dialog, get inputs

   if (zstat != 1) {
      zdialog_free(zd);                                                    //  canceled
      return;
   }
   
   for (ii = 0; ii < 4; ii++)                                              //  fetch lens data (revisions)
   {
      snprintf(text,20,"name%d",ii);
      zdialog_fetch(zd,text,lens4_name[ii],lens_cc);
      repl_1str(lens4_name[ii],lens4_name[ii]," ","_");                    //  replace blank with _
      snprintf(text,20,"mm%d",ii);
      zdialog_fetch(zd,text,lens4_mm[ii]);
      snprintf(text,20,"bow%d",ii);
      zdialog_fetch(zd,text,lens4_bow[ii]);
      snprintf(text,20,"radb%d",ii);                                       //  detect which is selected
      zdialog_fetch(zd,text,radb);
      if (radb) curr_lens = ii;
   }
   
   zdialog_free(zd);
   return;
}


/**************************************************************************/

//  set GUI language

void  m_lang(GtkWidget *, const char *)                                    //  v.6.2
{
   zdialog     *zd;
   int         ii, cc, err, zstat;

   char        lang[20], *pp, command[100], locmess[200];

   const char  *langs[10] = { "en  English", "de  German", "fr  French", 
                              "es  Spanish", "gl  Galacian", "el  Greek",
                              "zh_CN  Chinese", 0 };
   
   strcpy(locmess,"Available Translations");
   cc = strlen(locmess);

   for (ii = 0; langs[ii]; ii++)
   {
      strcpy(locmess+cc,"\n ");
      strcpy(locmess+cc+2,langs[ii]);
      cc += strlen(langs[ii]) + 2;
   }
   
   zd = zdialog_new("Set Language",mWin,Bapply,Bcancel,0);
   zdialog_add_widget(zd,"label","lab0","dialog",locmess,"space=5");
   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=10");
   zdialog_add_widget(zd,"label","lab1","hb1","language code");
   zdialog_add_widget(zd,"combo","combo","hb1",0);

   for (ii = 0; langs[ii]; ii++)                                           //  stuff languages into list
      zdialog_cb_app(zd,"combo",langs[ii]);

   zstat = zdialog_run(zd,0,0);                                            //  run dialog

   if (zstat != 1) {
      zdialog_free(zd);                                                    //  canceled
      return;
   }
   
   zdialog_fetch(zd,"combo",lang,20);                                      //  get user input
   zdialog_free(zd);
   
   pp = strchr(lang,' ');                                                  //  isolate lc_RC part
   *pp = 0;

   sprintf(command,"fotoxx -l %s &",lang);                                 //  restart fotoxx
   err = system(command);
   m_quit(0,0);

   return;
}


/**************************************************************************
      Image Tag and EXIF functions
***************************************************************************/

void  edit_tags_fixup(cchar * widgetname);                                 //  fixup tag selection widgets
void  edit_tags_mouse(GtkTextView *, GdkEventButton *, cchar *);           //  select tag via mouse click
int   add_unique_tag(cchar *tag, char *taglist, int maxcc);                //  add tag if unique and enough space
void  add_new_filetag();                                                   //  add tags_atag to tags_filetags
void  delete_filetag();                                                    //  remove tags_atag from tags_filetags
void  add_new_recentag();                                                  //  add tags_atag to tags_recentags

char     tags_pdate[12] = "";                                              //  image date, yyyymmdd
int      tags_pstars = 0;                                                  //  image rating in "stars"
char     tags_atag[maxtag1] = "";                                          //  one tag
char     tags_filetags[maxtag2] = "";                                      //  tags for one file
char     tags_asstags[maxtag3] = "";                                       //  all assigned tags
char     tags_searchfile[maxtagF] = "";                                    //  image search /path*/file*
char     tags_searchtags[maxtag4] = "";                                    //  image search tags
char     tags_recentags[maxtag5] = "";                                     //  recently added tags
int      tags_changed = 0;                                                 //  tags have been changed


/**************************************************************************/

//  edit tags menu function

void m_edit_tags(GtkWidget *, const char *)
{
   if (! Fexiftool) {                                                      //  exiftool is required
      zmessageACK(Bexiftoolmissing);
      return;
   }
   
   edit_tags_dialog();
   return;
}


//  activate edit tags dialog, stuff data from current file

void edit_tags_dialog()
{
   int edit_tags_dialog_event(zdialog *zd, const char *event);
   int edit_tags_dialog_compl(zdialog *zd, int zstat);

   char     *ppv, pstarsN[12];

   if (! image_file) return;

   if (! zdtags)                                                           //  (re) start tag edit dialog 
   {
      load_asstags();                                                      //  get all assigned tags

      zdtags = zdialog_new(ZTX("Edit Tags"),mWin,Bdone,Bcancel,0);         //  tag edit dialog

      zdialog_add_widget(zdtags,"hbox","hb1","dialog",0,"space=5");
      zdialog_add_widget(zdtags,"label","labfile","hb1",ZTX("file:"),"space=10");
      zdialog_add_widget(zdtags,"label","file","hb1");

      zdialog_add_widget(zdtags,"hbox","hb2","dialog",0,"space=5");
      zdialog_add_widget(zdtags,"label","lab21","hb2",ZTX("image date (yyyymmdd)"),"space=10");
      zdialog_add_widget(zdtags,"entry","pdate","hb2",0,"scc=12");

      zdialog_add_widget(zdtags,"hbox","hb3","dialog",0,"space=5");
      zdialog_add_widget(zdtags,"label","labstars","hb3",ZTX("image stars"),"space=10");
      zdialog_add_widget(zdtags,"vbox","vb3","hb3");
      zdialog_add_widget(zdtags,"hbox","hb31","vb3",0,"homog");
      zdialog_add_widget(zdtags,"hbox","hb32","vb3",0,"homog");
      zdialog_add_widget(zdtags,"label","lab30","hb31","0");
      zdialog_add_widget(zdtags,"label","lab31","hb31","1");
      zdialog_add_widget(zdtags,"label","lab32","hb31","2");
      zdialog_add_widget(zdtags,"label","lab33","hb31","3");
      zdialog_add_widget(zdtags,"label","lab34","hb31","4");
      zdialog_add_widget(zdtags,"label","lab35","hb31","5");
      zdialog_add_widget(zdtags,"radio","pstars0","hb32",0);
      zdialog_add_widget(zdtags,"radio","pstars1","hb32",0);
      zdialog_add_widget(zdtags,"radio","pstars2","hb32",0);
      zdialog_add_widget(zdtags,"radio","pstars3","hb32",0);
      zdialog_add_widget(zdtags,"radio","pstars4","hb32",0);
      zdialog_add_widget(zdtags,"radio","pstars5","hb32",0);

      zdialog_add_widget(zdtags,"hbox","hb4","dialog","space=5");
      zdialog_add_widget(zdtags,"label","lab4","hb4",ZTX("current tags"),"space=10");
      zdialog_add_widget(zdtags,"frame","frame4","hb4",0,"expand");
      zdialog_add_widget(zdtags,"edit","filetags","frame4",0,"expand");

      zdialog_add_widget(zdtags,"hbox","hb5","dialog","space=5");
      zdialog_add_widget(zdtags,"label","recent","hb5",ZTX("recently added"),"space=10");
      zdialog_add_widget(zdtags,"frame","frame5","hb5",0,"expand");
      zdialog_add_widget(zdtags,"edit","recentags","frame5",0,"expand");

      zdialog_add_widget(zdtags,"hbox","hb6","dialog",0,"space=5");
      zdialog_add_widget(zdtags,"button","add new tag","hb6",ZTX("create tag"),"space=10");
      zdialog_add_widget(zdtags,"entry","atag","hb6",0);

      zdialog_add_widget(zdtags,"hbox","hb7","dialog",0,"space=5");
      zdialog_add_widget(zdtags,"hbox","hb8","dialog");
      zdialog_add_widget(zdtags,"label","labasstags","hb8",ZTX("assigned tags"),"space=10");
      zdialog_add_widget(zdtags,"frame","frame8","dialog",0,"expand");
      zdialog_add_widget(zdtags,"edit","asstags","frame8",0,"expand");

      zdialog_resize(zdtags,400,300);
      zdialog_run(zdtags,edit_tags_dialog_event,edit_tags_dialog_compl);   //  start dialog
      
      edit_tags_fixup("filetags");                                         //  setup for mouse tag selection
      edit_tags_fixup("asstags");
      edit_tags_fixup("recentags");
   }

   load_filetags(image_file);                                              //  get file tags from EXIF data

   ppv = strrchr(image_file,'/');
   zdialog_stuff(zdtags,"file",ppv+1);                                     //  stuff dialog file name

   zdialog_stuff(zdtags,"pdate",tags_pdate);                               //  stuff dialog data
   sprintf(pstarsN,"pstars%d",tags_pstars);
   zdialog_stuff(zdtags,pstarsN,1);
   zdialog_stuff(zdtags,"filetags",tags_filetags);
   zdialog_stuff(zdtags,"asstags",tags_asstags);
   zdialog_stuff(zdtags,"recentags",tags_recentags);

   return;
}


//  setup tag display widget for tag selection using mouse clicks

void edit_tags_fixup(const char * widgetname)
{
   GtkWidget         *widget;
   GdkWindow         *gdkwin;

   widget = zdialog_widget(zdtags,widgetname);                             //  make widget wrap text
   gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(widget),GTK_WRAP_WORD);
   gtk_text_view_set_editable(GTK_TEXT_VIEW(widget),0);                    //  disable widget editing

   gdkwin = gtk_text_view_get_window(GTK_TEXT_VIEW(widget),textwin);       //  cursor for tag selection
   gdk_window_set_cursor(gdkwin,arrowcursor);

   gtk_widget_add_events(widget,GDK_BUTTON_PRESS_MASK);                    //  connect mouse-click event
   G_SIGNAL(widget,"button-press-event",edit_tags_mouse,widgetname)
}


//  edit tags mouse-click event function
//  get clicked tag and add to or remove from tags_filetags

void edit_tags_mouse(GtkTextView *widget, GdkEventButton *event, const char *widgetname)
{
   GtkTextIter    iter;
   int            mpx, mpy, tbx, tby, offset, cc;
   char           *ptext, *pp1, *pp2;

   if (event->type != GDK_BUTTON_PRESS) return;
   mpx = int(event->x);                                                    //  mouse click position
   mpy = int(event->y);

   gtk_text_view_window_to_buffer_coords(widget,GTK_TEXT_WINDOW_TEXT,mpx,mpy,&tbx,&tby);
   gtk_text_view_get_iter_at_location(widget,&iter,tbx,tby);
   offset = gtk_text_iter_get_offset(&iter);                               //  graphic position in widget text

   ptext = 0;   
   if (strEqu(widgetname,"filetags")) ptext = tags_filetags;               //  get corresponding text
   if (strEqu(widgetname,"asstags")) ptext = tags_asstags;
   if (strEqu(widgetname,"recentags")) ptext = tags_recentags;
   if (strEqu(widgetname,"asstags2")) ptext = tags_asstags;
   if (! ptext) return;

   pp1 = ptext + utf8_position(ptext,offset);                              //  graphic position to byte position
   if (! *pp1 || *pp1 == ' ') return;                                      //  reject ambiguity
   while (pp1 > ptext && *pp1 != ' ') pp1--;                               //  find preceeding delimiter
   if (*pp1 == ' ') pp1++;

   pp2 = strchr(pp1,' ');                                                  //  find delimiter following
   if (pp2) cc = pp2 - pp1;
   else cc = strlen(pp1);
   if (cc >= maxtag1) return;                                              //  reject tag too big
   strncpy0(tags_atag,pp1,cc+1);                                           //  tags_atag = selected tag
   
   if (strEqu(widgetname,"filetags")) {
      delete_filetag();                                                    //  remove tag from file tags
      zdialog_stuff(zdtags,"filetags",tags_filetags);                      //  update dialog widgets
   }
   
   if (strEqu(widgetname,"asstags")) {
      add_new_filetag();                                                   //  add assigned tag to file tags
      add_new_recentag();
      zdialog_stuff(zdtags,"filetags",tags_filetags);
   }

   if (strEqu(widgetname,"recentags")) {
      add_new_filetag();                                                   //  add recent tag to file tags
      zdialog_stuff(zdtags,"filetags",tags_filetags);
   }

   if (strEqu(widgetname,"asstags2")) {                                    //  search dialog:
      zdialog_fetch(zdtags,"searchtags",tags_searchtags,maxtag4);          //  add assigned tag to search tags
      strncatv(tags_searchtags,maxtag4," ",tags_atag,0);
      zdialog_stuff(zdtags,"searchtags",tags_searchtags);
   }

   return;
}


//  edit tags dialog event function

int edit_tags_dialog_event(zdialog *zd, const char *event)
{
   int      err;
   
   if (strEqu(event,"pdate")) {                                            //  image date revised
      err = zdialog_fetch(zd,"pdate",tags_pdate,11);
      if (err) return 1;
      tags_changed++;
   }

   if (strnEqu(event,"pstars",6)) {                                        //  stars revised
      tags_pstars = event[6] - '0';
      tags_changed++;
   }

   if (strEqu(event,"add new tag")) {
      err = zdialog_fetch(zd,"atag",tags_atag,maxtag1);                    //  add new tag to file
      if (err) return 1;                                                   //  reject too big tag
      add_new_filetag();
      add_new_recentag();
      zdialog_stuff(zd,"filetags",tags_filetags);                          //  update dialog widgets
      zdialog_stuff(zd,"asstags",tags_asstags);
      zdialog_stuff(zd,"atag","");
   }

   return 0;
}


int edit_tags_dialog_compl(zdialog *zd, int zstat)
{
   if (zstat == 1) update_filetags(image_file);                            //  done, update file EXIF tags
   zdialog_free(zdtags);                                                   //  kill dialog
   zdtags = null;
   return 0;
}


//  add input tag to output tag list if not already there and enough room
//  returns:   0 = added OK     1 = not unique (case ignored)
//             2 = overflow     3 = bad utf8 characters

int add_unique_tag(const char *tag, char *taglist, int maxcc)
{
   char     *pp1, *pp2, *ppv, temptag1[maxtag1], temptag2[maxtag1];
   int      atcc, cc1, cc2;

   strncpy0(temptag1,tag,maxtag1);                                         //  remove leading and trailing blanks
   atcc = strTrim2(temptag2,temptag1);
   if (! atcc) return 0;
   if (strEqu(temptag2,"(null)")) return 0;                                //  ignore the null tag

   for (ppv = temptag2; *ppv; ppv++)                                       //  replace imbedded blanks with '_'
      if (*ppv == ' ') *ppv = '_';
   
   if (utf8_check(temptag2)) {                                             //  check for valid utf8 encoding
      printf("bad utf8 characters: %s \n",temptag2);
      return 3;
   }
   
   pp1 = taglist;
   cc1 = strlen(temptag2);

   while (true)                                                            //  check if already in tag list
   {
      while (*pp1 == ' ') pp1++;
      if (! *pp1) break;
      pp2 = pp1 + 1;
      while (*pp2 && *pp2 != ' ') pp2++;
      cc2 = pp2 - pp1;
      if (cc2 == cc1 && strncaseEqu(temptag2,pp1,cc1)) return 1;
      pp1 = pp2;
   }
   
   cc2 = strlen(taglist);                                                  //  append to tag list if space enough
   if (cc1 + cc2 + 1 >= maxcc) return 2;
   strcpy(taglist + cc2,temptag2);
   strcpy(taglist + cc2 + cc1," ");
   return 0;
}


/**************************************************************************/

//  image file EXIF data >> tags_pdate, tags_pstars, tags_filetags in memory

void load_filetags(char *file)
{
   int         ii, jj, cc, err;
   const char  *pp;
   const char  *exifkeys[2] = { exif_date_key, exif_tags_key };
   char        **ppv, *imagedate, *imagetags;

   *tags_filetags = *tags_pdate = 0;
   tags_pstars = 0;
   
   ppv = exif_get(image_file,exifkeys,2);
   imagedate = ppv[0];
   imagetags = ppv[1];

   if (imagedate) {                       
      if (strlen(imagedate) > 9) {
         strncpy(tags_pdate,imagedate,4);                                  //  reformat yyyy:mm:dd
         strncpy(tags_pdate+4,imagedate+5,2);                              //    to  yyyymmdd
         strncpy(tags_pdate+6,imagedate+8,2);
         tags_pdate[8] = 0;
      }
      zfree(imagedate);
   }

   if (imagetags)
   {
      for (ii = 1; ; ii++)
      {
         pp = strField(imagetags,' ',ii);                                  //  assume blank delimited tags
         if (! pp) break;
         cc = strlen(pp);
         if (cc >= maxtag1) continue;                                      //  reject tags too big
         for (jj = 0; jj < cc; jj++)
            if (pp[jj] > 0 && pp[jj] < ' ') break;                         //  reject tags with control characters
         if (jj < cc) continue;
         
         if (strnEqu(pp,"stars=",6)) {                                     //  "stars=N" tag
            err = convSI(pp+6,tags_pstars,0,5);
            if (err > 1) tags_pstars = 0;
            continue;
         }

         strcpy(tags_atag,pp);                                             //  add to file tags if unique
         add_new_filetag();
      }

      zfree(imagetags);
   }
   
   tags_changed = 0;
   return;
}


/**************************************************************************/

//  tags_pdate, tags_pstars, tags_filetags in memory >> image file EXIF data

void update_filetags(char *file)
{
   const char  *exifkeys[2] = { exif_date_key, exif_tags_key };
   const char  *exifdata[2];
   char        pdate2[12];

   if (! tags_changed) return;

   strncpy(pdate2,tags_pdate,4);                                           //  yyyymmdd >> yyyy:mm:dd
   strncpy(pdate2+5,tags_pdate+4,2);
   strncpy(pdate2+8,tags_pdate+6,2);
   pdate2[4] = pdate2[7] = ':';
   pdate2[10] = 0;
   
   if (tags_pstars > 0) {                                                  //  add "stars=N" tag
      sprintf(tags_atag,"stars=%d",tags_pstars);
      add_new_filetag();
   }
   
   exifdata[0] = pdate2;
   exifdata[1] = tags_filetags;
   exif_set(file,exifkeys,exifdata,2);
   
   update_asstags(file);                                                   //  update assigned tags file
   tags_changed = 0;
   return;
}


//  add new tag to file tags, if not already and enough space.

void add_new_filetag()
{
   int         err;

   err = add_unique_tag(tags_atag,tags_filetags,maxtag2);
   if (err == 2) { 
      zmessageACK(ZTX("file tags exceed %d characters"),maxtag2);
      return;
   }
   
   tags_changed++;
   return;
}


//  add new tag to recent tags, if not already.
//  remove oldest to make space if needed.

void add_new_recentag()
{
   int         err;
   char        *ppv, temp_recentags[maxtag5];

   if (strnEqu(tags_atag,"stars=",6)) return;                              //  omit this tag

   err = add_unique_tag(tags_atag,tags_recentags,maxtag5);                 //  add tag to recent tags

   while (err == 2)
   {
      strncpy0(temp_recentags,tags_recentags,maxtag5);                     //  remove oldest to make room
      ppv = temp_recentags;
      while (*ppv && *ppv == ' ') ppv++;
      while (*ppv && *ppv != ' ') ppv++;
      while (*ppv && *ppv == ' ') ppv++;
      strcpy(tags_recentags,ppv);
      err = add_unique_tag(tags_atag,tags_recentags,maxtag5);
   }

   zdialog_stuff(zdtags,"recentags",tags_recentags);                       //  update dialog
   return;
}


//  delete a tag from file tags, if present

void delete_filetag()
{
   int         ii, ftcc, atcc;
   char        temp_filetags[maxtag2];
   const char  *pp;
   
   strncpy0(temp_filetags,tags_filetags,maxtag2);
   *tags_filetags = 0;
   ftcc = 0;
   
   for (ii = 1; ; ii++)
   {
      pp = strField(temp_filetags,' ',ii);
      if (! pp) break;
      if (strcaseEqu(pp,tags_atag)) continue;
      atcc = strlen(pp);
      strcpy(tags_filetags + ftcc, pp);
      ftcc += atcc;
      tags_filetags[ftcc] = ' ';
      ftcc++;
      tags_filetags[ftcc] = 0;
   }

   tags_changed++;
   return;
}


/**************************************************************************/

//  load assigned tags file >> tags_asstags in memory
//  create list of all assigned tags with no duplicates

void load_asstags()
{
   FILE        *fid;
   int         ntags = 0, ntcc, atcc, ii, err;
   char        *ppv, buff[1000];
   const char  *pp1;
   char        *tags[maxntags];
   
   ntcc = 0;
   *tags_asstags = 0;

   fid = fopen(asstagsfile,"r");
   if (! fid) return;                                                      //  no tags
   
   while (true)                                                            //  read assigned tags file
   {
      ppv = fgets_trim(buff,999,fid);
      if (! ppv) break;
      if (strnNeq(buff,"tags: ",6)) continue;

      for (ii = 1; ; ii++)                                                 //  add file tags to assigned tags
      {                                                                    //    unless already present
         pp1 = strField(buff+6,' ',ii);
         if (! pp1) break;
         if (strnEqu(pp1,"stars=",6)) continue;                            //  omit this tag
         err = add_unique_tag(pp1,tags_asstags,maxtag3);
         if (err == 2) goto overflow;
      }
   }

   err = fclose(fid);
   if (err) goto tagsfileerr;
   
   for (ii = 1; ; ii++)                                                    //  build sort list
   {
      pp1 = strField(tags_asstags,' ',ii);
      if (! pp1) break;
      tags[ntags] = strdupz(pp1);
      ntags++;
      if (ntags == maxntags) goto toomanytags;
   }
   
   HeapSort(tags,ntags);                                                   //  sort alphabetically
   
   ntcc = 0;
   *tags_asstags = 0;

   for (ii = 0; ii < ntags; ii++)                                          //  build sorted assigned tags list
   {
      atcc = strlen(tags[ii]);
      if (ntcc + atcc + 1 > maxtag3) goto overflow;
      strcpy(tags_asstags + ntcc,tags[ii]);
      ntcc += atcc;
      tags_asstags[ntcc] = ' ';
      ntcc++;
      zfree(tags[ii]);
   }

   tags_asstags[ntcc] = 0;
   
   return;

overflow:
   zmessageACK(ZTX("total tags exceed %d characters"),maxtag3);
   return;

toomanytags:
   zmessageACK(ZTX("too many tags: %d"),maxntags);
   return;

tagsfileerr:
   zmessLogACK(ZTX("assigned tags file error: %s"),strerror(errno));
   return;
}


/**************************************************************************/

//  update tags_asstags in memory from tags_filetags
//  update assigned tags file (add or replace changed file and its tags)

void update_asstags(char *file, int del)
{
   char        *ppv, temp_asstagsfile[1000], pdate2[12];
   char        datebuff[1000], tagsbuff[1000], filebuff[1000];
   const char  *pp1;
   int         ii, ntcc, err;
   FILE        *fidr, *fidw;

   ntcc = strlen(tags_asstags);
   
   if (! del)                                                              //  unless deleted
   {
      for (ii = 1; ; ii++)                                                 //  add file tags to assigned tags
      {                                                                    //    unless already present
         pp1 = strField(tags_filetags,' ',ii);
         if (! pp1) break;
         if (strnEqu(pp1,"stars=",6)) continue;                            //  omit this tag
         
         err = add_unique_tag(pp1,tags_asstags,maxtag3);
         if (err == 2) {
            zmessageACK(ZTX("total tags exceed %d characters"),maxtag3);
            break;
         }
      }
   }

   strcpy(temp_asstagsfile,asstagsfile);                                   //  temp tag file
   strcat(temp_asstagsfile,"_temp");

   fidr = fopen(asstagsfile,"r");                                          //  read tag file
   
   fidw = fopen(temp_asstagsfile,"w");                                     //  write temp tag file
   if (! fidw) goto tagserror;

   if (fidr) {   
      while (true)                                                         //  copy assigned tags file to temp
      {                                                                    //    file, omitting this image file
         ppv = fgets_trim(datebuff,999,fidr);
         if (! ppv) break;
         if (strnNeq(datebuff,"date: ",6)) continue;                       //  date added

         ppv = fgets_trim(tagsbuff,999,fidr);
         if (! ppv) break;
         if (strnNeq(tagsbuff,"tags: ",6)) continue;

         ppv = fgets_trim(filebuff,999,fidr);
         if (! ppv) break;
         if (strnNeq(filebuff,"file: ",6)) continue;

         if (strEqu(filebuff+6,file)) continue;                            //  if my file, skip copy

         fprintf(fidw,"%s\n",datebuff);                                    //  copy to temp file
         fprintf(fidw,"%s\n",tagsbuff);
         fprintf(fidw,"%s\n\n",filebuff);
      }
   }
   
   if (! del)                                                              //  unless deleted, append 
   {                                                                       //    revised file data to temp file
      strncpy(pdate2,tags_pdate,4);
      strncpy(pdate2+5,tags_pdate+4,2);                                    //  yyyymmdd >> yyyy:mm:dd
      strncpy(pdate2+8,tags_pdate+6,2);
      pdate2[4] = pdate2[7] = ':';
      pdate2[10] = 0;
      err = fprintf(fidw,"date: %s\n",pdate2);                             //  image date tag
      err = fprintf(fidw,"tags: %s\n",tags_filetags);                      //  file tags
      err = fprintf(fidw,"file: %s\n\n",file);                             //  filespec
      if (err <= 0) goto tagserror;
   }

   if (fidr) {
      err = fclose(fidr);
      if (err) goto tagserror;
   }

   err = fclose(fidw);
   if (err) goto tagserror;
   
   err = rename(temp_asstagsfile,asstagsfile);                             //  replace tag file with temp file
   if (err) goto tagserror;
   
   return;
   
tagserror:
   zmessLogACK(ZTX("assigned tags file error: %s"),strerror(errno));
   return;
}


/**************************************************************************/

//  search image tags for matching images

char     searchDateFrom[12] = "";                                          //  image search date range
char     searchDateTo[12] = "";
int      searchStarsFrom = 0;                                              //  image search stars range
int      searchStarsTo = 0;

void m_search_tags(GtkWidget *, const char *)
{
   int search_tags_dialog_event(zdialog*, const char *event);
   int search_tags_dialog_compl(zdialog*, int zstat);

   if (zdtags) {
      zdialog_free(zdtags);
      zdtags = null;
   }
   
   zdtags = zdialog_new(ZTX("Search Tags"),mWin,Bsearch,Bcancel,0);
   
   zdialog_add_widget(zdtags,"hbox","hb1","dialog",0,"space=2");
   zdialog_add_widget(zdtags,"vbox","vb1","hb1",0,"homog|space=5");
   zdialog_add_widget(zdtags,"vbox","vb2","hb1",0,"homog|space=2");

   zdialog_add_widget(zdtags,"label","labDR","vb1",ZTX("date range"));
   zdialog_add_widget(zdtags,"label","labSR","vb1",ZTX("stars range"));
   zdialog_add_widget(zdtags,"label","labF","vb1",ZTX("/path*/file*"));
   zdialog_add_widget(zdtags,"label","labT","vb1",ZTX("search tags"));

   zdialog_add_widget(zdtags,"hbox","hbDR","vb2",0,"space=5");             //  date range    yyyymmdd  yyyymmdd
   zdialog_add_widget(zdtags,"entry","datefrom","hbDR",0,"scc=10");        //  stars range   4  5
   zdialog_add_widget(zdtags,"entry","dateto","hbDR",0,"scc=10");          //  /path*/file*  /dname/dname*/fname*
   zdialog_add_widget(zdtags,"label","labDF","hbDR","(yyyymmdd)");         //  search tags   rosi  alaska

   zdialog_add_widget(zdtags,"hbox","hbSR","vb2",0);                       //  (o) match all tags  (o) match any
   zdialog_add_widget(zdtags,"entry","starsfrom","hbSR",0,"scc=2");
   zdialog_add_widget(zdtags,"entry","starsto","hbSR",0,"scc=2");          //  assigned tags

   zdialog_add_widget(zdtags,"hbox","hbF","vb2",0,"space=5");              //  list of all tags in a box
   zdialog_add_widget(zdtags,"entry","searchfile","hbF",0,"expand");
   zdialog_add_widget(zdtags,"button","fileclear","hbF",Bclear);

   zdialog_add_widget(zdtags,"hbox","hbT","vb2",0,"space=5");
   zdialog_add_widget(zdtags,"entry","searchtags","hbT",0,"expand");
   zdialog_add_widget(zdtags,"button","tagsclear","hbT",Bclear);

   zdialog_add_widget(zdtags,"hbox","hbM","dialog");
   zdialog_add_widget(zdtags,"radio","rmall","hbM",ZTX("match all tags"),"space=5");
   zdialog_add_widget(zdtags,"label","lspace","hbM","","space=10");
   zdialog_add_widget(zdtags,"radio","rmany","hbM",ZTX("match any tag"));

   zdialog_add_widget(zdtags,"hbox","hbsp","dialog",0,"space=1");

   zdialog_add_widget(zdtags,"hbox","hbAT","dialog");
   zdialog_add_widget(zdtags,"label","labasstags","hbAT",ZTX("assigned tags"));
   zdialog_add_widget(zdtags,"frame","frameAT","dialog",0,"expand");
   zdialog_add_widget(zdtags,"edit","asstags2","frameAT",0,"expand");

   zdialog_resize(zdtags,400,0);                                           //  start dialog
   zdialog_run(zdtags,search_tags_dialog_event,search_tags_dialog_compl);

   edit_tags_fixup("asstags2");                                            //  setup tag selection via mouse
   
   zdialog_stuff(zdtags,"datefrom",searchDateFrom);                        //  stuff previous date range
   zdialog_stuff(zdtags,"dateto",searchDateTo);
   zdialog_stuff(zdtags,"searchtags",tags_searchtags);                     //  stuff previous search tags
   zdialog_stuff(zdtags,"rmall",1);                                        //  default is match all tags
   zdialog_stuff(zdtags,"rmany",0);
   load_asstags();                                                         //  stuff assigned tags
   zdialog_stuff(zdtags,"asstags2",tags_asstags);
   
   return;
}

  
int search_tags_dialog_event(zdialog *zd, const char *event)               //  dialog event function
{
   if (strEqu(event,"fileclear")) zdialog_stuff(zd,"searchfile","");       //  v.6.6
   if (strEqu(event,"tagsclear")) zdialog_stuff(zd,"searchtags","");
   return 0;
}


int search_tags_dialog_compl(zdialog *zd, int zstat)                       //  dialog completion function
{
   const char  *pps, *ppf;
   char        resultsfile[200];
   char        *ppv, *file, *tags, rbuff[1000];
   char        date1[12], date2[12], lcfile[1000];
   int         err, nfiles, iis, iif, stars, nmatch, nfail;
   int         date1cc, date2cc, Fmall, Fdates, Ffiles, Ftags, Fstars;
   FILE        *fidr, *fidw;
   struct stat statbuf;

   if (zstat == 1) {
      zdialog_fetch(zd,"datefrom",searchDateFrom,10);                      //  get search date range
      zdialog_fetch(zd,"dateto",searchDateTo,10);
      zdialog_fetch(zd,"starsfrom",searchStarsFrom);                       //  get search stars range
      zdialog_fetch(zd,"starsto",searchStarsTo);
      zdialog_fetch(zd,"searchfile",tags_searchfile,maxtagF);              //  get search /path*/file*   v.6.6
      zdialog_fetch(zd,"searchtags",tags_searchtags,maxtag4);              //  get search tags
      zdialog_fetch(zd,"rmall",Fmall);                                     //  get match all/any option
   }
   
   zdialog_free(zdtags);                                                   //  kill dialog
   zdtags = null;
   if (zstat != 1) return 0;                                               //  cancelled

   strcpy(date1,"0000");                                                   //  defaults for missing dates  v.6.6
   strcpy(date2,"9999");                                                   //  (year 0000 to year 9999)
   date1cc = date2cc = 4;
   Fdates = 0;

   if (*searchDateFrom) {                                                  //  date from is given
      Fdates++;
      strncpy(date1,searchDateFrom,4);                                     //  convert format
      strncpy(date1+5,searchDateFrom+4,2);                                 //    yyyymmdd >> yyyy:mm:dd
      strncpy(date1+8,searchDateFrom+6,2);
      date1[4] = date1[7] = ':';
      date1cc = strlen(date1);
   }
   if (*searchDateTo) {                                                    //  date to is given
      Fdates++;
      strncpy(date2,searchDateTo,4);
      strncpy(date2+5,searchDateTo+4,2);
      strncpy(date2+8,searchDateTo+6,2);
      date2[4] = date2[7] = ':';
      date2cc = strlen(date2);
   }

   Fstars = 0;
   if (searchStarsFrom || searchStarsTo) Fstars = 1;                       //  stars given
   
   Ffiles = 0;
   if (! blank_null(tags_searchfile)) Ffiles = 1;                          //  search file* given     v.6.6

   Ftags = 0;
   if (! blank_null(tags_searchtags)) Ftags = 1;                           //  search tags given

   if (! Ffiles && ! Ftags && ! Fdates && ! Fstars) {                      //  no search criteria was given,
      strcpy(tags_searchtags,"(null)");                                    //    find images with no tags
      Ftags = 1;
   }
   
   if (Ffiles) strToLower(tags_searchfile);                                //  v.6.6
   if (Ftags) strToLower(tags_searchtags);
   
   snprintf(resultsfile,199,"%s/search_results",get_zuserdir());
   fidw = fopen(resultsfile,"w");                                          //  search results output file
   if (! fidw) goto writerror;

   fidr = fopen(asstagsfile,"r");                                          //  read assigned tags file
   if (! fidr) goto noasstags;
   
   nfiles = 0;                                                             //  count matching files found

   while (true)
   {
      ppv = fgets_trim(rbuff,999,fidr);                                    //  next assigned tags record
      if (! ppv) break;
      if (! strnEqu(ppv,"date: ",6)) continue;                             //  date: yyyy:mm:dd 

      if (Fdates) {      
         if (strncmp(ppv+6,date1,date1cc) < 0) continue;                   //  check search date range
         if (strncmp(ppv+6,date2,date2cc) > 0) continue;
      }

      ppv = fgets_trim(rbuff,999,fidr);                                    //  next record
      if (! ppv) break;
      if (! strnEqu(ppv,"tags: ",6)) continue;                             //  tags: xxxx xxxxx ...

      if (Ftags)
      {                                                                    //  tag search
         tags = ppv + 6;
         strToLower(tags);                                                 //  v.6.6

         nmatch = nfail = 0;

         for (iis = 1; ; iis++)
         {
            pps = strField(tags_searchtags,' ',iis);                       //  step thru search tags
            if (! pps) break;

            for (iif = 1; ; iif++)                                         //  step thru file tags
            {
               ppf = strField(tags,' ',iif);
               if (! ppf) { nfail++; break; }                              //  count matches and fails
               if (MatchWild(pps,ppf) == 0) { nmatch++; break; }           //  wildcard match   v.6.6
            }
         }

         if (nmatch == 0) continue;                                        //  no match to any tag
         if (Fmall && nfail) continue;                                     //  no match to all tags
      }
      
      if (Fstars)
      {                                                                    //  stars search
         nfail = 0;

         for (iif = 1; ; iif++)                                            //  step thru file tags
         {
            ppf = strField(ppv+6,' ',iif);
            if (! ppf) { nfail++; break; }
            if (! strnEqu(ppf,"stars=",6)) continue;
            stars = atoi(ppf+6);
            if (stars < searchStarsFrom) nfail++;
            if (searchStarsTo && stars > searchStarsTo) nfail++;
            break;
         }

         if (nfail) continue;
      }

      ppv = fgets_trim(rbuff,999,fidr,1);                                  //  next record
      if (! ppv) break;
      if (! strnEqu(ppv,"file: ",6)) continue;                             //  file: /dirks.../file.jpg
      file = ppv+6;
      strToLower(lcfile,file);                                             //  v.6.6
      
      if (Ffiles)                                                          //  test for path*/file* match
      {                                                                    //  v.6.6
         nmatch = 0;

         for (iis = 1; ; iis++)
         {
            pps = strField(tags_searchfile,' ',iis);                       //  step thru search files
            if (! pps) break;
            if (MatchWild(pps,lcfile) == 0) { nmatch++; break; }
         }
         
         if (nmatch == 0) continue;
      }

      err=stat(file,&statbuf);                                             //  check file exists
      if (err) continue;
      if (! S_ISREG(statbuf.st_mode)) continue;

      fprintf(fidw,"%s\n",file);                                           //  write matching file
      nfiles++;
   }

   fclose(fidr);
   
   err = fclose(fidw);
   if (err) goto writerror;
   
   if (! nfiles) {
      zmessageACK(ZTX("no matching images found"));
      return 0;
   }
   
   image_xthumbs(resultsfile,"initF",0,m_index2);                          //  generate index of matching files
   file = image_xthumbs(0,"first",0);
   f_open(file);                                                           //  show first matching file
   image_xthumbs(0,"paint1");                                              //  show new thumbnail index window
   Fsearchlist = 1;                                                        //  restricted updates     v.6.4

   return 0;

noasstags:
   zmessageACK(ZTX("no assigned tags index file"));
   return 0;

writerror:
   zmessLogACK(ZTX("search results file error %s"),strerror(errno));
   return 0;
}


/**************************************************************************/

//  list available EXIF data to popup window

void m_exif_list(GtkWidget *, const char *)
{
   char     command[1000];

   if (! image_file) return;

   if (! Fexiftool) {                                                      //  exiftool is required
      zmessageACK(Bexiftoolmissing);
      return;
   }

   snprintf(command,999,"exiftool -e \"%s\" ",image_file);
   popup_command(command,500,300);
   return;
}


/**************************************************************************/

//  get EXIF metadata for given image file and EXIF key(s)
//  returns array of pointers to corresponding key values
//  if a key is missing, corresponding pointer is null
//  returned strings belong to caller, are subject for zfree()
//  no more than 9 keynames may be requested per call
//  EXIF command: 
//       exiftool -exif:keyname1 -exif:keyname2 ... "file"
//  command output: 
//       keyname1: keyvalue1
//       keyname2: keyvalue2
//       ...

char ** exif_get(cchar *file, cchar **keys, int nkeys)
{
   char           command[1000], *pp;
   static char    *rettext[10];
   int            contx = 0, err, cc, ii;
   
   if (nkeys < 1 || nkeys > 9) appcrash("exif_get nkeys: %d",nkeys);

   strcpy(command,"exiftool -m -q -S -fast");
   
   for (ii = 0; ii < nkeys; ii++)
   {
      rettext[ii] = null;
      strncatv(command,999," -exif:",keys[ii],null);
   }

   strncatv(command,999," \"",file,"\"",null);

   while (true)
   {
      pp = command_output(contx,command);
      if (! pp) break;

      for (ii = 0; ii < nkeys; ii++)
      {
         cc = strlen(keys[ii]);
         if (strnEqu(pp,keys[ii],cc))
            rettext[ii] = strdupz(pp+cc+2);
      }

      zfree(pp);
   }
   
   err = command_status(contx);
   if (err) printf("exif_get failed \n");                                  //  v.6.9

   return rettext;
}


/**************************************************************************/

//  create or change EXIF metadata for given image file and key(s)
//  up to 9 keys may be processed
//  EXIF command: 
//    exiftool -overwrite_original -exif:keyname="keyvalue" ... "file"

int exif_set(cchar *file, cchar **keys, cchar **text, int nkeys)
{
   char     command[1000];
   int      ii, err;
   
   if (nkeys < 1 || nkeys > 9) appcrash("exif_set nkeys: %d",nkeys);
   
   strcpy(command,"exiftool -m -q -overwrite_original");
   
   for (ii = 0; ii < nkeys; ii++)
      strncatv(command,999," -exif:",keys[ii],"=\"",text[ii],"\"",null);
   strncatv(command,999," \"",file,"\"",null);

   err = system(command);
   if (err) printf(" exif_set: %s \n",wstrerror(err));
   return err;
}


/**************************************************************************/

//  copy EXIF data from original image file to new (edited) image file
//  if nkeys > 0, up to 9 keys may be replaced with new values
//  EXIF command:
//    exiftool -tagsfromfile "file1" -exif:keyname="keyvalue" ... "file2"

int exif_copy(cchar *file1, cchar *file2, cchar **keys, cchar **text, int nkeys)
{
   char     command[2000];
   int      ii, err;
   
   strcpy(command,"exiftool -m -q -overwrite_original -tagsfromfile");
   strncatv(command,1999," \"",file1,"\"",null);

   for (ii = 0; ii < nkeys; ii++)
      strncatv(command,1999," -exif:",keys[ii],"=\"",text[ii],"\"",null);

   strncatv(command,1999," \"",file2,"\"",null);

   err = system(command);
   if (err) printf(" exif_copy: %s \n",wstrerror(err));
   return err;
}


/**************************************************************************
      other support functions
***************************************************************************/

/**************************************************************************/

//  help menu function

void m_help(GtkWidget *, const char *menu)                                 //  add ZTX()   v.6.1
{
   if (strEqu(menu,ZTX("About"))) 
      zmessageACK(" %s \n %s \n %s \n %s \n %s",
                   fversion,flicense,fsource,ftranslators,fcredits);
      
   if (strEqu(menu,"FreeImage"))
      zmessageACK(FreeImage_GetCopyrightMessage());

   if (strEqu(menu,ZTX("User Guide"))) 
      showz_helpfile();

   if (strEqu(menu,"README"))
      showz_readme();

   if (strEqu(menu,ZTX("Change Log")))
      showz_changelog();

   if (strEqu(menu,ZTX("Translate")))
      showz_translations();

   return;
}


/**************************************************************************/

//  restore state data from prior session

int load_fotoxx_state()
{
   int            ww, hh, ii, np;
   FILE           *fid;
   char           buff[1000], text[100], *pp;
   float          parms[2];
   
   lens4_name[0] = strdupz("lens_1",lens_cc);
   lens4_name[1] = strdupz("lens_2",lens_cc);
   lens4_name[2] = strdupz("lens_3",lens_cc);
   lens4_name[3] = strdupz("lens_4",lens_cc);
   lens4_mm[0] = 30;
   lens4_mm[1] = 40;
   lens4_mm[2] = 50;
   lens4_mm[3] = 60;
   lens4_bow[0] = 0;
   lens4_bow[1] = 0;
   lens4_bow[2] = 0;
   lens4_bow[3] = 0;
   curr_lens = 1;

   snprintf(buff,999,"%s/saved_state",get_zuserdir());                     //  open saved state file
   fid = fopen(buff,"r");
   if (! fid) return 0;
   
   pp = fgets_trim(buff,999,fid,1);                                        //  read last image file

   if (pp && *pp == '/') {
      if (image_file) zfree(image_file);
      image_file = strdupz(pp);
   }

   pp = fgets_trim(buff,999,fid,1);                                        //  top directory
   if (pp && *pp == '/') topdirk = strdupz(pp);

   pp = fgets(buff,999,fid);                                               //  main window size
   if (pp) {
      ww = hh = 0;
      sscanf(buff," %d %d ",&ww,&hh);
      if (ww > 200 && ww < 3000) Mww = ww;
      if (hh > 200 && hh < 2000) Mhh = hh;
   }

   pp = fgets(buff,999,fid);                                               //  thumbnail index window size
   if (pp) {
      ww = hh = 0;
      sscanf(buff," %d %d ",&ww,&hh);
      if (ww > 200 && ww < 3000) image_navi::xwinW = ww;
      if (hh > 200 && hh < 2000) image_navi::xwinH = hh;
   }

   pp = fgets_trim(buff,999,fid,1);                                        //  thumbnail image size
   if (pp) {
      sscanf(buff," %d ",&ww);
      if (ww > 32 && ww < 256) image_navi::thumbsize = ww;
   }

   for (ii = 0; ii < 4; ii++)                                              //  4 sets of lens parameters
   {   
      pp = fgets_trim(buff,999,fid,1);
      if (! pp) break;
      np = sscanf(buff," %s %f %f ",text,&parms[0],&parms[1]);
      if (np != 3) break;
      strncpy0(lens4_name[ii],text,lens_cc);
      lens4_mm[ii] = parms[0];
      lens4_bow[ii] = parms[1];
   }
   
   pp = fgets_trim(buff,999,fid,1);                                        //  current lens
   if (pp) {
      sscanf(buff," %f ",&parms[0]);
      if (parms[0] >= 0 && parms[0] < 4) curr_lens = int(parms[0]);
   }

   pp = fgets_trim(buff,999,fid,1);                                        //  fotoxx version changed
   if (! pp || strNeq(pp,fversion)) { /* do nothing */ }

   fclose(fid);
   return 1;
}


/**************************************************************************/

//  save state data for next session

int save_fotoxx_state()
{
   FILE           *fid;
   char           buff[1000];
   int            ww, hh, ii;

   snprintf(buff,999,"%s/saved_state",get_zuserdir());                     //  open output file
   fid = fopen(buff,"w");
   if (! fid) return 0;
   
   if (image_file && *image_file == '/')                                   //  current image file 
      fputs(image_file,fid);
   fputs("\n",fid);

   if (topdirk && *topdirk == '/')                                         //  top image directory 
      fputs(topdirk,fid);
   fputs("\n",fid);

   gtk_window_get_size(GTK_WINDOW(mWin),&ww,&hh);
   snprintf(buff,20," %d %d \n",ww,hh);                                    //  window size
   fputs(buff,fid);

   snprintf(buff,20," %d %d \n",image_navi::xwinW,image_navi::xwinH);      //  thumbnail index window size
   fputs(buff,fid);

   snprintf(buff,20," %d \n",image_navi::thumbsize);                       //  thumbnail size
   fputs(buff,fid);
   
   for (ii = 0; ii < 4; ii++)                                              //  4 sets of lens parameters
   {
      snprintf(buff,100," %s %.1f %.2f \n",lens4_name[ii],lens4_mm[ii],lens4_bow[ii]);
      fputs(buff,fid);
   }
   
   snprintf(buff,100," %d \n",curr_lens);                                  //  current lens
   fputs(buff,fid);

   fputs(fversion,fid);                                                    //  fotoxx version
   fputs("\n",fid);
   fputs("\n",fid);
   
   fclose(fid);
   return 1;
}


/**************************************************************************/

//  free all resources associated with the current image file

void free_resources()
{
   char        *pp, command[200];
   int         ignore;

   mutex_lock(&bitmaps_lock);                                              //  lock bitmaps
   
   strcpy(command,"rm -f ");                                               //  delete all undo files
   strcat(command,undo_files);
   pp = strstr(command,"_undo_");                                          //  clone edit, bugfix  v.6.5
   strcpy(pp + 6,"*");
   ignore = system(command);
   Fmodified = Pundo = Pumax = 0;                                          //  reset undo/redo stack
   
   if (Fshutdown) {                                                        //  stop here if shutdown mode
      mutex_unlock(&bitmaps_lock);
      return;
   }
   
   if (image_file) {
      m_unselect(null,null);                                               //  clear select area
      update_filetags(image_file);                                         //  commit tag changes, if any
      zfree(image_file);                                                   //  free image file
      image_file = 0;
   }
   
   RGB_free(Frgb24);
   Frgb24 = 0;
   RGB_free(Frgb48);
   Frgb48 = 0;
   RGB_free(E1rgb48);
   E1rgb48 = 0;
   RGB_free(E3rgb48);
   E3rgb48 = 0;
   RGB_free(Mrgb24);
   Mrgb24 = 0;

   Fww = E1ww = E3ww = Mww = 0;                                            //  make unusable (crash)

   mutex_unlock(&bitmaps_lock);
   return;
}


/**************************************************************************/

//  ask user if modified image should be kept or discarded

int mod_keep()
{
   if (Fmodified == 0 && Pundo == 0) return 0;                             //  no mods
   if (zmessageYN(ZTX("Discard modifications?"))) return 0;                //  OK to discard
   return 1;
}


/**************************************************************************/

//  menu lock/unlock - some functions must not run concurrently

int menulock(int lock)
{
   static int     mlock = 0;

   if (lock && mlock) {
      zmessageACK("please wait for prior function to complete");
      return 0;
   }
   
   if (! lock && ! mlock) zappcrash("menu lock error");
   
   if (lock) mlock++;
   else mlock--;
   return 1;
}


/**************************************************************************/

//  Upright a turned image - not like an edit.
//  Rotate Frgb24 without setting the Fmodified flag.

void  upright_image(int angle)
{
   RGB * temp_bmp = RGB_rotate(Frgb24,angle);
   RGB_free(Frgb24);
   Frgb24 = temp_bmp;
   Fww = Frgb24->ww;
   Fhh = Frgb24->hh;
   Fzoom = 0;
   mwpaint();                                                              //  synch Mww etc.  v.6.8
   return;
}


/**************************************************************************/

//  FREEIMAGE error handler

void FI_error(FIF fif, const char *message)
{
   printf("FreeImage error: %s  file: %s \n",message,image_file);
   return;
}


/**************************************************************************
      bitmap conversion and rescale functions      revamped v.6.5
***************************************************************************/

//  initialize an RGB bitmap - allocate memory

RGB * RGB_make(int ww, int hh, int bpp)
{
   if (ww < 1 || hh < 1 || (bpp != 24 && bpp != 48))
      zappcrash("RGB_make() %d %d %d",ww,hh,bpp);
   
   RGB *rgb = (RGB *) zmalloc(sizeof(RGB));
   rgb->ww = ww;
   rgb->hh = hh;
   rgb->bpp = bpp;
   if (bpp == 24) rgb->bmp = zmalloc(ww * hh * 3);
   if (bpp == 48) rgb->bmp = zmalloc(ww * hh * 6);
   strcpy(rgb->wmi,"rgbrgb");
   return rgb;
}


//  free RGB bitmap

void RGB_free(RGB *rgb)
{
   if (! rgb) return;                                                      //  v.6.8
   if (! strEqu(rgb->wmi,"rgbrgb")) 
      zappcrash("RGB_free(), bad RGB");
   strcpy(rgb->wmi,"xxxxxx");
   zfree(rgb->bmp);
   zfree(rgb);
   return;
}


//  create a copy of an RGB bitmap

RGB * RGB_copy(RGB *rgb1)
{
   int      cc;
   RGB      *rgb2;

   rgb2 = RGB_make(rgb1->ww, rgb1->hh, rgb1->bpp);
   cc = rgb1->ww * rgb1->hh * rgb1->bpp / 8;
   memcpy(rgb2->bmp,rgb1->bmp,cc);
   return rgb2;
}


//   convert RGB bitmap from 24/48 to 48/24 bits per pixel

RGB * RGB_convbpp(RGB *rgb1)
{
   uint8          *bmp8, *pix8;
   uint16         *bmp16, *pix16;
   RGB            *rgb2 = 0;
   int            ww, hh, bpp, px, py;

   ww = rgb1->ww;
   hh = rgb1->hh;
   bpp = rgb1->bpp;

   if (bpp == 24)
   {
      rgb2 = RGB_make(ww,hh,48);
      bmp8 = (uint8 *) rgb1->bmp;
      bmp16 = (uint16 *) rgb2->bmp;
     
      for (py = 0; py < hh; py++) 
      {
         pix8 = bmp8 + py * ww * 3;
         pix16 = bmp16 + py * ww * 3;

         for (px = 0; px < ww; px++)
         {
            pix16[0] = pix8[0] << 8;
            pix16[1] = pix8[1] << 8;
            pix16[2] = pix8[2] << 8;
            pix8 += 3;
            pix16 += 3;
         }
      }
   }

   if (bpp == 48)
   {
      rgb2 = RGB_make(ww,hh,24);
      bmp8 = (uint8 *) rgb2->bmp;
      bmp16 = (uint16 *) rgb1->bmp;
     
      for (py = 0; py < hh; py++) 
      {
         pix8 = bmp8 + py * ww * 3;
         pix16 = bmp16 + py * ww * 3;

         for (px = 0; px < ww; px++)
         {
            pix8[0] = pix16[0] >> 8;
            pix8[1] = pix16[1] >> 8;
            pix8[2] = pix16[2] >> 8;
            pix8 += 3;
            pix16 += 3;
         }
      }
   }

   return rgb2;
}


//  convert RGB to FI bitmap, inverting row order

FIB * RGB_FIB(RGB *rgb)
{
   FIB         *fib = 0;
   uint8       *bmpf;
   uint8       *bmp1, *pix1, *pix2;
   uint16      *bmp2, *pix3, *pix4;
   int         ww, hh, bpp, px, py, pitch;
   
   ww = rgb->ww;
   hh = rgb->hh;
   bpp = rgb->bpp;

   if (bpp == 24)
   {   
      fib = FreeImage_Allocate(ww,hh,24);
      if (! fib) zappcrash("FIB allocation failure");

      bmp1 = (uint8 *) rgb->bmp;
      bmpf = FreeImage_GetBits(fib);
      pitch = FreeImage_GetPitch(fib);

      for (py = 0; py < hh; py++)
      {
         pix1 = bmp1 + (hh - py - 1) * ww * 3;
         pix2 = bmpf + py * pitch;
         
         for (px = 0; px < ww; px++)
         {
            pix2[RR] = pix1[0];
            pix2[GG] = pix1[1];
            pix2[BB] = pix1[2];
            pix1 += 3;
            pix2 += 3;
         }
      }
   }

   if (bpp == 48)
   {   
      fib = FreeImage_AllocateT(FIT_RGB16,ww,hh,48);
      if (! fib) zappcrash("FIB allocation failure");

      bmp2 = (uint16 *) rgb->bmp;
      bmpf = FreeImage_GetBits(fib);
      pitch = FreeImage_GetPitch(fib);
      
      for (py = 0; py < hh; py++)
      {
         pix3 = bmp2 + (hh - py - 1) * ww * 3;
         pix4 = (uint16 *) (bmpf + py * pitch);
         
         for (px = 0; px < ww; px++)
         {
            pix4[0] = pix3[0];
            pix4[1] = pix3[1];
            pix4[2] = pix3[2];
            pix3 += 3;
            pix4 += 3;
         }
      }
   }

   return fib;
}


//  convert FI to RGB bitmap, inverting row order

RGB * FIB_RGB(FIB *fib)
{
   RGB         *rgb;
   uint8       *bmpf;
   uint8       *bmp1, *pix1, *pix2;
   uint16      *bmp2, *pix3, *pix4;
   int         ww, hh, bpp, px, py, pitch;

   ww = FreeImage_GetWidth(fib);
   hh = FreeImage_GetHeight(fib);
   bpp = FreeImage_GetBPP(fib);
   pitch = FreeImage_GetPitch(fib);
   bmpf = FreeImage_GetBits(fib);
   
   rgb = RGB_make(ww,hh,bpp);

   if (bpp == 24)
   {   
      bmp1 = (uint8 *) rgb->bmp;

      for (py = 0; py < hh; py++)
      {
         pix1 = bmp1 + (hh - py - 1) * ww * 3;
         pix2 = bmpf + py * pitch;
         
         for (px = 0; px < ww; px++)
         {
            pix1[0] = pix2[RR];
            pix1[1] = pix2[GG];
            pix1[2] = pix2[BB];
            pix1 += 3;
            pix2 += 3;
         }
      }
   }

   if (bpp == 48)
   {   
      bmp2 = (uint16 *) rgb->bmp;

      for (py = 0; py < hh; py++)
      {
         pix3 = bmp2 + (hh - py - 1) * ww * 3;
         pix4 = (uint16 *) (bmpf + py * pitch);
         
         for (px = 0; px < ww; px++)
         {
            pix3[0] = pix4[0];
            pix3[1] = pix4[1];
            pix3[2] = pix4[2];
            pix3 += 3;
            pix4 += 3;
         }
      }
   }

   return rgb;
}


//  convert RGB-24 bitmap to GDK pixbuf bitmap (24)

PXB * RGB_PXB(RGB *rgb)                                                    //  new v.6.7.2
{
   int      ww, hh, bpp, px, py, rowst;
   uint8    *bmp1, *bmp2, *pix1, *pix2;
   PXB      *pxb;

   ww = rgb->ww;
   hh = rgb->hh;
   bpp = rgb->bpp;
   if (bpp != 24) zappcrash("RGB_PXB bpp %d",bpp);
   
   pxb = gdk_pixbuf_new(colorspace,0,8,ww,hh);
   if (! pxb) zappcrash("pixbuf allocation failure");

   bmp1 = (uint8 *) rgb->bmp;
   bmp2 = gdk_pixbuf_get_pixels(pxb);
   rowst = gdk_pixbuf_get_rowstride(pxb);

   for (py = 0; py < hh; py++)
   for (px = 0; px < ww; px++)
   {
      pix1 = bmp1 + (py * ww + px) * 3;
      pix2 = bmp2 + rowst * py + 3 * px;
      pix2[0] = pix1[0];
      pix2[1] = pix1[1];
      pix2[2] = pix1[2];
   }
   
   return pxb;
}


//  convert GDK pixbuf bitmap (24/32) to RGB-24 bitmap

RGB * PXB_RGB(PXB *pxb)
{
   RGB         *rgb;
   int         ww, hh, px, py, nch, rowst;
   uint8       *bmp1, *bmp2, *pix1, *pix2;

   ww = gdk_pixbuf_get_width(pxb);
   hh = gdk_pixbuf_get_height(pxb);
   nch = gdk_pixbuf_get_n_channels(pxb);
   rowst = gdk_pixbuf_get_rowstride(pxb);
   bmp1 = gdk_pixbuf_get_pixels(pxb);
   
   rgb = RGB_make(ww,hh,24);
   bmp2 = (uint8 *) rgb->bmp;

   for (py = 0; py < hh; py++)
   for (px = 0; px < ww; px++)
   {
      pix1 = bmp1 + rowst * py + nch * px;
      pix2 = bmp2 + (py * ww + px) * 3;
      pix2[0] = pix1[0];
      pix2[1] = pix1[1];
      pix2[2] = pix1[2];
   }

   return rgb;
}


//  rescale RGB bitmap to size ww2 x hh2

RGB * RGB_rescale(RGB *rgb1, int ww2, int hh2)
{
   void RGB_rescale_24(uint8*, uint8*, int, int, int, int);
   void RGB_rescale_48(uint16*, uint16*, int, int, int, int);

   RGB      *rgb2;
   int      ww1, hh1, bpp;
   uint8    *bmp1, *bmp2;
   uint16   *bmp3, *bmp4;

   ww1 = rgb1->ww;
   hh1 = rgb1->hh;
   bpp = rgb1->bpp;
   
   rgb2 = RGB_make(ww2,hh2,bpp);
   
   if (bpp == 24) {
      bmp1 = (uint8 *) rgb1->bmp;
      bmp2 = (uint8 *) rgb2->bmp;
      RGB_rescale_24(bmp1,bmp2,ww1,hh1,ww2,hh2);
   }

   if (bpp == 48) {
      bmp3 = (uint16 *) rgb1->bmp;
      bmp4 = (uint16 *) rgb2->bmp;
      RGB_rescale_48(bmp3,bmp4,ww1,hh1,ww2,hh2);
   }
   
   return rgb2;
}


//  rotate RGB bitmap through given angle in degrees (+ = clockwise)

RGB * RGB_rotate(RGB *rgb1, double angle)
{
   RGB * RGB_rotate_24(RGB *, double);
   RGB * RGB_rotate_48(RGB *, double);

   RGB      *rgb2 = 0;
   int      bpp;
   
   bpp = rgb1->bpp;
   if (bpp == 24) rgb2 = RGB_rotate_24(rgb1,angle);
   if (bpp == 48) rgb2 = RGB_rotate_48(rgb1,angle);
   return rgb2;
}


//  Load an image file into an RGB bitmap of 24 or 48 bits/pixel.
//  Also sets the following global variables:
//    file_MB = disk file megabytes
//    file_bpp = disk file bits per pixel (24 or 48)
//    file_type = "jpeg" or "tiff" or "other"

RGB * image_load(const char *filespec, int bpp)                            //  revamped   v.6.6.1
{
   FIF         fif;
   FIB         *fib = 0, *fib2 = 0;
   int         err, fibpp;
   RGB         *rgb24 = 0, *rgb48 = 0;
   struct stat fstat;
   
   if (bpp != 24 && bpp != 48)                                             //  requested bpp must be 24 or 48
      zappcrash("image_load bpp: %d",bpp);

   err = stat(filespec,&fstat);
   if (err) return 0;                                                      //  file not found
   if (! S_ISREG(fstat.st_mode)) return 0;                                 //  not a regular file

   fif = FreeImage_GetFileType(filespec,0);                                //  unsupported file type
   if (fif == FIF_UNKNOWN) goto ILreturn;

   fib = FreeImage_Load(fif,filespec,0);
   if (! fib) goto ILreturn;                                               //  cannot load file

   fibpp = FreeImage_GetBPP(fib);                                          //  disk file bits/pixel
   
   if (fibpp < 24) {
      fib2 = FreeImage_ConvertTo24Bits(fib);                               //  1 or 8
      if (! fib2) goto ILreturn;
      rgb24 = FIB_RGB(fib2);
      if (bpp == 48) rgb48 = RGB_convbpp(rgb24);
   }

   if (fibpp == 24) {                                                      //  24
      rgb24 = FIB_RGB(fib);
      if (bpp == 48) rgb48 = RGB_convbpp(rgb24);
   }
   
   if (fibpp == 32) {                                                      //  24 + alpha
      fib2 = FreeImage_ConvertTo24Bits(fib);
      if (! fib2) goto ILreturn;
      rgb24 = FIB_RGB(fib2);
      if (bpp == 48) rgb48 = RGB_convbpp(rgb24);
   }

   if (fibpp == 48) {                                                      //  48
      rgb48 = FIB_RGB(fib);
      if (bpp == 24) rgb24 = RGB_convbpp(rgb48);
   }
   
ILreturn:
   if (fib) FreeImage_Unload(fib);                                         //  free FI bitmaps
   if (fib2) FreeImage_Unload(fib2);
   if (! rgb24 && ! rgb48) {
      zmessageACK(Bsavetoedit);                                            //  unsupported file type
      return 0;
   }
   
   file_bpp = fibpp;                                                       //  disk file bits/pixel
   file_MB = 1.0 * fstat.st_size / mega;                                   //  disk file megabytes
   if (fif == FIF_JPEG) strcpy(file_type,"jpeg");                          //  disk file type
   else if (fif == FIF_TIFF) strcpy(file_type,"tiff");
   else strcpy(file_type,"other");

   if (bpp == 24) {
      RGB_free(rgb48);                                                     //  return RGB-24 bitmap
      return rgb24;
   }

   RGB_free(rgb24);                                                        //  return RGB-48 bitmap
   return rgb48;
}


/**************************************************************************

   Rescale 24 bpp image (3 x 8 bits per color) to new width and height.
   The scale ratios may be different for width and height.

   Method: 
   The input and output images are overlayed, stretching or shrinking the
   output pixels as needed. The contribution of each input pixel overlapping
   an output pixel is proportional to the area of the output pixel covered by
   the input pixel. The contributions of all overlaping input pixels are added.
   The work is spread among 4 threads to reduce the elapsed time on modern 
   computers having multiple SMP processors.

   Example: if the output image is 40% of the input image, then:
     outpix[0,0] = 0.16 * inpix[0,0] + 0.16 * inpix[1,0] + 0.08 * inpix[2,0]
                 + 0.16 * inpix[0,1] + 0.16 * inpix[1,1] + 0.08 * inpix[2,1]
                 + 0.08 * inpix[0,2] + 0.08 * inpix[1,2] + 0.04 * inpix[2,2]

*********/

typedef struct {
   uint8    *bitmap1;                                                      //  file scope data for threads
   uint8    *bitmap2;
   int      ww1;
   int      hh1;
   int      ww2;
   int      hh2;
   int      *py1L;
   int      *px1L;
   float    *pymap;
   float    *pxmap;
   int      maxmapx;
   int      maxmapy;
   int      index;
   int      done;
} rs24poop;

void RGB_rescale_24(uint8 *bitmap1, uint8 *bitmap2, int ww1, int hh1, int ww2, int hh2)
{
   int         px1, py1, px2, py2;
   int         pxl, pyl, pxm, pym, ii;
   int         maxmapx, maxmapy;
   float       scalex, scaley;
   float       px1a, py1a, px1b, py1b;
   float       fx, fy;

   scalex = 1.0 * ww1 / ww2;                                               //  compute x and y scales
   scaley = 1.0 * hh1 / hh2;
   
   if (scalex <= 1) maxmapx = 2;                                           //  compute max input pixels
   else maxmapx = int(scalex + 2);                                         //    mapping into output pixels
   maxmapx += 1;                                                           //      for both dimensions
   if (scaley <= 1) maxmapy = 2;                                           //  (pixels may not be square)
   else maxmapy = int(scaley + 2);
   maxmapy += 1;
   
   memset(bitmap2, 0, ww2 * hh2 * 3);                                      //  clear output bitmap
   
   int *py1L = (int *) zmalloc(hh2 * sizeof(int));                         //  maps first (lowest) input pixel
   int *px1L = (int *) zmalloc(ww2 * sizeof(int));                         //    per output pixel

   float *pymap = (float *) zmalloc(hh2 * maxmapy * sizeof(float));        //  maps overlap of < maxmap input
   float *pxmap = (float *) zmalloc(ww2 * maxmapx * sizeof(float));        //    pixels per output pixel

   for (py2 = 0; py2 < hh2; py2++)                                         //  loop output y-pixels
   {
      py1a = py2 * scaley;                                                 //  corresponding input y-pixels
      py1b = py1a + scaley;
      if (py1b >= hh1) py1b = hh1 - 0.001;                                 //  fix precision limitation
      pyl = int(py1a);
      py1L[py2] = pyl;                                                     //  1st overlapping input pixel

      for (py1 = pyl, pym = 0; py1 < py1b; py1++, pym++)                   //  loop overlapping input pixels
      {
         if (py1 < py1a) {                                                 //  compute amount of overlap
            if (py1+1 < py1b) fy = py1+1 - py1a;                           //    0.0 to 1.0 
            else fy = scaley;
         }
         else if (py1+1 > py1b) fy = py1b - py1;
         else fy = 1;

         ii = py2 * maxmapy + pym;                                         //  save it
         pymap[ii] = 0.9999 * fy / scaley;
      }
      ii = py2 * maxmapy + pym;                                            //  set an end marker after
      pymap[ii] = -1;                                                      //    last overlapping pixel
   }
   
   for (px2 = 0; px2 < ww2; px2++)                                         //  do same for x-pixels
   {
      px1a = px2 * scalex;
      px1b = px1a + scalex;
      if (px1b >= ww1) px1b = ww1 - 0.001;
      pxl = int(px1a);
      px1L[px2] = pxl;

      for (px1 = pxl, pxm = 0; px1 < px1b; px1++, pxm++)
      {
         if (px1 < px1a) {
            if (px1+1 < px1b) fx = px1+1 - px1a;
            else fx = scalex;
         }
         else if (px1+1 > px1b) fx = px1b - px1;
         else fx = 1;

         ii = px2 * maxmapx + pxm;
         pxmap[ii] = 0.9999 * fx / scalex;
      }
      ii = px2 * maxmapx + pxm;
      pxmap[ii] = -1;
   }
   
   rs24poop threaddata[4];                                                 //  data for threads
   void  * RGB_rescale_24_thread(void *arg);

   for (ii = 0; ii < 4; ii++)                                              //  start 4 threads
   {
      threaddata[ii].bitmap1 = bitmap1;
      threaddata[ii].bitmap2 = bitmap2;
      threaddata[ii].ww1 = ww1;
      threaddata[ii].hh1 = hh1;
      threaddata[ii].ww2 = ww2;
      threaddata[ii].hh2 = hh2;
      threaddata[ii].py1L = py1L;
      threaddata[ii].px1L = px1L;
      threaddata[ii].pymap = pymap;
      threaddata[ii].pxmap = pxmap;
      threaddata[ii].maxmapx = maxmapx;
      threaddata[ii].maxmapy = maxmapy;
      threaddata[ii].index = ii;
      threaddata[ii].done = 0;
      start_detached_thread(RGB_rescale_24_thread,&threaddata[ii]);
   }
   
   for (ii = 0; ii < 4; ii++)                                              //  wait for all done
      while (threaddata[ii].done == 0) zsleep(0.01);

   zfree(py1L);
   zfree(px1L);
   zfree(pymap);
   zfree(pxmap);
   return;
}

void * RGB_rescale_24_thread(void *arg)                                    //  thread function
{
   rs24poop * threaddata = (rs24poop *) arg;

   uint8       *bitmap1 = threaddata->bitmap1;
   uint8       *bitmap2 = threaddata->bitmap2;
   int         ww1 = threaddata->ww1;
   int         ww2 = threaddata->ww2;
   int         hh2 = threaddata->hh2;
   int         *py1L = threaddata->py1L;
   int         *px1L = threaddata->px1L;
   float       *pymap = threaddata->pymap;
   float       *pxmap = threaddata->pxmap;
   int         maxmapx = threaddata->maxmapx;
   int         maxmapy = threaddata->maxmapy;
   int         index = threaddata->index;
   
   int         px1, py1, px2, py2;
   int         pxl, pyl, pxm, pym, ii;
   uint8       *pixel1, *pixel2;
   float       fx, fy, ftot;
   float       red, green, blue;

   for (py2 = index; py2 < hh2; py2 += 4)                                  //  loop output y-pixels
   {
      pyl = py1L[py2];                                                     //  corresp. 1st input y-pixel

      for (px2 = 0; px2 < ww2; px2++)                                      //  loop output x-pixels
      {
         pxl = px1L[px2];                                                  //  corresp. 1st input x-pixel

         red = green = blue = 0;                                           //  initz. output pixel

         for (py1 = pyl, pym = 0; ; py1++, pym++)                          //  loop overlapping input y-pixels
         {
            ii = py2 * maxmapy + pym;                                      //  get y-overlap
            fy = pymap[ii];
            if (fy < 0) break;                                             //  no more pixels

            for (px1 = pxl, pxm = 0; ; px1++, pxm++)                       //  loop overlapping input x-pixels
            {
               ii = px2 * maxmapx + pxm;                                   //  get x-overlap
               fx = pxmap[ii];
               if (fx < 0) break;                                          //  no more pixels

               ftot = fx * fy;                                             //  area overlap = x * y overlap
               pixel1 = bitmap1 + (py1 * ww1 + px1) * 3;
               red += pixel1[0] * ftot;                                    //  add input pixel * overlap
               green += pixel1[1] * ftot;
               blue += pixel1[2] * ftot;
            }

            pixel2 = bitmap2 + (py2 * ww2 + px2) * 3;                      //  save output pixel
            pixel2[0] = int(red);
            pixel2[1] = int(green);
            pixel2[2] = int(blue);
         }
      }
   }

   threaddata->done = 1;
   pthread_exit(0);
}


/**************************************************************************

   Rescale 48 bpp image (3 x 16 bits per color) to new width and height.
   Identical to RGB_rescale_24 except for the following: 
      uint8 >> uint16
      xxx24 >> xxx48
      3 >> 6
      memset 3 >> 6.

*******/

typedef struct {
   uint16   *bitmap1;                                                      //  file scope data for threads
   uint16   *bitmap2;
   int      ww1;
   int      hh1;
   int      ww2;
   int      hh2;
   int      *py1L;
   int      *px1L;
   float    *pymap;
   float    *pxmap;
   int      maxmapx;
   int      maxmapy;
   int      index;
   int      done;
} rs48poop;

void RGB_rescale_48(uint16 *bitmap1, uint16 *bitmap2, int ww1, int hh1, int ww2, int hh2)
{
   int         px1, py1, px2, py2;
   int         pxl, pyl, pxm, pym, ii;
   int         maxmapx, maxmapy;
   float       scalex, scaley;
   float       px1a, py1a, px1b, py1b;
   float       fx, fy;

   scalex = 1.0 * ww1 / ww2;                                               //  compute x and y scales
   scaley = 1.0 * hh1 / hh2;
   
   if (scalex <= 1) maxmapx = 2;                                           //  compute max input pixels
   else maxmapx = int(scalex + 2);                                         //    mapping into output pixels
   maxmapx += 1;                                                           //      for both dimensions
   if (scaley <= 1) maxmapy = 2;                                           //  (pixels may not be square)
   else maxmapy = int(scaley + 2);
   maxmapy += 1;
   
   memset(bitmap2, 0, ww2 * hh2 * 6);                                      //  clear output bitmap
   
   int *py1L = (int *) zmalloc(hh2 * sizeof(int));                         //  maps first (lowest) input pixel
   int *px1L = (int *) zmalloc(ww2 * sizeof(int));                         //    per output pixel

   float *pymap = (float *) zmalloc(hh2 * maxmapy * sizeof(float));        //  maps overlap of < maxmap input
   float *pxmap = (float *) zmalloc(ww2 * maxmapx * sizeof(float));        //    pixels per output pixel

   for (py2 = 0; py2 < hh2; py2++)                                         //  loop output y-pixels
   {
      py1a = py2 * scaley;                                                 //  corresponding input y-pixels
      py1b = py1a + scaley;
      if (py1b >= hh1) py1b = hh1 - 0.001;                                 //  fix precision limitation
      pyl = int(py1a);
      py1L[py2] = pyl;                                                     //  1st overlapping input pixel

      for (py1 = pyl, pym = 0; py1 < py1b; py1++, pym++)                   //  loop overlapping input pixels
      {
         if (py1 < py1a) {                                                 //  compute amount of overlap
            if (py1+1 < py1b) fy = py1+1 - py1a;                           //    0.0 to 1.0 
            else fy = scaley;
         }
         else if (py1+1 > py1b) fy = py1b - py1;
         else fy = 1;

         ii = py2 * maxmapy + pym;                                         //  save it
         pymap[ii] = 0.9999 * fy / scaley;
      }
      ii = py2 * maxmapy + pym;                                            //  set an end marker after
      pymap[ii] = -1;                                                      //    last overlapping pixel
   }
   
   for (px2 = 0; px2 < ww2; px2++)                                         //  do same for x-pixels
   {
      px1a = px2 * scalex;
      px1b = px1a + scalex;
      if (px1b >= ww1) px1b = ww1 - 0.001;
      pxl = int(px1a);
      px1L[px2] = pxl;

      for (px1 = pxl, pxm = 0; px1 < px1b; px1++, pxm++)
      {
         if (px1 < px1a) {
            if (px1+1 < px1b) fx = px1+1 - px1a;
            else fx = scalex;
         }
         else if (px1+1 > px1b) fx = px1b - px1;
         else fx = 1;

         ii = px2 * maxmapx + pxm;
         pxmap[ii] = 0.9999 * fx / scalex;
      }
      ii = px2 * maxmapx + pxm;
      pxmap[ii] = -1;
   }
   
   rs48poop threaddata[4];                                                 //  data for threads
   void  * RGB_rescale_48_thread(void *arg);

   for (ii = 0; ii < 4; ii++)                                              //  start 4 threads
   {
      threaddata[ii].bitmap1 = bitmap1;
      threaddata[ii].bitmap2 = bitmap2;
      threaddata[ii].ww1 = ww1;
      threaddata[ii].hh1 = hh1;
      threaddata[ii].ww2 = ww2;
      threaddata[ii].hh2 = hh2;
      threaddata[ii].py1L = py1L;
      threaddata[ii].px1L = px1L;
      threaddata[ii].pymap = pymap;
      threaddata[ii].pxmap = pxmap;
      threaddata[ii].maxmapx = maxmapx;
      threaddata[ii].maxmapy = maxmapy;
      threaddata[ii].index = ii;
      threaddata[ii].done = 0;
      start_detached_thread(RGB_rescale_48_thread,&threaddata[ii]);
   }
   
   for (ii = 0; ii < 4; ii++)                                              //  wait for all done
      while (threaddata[ii].done == 0) zsleep(0.01);

   zfree(py1L);
   zfree(px1L);
   zfree(pymap);
   zfree(pxmap);
   return;
}

void * RGB_rescale_48_thread(void *arg)                                    //  thread function
{
   rs48poop * threaddata = (rs48poop *) arg;

   uint16      *bitmap1 = threaddata->bitmap1;
   uint16      *bitmap2 = threaddata->bitmap2;
   int         ww1 = threaddata->ww1;
   int         ww2 = threaddata->ww2;
   int         hh2 = threaddata->hh2;
   int         *py1L = threaddata->py1L;
   int         *px1L = threaddata->px1L;
   float       *pymap = threaddata->pymap;
   float       *pxmap = threaddata->pxmap;
   int         maxmapx = threaddata->maxmapx;
   int         maxmapy = threaddata->maxmapy;
   int         index = threaddata->index;
   
   int         px1, py1, px2, py2;
   int         pxl, pyl, pxm, pym, ii;
   uint16      *pixel1, *pixel2;
   float       fx, fy, ftot;
   float       red, green, blue;

   for (py2 = index; py2 < hh2; py2 += 4)                                  //  loop output y-pixels
   {
      pyl = py1L[py2];                                                     //  corresp. 1st input y-pixel

      for (px2 = 0; px2 < ww2; px2++)                                      //  loop output x-pixels
      {
         pxl = px1L[px2];                                                  //  corresp. 1st input x-pixel

         red = green = blue = 0;                                           //  initz. output pixel

         for (py1 = pyl, pym = 0; ; py1++, pym++)                          //  loop overlapping input y-pixels
         {
            ii = py2 * maxmapy + pym;                                      //  get y-overlap
            fy = pymap[ii];
            if (fy < 0) break;                                             //  no more pixels

            for (px1 = pxl, pxm = 0; ; px1++, pxm++)                       //  loop overlapping input x-pixels
            {
               ii = px2 * maxmapx + pxm;                                   //  get x-overlap
               fx = pxmap[ii];
               if (fx < 0) break;                                          //  no more pixels

               ftot = fx * fy;                                             //  area overlap = x * y overlap
               pixel1 = bitmap1 + (py1 * ww1 + px1) * 3;
               red += pixel1[0] * ftot;                                    //  add input pixel * overlap
               green += pixel1[1] * ftot;
               blue += pixel1[2] * ftot;
            }

            pixel2 = bitmap2 + (py2 * ww2 + px2) * 3;                      //  save output pixel
            pixel2[0] = int(red);
            pixel2[1] = int(green);
            pixel2[2] = int(blue);
         }
      }
   }

   threaddata->done = 1;
   pthread_exit(0);
}


/**************************************************************************

      RGB *rgb2 = RGB_rotate_24(RGB *rgb1, double angle)

      Rotate RGB-24 bitmap through an arbitrary angle (degrees).

      The returned image has the same size as the original, but the
      bitmap size is increased to accomodate the rotated image.
      (e.g. a 100x100 image rotated 45 deg. needs a 142x142 bitmap).
      The parameters ww and hh are the dimensions of the input
      bitmap, and are updated to the dimensions of the output bitmap.

      The space added around the rotated image is black (RGB 0,0,0).
      Angle is in degrees. Positive direction is clockwise.
      Speed is about 3 million pixels/sec/thread for a 2.4 GHz CPU.
      Loss of resolution is less than 1 pixel.

      Algorithm:
      create output bitmap big enough for rotated input bitmap
      loop all output pixels:
         get next output pixel (px2,py2)
         compute (R,theta) from center of pixbuf
         rotate theta by -angle
         (R,theta) is now within the closest input pixel
         convert to input pixel (px1,py1)
         if outside of bitmap
            output pixel = black
            continue
         for 4 input pixels based at (px0,py0) = (int(px1),int(py1))
            compute overlap (0 to 1) with (px1,py1)
            sum RGB values * overlap
         output aggregate RGB to pixel (px2,py2)

******/

#define  rotrgb24_NT  4                                                    //  no. working threads

int      rotrgb24_done[rotrgb24_NT];                                       //  args for threads
uint8    *rotrgb24_bitmap1;
uint8    *rotrgb24_bitmap2;
int      rotrgb24_ww1;
int      rotrgb24_hh1;
int      rotrgb24_ww2;
int      rotrgb24_hh2;
double   rotrgb24_angle;


RGB * RGB_rotate_24(RGB *rgb1, double angle)
{
   void     *RGB_rotate_24_thread(void *);

   int      ww1, hh1, ww2, hh2, cc;
   double   pi = 3.141592654;
   int      ii, index[rotrgb24_NT];
   uint8    *bitmap1, *bitmap2;
   RGB      *rgb2;
   
   ww1 = rgb1->ww;
   hh1 = rgb1->hh;
   bitmap1 = (uint8 *) rgb1->bmp;

   while (angle < -180) angle += 360;                                      //  normalize, -180 to +180
   while (angle > 180) angle -= 360;
   angle = angle * pi / 180;                                               //  radians, -pi to +pi
   
   if (fabs(angle) < 0.001) {                                              //  angle = 0 within my precision
      rgb2 = RGB_make(ww1,hh1,24);                                         //  return a copy of the input
      bitmap2 = (uint8 *) rgb2->bmp;
      cc = ww1 * hh1 * 3;
      memcpy(bitmap2,bitmap1,cc);
      return rgb2;
   }

   ww2 = int(ww1*fabs(cos(angle)) + hh1*fabs(sin(angle)));                 //  rectangle containing rotated image
   hh2 = int(ww1*fabs(sin(angle)) + hh1*fabs(cos(angle)));
   
   rgb2 = RGB_make(ww2,hh2,24);
   bitmap2 = (uint8 *) rgb2->bmp;

   rotrgb24_bitmap1 = bitmap1;
   rotrgb24_bitmap2 = bitmap2;
   rotrgb24_ww1 = ww1;
   rotrgb24_hh1 = hh1;
   rotrgb24_ww2 = ww2;
   rotrgb24_hh2 = hh2;
   rotrgb24_angle = angle;

   for (ii = 0; ii < rotrgb24_NT; ii++)                                    //  start NT threads
   {
      index[ii] = ii;
      rotrgb24_done[ii] = 0;
      start_detached_thread(RGB_rotate_24_thread,&index[ii]);
   }

   for (ii = 0; ii < rotrgb24_NT; ii++)                                    //  wait for all done
      while (rotrgb24_done[ii] == 0) zsleep(0.01);

   return rgb2;
}


void * RGB_rotate_24_thread(void *arg)
{
   int      index = *((int *) (arg));
   int      ww1, hh1, ww2, hh2;
   int      px2, py2, px0, py0;
   uint8    *ppix1, *ppix2, *pix0, *pix1, *pix2, *pix3;
   double   rx1, ry1, rx2, ry2, R, theta, px1, py1;
   double   f0, f1, f2, f3, red, green, blue;
   double   angle, pi = 3.141592654;

   ppix1 = rotrgb24_bitmap1;                                               //  input pixel array
   ppix2 = rotrgb24_bitmap2;                                               //  output pixel array
   ww1 = rotrgb24_ww1;
   hh1 = rotrgb24_hh1;
   ww2 = rotrgb24_ww2;
   hh2 = rotrgb24_hh2;
   angle = rotrgb24_angle;
   
   for (py2 = index; py2 < hh2; py2 += rotrgb24_NT)                        //  loop through output pixels
   for (px2 = 0; px2 < ww2; px2++)                                         //  outer loop y
   {
      rx2 = px2 - 0.5 * ww2;                                               //  (rx2,ry2) = center of pixel
      ry2 = py2 - 0.5 * hh2;
      R = sqrt(rx2*rx2 + ry2*ry2);                                         //  convert to (R,theta)
      if (R < 0.1) theta = 0;
      else theta = qarcsine(ry2 / R);                                      //  quick arc sine
      if (rx2 < 0) {
         if (theta < 0) theta = - pi - theta;                              //  adjust for quandrant
         else theta = pi - theta;
      }

      theta = theta - angle;                                               //  rotate theta backwards
      if (theta > pi) theta -= 2 * pi;                                     //  range -pi to +pi
      if (theta < -pi) theta += 2 * pi;

      rx1 = R * qcosine(theta);                                            //  quick cosine, sine
      ry1 = R * qsine(theta);
      px1 = rx1 + 0.5 * ww1;                                               //  (px1,py1) = corresponding
      py1 = ry1 + 0.5 * hh1;                                               //    point within input pixels

      px0 = int(px1);                                                      //  pixel containing (px1,py1)
      py0 = int(py1);
      
      if (px1 < 0 || px0 >= ww1-1 || py1 < 0 || py0 >= hh1-1) {            //  if outside input pixel array
         pix2 = ppix2 + (py2 * ww2 + px2) * 3;                             //    output is black
         pix2[0] = pix2[1] = pix2[2] = 0;
         continue;
      }

      pix0 = ppix1 + (py0 * ww1 + px0) * 3;                                //  4 input pixels based at (px0,py0)
      pix1 = pix0 + ww1 * 3;
      pix2 = pix0 + 3;
      pix3 = pix1 + 3;

      f0 = (px0+1 - px1) * (py0+1 - py1);                                  //  overlap of (px1,py1)
      f1 = (px0+1 - px1) * (py1 - py0);                                    //    in each of the 4 pixels
      f2 = (px1 - px0) * (py0+1 - py1);
      f3 = (px1 - px0) * (py1 - py0);
   
      red =   f0 * pix0[0] + f1 * pix1[0] + f2 * pix2[0] + f3 * pix3[0];   //  sum the weighted inputs
      green = f0 * pix0[1] + f1 * pix1[1] + f2 * pix2[1] + f3 * pix3[1];
      blue =  f0 * pix0[2] + f1 * pix1[2] + f2 * pix2[2] + f3 * pix3[2];
      
      pix2 = ppix2 + (py2 * ww2 + px2) * 3;                                //  output pixel
      pix2[0] = int(red);
      pix2[1] = int(green);
      pix2[2] = int(blue);
   }
   
   rotrgb24_done[index] = 1;
   pthread_exit(0);
}


/**************************************************************************

   RGB *rgb2 = RGB_rotate_48(RGB *rgb1, double angle)
   Rotate RGB-48 bitmap through an arbitrary angle (degrees).
   Identical to RGB_rotate_24() except for:
      uint8 >> uint16
      rotrgb24 >> rotrgb48
      24 >> 48   
      3 >> 6 

**********/

#define  rotrgb48_NT  4                                                    //  no. working threads

int      rotrgb48_done[rotrgb48_NT];                                       //  args for threads
uint16   *rotrgb48_bitmap1;
uint16   *rotrgb48_bitmap2;
int      rotrgb48_ww1;
int      rotrgb48_hh1;
int      rotrgb48_ww2;
int      rotrgb48_hh2;
double   rotrgb48_angle;


RGB * RGB_rotate_48(RGB *rgb1, double angle)
{
   void     *RGB_rotate_48_thread(void *);

   int      ww1, hh1, ww2, hh2, cc;
   double   pi = 3.141592654;
   int      ii, index[rotrgb48_NT];
   uint16   *bitmap1, *bitmap2;
   RGB      *rgb2;
   
   ww1 = rgb1->ww;
   hh1 = rgb1->hh;
   bitmap1 = (uint16 *) rgb1->bmp;

   while (angle < -180) angle += 360;                                      //  normalize, -180 to +180
   while (angle > 180) angle -= 360;
   angle = angle * pi / 180;                                               //  radians, -pi to +pi
   
   if (fabs(angle) < 0.001) {                                              //  angle = 0 within my precision
      rgb2 = RGB_make(ww1,hh1,48);                                         //  return a copy of the input
      bitmap2 = (uint16 *) rgb2->bmp;
      cc = ww1 * hh1 * 6;
      memcpy(bitmap2,bitmap1,cc);
      return rgb2;
   }

   ww2 = int(ww1*fabs(cos(angle)) + hh1*fabs(sin(angle)));                 //  rectangle containing rotated image
   hh2 = int(ww1*fabs(sin(angle)) + hh1*fabs(cos(angle)));
   
   rgb2 = RGB_make(ww2,hh2,48);
   bitmap2 = (uint16 *) rgb2->bmp;

   rotrgb48_bitmap1 = bitmap1;
   rotrgb48_bitmap2 = bitmap2;
   rotrgb48_ww1 = ww1;
   rotrgb48_hh1 = hh1;
   rotrgb48_ww2 = ww2;
   rotrgb48_hh2 = hh2;
   rotrgb48_angle = angle;

   for (ii = 0; ii < rotrgb48_NT; ii++)                                    //  start NT threads
   {
      index[ii] = ii;
      rotrgb48_done[ii] = 0;
      start_detached_thread(RGB_rotate_48_thread,&index[ii]);
   }

   for (ii = 0; ii < rotrgb48_NT; ii++)                                    //  wait for all done
      while (rotrgb48_done[ii] == 0) zsleep(0.01);

   return rgb2;
}


void * RGB_rotate_48_thread(void *arg)
{
   int      index = *((int *) (arg));
   int      ww1, hh1, ww2, hh2;
   int      px2, py2, px0, py0;
   uint16   *ppix1, *ppix2, *pix0, *pix1, *pix2, *pix3;
   double   rx1, ry1, rx2, ry2, R, theta, px1, py1;
   double   f0, f1, f2, f3, red, green, blue;
   double   angle, pi = 3.141592654;

   ppix1 = rotrgb48_bitmap1;                                               //  input pixel array
   ppix2 = rotrgb48_bitmap2;                                               //  output pixel array
   ww1 = rotrgb48_ww1;
   hh1 = rotrgb48_hh1;
   ww2 = rotrgb48_ww2;
   hh2 = rotrgb48_hh2;
   angle = rotrgb48_angle;
   
   for (py2 = index; py2 < hh2; py2 += rotrgb48_NT)                        //  loop through output pixels
   for (px2 = 0; px2 < ww2; px2++)                                         //  outer loop y
   {
      rx2 = px2 - 0.5 * ww2;                                               //  (rx2,ry2) = center of pixel
      ry2 = py2 - 0.5 * hh2;
      R = sqrt(rx2*rx2 + ry2*ry2);                                         //  convert to (R,theta)
      if (R < 0.1) theta = 0;
      else theta = qarcsine(ry2 / R);                                      //  quick arc sine
      if (rx2 < 0) {
         if (theta < 0) theta = - pi - theta;                              //  adjust for quandrant
         else theta = pi - theta;
      }

      theta = theta - angle;                                               //  rotate theta backwards
      if (theta > pi) theta -= 2 * pi;                                     //  range -pi to +pi
      if (theta < -pi) theta += 2 * pi;

      rx1 = R * qcosine(theta);                                            //  quick cosine, sine
      ry1 = R * qsine(theta);
      px1 = rx1 + 0.5 * ww1;                                               //  (px1,py1) = corresponding
      py1 = ry1 + 0.5 * hh1;                                               //    point within input pixels

      px0 = int(px1);                                                      //  pixel containing (px1,py1)
      py0 = int(py1);
      
      if (px1 < 0 || px0 >= ww1-1 || py1 < 0 || py0 >= hh1-1) {            //  if outside input pixel array
         pix2 = ppix2 + (py2 * ww2 + px2) * 3;                             //    output is black
         pix2[0] = pix2[1] = pix2[2] = 0;
         continue;
      }

      pix0 = ppix1 + (py0 * ww1 + px0) * 3;                                //  4 input pixels based at (px0,py0)
      pix1 = pix0 + ww1 * 3;
      pix2 = pix0 + 3;
      pix3 = pix1 + 3;

      f0 = (px0+1 - px1) * (py0+1 - py1);                                  //  overlap of (px1,py1)
      f1 = (px0+1 - px1) * (py1 - py0);                                    //    in each of the 4 pixels
      f2 = (px1 - px0) * (py0+1 - py1);
      f3 = (px1 - px0) * (py1 - py0);
   
      red =   f0 * pix0[0] + f1 * pix1[0] + f2 * pix2[0] + f3 * pix3[0];   //  sum the weighted inputs
      green = f0 * pix0[1] + f1 * pix1[1] + f2 * pix2[1] + f3 * pix3[1];
      blue =  f0 * pix0[2] + f1 * pix1[2] + f2 * pix2[2] + f3 * pix3[2];
      
      pix2 = ppix2 + (py2 * ww2 + px2) * 3;                                //  output pixel
      pix2[0] = int(red);
      pix2[1] = int(green);
      pix2[2] = int(blue);
   }
   
   rotrgb48_done[index] = 1;
   pthread_exit(0);
}


